/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.*;
import java.io.File;

import org.mitre.pam.interfaces.*;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.Module;
import org.mitre.pam.util.EnabledTable;
import org.mitre.pam.daemon.DaemonBase;
import org.mitre.pam.executive.Command;
import org.mitre.pam.executive.CommandParallelInputs;
import org.mitre.pam.executive.CommandParallelOutputs;

/**
* A Juggler base class for a Juggler that has schedules.
*
* @author Nazario Irizarry, Jr.
*/
abstract public class ChanSchedBase extends DaemonBase implements Juggler {
/**
* product recepient is a reference to the one sink or null (all sinks)
*/
	protected Sink productRecepient;
/**
* The list of Sources
*/
	protected EnabledTable sourceList = new EnabledTable();
/**
* The list of Sinks
*/
	protected EnabledTable sinkList = new EnabledTable();
/**
* This form of the constructor also invokes init()
*/
	protected ChanSchedBase () {
		super();
		initChanSchedBase(false);
	}
/**
* This form of the constructor does not invoke init()
*/
	protected ChanSchedBase (boolean ignore, boolean ignore2) {
		super(false,false);
	}
	
/**
* Init actually fills in the Memento and sets default attributes
*/
	protected void initChanSchedBase(boolean initParent) {
		if (initParent)
			super.initDaemonBase(true);
		setProperty(HAS_OUTPUTS,true);
		setProperty(HAS_INPUTS,true);
		setProperty(Schedulable.HAS_SCHEDULE, true);
	}
	
/**
* Callback to receive a command, a juggler must always return to
* state= ExecGlobal.idle when it is not waiting for pims or dims
*/
	public void doCommittedTask(){
		Task task = taskHelper.getCommittedTask();
		String command = task.getCommand();

		reporter.debugMsg(this,"task:",task,null,ExecGlobal.TRACE_MSG_FLOW);
		if ( command == Command.SEND_PRODUCT ) {
			pleasePutProduct(task.getProduct(), task);
		} 
		else if ( command == Command.ACTIVATE ) {
			doGetFromAllForAll(getSourceList().elements(), task, getSinkList().elements());
		} 
		else if ( command == Command.GET_PRODUCT ) {
			if (getActivator() == null) {
				doGetFromAllForOne(getSourceList().elements(), task.getPutToken(), task.getSink());
			}
			else {
				doPutExistingProductToOne(getProduct(), task.getSink(), task.getPutToken());
			}
		} 
		else {
			warning("unexpected task, "+task);
			declareEnd(false); // clear committed task
		}
	}

/**
* Callback that indicates when the command was done.  
* @param task is the task that was completed
*/
	public void taskDone(Task doneTask){
		if (doneTask == asyncOutTask) {
		} // ignore
		else if (doneTask instanceof CommandParallelInputs) {
			reporter.debugMsg(this,"input command done",null,null,ExecGlobal.TRACE_MSG_FLOW);
		/*
		* It is possible that the inputs were done but had no product. 
		* In that case the PIMS would still be outstanding.  If there
		* had been product then our current state would be DIMS_OUTSTANDING
		* because we would already have dispatched the output in response
		* to the putProduct()
		*/
			if (getState() == STATE_PIMS_OUTSTANDING) {
				declareEnd(false);
			}
		}
		else {
			reporter.warning("unexpected taskDone, "+doneTask);
		}
	}	
	
/*
* Set run state to end of update and notify observers that we are done - Doer.DONE
*/
	public void declareEnd(boolean hasProduct) {

		if (logActivity) {
			if (hasProduct) {
				reporter.log("end juggler,\tproduct");
			}
			else {
				reporter.log("end juggler,\t[]");
			}
		}

		updateComputer.setLastUpdateTime(System.currentTimeMillis());
		if (enabled)
			updateComputer.needRecompute();
		setState(ExecGlobal.IDLE); //clears committed task
	}
	
/**
* Log the fact that we are starting and switch our execution state
*/
	public void declareStart(String newState) {
		String state = getState();
		if ((state == ExecGlobal.IDLE || state == ExecGlobal.COMMITTED) && logActivity) {
			reporter.log("begin juggler");
		}
		setState(newState);
	}
	
/**
* @return a table of enabled sources.
*/
	public Hashtable getSourceList(){
		return sourceList;
	}
	
/**
* get enumeration of all Sources including disabled items
*/
	public Enumeration getFullSourceElementEnumeration(){
		return sourceList.allElements();
	}
/**
* get enumeration of all Sinks including disabled items
*/
	public Enumeration getFullSinkElementEnumeration(){
		return sinkList.allElements();
	}
	
/**
* @return a table of enabled Sinks.
*/
	public Hashtable getSinkList(){
		return sinkList;
	}

/**
* 
*/
	final public void putProduct(Product product){
		doPutProductToAll(product, getSinkList().elements());
	}

/**
* Checks the specified putToken
* to see if it is valid, before putting the product.
* This prevents asyncrhonouse sources from sneaking product
* in while the sink is in the middle of another put or task.
*
* @param product to put
* @param the token that grants authorization to put
*/
	final public void pleasePutProduct(Product product, Task token){
		taskHelper.pleasePutProduct(product, token);
	}
/**
* Notify an agents clients that there is new product.
* Only used by agents that have output.
* The agents are notified asynchronously.  They
* will be run by the exec at a later time.
*/
	public void notifyMyClients(Task task) {
		dispatcher.notifyMyClients(this, task);
	}
/**
* Tell the exec to activate the sources to this module to ensure that
* they have the most current product.  Usually causes PIMs to get
* their product.
* Only used by agents that have input.
* The agents are activated asynchronously, at a later time.
*/
	public void executeMySources(Task task) {
		dispatcher.executeMySources(this, task);
	}
}
