/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon;

import java.util.*;
import java.io.*;

import org.mitre.pam.interfaces.*;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.Command;
import org.mitre.pam.executive.Module;


/**
* Base class for Daemons.
*
* @author Nazario Irizarry, Jr.
*/
public abstract class DaemonBase extends Module implements Daemon {
/**
* An agent to compute update times for us
*/
	protected UpdateComputer updateComputer;
	
/**
* This form of the constructor also invokes init()
*/
	protected DaemonBase() {
		super();
		initDaemonBase(false);
	}
/**
* This form of the constructor does not create a Memento or invoke init()
*/
	protected DaemonBase(boolean ignored, boolean ignoredAlso) {
		super(false, false);
	}
/**
* Invokes initModule() and fills in daemon specific 
* attributes: schedule, period, last update time
*/
	protected void initDaemonBase(boolean initParent) {
		if (initParent)
			super.initModule();
		stateMemento.setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_PERIODIC);
		updateComputer = new UpdateComputer(this, stateMemento);
		setPeriod(ScheduleAttributes.ONE_DAY);
		updateComputer.setLastUpdateTime(System.currentTimeMillis()); 
		setProperty(Schedulable.HAS_SCHEDULE, true);
	}
/**
* @deprecated
*/
	final public void doCommand(String command, Object arg){
	}
/**
* Callback to receive a command, a daemon must always return to
* state= ExecGlobal.idle when idle
*/
	public void doCommittedTask(){
		String command = taskHelper.getCommittedTask().getCommand();
		if ( command == Task.ACTIVATE ) {
			if (logActivity) 
				reporter.log("begin daemon");
			try {
				doDaemonTask();
			}
			catch (Exception e) {
				reportException(e);
			}
		}
		declareEnd(false);
	}

	protected void doDaemonTask() {
	}
/*
* set the conditions at end of update
*/
	public void declareEnd(boolean ignored) {
		if (logActivity)
			reporter.log( "end daemon");
		updateComputer.setLastUpdateTime(System.currentTimeMillis());
		if (enabled)
			updateComputer.needRecompute();

		setState(ExecGlobal.IDLE); // clears committed task
	}
/*
* set the conditions at start - unused in this base class
*/
	public void declareStart(String unused) {
	}
/**
* This is not implemented for this object.
*/
	public void taskDone(Task task){
		throw new RuntimeException("not implemented");
	}

/**
* Enable or disable this item (default = true on construction)
*/
	public void setEnabled(boolean enabled) {
		if (enabled && updateComputer != null)
			updateComputer.needRecompute();
		super.setEnabled(enabled);
	}
/**
* @return the next command/time to be executed
* @return null if  not enabled
*/
	public ScheduledCommand getNextScheduleCommand() {
		if (!isEnabled() || !isIdle())
			return null;
		return updateComputer.getNextScheduleCommand();		
	}

/**
* Check for schedule related properties and give
* them to our updateComputer.  Otherwise give them to super().setProperty().
* Ingore set of our workingState attribute.
*/
	public void setProperty(String attr, String value) {
		try {
			if (attr.equalsIgnoreCase(ExecGlobal.STATE)) {
				return;
			}
			else if (updateComputer == null || 
						!updateComputer.handleSetUpdateAttributes(attr, value)) { 
				super.setProperty(attr, value);
			}
		}
		catch (Exception e) {
			reportException(e);
		}
	}
/**
* Set period
*/
	protected void setPeriod(long period) {
		updateComputer.setPeriod(period);
	}
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to find out what to save
* on system shutdown.
*/
	public Product getPersistentProduct(){
		return null;
	}
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to restore product during
* system restart
*/
	public void setPersistentProduct(Product p){
	}
/**
* Tell the agent that it needs to active as soon as possible.
* Will only work if the agent is currently idle.
*/
	public void activateASAP(){
		if (isIdle())
			updateComputer.scheduleNow();
	}
/**
* Tell the agent that it needs to active at the specified time.
*
* @param timeInMillis time in milliseconds at which to activate
*/
	public void activateAt(long timeInMillis){
		if (isIdle())
			updateComputer.scheduleAt(timeInMillis);
	}
/**
* An activator is the thing that gets an agent to do something.  
* The activator may be a schedule. This is currently not implemented.
*/
	public Object getActivator(){
		String scheduleType = getProperty(ScheduleAttributes.SCHEDULE_VERB);
		
		if (scheduleType == ScheduleAttributes.SCHEDULE_NONE)
			return null;
		
		return updateComputer;
	}

}