/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon;
import java.util.Observable;
import java.util.Random;

import org.mitre.pam.interfaces.*;
import org.mitre.pam.executive.TimedCommand;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.PAMDate;

/**
* Compute the next update time on behalf of a juggler
*
* @author Nazario Irizarry, Jr.
*/
public class UpdateComputer {

	Schedulable juggler;
	
	public long scheduledUpdateTime = 0;
	public long lastUpdateTime = 0;
	public long period;
	
	public Memento theMemento;
	public TimedCommand scheduledUpdateCommand;
	public String scheduleType;
	

	private boolean recomputeFlag;
	private boolean scheduleNowFlag;
	private int minuteOfHour = 3;
	private int hourOfDay = 7;
	private int dayOfWeek = 1;
	private static ExecCallbacks exec;

/**
* Used for computing some update intervals
*/
	static protected Random rand = new Random();

	public UpdateComputer (Schedulable juggler, Memento theMemento) {
		this.juggler = juggler;
		this.theMemento = theMemento;
		this.scheduleType = ScheduleAttributes.SCHEDULE_PERIODIC;
		if (exec == null)
			exec = ExecGlobal.getExec();
	}
	
/**
* Compute next scheduled update time.  If the computed time
* is in the past, then assume that at least
* one update interval has passed and set the desired update time to
* 5 to 10 seconds from now (random over interval).  This is
* so that when the system first comes up it will not schedule everything
* to update at exactly the same time.  This method ensures that the
* modules NEXT_UPDATE_TIME and NEXT_UPDATE_LONG properties are properly
* set.
*/
	private void computeNextSchedTime() {
		if (scheduleType == ScheduleAttributes.SCHEDULE_NONE || scheduleType == ScheduleAttributes.SCHEDULE_ONCE) {
			/*
			* If there is already a command then set its time into the future
			* in case it is the next command already to execute on the scheduler's queue.
			* In that case the scheduler will be observing the command and will
			* get notified of the change
			*/
			if (scheduledUpdateCommand != null) {
				scheduledUpdateCommand.setTime(Long.MAX_VALUE);
			}
			scheduledUpdateCommand = null;
			scheduledUpdateTime = Long.MAX_VALUE;
			String status = (scheduleType == ScheduleAttributes.SCHEDULE_NONE) ? "never" :ScheduleAttributes.ALREADY_RAN_ONCE; 
			theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_TIME, status);
			theMemento.removeProperty(ScheduleAttributes.NEXT_UPDATE_LONG);
		}
		else {
			scheduledUpdateTime = nextTime(lastUpdateTime);
			if (scheduledUpdateTime < System.currentTimeMillis() ) {
				int interval = rand.nextInt();
				if (interval < 0)
					interval = - interval;
				interval %= 5000;
				scheduledUpdateTime = System.currentTimeMillis() + 5000 + interval;
			}
			
			/*
			* Create a new command if none exists,
			* otherwise just change the time on the current one in case
			* it is the next command to execute on the schedulers queue
			*/
			if (scheduledUpdateCommand == null) {
				scheduledUpdateCommand =  new TimedCommand(scheduledUpdateTime, null, juggler, Task.ACTIVATE);
			}
			else {
				scheduledUpdateCommand.setTime(scheduledUpdateTime);
			}
			theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_TIME, (new PAMDate(scheduledUpdateTime)).toString());
			theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_LONG, Long.toString(scheduledUpdateTime));
		}
		recomputeFlag = false;
	}
/**
* Given the last update time compute the delta to the next update time
* based on the style of scheduling, periodic, daily, hourly, weekly.
* Does not have to deal with SCHEDULE_NONE or SCHEDULE_ONCE
*/
	private long nextTime(long lastTime) {
		String style = theMemento.getProperty(ScheduleAttributes.SCHEDULE_VERB);
		if (style == null)
			return lastTime + 1000*period;
		if (style == ScheduleAttributes.SCHEDULE_NONE || style == ScheduleAttributes.SCHEDULE_ONCE){
			return Long.MAX_VALUE;
		}
		else if (style == ScheduleAttributes.SCHEDULE_HOURLY){
			PAMDate next = new PAMDate(lastTime + ScheduleAttributes.ONE_HOUR);
			next.setMinutes(minuteOfHour);
			long delta = next.getTime() - lastTime;
			if (delta < 5*ScheduleAttributes.ONE_MINUTE) {
				next = new PAMDate(lastTime + 2* ScheduleAttributes.ONE_HOUR);
				next.setMinutes(minuteOfHour);
			}
			return next.getTime();
		}
		else if (style == ScheduleAttributes.SCHEDULE_DAILY){
			PAMDate next = new PAMDate(lastTime + ScheduleAttributes.ONE_DAY);
			next.setMinutes(minuteOfHour);
			next.setHours(hourOfDay);
			long delta = next.getTime() - lastTime;
			if (delta < ScheduleAttributes.ONE_HOUR) {
				next = new PAMDate(lastTime + 2 * ScheduleAttributes.ONE_DAY);
				next.setMinutes(minuteOfHour);
				next.setHours(hourOfDay);
			}
			return next.getTime();
		}
		else if (style == ScheduleAttributes.SCHEDULE_WEEKLY){
			lastTime += ScheduleAttributes.ONE_DAY;
			PAMDate next = new PAMDate(lastTime);
			next.setMinutes(minuteOfHour);
			next.setHours(hourOfDay);
			while (next.getDay() != dayOfWeek) {
				lastTime += ScheduleAttributes.ONE_DAY;
				next = new PAMDate(lastTime);
				next.setMinutes(minuteOfHour);
				next.setHours(hourOfDay);
			}

			return next.getTime();
		}
		return lastTime + 1000*period;
	}
/**
* Set last update time
*/
	public void setLastUpdateTime(long when) {
		scheduleNowFlag = false;
		lastUpdateTime = when;
		theMemento.setProperty(ScheduleAttributes.LAST_UPDATE_TIME, (new PAMDate(when)).toString());
		theMemento.setProperty(ScheduleAttributes.LAST_UPDATE_LONG, Long.toString(when));
		needRecompute();
	}
/**
* @return the next command/time to be executed
*/
	public ScheduledCommand getNextScheduleCommand() {

		if ( scheduleNowFlag) {
			return scheduledUpdateCommand;
		}
		if (recomputeFlag ||
//			System.currentTimeMillis() > scheduledUpdateTime || 
			scheduledUpdateCommand == null) {
			computeNextSchedTime();
		}

		return scheduledUpdateCommand;		
	}

	public boolean handleSetUpdateAttributes (String attr, String value) throws NumberFormatException {

		try {
			if (attr.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_VERB)) {
	//			System.out.println("schedule "+value);
				if (value.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_NONE)) {
					setScheduleType( ScheduleAttributes.SCHEDULE_NONE);
					theMemento.removeProperty(ScheduleAttributes.HOUR_OF_DAY);
					theMemento.removeProperty(ScheduleAttributes.MINUTE_OF_HOUR);
					theMemento.removeProperty(ScheduleAttributes.DAY_OF_WEEK);
					theMemento.removeProperty(ScheduleAttributes.PERIOD);
				}
				else if (value.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_HOURLY)) {
					setScheduleType( ScheduleAttributes.SCHEDULE_HOURLY);
					String minute = theMemento.getProperty(ScheduleAttributes.MINUTE_OF_HOUR);
					if (minute == null || (minute.equals("")))
						setMinute(3);
				}
				else if (value.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_DAILY)) {
					setScheduleType( ScheduleAttributes.SCHEDULE_DAILY);
					String hour = theMemento.getProperty(ScheduleAttributes.HOUR_OF_DAY);
					if (hour == null || (hour.equals("")))
						setHour(7);
					String minute = theMemento.getProperty(ScheduleAttributes.MINUTE_OF_HOUR);
					if (minute == null || (minute.equals("")))
						setMinute(3);
				}
				else if (value.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_WEEKLY)) {
					setScheduleType( ScheduleAttributes.SCHEDULE_WEEKLY);
					String hour = theMemento.getProperty(ScheduleAttributes.HOUR_OF_DAY);
					if (hour == null || (hour.equals("")))
						setHour(7);
					String minute = theMemento.getProperty(ScheduleAttributes.MINUTE_OF_HOUR);
					if (minute == null || (minute.equals("")))
						setMinute(3);
					String day = theMemento.getProperty(ScheduleAttributes.DAY_OF_WEEK);
					if (day == null || (day.equals("")))
						setDay(1);
				}
				else if (value.equalsIgnoreCase(ScheduleAttributes.SCHEDULE_ONCE)) {
					theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_TIME, "soon"); // make sure this is not already saying we already ran once
					setScheduleType( ScheduleAttributes.SCHEDULE_ONCE);
					theMemento.removeProperty(ScheduleAttributes.HOUR_OF_DAY);
					theMemento.removeProperty(ScheduleAttributes.MINUTE_OF_HOUR);
					theMemento.removeProperty(ScheduleAttributes.DAY_OF_WEEK);
					theMemento.removeProperty(ScheduleAttributes.PERIOD);
					scheduleDelayed(10000); // schedule in 10 seconds
				}
				else {
					setScheduleType( ScheduleAttributes.SCHEDULE_PERIODIC);
					String period = theMemento.getProperty(ScheduleAttributes.PERIOD);
					if (period == null || (period.equals("")))
						setPeriod(ScheduleAttributes.ONE_DAY);
				}
					
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.DAY_OF_WEEK)) {
				for (int d = 0; d<= 6; d++) {
					if (value.equalsIgnoreCase(ScheduleAttributes.dayNames[d])){
						setDay(d);
					}
				}		
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.PERIOD)) {
				long newPeriod = Long.parseLong(value);
				setPeriod(newPeriod);
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.LAST_UPDATE_TIME)) {
				// do nothing
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.NEXT_UPDATE_TIME)) {
				// do nothing
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.NEXT_UPDATE_LONG)) {
				// do nothing
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.LAST_UPDATE_LONG)) {
				setLastUpdateTime(Long.parseLong(value));
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.HOUR_OF_DAY)) {
				setHour(Integer.parseInt(value));
			}
			else if (attr.equalsIgnoreCase(ScheduleAttributes.MINUTE_OF_HOUR)) {
				setMinute(Integer.parseInt(value));
			}
			else {
				return false;
			}
		}
		catch (NumberFormatException e) {
			juggler.getReporter().warning("bad number format -   attr="+attr+", value="+value);
			throw e;
		}
		return true;
	}

	protected void setHour(int hour) {
		if (hour > 23)
			hour = 23;
		if (hour < 0)
			hour = 0;
		hourOfDay = hour;
		theMemento.setProperty(ScheduleAttributes.HOUR_OF_DAY, Integer.toString(hourOfDay));
		needRecompute();
	}

	protected void setDay(int day) {
		if (day > 6)
			day = 6;
		if (day < 0)
			day = 0;
		dayOfWeek = day;
		theMemento.setProperty(ScheduleAttributes.DAY_OF_WEEK, ScheduleAttributes.dayNames[dayOfWeek]);
		needRecompute();
	}

	protected void setMinute(int minute) {
		if (minute > 59)
			minute = 59;
		if (minute < 0)
			minute = 0;
		minuteOfHour = minute;
		theMemento.setProperty(ScheduleAttributes.MINUTE_OF_HOUR, Integer.toString(minuteOfHour));
		needRecompute();
	}

	protected void setPeriod(long period) {
		this.period = period;
		theMemento.setProperty(ScheduleAttributes.PERIOD, Long.toString(period));
		needRecompute();
	}

	private void setScheduleType(String scheduleType) {
		this.scheduleType = scheduleType;
		theMemento.setProperty(ScheduleAttributes.SCHEDULE_VERB, scheduleType);
		needRecompute();
	}
	
	public void needRecompute() {
		recomputeFlag = true;
		if (juggler == null || juggler.getKind() == null || juggler.getId() == null)
			return; // too early in initialization phase
		if (juggler instanceof Juggler || juggler instanceof Daemon)
			exec.myScheduleChanged((Schedulable)juggler);
	}
	public void scheduleNow() {
		scheduleDelayed(100);
	}
	public void scheduleDelayed(long deltaFromNowMillis) {
		scheduleAt( System.currentTimeMillis()+deltaFromNowMillis );			
	}
	public void scheduleAt(long timeMillis) {
		scheduleNowFlag = true;
		scheduledUpdateTime = timeMillis;			
		/*
		* Create a new command if none exists,
		* otherwise just change the time on the current one in case
		* it is the next command to execute on the schedulers queue
		*/
		if (scheduledUpdateCommand == null) {
			scheduledUpdateCommand =  new TimedCommand(scheduledUpdateTime, null, juggler, Task.ACTIVATE);
		}
		else {
			scheduledUpdateCommand.setTime(scheduledUpdateTime);
		}
		theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_TIME, (new PAMDate(scheduledUpdateTime)).toString());
		theMemento.setProperty(ScheduleAttributes.NEXT_UPDATE_LONG, Long.toString(scheduledUpdateTime));

		if (juggler == null || juggler.getKind() == null || juggler.getId() == null)
			return; // too early in initialization phase
		if (juggler instanceof Juggler || juggler instanceof Daemon)
			exec.myScheduleChanged((Schedulable)juggler);
	}
}
