/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.io.File;
import java.util.Observable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;


import org.mitre.pam.interfaces.*;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.util.PAMDate;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.DoTaskHelper;
import org.mitre.pam.executive.CompletionSink;

/**
* A base class for storable modules, like PIMs, DIMs, and Jugglers. <P>
*
* @author Nazario Irizarry, Jr.
*/
public abstract class Module extends Observable implements Agent, Tasker  {
/**
* A flag to indicate that the startService() method was called
*/
	boolean isAsyncService;
/**
* A flag to indicate that the enableService() method was called
*/
	boolean serviceEnabled;
/**
* A flag used by the exec that indicates that async agent has been started
*/
	boolean asyncServiceStarted;
/**
* A help with task synchronization
*/
	protected DoTaskHelper taskHelper = new DoTaskHelper(this);
/**
* The flow dispatcher controlls how inputs or outputs are activated
*/
	protected FlowDispatcher dispatcher;
/**
* Each module instance has its own reporter
*/
	protected Reporter reporter;
/**
* The debug keyword to cause complying PIMs to show their
* query strings  LOG_QUERIES = "logQueries"
*/
	public final static String LOG_QUERIES = "logQueries";
/**
* The debug keyword to cause complying DIMs to log 
* commands issued to external agents or services LOG_COMMANDS = "logCommands"
*/
	public final static String LOG_COMMANDS = "logCommands";
/**
* Agent property that enables making entries into the log file when agents start and stop
*/
	public final static String LOG_ACTIVITY = "logActivity";
/**
* SMTP host name attribute
*/
	public final static String SMTP_HOST_NAME = "smtpHostName";

/**
* An in memory copy of the logActivity property
*/
	protected boolean logActivity;
/**
* Name of module (aka id)
*/
	protected String name;
/**
* Kind of module, typically an alias not the same as the java class name
*/
	protected String kind;
/**
* State of module
*/
	protected Memento stateMemento;
/**
* An in memory copy of our enabled property state
*/
	protected boolean enabled;
/**
* A reference to the exec
*/
	transient protected static ExecCallbacks exec;
/**
* The flow supervisor
*/
	transient protected Vector flowListenerList = new Vector(4);
/**
* Flag indicates that log messages should get written to 
* show when a module is created or destroyed
*/
	private final String MODULE_LIFE_CREATE_DELETE_LOGGING = "moduleNewFinalizeLogging";
/**
* The agent's product (if there is any)
*/
	protected Product gAgentsProduct;

/**
* This form of the constructor creates a stateMemento and invokes init().
*/
	protected Module () {
		this(false,false);
		stateMemento = exec.makeNewMemento("generic", "module");
		initModule();
	}
/**
* This form of the constructor does not create a stateMemento or invoke init().
*/
	protected Module (boolean ignored, boolean ignored2) {
		reporter = exec.getExecAsAgent().getReporter(); // use exec reporter until we get our own name
	}
/**
* Initializes key attributes:
* working, description, and enabled
*/
	protected void initModule () {
		stateMemento.setProperty(ExecGlobal.STATE, ExecGlobal.IDLE); // ensure we have an externally legible property
		setState(ExecGlobal.IDLE);
		setEnabled(true);
		setProperty(ConnectedAgent.HAS_INPUTS,"false");
		setProperty(ConnectedAgent.HAS_OUTPUTS,"false");
		setProperty(Schedulable.HAS_SCHEDULE,"false");

		setDescription("");
		metaStringInfo(DESCRIPTION, "", "What this agent does");

		setProperty(LOG_ACTIVITY, "true");
		metaBooleanInfo(LOG_ACTIVITY, true, "When set will cause a log message to be generated to show activity");

		setProperty(CREATOR, "");
		metaStringInfo(CREATOR, "", "Identifies the owner or owning configuration");
	}
	
/**
* Get the token to issue a task.
* 
* This class will return false to anybody if its state is not idle 
* (getState() != ExecGlobal.IDLE).  If a requestor that already has permission requests
* again (while the agent is idle, then it will get true (permission).
*
* @param task is what is to be done
* @return false if not idle or if already committed to another task
*/
	final public boolean getTaskingPermission(Task task){
		return taskHelper.getTaskingPermission(task);
	}
/**
* Are you idle?  Can only be idle if runstate is idle and if
* no taskers have reserved the right to task.
*
* @return true if idle
*/
	public boolean isIdle(){
		return taskHelper.isIdle();
	}
/**
* Release tasking permission.  Only the tasker with current permission
* can release tasking permission.
*/
	final public void releaseTaskingPermission(Task task){
		taskHelper.releaseTaskingPermission(task);
	}
/**
* Get a boolean-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public boolean getBooleanProperty(String property, boolean defaultValue){
		return stateMemento.getBooleanProperty(property, defaultValue);
	}
/**
* Get a double-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public double getDoubleProperty(String property, double defaultValue){
		return stateMemento.getDoubleProperty(property, defaultValue);
	}
/**
* Get a integer-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public int getIntegerProperty(String property, int defaultValue){
		return stateMemento.getIntegerProperty(property, defaultValue);
	}
/**
* Get a single string-valued property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public String getStringProperty(String property, String defaultValue){
		return stateMemento.getStringProperty(property, defaultValue);
	}
/**
* Temporary until the interface HasProperties gets this method
*/
    public long getLongProperty(String name, long defaultValue) {
	String s = stateMemento.getProperty(name);
	if (s != null && s.length() != 0) {
	    try {
		return Long.parseLong(s);
	    } catch (NumberFormatException nfe) {}
	}
	return defaultValue;
    }

/**
* @return true if the property exists
*/
	public boolean isProperty(String propertyName){
		return stateMemento.isProperty(propertyName);
	}
/**
* Set the id of this module.  The exec always calls this after
* the call to setKind().
* @param id the identifier of this module (should be unique)
*/
	final public void setId(String id) {
		reporter.assert(kind != null, "setKind() must be called before setId(), kind cannot be null, attempted id = ",id,null,null);
		reporter.assert(this.name == null, "setId() called after id was set, id = ",id,null,null);
		this.name = id;
		stateMemento.setParentInfo(kind,id);
		reporter = exec.makeReporter(id); // now we get our own reporter
		if (reporter.isDebugKey(MODULE_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("created");
		}
	}
/**
* Set the id of this module
* @param kind the generic type by which this class is known
* @param name the identifier of this module (should be unique)
*/
	final public void setKind(String kind) {
		this.kind = kind;
		setProperty(Memento.MEMENTO_KIND, kind);
	}
	
	protected void finalize()// throws Throwable
	{
		if (reporter.isDebugKey(MODULE_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("disposed");
		}
		gAgentsProduct = null;
//		super.finalize();
	}

/**
* Get the PIMs memento which must contain any parameters
* required for search/fetch.
*/
	public Memento getMemento (){
		return stateMemento;
	}
/**
* Allow the exec to set the kind of dispatcher to use (for Jugglers ony)
*/
	void setFlowDispatcher(FlowDispatcher dispatcher) {
		this.dispatcher = dispatcher;
	}
	
/**
* Accept a memento at system startup and initialize our state from it.
* The memento only describes the module, not its
* connections or sources.  Also ignore requests to set our workingState attribute.
* This method does not define the product during startup.
*/
	public void setMemento (Memento mem){
		Enumeration props = mem.getProperties();
		while(props.hasMoreElements()) {
			String attr = (String) props.nextElement();
			
			try {
				if (attr.equalsIgnoreCase(ExecGlobal.STATE)) {
					// do nothing
				}
				else if (attr.startsWith("meta_")) {
					// do nothing
				}
				else {
					copyPropertyFromAs(attr, mem, attr);
				}
			}
			catch (NumberFormatException e) {
				reporter.reportException(e);
			}
		}
	}
/**
* Accept a hashtable at system startup and initialize our state from it.
* The hashtable only describes the module, not its
* connections or sources.  Ignore requests to set our workingState attribute.
*/
	public void setAttributesFromHashtable(Hashtable hash) {
		
		for (Enumeration e = hash.keys() ; e.hasMoreElements() ;) {
		    String attribute = (String)e.nextElement();

			if (!attribute.equalsIgnoreCase(ExecGlobal.STATE) &&
				!attribute.startsWith("meta_")) {
			
				Object val = hash.get(attribute);
				if (val == null) {
					setProperty(attribute, "");
				}
				else if (val instanceof ValueList) {
					setPropertyValues(attribute, (ValueList)val);
				}
				else {
					String sval = val.toString();
					if (sval.equals(PropertySet.DELETED)) {
						removeProperty(attribute);
					}
					else {
						setProperty(attribute, sval);
					}
				}
			}
		}
	}

/**
* Sets an attribute in the modules memento.  
* The default behavior is: <p>
*
*	stateMemento.setProperty(attribute, value);
*/
	public void setProperty(String attribute, String value){
		moduleSetProperty(attribute, value);
	}
	
/**
* Sets an attribute in the modules memento.  
* The default behavior is: <p>
*
*	stateMemento.setProperty(attribute, value);
*/
	protected void moduleSetProperty(String attribute, Object oValue){
		String value = oValue.toString();
		if (attribute.equals(LOG_ACTIVITY)) {
			logActivity = value.equalsIgnoreCase("true");
		}
		else if (attribute.equals(ENABLED)) {
			setEnabled(value.equalsIgnoreCase("true"));
			return;
		}
		else if (attribute.equals(ExecGlobal.STATE)) {
			setState(value);
			return;
		}

		stateMemento.setProperty(attribute, value);
	}
/**
* Set a boolean-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(String property, boolean value){
		moduleSetProperty(property, value?"true":"false");
	}
/**
* Set a single integer-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, int value){
		moduleSetProperty(property, Integer.toString(value));
	}
/**
* Set a single long-valued property
* @param property name of the property
* @param value which is the new property value to use
*/
	public void setProperty(String property, long value){
		moduleSetProperty(property, Long.toString(value));
	}
/**
* Set a single double-valued property
* @param property name of the property
* @param value which is the new property value to use
*/
	public void setProperty(String property, double value){
		moduleSetProperty(property, Double.toString(value));
	}
/**
* Get list of property names
*/
	public Enumeration getProperties() {
		return stateMemento.getProperties();
	}
/**
*Get the value of the property
*/
	public String getProperty(String attribute){
		return stateMemento.getProperty(attribute);
	}
/**
*Get the value of the property
*/
	public Object getPropertyAsObject(String attribute){
		return stateMemento.getPropertyAsObject(attribute);
	}
/**
* @return 0 if property does not exist or is an empty valuelist
* @return 1 if string or valuelist with 1 object
* @return >1 if valuelist with more than one value
*/
	public int getCurrentDimension(String propertyName){
		return stateMemento.getCurrentDimension(propertyName);
	}
/**
* Copy the property and value potentially with a different name
*/
	public void copyPropertyFromAs(String fromName, HasProperties fromOther, String newName){
		int dim = fromOther.getCurrentDimension(fromName);

		if (dim == 0)
			removeProperty(fromName);
		else if (dim == 1)
			setProperty(newName, fromOther.getProperty(fromName));
		else
			setPropertyValues(newName, fromOther.getPropertyValues(fromName));
	}
/**
* Set a property with multiple values
*/
	public void setPropertyValues(String property, ValueList values){
		stateMemento.setPropertyValues(property, values);
	}
/**
* Get a multivalued property.
* @return always of type ValueList, never null
*/
	public ValueList getPropertyValues(String attribute){
		return stateMemento.getPropertyValues(attribute);
	}
/**
* Remove a module's attribute
*/
	public void removeProperty(String attribute){
		stateMemento.removeProperty(attribute);
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*/
	public void removePropertyAndRemoteCopies(String property){
		stateMemento.removeProperty(property);
	}
	
/**
* Get the identifier for this pim
*/
	final public String getId(){
		return name;
	}
/**
* Get the kind of agent
* @return a string that is the common name of this type of agent
*/
	final public String getKind(){
		return kind;
	}
/**
* Get state
*/
	public String getState() {
		return taskHelper.getState();
	}
/**
* Set module state.  If the state == ExecGlobal.IDLE then clear the
* tasker reservation lock.
*/
	protected void setState(String newState) {
		taskHelper.setState(newState);
	}
/**
* indicate progress by setting the runState in our memento, for
* now this is implemented the same as setState()
*/
	protected void setProgress(String p) {
		stateMemento.setProperty(ExecGlobal.STATE,p);
	}
/**
* Build a simple descriptive string
*/	
	final public String getDescription() {
		String description = getProperty(DESCRIPTION);
		return (description == null) ? "" : description;
	}
/**
* set description
*/
	protected void setDescription(String description){
		stateMemento.setProperty(DESCRIPTION, description);
	}
/**
* make the working directory for this module
*/
	final public File getWorkingDir() {
		return ExecGlobal.g_getWorkingDir(stateMemento.getUid());
	}
/**
* Get the SMTP host name for this module.  First check to see if one
* is defined for the agent, if not then check if one is defined for the exec
*/
	public String getSmtpHostName() {
		String agentHostSelection = getStringProperty(SMTP_HOST_NAME,"");
		if (agentHostSelection.length() > 0)
			return agentHostSelection;
		return ExecGlobal.getExec().getExecAsAgent().getStringProperty(ExecGlobal.DEFAULT_SMTP_HOST_NAME,"");
	}
/**
* get the string which is a url based on our current peer 
* server such as "http://x.x.x:9090/WorkingDir". 
* Does not include the final "/"
*/
	final public String getFullWorkingDirURL() {
	    return ExecGlobal.getPeerProtocolHostAndPort()  + ExecGlobal.getWorkingDirURLRoot() + "/" +
	    stateMemento.getUid();
	}
/**
* Does not return until the module is idle. After 10 minutes
* automatically logs warnings that the module has been delayed for a long time.
* However, if the agent has the attribute Agent.EXTREME_EXECUTION_TIME (which
* is "extremeExecutionTimeIsSecs") then that time is tolerated instead of the 10 minutes.
*
* @param waitIfNotIdle true if call should block until idle
*/
	final public void waitTillIdle(){
		long delay = 5171; // test every 5 seconds
		
		if (isIdle())
			return;
		
		long startTime = System.currentTimeMillis();
		long thresholdTime1 = startTime + 60*1000; // expect to be done within 1 minutes
		long expectedMax = getIntegerProperty(EXTREME_EXECUTION_TIME,9*60);
		long thresholdTime2 = startTime + expectedMax*1000; // expect to be done within 10 minutes
		PAMDate pamDate=null;
		
		while (!isIdle() && System.currentTimeMillis() < thresholdTime1) {
			try {
				taskHelper.waitTillIdle(delay);
			}
			catch (Exception ignore) {}
		}
		delay = 60*1000; // new delay every 1 minutes
		while (!isIdle() && exec.getState() != ExecGlobal.SHUTTING_DOWN && System.currentTimeMillis() < thresholdTime2) {
			if (pamDate == null)
				pamDate = new PAMDate(startTime);
			reporter.warning("has been busy since "+pamDate.toString());
			try {
				taskHelper.waitTillIdle(delay);
			}
			catch (Exception ignore) {}
		}
		if (isIdle())
			return;
		reporter.warning("remained busy, will continue but DISABLE");
		setEnabled(false);
		taskHelper.setState(ExecGlobal.IDLE); // will notifyAll()
	}

/**
* An activator is the thing that gets an agent to do something.  
* The activator may be a schedule. This is only marginally implemented.
*/
	abstract public Object getActivator();
/**
* Enable or disable this item (default = true on construction)
*/
	public void setEnabled(boolean which) {
		enabled = which;
		stateMemento.setProperty(ENABLED, which);
		if (which)
			startIfAppropriate();
		else
			stopIfAppropriate();
	}
/**
* @return true if this module is enabled
*/
	final public boolean isEnabled() {
		return enabled;
	}
/**
* Descriptive string for debugging
*/
	public String toString() {
		return "module "+name+" ("+kind+")";
	}

/**
* Sets the attributes on a product from attributes that start with "product_".
*/
	public static void copyAttributesToProduct(Memento mem, Product product) {
		if (product == null)
			return;
			
		Enumeration attributes = mem.getProperties();
		
		while(attributes.hasMoreElements()) {
			String attr = (String)attributes.nextElement();
			if ( attr.startsWith(Product.PRODUCT_PREFIX)) {
				if (mem.getStringProperty(attr,"").length() > 0) {
					String productProperty = attr.substring(Product.PRODUCT_PREFIX.length());
					product.copyPropertyFromAs(attr, mem, productProperty);
				}
			}
		}
	}
/**
* Add the listener
*/
	public void addFlowListener(FlowListener listener){
		if (flowListenerList.contains(listener))
			return;
		flowListenerList.addElement(listener);
	}
/**
* Remove the listener
*/
	public void removeFlowListener(FlowListener listener){
		flowListenerList.removeElement(listener);
	}
/**
* Generic call, no-op
*/
	public void taskDone(Task ignore) {
		
	}
/**
* For agents that have an input this gets a list
* of all inputs, even if disabled.
*/
	public Enumeration getFullSourceElementEnumeration() {
		throw new RuntimeException("cant call this");
	}
/**
* For agents that have an output this gets a list
* of all outputs, even if disabled.
*/
	public Enumeration getFullSinkElementEnumeration() {
		throw new RuntimeException("cant call this");
	}
/**
* For agents that have an input this gets the
* hashtable of outputs.  However, this only includes enabled outputs.
*/
	public Hashtable getSourceList() {
		throw new RuntimeException("cant call this");
	}
/**
* For agents that have an output this gets the
* hashtable of outputs.  However, this only includes enabled outputs.
*/
	public Hashtable getSinkList() {
		throw new RuntimeException("cant call this");
	}
/**
* Report an exception
*/
	protected void reportException(Exception e) {
		reporter.reportException(e);
	}
/**
* Make an entry into the log
*/
	protected void log(String s) {
		reporter.log(s);
	}
/**
* Report an warning
*/
	protected void warning(String msg) {
		reporter.warning(msg);
	}
/**
* Report an assertion failure
*/
	protected void assert(boolean cond, Object msg1, Object msg2, Object msg3, Object msg4 ) {
		reporter.assert(cond, msg1, msg2, msg3, msg4);
	}
/**
* Return a reference to our reporter
*/
	final public Reporter getReporter() {
		return reporter;
	}

/**
* Set metadata for Integer attributes
*/
	public void metaIntegerInfo(String attribute, Integer defaultValue, String help){
		metaIntegerInfo(attribute, defaultValue.intValue(),  help);
	}
/**
* Set metadata for integer attributes
*/
	public void metaIntegerInfo(String attribute, int defaultValue, String help){
	
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_INTEGER);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
		v.addElement(Integer.toString(defaultValue));

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Set metadata for Long attributes
*/
	public void metaLongInfo(String attribute, Long defaultValue, String help){
		metaLongInfo(attribute, defaultValue.longValue(),  help);
	}
/**
* Set metadata for long attributes
*/
	public void metaLongInfo(String attribute, long defaultValue, String help){
	
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_LONG);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
		v.addElement(Long.toString(defaultValue));

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Set metadata for Double attributes
*/
	public void metaDoubleInfo(String attribute, Double defaultValue, String help){
		metaDoubleInfo(attribute, defaultValue.doubleValue(),  help);
	}
/**
* Set metadata for double attributes
*/
	public void metaDoubleInfo(String attribute, double defaultValue, String help){
	
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_DOUBLE);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
		v.addElement(Double.toString(defaultValue));

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Set metadata for Boolean attributes
*/
	public void metaBooleanInfo(String attribute, Boolean defaultValue, String help){
		metaBooleanInfo(attribute, defaultValue.booleanValue(), help);
	}
/**
* Set metadata for boolean attributes
*/
	public void metaBooleanInfo(String attribute, boolean defaultValue, String help){
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_BOOLEAN);

		v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
		v.addElement(defaultValue?"t":"f");

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Set metadata for string attributes
*/
	public void metaStringInfo(String attribute, String defaultValue, String help){
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_STRING);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		if (defaultValue != null) {
			v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
			v.addElement(defaultValue);
		}

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Indicates that this property is multivalued.
*/
	public void metaMultiStringInfo(String attribute, ValueList defaultValue, String help){
	

		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue.elementAt(0));
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_STRING);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		v.addElement(META_ATTRIBUTE_MULTIVALUED);

		if (defaultValue != null && defaultValue.size() > 0) {
			v.addElement(META_ATTRIBUTE_VALUE_DEFAULTS);
			v.addElement(Integer.toString(defaultValue.size()));
			Enumeration e = defaultValue.elements();
			while (e.hasMoreElements()) {
				v.addElement(e.nextElement().toString());
			}
		}

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Set metadata for enumeration attributes
*/
	public void metaEnumerationInfo(String attribute, String defaultValue, String help, String[] allowableValues){
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, defaultValue);
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_ENUMERATION);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		if (defaultValue != null) {
			v.addElement(META_ATTRIBUTE_VALUE_DEFAULT);
			v.addElement(defaultValue);
		}

		v.addElement(META_ATTRIBUTE_ENUMERATED_VALUES);
		v.addElement(Integer.toString(allowableValues.length));
		for (int i = 0; i<allowableValues.length; i++) {
			v.addElement(allowableValues[i]);
		}

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Identify this attribute as a password. Should not be used
* in conjunction with metaBooleanInfo, metaStringInfo, metaEnumerationInfo,
* etc.  In other words a password is a type of field.
*/
	public void metaSetIsPassword(String attribute, String help){
		if (stateMemento.getPropertyAsObject(attribute) == null) {
			setProperty(attribute, "");
		}
		Vector v = new Vector(16);

		v.addElement(META_ATTRIBUTE_VALUE_TYPE);
		v.addElement(META_ATTRIBUTE_TYPE_PASSWORD);
		v.addElement(META_PASSWORD_ALGORITHM_DEFAULT);

		v.addElement(META_ATTRIBUTE_HELP_SHORT);
		v.addElement(help);

		setPropertyValues(META_ATTRIBUTE_PREFIX+attribute, new ValueList(v));
	}
/**
* Overridden by Sinks
*/
	public void putProduct(Product product){}
/**
* Overridden by Jugglers
*/
	protected Product buildProduct(Product product){
		return null;
	}
/**
* Overridden by all agents
*/
	abstract public void declareStart(String state);
/**
* Overridden by all agents
*/
	abstract public void declareEnd(boolean hasProduct);
/**
* Overriden by agents with sources
*/
	public void notifyMyClients(Task task) {
		throw new RuntimeException("method not implemented");
	}
/**
* Overriden by agents with sinks
*/
	public void executeMySources(Task task) {
		throw new RuntimeException("method not implemented");
	}
/**
* Push product to all clients.
* For use by Asynchonous pims, trying to asynchronously send 
* product to outputs.
* Not for use in response to an activate task.
*/
	protected CommandParallelOutputs asyncOutTask;
	final protected void pushProductToAllClients(Product p) {
		synchronized (this) {
			if (asyncOutTask == null) {
				 asyncOutTask = new CommandParallelOutputs(this);
			}
		}

		if (p != null) {
			try {
				copyAttributesToProduct(getMemento(), p);
				synchronized(asyncOutTask) {
					asyncOutTask.waitTillDone(this);
					asyncOutTask.setBusy();
					asyncOutTask.setSinks(getSinkList().elements());
					asyncOutTask.setProduct(p);
				}
				notifyMyClients(asyncOutTask);
			}
			catch (Exception e) {
				reportException(e);
			}
		}

	}
/**
* Used by jugglers and handlers.
* Gets from all inputs, calls the juggler/handlers buildProduct, then outputs
* to a specific sink.
*/
	final protected void doGetFromAllForOne(Enumeration sources, Task token, Sink productRecepient) {
		declareStart(Juggler.STATE_PIMS_OUTSTANDING);
		if ( sources.hasMoreElements() ) {
			reporter.debugMsg(this,"get inputs for one",null,null,ExecGlobal.TRACE_MSG_FLOW);
			CompletionSink completionHelper = new CompletionSink(this, productRecepient, token);
			executeMySources(new CommandParallelInputs(this, completionHelper, token, sources));
		}
		else {
			declareEnd(false);
		}
	}
/**
* Check if we have any sources.  If we don't then the command is already done.
* If we do then queue a set of jobs to execute
* This puts a new product so we have to call buildProduct()
* in the completion helper.
*/
	final protected void doGetFromAllForAll(Enumeration sources, Task token, Enumeration sinks) {
		declareStart(Juggler.STATE_PIMS_OUTSTANDING);
		reporter.debugMsg(this,"get inputs for all",null,null,ExecGlobal.TRACE_MSG_FLOW);
		if ( sources.hasMoreElements() ) {
			CompletionSinkN completionHelper = new CompletionSinkN(this, getSinkList().elements());
			executeMySources(new CommandParallelInputs(this, completionHelper, token, sources));
		}
		else {
			declareEnd(false);
		}
	}
/**
* Used by Jugglers.
* This puts a new product so we have to call buildProduct()
* in the completion helper.
*/
	final protected void doPutProductToAll(Product p, Enumeration sinks) {
		declareStart(Juggler.STATE_DIMS_OUTSTANDING);
		reporter.debugMsg(this,"send product to all",null,null,ExecGlobal.TRACE_MSG_FLOW);
		CompletionSinkN completionHelper = new CompletionSinkN(this, sinks);
		completionHelper.pleasePutProduct(p, null);
	}
/**
* Overridden by Jugglers and DIMs that have a separate
* input queue and implement the ManagesOwnInputQueue interface.
*/
	protected void putQueuedProduct(Product p){
	}
/**
* Used by Handlers and Jugglers. This can only
* be used in response to a GET_PRODUCT command because
* it assumes that the sink is ready to accept the product.
* This puts and EXISTING product so we do not have to call
* buildProduct() again.
*/
	protected void doPutExistingProductToOne(Product p, Sink productRecepient, Task putToken) {
		declareStart(Juggler.STATE_DIMS_OUTSTANDING);
		reporter.debugMsg(this,"send product to one:",productRecepient,null,ExecGlobal.TRACE_MSG_FLOW);
		try {
			if (p != null) {
				productRecepient.pleasePutProduct(p, putToken);
			}
		}
		catch (Exception e) {
			reportException(e);
		}
		declareEnd(p != null);
	}
/**
* Maintained for backward compatibility.  Do not override.
* Implement enableService method of AsynchronousAgent if
* you need to.
*/
	public void startService(){
		isAsyncService = true;
		startIfAppropriate();
	}

/**
* Maintained for backward compatibility.  Do not override.
* Implement disableService method of AsynchronousAgent if
* you need to.
*/
	public void stopService(){
		isAsyncService = false;
		stopIfAppropriate();
	}
/**
* Calls enableService() if this is a service (must implement AsynchronousAgent)
* whenever we transition from disabled to enabled.
*/
	protected void startIfAppropriate() {
		if (isAsyncService && enabled && !serviceEnabled) {
			serviceEnabled = true;
			enableService();
		}
	}
/**
* Calls disableService() on shutdown or whenever the agent transitions from enabled to disabled.
*/
	protected void stopIfAppropriate() {
		if(serviceEnabled) {
			serviceEnabled = false;
			disableService();
		}
	}
/**
* Overridden by AsynchronousAgent
*/
	public void pauseService(){
	}
	
/**
* Overridden by AsynchronousAgent
*/
	public void resumeService(){
	}
/**
* Called if this is a service (must implement AsynchronousAgent)
* whenever we transition from disabled to enabled.
*/
	public void enableService(){
	}

/**
* Called on shutdown or whenever the agent transitions from enabled to disabled.
*/
	public void disableService(){
	}
/**
* Used by the exec only
*/
	synchronized void execSetAsync(boolean on) {
		if (asyncServiceStarted && !on) {
			asyncServiceStarted = false;
			stopService();
		}
		
		else if (!asyncServiceStarted && on) {
			asyncServiceStarted = true;
			startService();
		}
	}
/**
* May be overridden by agents that have product
*/
	public Product getProduct (){
		return gAgentsProduct;
	}
/**
* Identifies the persistent product for agents that have persistent product.
* This is the method used by the exec to find out what to save
* on system shutdown.
*/
	public Product getPersistentProduct(){
		return gAgentsProduct;
	}
/**
* Allows the exec to restore persistend product during system restart
*/
	public void setPersistentProduct(Product p){
		gAgentsProduct =  p;
	}
/**
* 
*/
	public void setProduct(Product product){
		gAgentsProduct = product;
	}
}
