/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;

import org.mitre.pam.interfaces.*;

import org.mitre.pam.store.SourceSinkInfo;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pam.product.view.ProductStore;
import org.mitre.pam.util.SortedHash;
import org.mitre.pam.util.SortedLink;

import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.util.pdifHashtables.PDIFRead;

/**
* Saves system state. <P>
*
* @author Nazario Irizarry, Jr.
*/
public final class PAMSave {
	private ExecCallbacks exec;
	private PropertySet execOptions;
	private SourceSinkInfo sourceSinkInfo;
	private SortedHash listOfThingsWithoutProduct[];
	private SortedHash listOfThingsWithProduct[];
	private Reporter reportingAgent;
	private String filterProperty;
	private String filterValue;
	private String configurationName;
	public static final String BULID_NUMBER_PREFIX = "psvsnb";
	private Hashtable idsSaved;
	private PDIFWriter pdifWriter;

/**
* Needs a few references to exec state items
*/
	PAMSave (
			ExecCallbacks exec, 
			Memento execOptions, 
			SourceSinkInfo sourceSinkInfo,
			SortedHash listOfThingsWithoutProduct[], 
			SortedHash listOfThingsWithProduct[]
			) {
		this.exec = exec;
		this.execOptions = (PropertySet) execOptions;
		this.sourceSinkInfo = sourceSinkInfo;
		this.listOfThingsWithoutProduct = listOfThingsWithoutProduct;
		this.listOfThingsWithProduct = listOfThingsWithProduct;
		this.reportingAgent = exec.getExecAsAgent().getReporter();
	}

/**
* Save exec state, module state, and product state
*/
	void save(PDIFWriter pout, PrintWriter prodOut) throws IOException {		
		pout.writeHeaderAttribute(BULID_NUMBER_PREFIX,Integer.toString(exec.getBuildNumber(0)));
		checkPointObjects( pout, prodOut);
		checkPointConnections( pout);
	}

/**
* Save exec state and module state only
*/
	void save(PDIFWriter pout) throws IOException {
		save(pout, null);
	}
/**
* Turn on redundancy suppression.  This method must
* be called before doing any saving.
*/
	void setSuppressRedundantAttributes(boolean flag) {
		if (idsSaved != null)
			idsSaved.clear();
		if (flag) {
			idsSaved = new Hashtable();
		}
		else {
			idsSaved = null;
		}
	}
	
/**
* Save exec state and module state for agents
* with the selected attribute and value.
*/
	void save(PDIFWriter pout, String attribute, String value) throws IOException {
		if (attribute != null && attribute.length()!=0 &&
				value != null && value.length()!=0) {
			filterProperty = attribute;
			filterValue = value;
		}
		save(pout, null);
	}

/**
* Save exec state and module state for agents
* that are part of the specified configuration. 
* The agents must have exposure=public or they
* must have been created by the "configurationName"
*/
	void saveConfiguration(PDIFWriter pout, String configurationName) throws IOException {
		if (configurationName != null && configurationName.length()!=0) {
			this.configurationName = configurationName;
		}
		save(pout, null);
	}
	
	
	private void checkPointObjects (
			PDIFWriter out, 
			PrintWriter prodOut) throws IOException{
		pdifWriter = out;
	/*
	* save exec
	*/
		saveNode(execOptions,"exec","exec");
		storeAgentTemplates();
	/*
	* save items that have product
	*/

		ProductStore prodStore = null;
		
		if (prodOut != null)
			prodStore = new ProductStore();
			
		for (int i=0; i<listOfThingsWithProduct.length; i++) {
			Enumeration things = listOfThingsWithProduct[i].elements();
			while (things.hasMoreElements()) {
				Agent mb = (Agent) things.nextElement();
				if (prodOut == null) {
					mb.removeProperty(ExecGlobal.PERSISTENT_PRODUCT);
				}
				else {
					Product prod = mb.getPersistentProduct();
					if (prod == null) 
						mb.removeProperty(ExecGlobal.PERSISTENT_PRODUCT);
					else {
						mb.setProperty(ExecGlobal.PERSISTENT_PRODUCT, prod.getUid());
						prodStore.addToStore(prod);			
					}
				}
				
				if (passesFilterCriteria(mb)) {
					try {
						PropertySet mem = (PropertySet) mb.getMemento();
						saveNode(mem, mb.getKind(), mb.getId());
					}
					catch (Exception e) {
						reportingAgent.reportException(e);
					}
				}
			}
		}

		
		if (prodOut != null) {
			prodStore.store(prodOut);
			prodOut.close();
		}
	


	/*
	* save items that don't have product
	*/
		for (int i=0; i<listOfThingsWithoutProduct.length; i++) {
			Enumeration things = listOfThingsWithoutProduct[i].elements();
			while (things.hasMoreElements()) {
				Agent mb = (Agent) things.nextElement();
				mb.removeProperty(ExecGlobal.PERSISTENT_PRODUCT);
				if (passesFilterCriteria(mb)) {
					try {
						PropertySet mem = (PropertySet) mb.getMemento();
						saveNode(mem,
							sourceSinkInfo.getNameFromJavaClass(mb.getClass().getName()), 
							mb.getId());
						}
					catch (Exception e) {
						reportingAgent.reportException(e);
					}
				}
			}
		}
		if (idsSaved != null) {
			idsSaved.clear();
		}
	}		



	private void checkPointConnections (PDIFWriter out) throws IOException {
	
//		out.writeComment("Relations are getting written two ways for backward compatibility");
		for (int i = 0; i< listOfThingsWithProduct.length; i++ ) {
			Enumeration items = listOfThingsWithProduct[i].elements();
			while (items.hasMoreElements()) {
				Agent agent = (Agent) items.nextElement();
				if (passesFilterCriteria(agent)) {
					if (agent instanceof Sink) {
						Enumeration pulls = ((ConnectedAgent)agent).getFullSourceElementEnumeration();
						while (pulls.hasMoreElements()) {
							Agent pull = (Agent) pulls.nextElement();
							if (passesFilterCriteria(pull)) {
								out.writeRelation(typeof(pull), "from", pull.getId(), typeof(agent), "to", agent.getId());
							}
						}
					}
				}
			}
		}
	}

/*
* @return true if there are no filter  criteria 
* @return true if there is criteria and it passes
* @return false if the agent implements IDictionary
* @return false if there is criteria and it fails
*/
	private boolean passesFilterCriteria(Agent agent) {
		if (agent instanceof IDictionary)
			return false;
		if (!agent.getBooleanProperty(Agent.META_PERSIST, true))
			return false;
		if (configurationName != null)
			return passesFilterCriteriaConfig(agent);
		if (filterProperty != null)
			return passesFilterCriteriaProperty(agent);
		return true;
	}
/*
* @return true if there are no filter  criteria 
* @return true if there is criteria and it passes
* @return false if the agent implements IDictionary
* @return false if there is criteria and it fails
*/
	private boolean passesFilterCriteriaProperty(Agent agent) {
		String value = agent.getProperty(filterProperty);
		if (value == null)
			return true;
		if (value.length() == 0)
			return true;
		return value.equals(filterValue);
	}

/*
* @return true if there are no filter  criteria 
* @return true if there is criteria and it passes
* @return false if the agent implements IDictionary
* @return false if there is criteria and it fails
*/
	private boolean passesFilterCriteriaConfig(Agent agent) {
		String exposure = agent.getStringProperty(Agent.EXPOSURE_VERB, Agent.EXPOSURE_PUBLIC);
		boolean exposurePublic = exposure.length() == 0 || exposure.equals("public");
		if (exposurePublic) {
			return true;
		}
		return agent.getProperty(Agent.CREATOR).equals(configurationName);
	}
	
	private String typeof(Object foo) {
		String type = sourceSinkInfo.getNameFromJavaClass(foo.getClass().getName());
			
		reportingAgent.assert(type != null, "Exec unknown type : ",foo.getClass().getName(), null, null);
		
		if (type == null)
			return "<unknown>";
			
		return type;
	}
	private String typeofOld(Object foo) {
		if (foo instanceof PIM)
			return "PIM";
		if (foo instanceof DIM)
			return "DIM";
		if (foo instanceof Handler)
			return "juggler";
			
		reportingAgent.assert(false, "Exec unknown old type : ", foo.getClass().getName(), null, null);
			return "<unknown>";
	}

/**
* Stores properties as a sequence of parenthesized tag/value pairs. 
* This code used to be in PropertySet but was moved here so the mementos
* knew nothing of the serialization.
* @param hash is the hashtable we want to output
* @param out the output stream
*/

	private void saveNodeBehavior(PropertySet hash, PropertySet parentProperties) throws IOException {		
		for (Enumeration e = hash.keys(); e.hasMoreElements() ;) {
		    String key = (String)e.nextElement();
		    Object val = hash.getPropertyAsObject(key);
		    if ( (val != null) && 
		    	(parentProperties == null || !(val.equals(parentProperties.getPropertyAsObject(key)))))
	    		pdifWriter.writeAttributeValue(key, val);
		}
	}
/**
* Stores properties as a sequence of parenthesized tag/value pairs. 
* This code used to be in PropertySet but was moved here so the mementos
* knew nothing of the serialization.
* @param hash is the hashtable we want to output
* @param out the output stream
*/

	private void saveNodeBehavior(Hashtable hash) throws IOException{
		for (Enumeration e = hash.keys(); e.hasMoreElements() ;) {
		    String key = (String)e.nextElement();
	    	pdifWriter.writeAttributeValue(key, hash.get(key));
		}		
	}

/**
* Stores properties as a sequence of parenthesized tag/value pairs. 
* This code used to be in PropertySet but was moved here so the mementos
* knew nothing of the serialization.
* @param kind the kind of the class that this is a memento for
* @param id the id of the memento "parent" 
* @param out the output stream
*/

	private void saveNode(PropertySet pSet, String kind, String id) throws IOException {
		pdifWriter.writeNode(kind, id);

		PropertySet parentProperties = (idsSaved == null) ? null : (PropertySet)idsSaved.get(kind);
		if (parentProperties != null) {
			pdifWriter.writeAttributeValue(PDIFRead.META_NOT_REDUNDANT,"true");
		}

		saveNodeBehavior(pSet,  parentProperties);
		if (pSet.addOn != null)
			saveNodeBehavior(pSet.addOn);
		if (idsSaved != null)
			idsSaved.put(id, pSet);
	}

/**
* Write out the PDIF for the generic templates
*/
	private void storeAgentTemplates() throws IOException {
		Enumeration e = sourceSinkInfo.getListOfJavaClassNames();
		while(e.hasMoreElements()) {
			String className = (String) e.nextElement();
			Agent agent = sourceSinkInfo.getUnusedInstanceFromJavaClassName(className);
			storeOneTemplate(agent);
		}
	}
/**
* Write one element out to a printStream
*/
	private void storeOneTemplate(Agent agent) throws IOException{
	/*
	* Other attributes
	*/
		if (agent != null) {
			PropertySet memento = (PropertySet) agent.getMemento();
			saveNode(memento, agent.getKind(), agent.getId());
			pdifWriter.writeAttributeValue(SourceSinkInfo.JAVA_CLASS_STR, agent.getClass().getName());
		}
	}

}
