/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.interfaces;

import org.mitre.pam.product.view.BufferEmptyException;

/**
* Inteface for a set of products that can be recycled
* so we can reduce the reliance on Java GC
* for near realtime applications.
*
* @author Nazario Irizarry, Jr.
*/
public interface ProductPool {

/**
* Get an unused product from the pool.  The actual type of the product
* depends on the Product class that implements this interface method.
* If an empty product is not available then the call blocks until one
* is available.
* This call also increments the usage count on the allocated product - so it must
* be decremented when done with this product with an explicit
* call to doneWithProduct() @see doneWithProduct;
*
* @param originator a reference to the agent that gets the product from the pool
*
* @return an unused Product
* @throws InterruptedException if it was blocking and the block was interrupted
*/
	public Product getBlocking(Object originator) throws InterruptedException;

/**
* Get an unused product from the pool. 
* This call also increments the usage count on the allocated product - so it must
* be decremented when done with this product with an explicit
* call to doneWithProduct() @see doneWithProduct;
*
* @param originator a reference to the agent that gets the product from the pool
*
* @return an empty BasicProduct 
* @throws BufferEmptyException if nothing is available, ignore the stacktrace
* on this exception
*/
	public Product getWithoutBlocking(Object originator) throws BufferEmptyException;

/**
* Flags that the agent that allocated the product has handed off the product
* and is now done with this product.
*
* @param originator a reference to the agent that got the product from the pool
* @param product that originator is done with
*/
	public void doneWithProduct(Object originator, Product p);
/**
* Free up all space used by the product pool.
* If products are currently in use they will be freed when they
* they are no longer being used by any agents.
*/
	public void dispose();
}