/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.*;
import java.io.IOException;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.executive.ExecGlobal;

import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pdif.eventParser.PDIFItem;




/**
* Has a list of keyed contents.<p>
* Do not use these.  They were created but found
* awkward to use and to maintain!  They
* will go away in time.
*
* @author Nazario Irizarry, Jr.
*/
public class HashtableProduct extends EnumeratedProduct implements ContainerProduct {
	Hashtable hash = new Hashtable(20);

/**
* only need id and description to have a product of some sort
* @deprecated
*/
	public HashtableProduct (String id, String description) {
		super(id, description);
	}

/**
* used only by persistence subsystem
*/
	protected HashtableProduct () {
		super();
	}
/**
* Add to contents. 
* The products Uid is used as the key for this hashtable.
*/
	public void addElement(Product element) {
		assert(element != null, "HashtableProduct.addElement(null) call");
		hash.put(element.getUid(), element);
	}

/*
* Add to contents, and stamp element with a PROPERTY_INDEX attribute.
* The products Uid is used as the key for this hashtable.
* Never used
*/
/*
	public void addIndexElement(Product element) {
		assert(element != null, "HashtableProduct.addIndexElement(null) call");
		element.setProperty(PROPERTY_INDEX, PROPERTY_INDEX);
		hash.put(element.getUid(), element);
	}
*/

/**
* Add to contents, and stamp element with a PROPERTY_ROOT attribute
* The products Uid is used as the key for this hashtable.
*/
	public void addRootElement(Product element) {
		assert(element != null, "HashtableProduct.addRootElement(null) call");
		element.setProperty(PROPERTY_ROOT, PROPERTY_ROOT);
		hash.put(element.getUid(), element);
	}

/**
* Add an element to the product using an arbitrary String value as the key.<p>
*
* @param key  The key to store the value under.  This doesn't have to be
*  related to the product.
* @param element  The product to store in the composite product.
*/
	public void addElement(String key, Product element) {
		assert(element != null, "HashtableProduct.addElement(key,null) call");
		assert(key != null, "HashtableProduct.addElement(null,element) call");
		hash.put(key, element);
	}  
/**
* @return true if element is already contained
*/
	public boolean containsElement(Product element) {
		return hash.contains(element);
	}
/**
* @return true if element is already contained
*/
	public boolean isEmpty() {
		return hash.isEmpty();
	}

/**
* Add to contents, and stamp element with a PROPERTY_INDEX attribute
* Never used
*/
/*
	public void addIndexElement(String key, Product element) {
		assert(element != null, "HashtableProduct.addIndexElement(key,null) call");
		assert(key != null, "HashtableProduct.addIndexElement(null,element) call");
		element.setProperty(PROPERTY_INDEX, PROPERTY_INDEX);
		hash.put(key, element);
	}
*/
/**
* Add to contents, and stamp element with a PROPERTY_ROOT attribute
*/
	public void addRootElement(String key, Product element) {
		assert(element != null, "HashtableProduct.addRootElement(key,null) call");
		assert(key != null, "HashtableProduct.addRootElement(null,element) call");
		element.setProperty(PROPERTY_ROOT, PROPERTY_ROOT);
		hash.put(key, element);
	}

/**
* Remove an element from the product.
*/
	public void removeElement(Product product) {
		String key = (String) product.getUid();
		if (hash.containsKey(key)) {
			hash.remove(key);
			return;
		}
		
		key = getKey(product);
		if (hash.containsKey(key)) {
			hash.remove(key);
			return;
		}
		
		assert(false, "HashtableProduct.removeElement can't find element");
	}

/**
* Remove an element from the product using a specified key.
* @param String key  The key used when adding the product.
*/
	public void removeElement(String key) {
		hash.remove(key);
	}

/**
* Get a specific element
*/
	public Product getElement(String key) {
		return (Product) hash.get(key);
	}

/**
* Get an elements key.  Assumes all elements appear by only one key.
*/
	public String getKey(Product product) {
		Enumeration keys = hash.keys();
		while (keys.hasMoreElements()) {
			Object key =  keys.nextElement();
			Object element = hash.get(key);
			if (element == product)
				return (String) key;
		}
		return null;
	}

/**
* Get the enumeration of contents
*/
	public Object getContent() {
		return hash.elements();
	}
/**
* Resets content to no elements.
*/
	public void clearContent(){
		hash.clear();
	}

/**
* Get size of vector
*/
	public int size() {
		return hash.size();
	}
/**
* Get the size of the product
*/
	public double getSizeInBytesAsDouble(){
		double size = 0.0;
		Enumeration e = hash.elements();
		while (e.hasMoreElements()) {
			size += ((Product)e.nextElement()).getSizeInBytesAsDouble();
		}
		return size;
	}
/**
* Get the keys
*/
	public Enumeration keys() {
		return hash.keys();
	}

/**
* Used by persistence system
*/
	public void storeRelations(PDIFWriter pdifOut) throws IOException{
		Enumeration keys = hash.keys();
		
		if (!keys.hasMoreElements())
			return;
			
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			PDIFItem toProduct = (PDIFItem) hash.get(key);
			pdifOut.writeRelation(
				getKind(), "from", getUid(),
				toProduct.getKind(), key, toProduct.getUid());
		}
	}
/**
* Returns a copy of this product with String or URL or FileRef
* contents, but without enumerated contents.  
* Also clones a copy of product properties, id, topicVector, 
* modification date, description
*/
	public Product cloneShallow(String newId, String newDescription){
		HashtableProduct clone = new HashtableProduct(newId, newDescription);
		cloneTo(clone, newId, newDescription);
		return clone;
	}
/**
* Invokes cloneShallow() and then copies the first level references
* to other products
*/
	public Product cloneWithFirstLevelReferences(String newId, String newDescription){
		HashtableProduct clone = (HashtableProduct) cloneShallow(newId, newDescription);
		Enumeration enum = hash.keys();
		while(enum.hasMoreElements()) {
			String key = (String)enum.nextElement();
			clone.addElement(key, (Product)hash.get(key));
		}
		
		return clone;
	}
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper(String newId, String newDescription) {
		HashtableProduct clone = (HashtableProduct) cloneShallow(newId, newDescription);
		Enumeration keys = hash.keys();
		while(keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			Product p = (Product) hash.get(key);
			clone.addElement(key, p.cloneDeeper());
		}
		
		return clone;
	}
/**
* Tells containers to eliminate branches that have no children
* @return true if this ContainerProduct should be kept; false if it should be removed
*/
	public boolean prune(){
		if (hash.size() == 0)
			return false;
		
		
		Enumeration keys = hash.keys();
		while(keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			Product p = (Product) hash.get(key);
			if (p instanceof ContainerProduct) {
				if (!((ContainerProduct)p).prune())
					hash.remove(key);
			}
		}

		if (hash.size() == 0)
			return false;
		return true;
	}
/**
* Apply the filter to everything in the hashtable. Breadth first application.
*/
	protected boolean mapply(ProductFilter filter){			
		Enumeration enum = (Enumeration) getContent();
		while (enum.hasMoreElements()) {
			if (!filter.filter(this, (ProductShell) enum.nextElement()))
				return false;
		}
		enum = (Enumeration) getContent();
		while (enum.hasMoreElements()) {
			if (!((ProductShell) enum.nextElement()).mapply(filter))
				return false;
		}
		return true;
	}
/**
* Increment the reference count for all products
* in this container
*/
	public void incrementUsage(Object user){
		if (ProductFactoryImpl.productPools == null)
			return;
		Enumeration enum = hash.elements();
		while (enum.hasMoreElements()) {
			((Product) enum.nextElement()).incrementUsage(user);
		}
	}

/**
* Decrement the reference count for all products
* in this container
*/
	public void decrementUsage(Object user){
		if (ProductFactoryImpl.productPools == null)
			return;
		Enumeration enum = hash.elements();
		while (enum.hasMoreElements()) {
			((Product) enum.nextElement()).decrementUsage(user);
		}
	}
/**
* Does nothing
*/
	public Product elementAt(int n) {
		throw new RuntimeException("don't call this method");
	}

}