package org.mitre.pam.putter;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;

import org.mitre.pam.productVisitor.AsFlatDictList;

import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

/**
* This class represents a base class for any DIM that will be sending DictionaryProducts.
*
* @author John R. Fantry
*
* Note: Some code copied from source originally written by Jones Leung.
*/
public abstract class DictionaryBaseDIM extends DIMShell1 implements AsynchronousAgent {
  // The max amount of time (in seconds) a dictionary is allowed to age before being resent
  protected static final String DICT_TIMEOUT = "dictTimeout";
  protected static final int DEFAULT_TIMEOUT = 600;

  // The percentage of dictTimeout that will be the upper bound in updateDictionaryTime()
  protected static final double DICT_TIMEOUT_PERCENT = 0.25;

  protected transient SendDictionaryThread sendDictionary = null;

  // Key = IDictionary, Val = Date (time this dictionary was last sent)
  protected Hashtable dictionaries = new Hashtable();

  public DictionaryBaseDIM() {
    metaIntegerInfo(DICT_TIMEOUT, DEFAULT_TIMEOUT, "the max amount of time (in seconds) a dictionary is allowed to age");
  }

  /**
  * Begin the asynchronous portion of this DIM.
  * Start the thread which sends "dictionaries" when they time out
  */
  public void startService() {
    sendDictionary = new SendDictionaryThread(this);
    sendDictionary.setDaemon(true);
    sendDictionary.start();
  }

  /**
  * Stop the asynchronous portion of this DIM.
  * Stop the thread which sends "dictionaries" when they time out
  */
  public void stopService() {
    if(sendDictionary != null) {
      sendDictionary.running = false;
      sendDictionary = null;
    }
  }

  /**
  * This DIM doesn't inject asynchronous events into iPAM, so there are none that
  * need to be paused.
  */
  public void pauseService() {
    // do nothing
  }

  /**
  * This DIM doesn't inject asynchronous events into iPAM, so there are none that
  * need to be resumed.
  */
  public void resumeService() {
    // do nothing
  }

  public void putProduct(Product p) {
    // generate list of all dictionaries referenced in this product
    AsFlatDictList visitor = new AsFlatDictList();
    visitor.visit(p);
    Vector dictionaryList = visitor.getDictVector();

    for(int i=0; i<dictionaryList.size(); i++) {
      IDictionary dictionary = (IDictionary)dictionaryList.elementAt(i);

      // if this dictionary has never been sent before, send it now
      // dictionaries must be sent before the products that reference them
      if(!dictionaries.containsKey(dictionary)) {
        sendDictionary(dictionary);
      }
    }

    sendObject(p);
  }

  /**
  * Send all dictionaries that have timed out.
  */
  public void sendDictionaries() {
    int dictTimeout = getIntegerProperty(DICT_TIMEOUT,DEFAULT_TIMEOUT);  // seconds
    Enumeration enum = dictionaries.keys();
    while(enum.hasMoreElements()) {
      IDictionary dict = (IDictionary)enum.nextElement();
      Date lastSentTime = (Date)dictionaries.get(dict);
      Date currentTime = new Date();
      if (lastSentTime == null || (currentTime.getTime() - lastSentTime.getTime() >= dictTimeout * 1000)) {
        sendDictionary(dict);
      }
    }
  }

  /**
  * Send a single dictionary
  * @param the Dictionary object to be sent
  */
  public void sendDictionary(IDictionary dictionary) {
    if(dictionary == null) {
      return;
    }

    sendObject(dictionary);
    updateDictionaryTime(dictionary);
  }

  /**
  * Where is your data going, and how is it getting there?
  *
  * @param the object to be sent
  */
  protected abstract void sendObject(Object obj);

  /**
  * This function should be called after a dictionary has been sent.  Because we want
  * to stagger the times at which different dictionaries are sent, the timeSent variable
  * will be augmented by a random amount of time that falls between 0 and the upperBound.
  *
  * @param the dictionary that has just been sent
  */
  protected void updateDictionaryTime(IDictionary dict) {
    Date timeSent = new Date();

    Random generator = new Random(timeSent.getTime());
    int dictTimeout = getIntegerProperty(DICT_TIMEOUT, DEFAULT_TIMEOUT);  // seconds
    long upperBound = (long)(dictTimeout * DICT_TIMEOUT_PERCENT) * 1000;  // milliseconds

    long offset = upperBound + 1;
    while(offset > upperBound) {
      offset = Math.abs(generator.nextLong()) % offset;
    }

     dictionaries.put(dict, new Date(timeSent.getTime() - offset));
  }

  /**
  * This funciton is called by the sendDictionary thread to determine how long it needs
  * to sleep until a dictionary has timed out.  If there are no dictionaries currently
  * in the list, the function will return the property DICT_TIMEOUT.
  *
  * @return the number of milliseconds to sleep
  */
  public long nextSleepTime() {
    int dictTimeout = getIntegerProperty(DICT_TIMEOUT,DEFAULT_TIMEOUT);  // seconds

    if(dictionaries.isEmpty()) {
      return dictTimeout * 1000;
    }

    Enumeration lastUpdateTimes = dictionaries.elements();
    Date oldestUpdateTime = (Date)lastUpdateTimes.nextElement();

    // find the oldest date
    while(lastUpdateTimes.hasMoreElements()) {
      Date candidate = (Date)lastUpdateTimes.nextElement();

      if(candidate.before(oldestUpdateTime)) {
        oldestUpdateTime = candidate;
      }
    }

    // determine the date when the oldest dictionary will be sent
    Date nextUpdateTime = new Date(oldestUpdateTime.getTime() + (dictTimeout * 1000));

    Date now = new Date();
    long sleepTime = nextUpdateTime.getTime() - now.getTime();

    if(sleepTime > 0) {
      return sleepTime;
    }

    return dictTimeout * 1000;
  }

/**
* Sends dictionaries at specified intervals.
*/
public final class SendDictionaryThread extends Thread {
  Reporter reporter;
  DictionaryBaseDIM agent;
  boolean running = true;

  public SendDictionaryThread(DictionaryBaseDIM agent) {
    this.agent = agent;
    this.reporter = agent.getReporter();
  }

  public void run() {
    while (running) {
      try {
        agent.sendDictionaries();
        long sleepTime = agent.nextSleepTime();
        sleep(sleepTime);
      } catch (Exception e) {
        reporter.reportException(e);
      }
    }
  }
}
}