/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.store;

import java.io.*;
import java.util.Random;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.StringTokenizer;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pdif.eventParser.ParseEvent2;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.util.smtp.MIMEc;
import org.mitre.pam.util.StreamCopy;

/**
 * A file ref points to persistent files.
 *
 * @author Jones Leung, 1.1, added getKeepFile()
 * @author Nazario Irizarry, Jr. 1.0
 * @author John R. Fantry - added support for Java serialization
 */
public class FileRef implements Serializable
{
	protected static final int MAX_READ_SIZE = 65000;
	
	transient protected static ExecCallbacks exec;

	/**
	 * We need to refer to the master list of files
	 */
	protected transient FileRefList fileRefPeer;

	/**
	 * used for generating unique file ids
	 */
	static Random randFileId = new Random(System.currentTimeMillis());

	/**
	 * A global pointer back to the exec
	 */
	private transient static Reporter reporter;

	/**
	 * modification date refers to its contents, not its attributes
	 */
	protected long modificationDate;

	/**
	 * mime type of file, not always known
	 */
	protected String MIMEType = new String();

	/**
	 * name of file from the source
	 */
	protected String sourceFileName;

  // the following hashtables are used during serialization/deserialization
	transient private static Hashtable serialFileList = new Hashtable();
	transient private static Hashtable rootDirectories = new Hashtable();
	transient private static Hashtable serializedFiles = new Hashtable();
	transient private static Hashtable deserializedFiles = new Hashtable();

	/**
	 * Build a file ref from a File <P>
	 *
	 * @param file is the local File that this FileRef describes
	 * @param date is the file modification date in milliseconds
	 * @param soureFileName the original file name of this file
	 */
	public FileRef(File file, long date, String sourceFileName)
	{
		fileRefPeer = getPeer(file);
		this.sourceFileName = sourceFileName;
		this.modificationDate = date;
		MIMEType = MIMEc.getMIMEtype(file);
		if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("created\t"+this);
		}
	}

	/**
	 * Used only by other members of the Store package
	 */
	FileRef()
	{
		this.modificationDate = System.currentTimeMillis();
	}

	/**
	 * constructs a fileref from its serialized string
	 */
	public FileRef(String serialized)
	{
		this.modificationDate = System.currentTimeMillis();
		serialIn(serialized);
	}

	/**
	 * Set the keep file flag, can only be set, never reset
	 */
	public void setKeepFile(boolean keepFlag)
	{
		fileRefPeer.keepFlag = keepFlag;
	}

	/**
	 * Change the parent on this file ref.  Used during product
	 * reconstruction when the product is "sent" to another
	 * site.
	 */
	public void resetParent(File parent)
	{
		fileRefPeer.resetParent(parent);
	}


	/**
	 * Mark this with the current hold token
	 */
	public void mark()
	{
		fileRefPeer.mark();
	}

	/**
	 * Get the File object that is wrapped by this FileRef
	 */
	public File getFile()
	{
		if(fileRefPeer == null)
		{
			throw new IllegalArgumentException("FileRef " + sourceFileName + " has no corresponding peer");
		}
		return fileRefPeer.thisFile;
	}

	/**
	 * Get the original name of the file
	 */
	public String getSourceName()
	{
		return sourceFileName;
	}

	/**
	 * @return Date when file was last modified
	 */
	public Date getModificationDate()
	{
		return new Date(modificationDate);
	}

	/**
	 * @return long date when file was last modified
	 */
	public long getModificationDateLong()
	{
		return modificationDate;
	}

	/**
	 * @return file mime type
	 */
	public String getMIMEType()
	{
		return MIMEType;
	}

	/**
	 * Set file mime type
	 */
	public void setMIMEType(String type)
	{
		MIMEType = type;
	}

	/**
	 * Set our exec
	 */
	public static void setExec(ExecCallbacks execRef)
	{
		exec = execRef;
		reporter = execRef.makeReporter("FileRef");
		FileRefList.setExec(execRef);
	}

	/**
	 * Serialize out
	 */
	public String serialOut()
	{
		return ParseEvent2.fancyQuote(sourceFileName) + '+' + modificationDate + '+' + MIMEType + '+' + ParseEvent2.fancyQuote(getFile().getAbsolutePath()) + '+' + (fileRefPeer.keepFlag ? "keep" : "nokeep") + '+';
	}

	/**
	 * Serialize in
	 */
	protected void serialIn(String in)
	{
		StringBuffer workingSpace = new StringBuffer(80);
		int d1 = in.indexOf('+',0);
		sourceFileName = ParseEvent2.fancyUnquote(in.substring(0,d1), workingSpace);
		//			System.out.println("source "+sourceFileName);
		d1++;
		int d2 = in.indexOf('+',d1);
		try
		{
			modificationDate = Long.parseLong(in.substring(d1,d2));
		//				System.out.println("long date "+in.substring(d1,d2));
		}
		catch(NumberFormatException e)
		{
			reporter.reportException(e);
		}
		d2++;
		int d3 = in.indexOf('+',d2);
		MIMEType = in.substring(d2,d3);
		//			System.out.println("MIMEType "+MIMEType);
		d3++;
		int d4 = in.indexOf('+',d3);
		String localName = ParseEvent2.fancyUnquote(in.substring(d3,d4), workingSpace);
		fileRefPeer = getPeer(new File(localName));
		//			System.out.println("localName "+localName);
		d4++;
		int d5 = in.indexOf('+',d4);
		String keepStr = ParseEvent2.fancyUnquote(in.substring(d4,d5), workingSpace);
		fileRefPeer.keepFlag = keepStr.equals("keep");
	//			System.out.println("keep "+keepFlag);
		if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("reconstituted\t"+this);
		}
	}

	/**
	 * Returns a full URL  for the
	 * file provided. If the file does not exist within the support or working directory
	 * structure of the exec then an exception is thrown. <p>
	 *
	 * @return A string that specifies the URL for the given file (minus the
	 *  protocol and host - Call ExecGlobal.getProtocolAndHost() to get that portion).
	 *
	 */
	public String getReferenceAsFullURLString() throws FileNotLocalException
	{
		URL filePart = null;
		try
		{
			URL hostPart = new URL(ExecGlobal.getPeerProtocolHostAndPort());
			filePart = new URL(hostPart,ExecGlobal.getURLFromFile(this.getFile()));
		}
		catch(MalformedURLException e)
		{
			reporter.reportException(e);
		}
		/*
		 * URLs should not have any spaces in them
		 */
		return encodeURL(filePart.toExternalForm());
	}

	//Similar to URLEncode.encode(), but turns ' 's into
	// "%20"s and assumes that urlString is already
	// in external form (it ignores characters such
	// as ':','/', etc.)
	private String encodeURL(String urlString)
	{
		StringBuffer retVal = new StringBuffer();
		int length = urlString.length();
		int c = -1;
		for(int i = 0;i < length;i++)
		{
			c = (int)urlString.charAt(i);
			if((c >= 'a') && (c <= 'z'))
			{
				retVal.append((char)c);
			}
			else
				if((c >= 'A') && (c <= 'Z'))
				{
					retVal.append((char)c);
				}
				else
					if((c >= '0') && (c <= '9'))
					{
						retVal.append((char)c);
					}
					else
						if(":/_-.*?&=+".indexOf(c) >= 0)
						{
							retVal.append((char)c);
						}
						else
						{
							// note that ' ' *IS* being encoded!
							retVal.append("%");
							if(c < 16)
							{
								retVal.append("0");
							}
							retVal.append(Integer.toHexString(c));
						}
		}
		return retVal.toString();
	}

	/**
	 * Identifies whether a file can have a URL (is servicable via http). <p>
	 */
	public boolean hasURL()
	{
		return ExecGlobal.fileHasURL(this.getFile());
	}

/**
* Copy one file to another. Just delegates to StreamCopy.copyFile.
* This entry point was left here for backward compatibility.  It
* will need to get cleaned up later.
*/
	public static void copyFile(File from, File to) throws IOException {
		StreamCopy.copyFile(from,to);
	}


	/**
	 * Get the name of the parent directory for this file
	 */
	//	private File getParentDir() {return null;}
	/**
	 * cleanup files during working operation
	 */
	protected void finalize() throws Throwable
	{
		if (exec.getState() != ExecGlobal.SHUTTING_DOWN) {
			if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
				reporter.log("finalize\t"+this);
			}
			fileRefPeer.decrementRefCount();
		}
		else {
			if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
				reporter.log("finalize during shutdown ignored\t"+this);
			}
		}

		fileRefPeer = null;
		super.finalize();
	}

	/*
	 * Return a unique file id with a specified prefix and extension.  The final
	 * string returned consists of the prefix, followed by '-',
	 * followed by a random number followed by a '.' followed by an extension
	 */
	static public String getNewFileId(String prefix, String extension)
	{
		long id = randFileId.nextLong();
		if(id < 0)
			id = -id;
		id %= 1000000;
		StringBuffer composite = new StringBuffer(40);
		if(prefix != null && !prefix.equals(""))
		{
			composite.append(prefix);
			composite.append('-');
		}
		composite.append(Long.toString(id));
		if(extension != null && !extension.equals(""))
		{
			composite.append('.');
			composite.append(extension);
		}
		return composite.toString();
	}

	/*
	 * Return a unique file id with a specified prefix.  The final
	 * string returned consists of the prefix, followed by '-',
	 * followed by a random number
	 */
	static public String getNewFileId(String prefix)
	{
		return getNewFileId(prefix,null);
	}

	/*
	 * Return a unique file id
	 */
	static public String getNewFileId()
	{
		return getNewFileId("fid",null);
	}

	/**
	 * Get the keep file flag
	 */
	public boolean getKeepFile()
	{
		return fileRefPeer.keepFlag;
	}

	/**
	 * Converts this FileRef to a <code>String</code>.
	 *
	 * @return  a string representation of this FileRef.
	 */
	public String toString()
	{
		return "FileRef[sourceFileName=" + sourceFileName + ",modificationDate=" + modificationDate + ",fileRefPeer=" + fileRefPeer + ",MIMEType=" + MIMEType + "]";
	}

	/**
	 * Builds a valid path name with the three arguments.  The directory hierarchy is
	 * created if it does not exist.
	 */
	protected File buildPathName(File rootDir, String path, String fileName)
	{
		StringBuffer pathName;
		if(rootDir.getAbsolutePath().endsWith(File.separator))
		{
			pathName = new StringBuffer(rootDir.getAbsolutePath());
		}
		else
		{
			pathName = new StringBuffer(rootDir.getAbsolutePath() + File.separator);
		}
		if(path.startsWith(File.separator))
		{
			pathName.append(path.substring(1));
		}
		else
		{
			pathName.append(path);
		}
		File dir = new File(pathName.toString());
		if(!dir.exists())
		{
			dir.mkdirs();
		}
		if(pathName.toString().endsWith(File.separator))
		{
			pathName.append(fileName);
		}
		else
		{
			pathName.append(File.separator + fileName);
		}
		return new File(pathName.toString());
	}

	/**
	 * After using default Java serialization, this function will write additional data
	 * including the contents of the file if needed.
	 */
	private void writeObject(ObjectOutputStream out) throws IOException
	{
		out.defaultWriteObject();
		String srcPath = fileRefPeer.thisFile.getAbsolutePath();
		out.writeObject(srcPath);
		// write the file to the stream if we need to
		if(addSerializedFile(out,fileRefPeer.thisFile))
		{
			String dstPath = getSerialFilePath(out,this);
			out.writeObject(dstPath);
			out.writeLong(fileRefPeer.thisFile.length());
			byte[] bbuf = new byte[MAX_READ_SIZE];
			FileInputStream fis = new FileInputStream(fileRefPeer.thisFile);
			try
			{
				int bytesRead = fis.read(bbuf);
				while(bytesRead != -1)
				{
					out.write(bbuf,0,bytesRead);
					bytesRead = fis.read(bbuf);
				}
			}
			finally
			{
				try
				{
					fis.close();
				}
				catch(Exception ignore)
				{
				}
			}
		}
	}

	/**
	 * After using default Java serialization, this function restores the file, if needed, and
	 * fileRefPeer.
	 */
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		in.defaultReadObject();
		String srcPath = (String)in.readObject();
		// try and get a handle to this file if it has been previously deserialized
		fileRefPeer = getDeserializedFile(in,srcPath);
		// if null then the file must be restored
		if(fileRefPeer == null)
		{
			File rootDir = getSerialRootDir(in);
			String dstPath = (String)in.readObject();
			long fileLength = in.readLong();
			String fileName;
			if(srcPath.indexOf(File.separator) != -1)
			{
				fileName = srcPath.substring(srcPath.lastIndexOf(File.separator) + File.separator.length());
			}
			else
			{
				fileName = srcPath;
			}
			File file = buildPathName(rootDir,dstPath,fileName);
			FileOutputStream out = new FileOutputStream(file);
			try
			{
				byte[] bbuf;
				if(MAX_READ_SIZE > fileLength)
				{
					bbuf = new byte[(int)fileLength];
					in.readFully(bbuf);
					out.write(bbuf);
				}
				else
				{
					bbuf = new byte[MAX_READ_SIZE];
					long bytesRemaining = fileLength;
					while(bytesRemaining > 0)
					{
						if(bytesRemaining >= MAX_READ_SIZE)
						{
							in.readFully(bbuf);
							out.write(bbuf);
							bytesRemaining -= bbuf.length;
						}
						else
						{
							bbuf = new byte[(int)bytesRemaining];
							in.readFully(bbuf);
							out.write(bbuf);
							bytesRemaining -= bbuf.length;
						}
					}
				}
				out.flush();
			}
			finally
			{
				try
				{
					out.close();
				}
				catch(Exception ignore)
				{
				}
			}
			fileRefPeer = getPeer(file);
			addDeserializedFile(in,srcPath,fileRefPeer);
		}
	}

	private FileRefList getPeer(File file) {
		FileRefList fileRefList = FileRefList.getPeer(file);
		fileRefList.incrementRefCount();
		return fileRefList;
	}
	
  // used by Serializer.serialize
  public static void addSerialFileList(ObjectOutputStream out, Hashtable fileList) {
    serialFileList.put(out, fileList);
  }

  // used by FileRef.writeObject
  public static String getSerialFilePath(ObjectOutputStream out, FileRef file) {
    Hashtable files = (Hashtable)serialFileList.get(out);
    String path = (String)files.get(file);

    return path;
  }

  // used by Serializer.serialize
  public static void removeSerialFileList(ObjectOutputStream out) {
    serialFileList.remove(out);
  }

  // used by Serializer.unSerialize
  public static void addSerialRootDir(ObjectInputStream in, File dir) {
    rootDirectories.put(in, dir);
  }

  // used by FileRef.readObject
  public static File getSerialRootDir(ObjectInputStream in) {
    return (File)rootDirectories.get(in);
  }

  // used by Serializer.unSerialize
  public static void removeSerialRootDir(ObjectInputStream in) {
    rootDirectories.remove(in);
  }

  /**
   * This function returns true if this file has not been serialized to this stream,
   * false otherwise.  Used by FileRef.writeObject
   */
  public static boolean addSerializedFile(ObjectOutputStream out, File file) {
    Vector files = (Vector)serializedFiles.get(out);

    if(files == null) {
      files = new Vector();
    }

    if(files.contains(file)) {
      return false;
    }

    files.addElement(file);
    serializedFiles.put(out, files);

    return true;
  }

  // used by Serializer.serialize
  public static void removeSerializedFiles(ObjectOutputStream out) {
    serializedFiles.remove(out);
  }

  /**
  * This function returns a handle to a FileRefList object if it exists.
  * Used by FileRef.readObject
  */
  public static FileRefList getDeserializedFile(ObjectInputStream in, String srcPath) {
    Hashtable files = (Hashtable)deserializedFiles.get(in);

    if(files == null) {
      files = new Hashtable();
    }

    return (FileRefList)files.get(srcPath);
  }

  // used by FileRef.readObject
  public static void addDeserializedFile(ObjectInputStream in, String srcPath, FileRefList frl) {
    Hashtable files = (Hashtable)deserializedFiles.get(in);

    if(files == null) {
      files = new Hashtable();
    }

    files.put(srcPath, frl);
    deserializedFiles.put(in, files);
  }

  // used by Serializer.unSerialize
  public static void removeDeserializedFiles(ObjectInputStream in) {
    deserializedFiles.remove(in);
  }
}

