/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.store;

import java.io.*;
import java.util.*;


import org.mitre.pam.interfaces.*;
import org.mitre.pdif.eventParser.ParseEvent3;
import org.mitre.pdif.eventParser.BadPDIFException;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pam.executive.PropertySet;

/**
* A "database" record
*/
class SSRecord {
	String commonName = new String("");
	String javaClassName = new String("");
	Agent instance;
	
	public String toString() { 
		return "[SSRecord commonName = "+commonName+" class = "+javaClassName+" agent = "+instance+" ]" ;
	}
}
/**
* Provides information about the PIM, DIM, Juggler classes that
* are available <P>
*
* @author Nazario Irizarry, Jr.
*/
public final class SourceSinkInfo {
/**
* reference to the exec
*/
//	private ExecCallbacks exec;
/**
* commonName is a key field
*/
	private Hashtable commonName = new Hashtable(40);

/**
* javaClassName is a key field
*/
	private Hashtable javaClassName = new Hashtable(40);
/**
* A list of DIMs
*/
	private Hashtable unusedDIMs = new Hashtable(40);
/**
* A list of PIMs
*/
	private Hashtable unusedPIMs = new Hashtable(40);
/**
* A list of Jugglers
*/
	private Hashtable unusedJugglers = new Hashtable(40);
/**
* A list of Daemons
*/
	private Hashtable unusedDaemons = new Hashtable(40);
/**
* A list of lists
*/
	private Hashtable[] unusedModuleList;
/**
* constants used in output files
*/
	private static final String SOURCE_STR = "source";
	private static final String SINK_STR = "sink";
	public static final String PIM_STR = "pimType";
	public static final String DAEMON_STR = "daemonType";
	public static final String DIM_STR = "dimType";
	public static final String JUGGLER_STR = "jugglerType";
	public static final String MODULE_STR = "moduleType";
	public static final String JAVA_CLASS_STR = "javaClass";
	private static final String COMMON_NAME_STR = "commonName";
	Reporter reporter;

/**
* used for generating unique ids
*/	
	public SourceSinkInfo (ExecCallbacks exec) {
		this.reporter = exec.getExecAsAgent().getReporter();
		unusedModuleList = new Hashtable[4];
		unusedModuleList[ExecToMonitor.DIM] = unusedDIMs;
		unusedModuleList[ExecToMonitor.PIM] = unusedPIMs;
		unusedModuleList[ExecToMonitor.JUGGLER] = unusedJugglers;
		unusedModuleList[ExecToMonitor.DAEMON] = unusedDaemons;
	}
	
/**
* Load and return a product from persistent storage.
*/
	public void load(Reader in) {
		try {
						
			ParseEvent3 pEvent = new ParseEvent3(in);
			
			try {
				while ( true ) {
					int type = pEvent.next();
					
					if (type == ParseEvent3.EOF)
						break;

					if ( type == ParseEvent3.NODE_ENTER ) {

						if (pEvent.node.kind.equals(PIM_STR)) {
							loadOne(pEvent);
						}
						else if (pEvent.node.kind.equals(DIM_STR)) {
							loadOne(pEvent);
						}
						else if (pEvent.node.kind.equals(JUGGLER_STR)) {
							loadOne(pEvent);
						}
						else if (pEvent.node.kind.equals(DAEMON_STR)) {
							loadOne(pEvent);
						}
						else if (pEvent.node.kind.equals(MODULE_STR)) {
							loadOne(pEvent);
						}
					}
				}
			
			}
			catch (BadPDIFException e) {
				reporter.reportException(e);
			}
			
		}
		catch (IOException e) {
			reporter.reportException(e);
		}

	}

/**
* Load and return a product from persistent storage.
*/
	public void loadOne(ParseEvent3 pEvent) throws IOException, BadPDIFException {
		SSRecord currentRecord = null;

		int type = pEvent.eventType;
			

		while ( true  ) {
			
			if (type == ParseEvent3.EOF)
				return;

			if ( type == ParseEvent3.NODE_ENTER ) {
				currentRecord = new SSRecord();
				currentRecord.commonName = pEvent.node.id;
			}
			else if (type == ParseEvent3.NODE_ATTRIBUTE && (currentRecord != null)  ) {
				
				if (pEvent.attribute.name.equals(JAVA_CLASS_STR)) {
					currentRecord.javaClassName = pEvent.attribute.getValue();
				}
			}
			else if ((type == ParseEvent3.NODE_LEAVE) && 
				(currentRecord != null) ) {

				if (currentRecord.javaClassName == null || currentRecord.javaClassName.equals("")) {
					throw new BadPDIFException("Attempt to define new "+pEvent.node.id+" without a java class name");
				}
				defineModuleAlias(currentRecord);
				return;
			}
			type = pEvent.next();
		}
	}

/**
* Get list of available modules of the specified type.  
* 
* @param type one of ExecCallbacks.DIM, ExecCallbacks.PIM, etc
* @return an enumeration of unused instances of DIM for use as a template
*/
	public Enumeration getListOfAvailableModules(int type){
		return unusedModuleList[type].elements();
	}

/**
* Get a list of the java class names that we know of
*/
	public Enumeration getListOfJavaClassNames() {
		return javaClassName.keys();
	}
/**
* Get an unused instance from the java class name
*/
	public Agent getUnusedInstanceFromJavaClassName(String className) {
		SSRecord r = (SSRecord) javaClassName.get(className);
		return ((r == null)? null : r.instance);
	}
/**
* Get the java class name given the common name
*/
	public String getClassNameFromCommonName(String name) {
		SSRecord r = (SSRecord) commonName.get(name);
		return ((r == null)? null : r.javaClassName);
	}
/**
* Get an unused instance given the common name
*/
	public Agent getUnusedInstanceFromCommonName(String name) {
		SSRecord r = (SSRecord) commonName.get(name);
		return ((r == null)? null : r.instance);
	}
/**
* Get the name given the java class name
*/
	public String getNameFromJavaClass(String name) {
		SSRecord r = (SSRecord) javaClassName.get(name);
		return ((r == null)? null : r.commonName);
	}

/**
* Get a string which is the kind of the module
*/
	public String kindOf(Agent agent) {
		String kind = "unknown";
		if (agent instanceof PIM) 
			kind = PIM_STR;
		else if (agent instanceof DIM) 
			kind = DIM_STR;
		else if (agent instanceof Handler) 
			kind = JUGGLER_STR;
		else if (agent instanceof Daemon) 
			kind = DAEMON_STR;
		else {
			reporter.assert(false,  "SourceSinkInfo cant determine type of ",agent,null,null);
		}
		return kind;
	}
/**
* Allows a controller to define aliases for pims, dims, or jugglers
*
* @param alias the name to be used as an alias (has pdif visibility)
* @param javaClassName must match the javaClassName, used for dynamic loading
*/
	public void defineModuleAlias(String alias, String javaClassName){
		SSRecord currentRecord = new SSRecord();
		currentRecord.commonName = alias;
		currentRecord.javaClassName = javaClassName;
		defineModuleAlias(currentRecord);
	}
/**
* Allows a controller to define aliases for pims, dims, or jugglers.
* The first definition of a java class will result in the creation of that class.
* Subsequent aliases may refer to the same class but will result in multiple
* aliases to the same reference instantiation.  The first (the reference)
* is the one who's name is used when the system checkpoints.  This strategy
* supports backward compatible rename of agents because the iPAM
* plugins get to define their stuff before the checkpoint file is read.
* Any definitions in the checkpoint file will be understood, but the plugins
* name will be used on the next save.
*
* @param alias the name to be used as an alias (has pdif visibility)
* @param javaClassName must match the javaClassName, used for dynamic loading
*/
	private void defineModuleAlias(SSRecord currentRecord){
	
		boolean succeed = true;
		
		if (commonName.containsKey(currentRecord.commonName)) {
			return;
		}
		SSRecord existingRecord = (SSRecord) javaClassName.get(currentRecord.javaClassName);
		if (existingRecord != null) {
			commonName.put(currentRecord.commonName, existingRecord);
//			reporter.log("duplicate classname "+currentRecord.commonName);
		}
		else {
			try {
				Class itemClass = Class.forName(currentRecord.javaClassName);
				Agent module = (Agent)itemClass.newInstance();
	
				module.setKind(kindOf(module));
				module.setId(currentRecord.commonName);
	
				currentRecord.instance = module;
	
				if (module instanceof DIM) {
					unusedDIMs.put(currentRecord.commonName, currentRecord.instance);
				}
				else if (module instanceof PIM) {
					unusedPIMs.put(currentRecord.commonName, currentRecord.instance);
				}
				else if (module instanceof Handler) {				
					unusedJugglers.put(currentRecord.commonName, currentRecord.instance);
				}
				else if (module instanceof Daemon) {
					unusedDaemons.put(currentRecord.commonName, currentRecord.instance);
				} else if (module instanceof AsynchronousAgent) {
					unusedDaemons.put(currentRecord.commonName, currentRecord.instance);
				}
				else 
					return;
	
				commonName.put(currentRecord.commonName, currentRecord);
				javaClassName.put(currentRecord.javaClassName, currentRecord);
					
			}
			catch (ClassNotFoundException skipIt) {
				reporter.error("Unable to load java class: "+currentRecord.javaClassName+", "+skipIt.getMessage());
			}
			catch (IllegalAccessException skipIt) {}
			catch (InstantiationException skipIt) {}
			catch (NoClassDefFoundError skipIt) {
				reporter.error("Unable to load java class: "+currentRecord.javaClassName+", "+skipIt.getMessage());
			}
		}
	}

}
