/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util;

/*
 * @(#)Codecs.java					0.2-2 23/03/1997
 *
 *  This file is part of the HTTPClient package
 *  Copyright (C) 1996,1997  Ronald Tschalaer
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 *
 *  For questions, suggestions, bug-reports, enhancement-requests etc.
 *  I may be contacted at:
 *
 *  ronald@innovation.ch
 *  Ronald.Tschalaer@psi.ch
 *
 *  -------------------------------------------------------------------
 *
 *  UUENCODE and UUDECODE were written on 7/22/97 by John Fantry
 *  of the Mitre Corporation.
 *
 *  Jason Mathews 10/18/99 - Added HTMLEncode and urlPathEncode methods.
 */

import java.util.BitSet;

/**
 * This class collects various encoders and decoders.
 *
 * @version	1 7/22/97
 * @author	John Fantry added UUENCODE and UUDECODE
 */
/**
 * This class collects various encoders and decoders.
 *
 * @version	0.2 (bug fix 2)  23/03/1997
 * @author	Ronald Tschal&auml;r
 */

public class Codecs
{
    private static BitSet  BoundChar;
    private static byte[]  Base64DecMap;
    private static char[]  Base64EncMap;

    // Class Initializer

    static
    {
	// rfc-1521: bcharsnospace - used for multipart codings
	BoundChar = new BitSet(256);
	for (int ch='0'; ch <= '9'; ch++)  BoundChar.set(ch);
	for (int ch='A'; ch <= 'Z'; ch++)  BoundChar.set(ch);
	for (int ch='a'; ch <= 'z'; ch++)  BoundChar.set(ch);
	BoundChar.set('\'');
	BoundChar.set('(');
	BoundChar.set(')');
	BoundChar.set('+');
	BoundChar.set(',');
	BoundChar.set('-');
	BoundChar.set('.');
	BoundChar.set('/');
	BoundChar.set(':');
	BoundChar.set('=');
	BoundChar.set('?');
	BoundChar.set('_');

	// rfc-521: Base64 Alphabet
	char[] map = 
	    {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	     'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	     'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	     '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'};
	Base64EncMap = map;
	Base64DecMap = new byte[128];
	for (int idx=0; idx<Base64EncMap.length; idx++)
	    Base64DecMap[Base64EncMap[idx]] = (byte) idx;
    }


    // Constructors

    /**
     * This class isn't meant to be instantiated.
     */
    private Codecs() {}


    // Methods

    /**
     * This method encodes the given string using the base64-encoding
     * specified in RFC-1521 (Section 5.2). It's used for example in the
     * "Basic" authorization scheme.
     *
     * @param  str the string
     * @return the base64-encoded <var>str</var>
     */
    public final static String base64Encode(String str)
    {
      if (str == null)  return  null;

      byte data[] = str.getBytes();

      return new String(base64Encode(data));
    }


    /**
     * This method encodes the given byte[] using the base64-encoding
     * specified in RFC-1521 (Section 5.2).
     *
     * @param  data the data
     * @return the base64-encoded <var>data</var>
     */
    public final static byte[] base64Encode(byte[] data)
    {
      if (data == null)  return  null;

      int sidx, didx;
      byte dest[] = new byte[((data.length+2)/3)*4];


      // 3-byte to 4-byte conversion + 0-63 to ascii printable conversion
      for (sidx=0, didx=0; sidx < data.length-2; sidx += 3)
      {
        dest[didx++] = (byte)Base64EncMap[(data[sidx] >>> 2) & 077];
        dest[didx++] = (byte)Base64EncMap[(data[sidx+1] >>> 4) & 017 |
					(data[sidx] << 4) & 077];
        dest[didx++] = (byte)Base64EncMap[(data[sidx+2] >>> 6) & 003 |
					(data[sidx+1] << 2) & 077];
        dest[didx++] = (byte)Base64EncMap[data[sidx+2] & 077];
      }
      if (sidx < data.length)
      {
        dest[didx++] = (byte)Base64EncMap[(data[sidx] >>> 2) & 077];
        if (sidx < data.length-1)
        {
          dest[didx++] = (byte)Base64EncMap[(data[sidx+1] >>> 4) & 017 |
					    (data[sidx] << 4) & 077];
          dest[didx++] = (byte)Base64EncMap[(data[sidx+1] << 2) & 077];
        }
	    else
        dest[didx++] = (byte)Base64EncMap[(data[sidx] << 4) & 077];
      }

      // add padding
      for ( ; didx < dest.length; didx++)
        dest[didx] = (byte)'=';

      return dest;
    }


    /**
     * This method decodes the given string using the base64-encoding
     * specified in RFC-1521 (Section 5.2).
     *
     * @param  str the base64-encoded string.
     * @return the decoded <var>str</var>.
     */
    public final static String base64Decode(String str)
    {
      if (str == null)  return  null;

      byte data[] = str.getBytes();

      return new String(base64Decode(data));
    }


    /**
     * This method decodes the given byte[] using the base64-encoding
     * specified in RFC-1521 (Section 5.2).
     *
     * @param  data the base64-encoded data.
     * @return the decoded <var>data</var>.
     */
    public final static byte[] base64Decode(byte[] data)
    {
      if (data == null)  return  null;

      int tail = data.length;
      while (data[tail-1] == '=')  tail--;

      byte dest[] = new byte[tail - data.length/4];


      // ascii printable to 0-63 conversion
      for (int idx = 0; idx <data.length; idx++)
        data[idx] = Base64DecMap[data[idx]];

      // 4-byte to 3-byte conversion
      int sidx, didx;
      for (sidx = 0, didx=0; didx < dest.length-2; sidx += 4, didx += 3)
      {
        dest[didx]   = (byte) ( ((data[sidx] << 2) & 255) |
			    ((data[sidx+1] >>> 4) & 003) );
        dest[didx+1] = (byte) ( ((data[sidx+1] << 4) & 255) |
			    ((data[sidx+2] >>> 2) & 017) );
        dest[didx+2] = (byte) ( ((data[sidx+2] << 6) & 255) |
			    (data[sidx+3] & 077) );
      }
      if (didx < dest.length)
        dest[didx]   = (byte) ( ((data[sidx] << 2) & 255) |
			    ((data[sidx+1] >>> 4) & 003) );
      if (++didx < dest.length)
        dest[didx]   = (byte) ( ((data[sidx+1] << 4) & 255) |
			    ((data[sidx+2] >>> 2) & 017) );

      return dest;
    }

    /**
     * UUENCODES the given byte array.  The first char
     * in the returned array is the length of the array offset by decimal 32
     *
     * @param   data    The byte array to be encoded.
     * @return          Returns the UUENCODED byte array.
     */
    public final static byte[] uuencode(byte[] data) {
      if ((data == null) || (data.length == 0)) {
        return null;
      }

      int p = data.length + (data.length/3) + 1;  // calculate the size of the destination array
      byte dest[] = new byte[p];                  // destination array will be 1/3 bigger plus one additional
                                                  // character representing the number of uuencoded bytes
//      char z = 15;
//      char y = 3;
      dest[0] = (byte)(data.length + 32);
      int offset = 1;

      for (int i=0; i<data.length; i+=3) {
        if ((i+2) < data.length) {      // ensures that we have 3 bytes to be encoded
          byte a = data[i];
          byte b = data[i+1];
          byte c = data[i+2];

          if ((offset+i) < dest.length)
            dest[offset + i] = (byte)(' ' + ((a >> 2) & 0x3F));

          if ((offset+i+1) < dest.length)
            dest[offset + i + 1] = (byte)(' ' + (((a << 4) | ((b >> 4) & 0x0F)) & 0x3F));

          if ((offset+i+2) < dest.length)
            dest[offset + i + 2] = (byte)(' ' + (((b << 2) | ((c >> 6) & 0x03)) & 0x3F));

          if ((offset+i+3) < dest.length)
            dest[offset + i + 3] = (byte)(' ' + ((c) & 0x3F));

          offset++;
        }
      }
      return dest;
    }

    /**
     * UUDECODES the given byte array. 
     * The first char
     * in the input array is the length of the array offset by decimal 32
     * @param   data    The byte array to be decoded.
     * @return          Returns the UUDECODED byte array.
     */
    public final static byte[] uudecode(byte[] data) {
      int j = data[0] - 32;             // calculate the size of the destination array
      byte dest[] = new byte[j];

      int offset = 1;

		for (int i=1; i<data.length; i+=4) {
			if (((i-offset) < dest.length) && ((i+1) < data.length))
			{
				int a = data[i] - ' ';
				int b = data[i+1] - ' ';
				dest[i-offset] = (byte)((a << 2) | (b >> 4));

				if (((i-offset+1) < dest.length) && ((i+2) < data.length))
				{
					int c = data[i+2] - ' ';
					dest[i-offset+1] = (byte)((b  << 4) | ((c >> 2) & 15));

					if (((i-offset+2) < dest.length) && ((i+3) < data.length)) {
						int d = data[i+3] - ' ';
						dest[i-offset+2] = (byte)(((c << 6) & 192) | d );
					}
				}
			}

        offset++;
      }
      return dest;
    }

    /**
     * Decodes a string that has been URL "%" encoded (see RFC 1738)
     */
    public final static String URLdecode(String s) {
      if (s == null) return null;
      int len = s.length();
      if (len < 3) return s; // if less than 3 chars then it cannot be url encoded
      StringBuffer buf = new StringBuffer();
      for (int i = 0; i < len; i++) {
        char c = s.charAt(i);
	// convert +'s  back into spaces
	if (c == '+')
            buf.append(' ');
	// '%' must be followed by 2 hex characters
        else if (c == '%' && i+2 < len) {
          try {
            int n = Integer.valueOf(s.substring(i+1, i+3), 16).intValue();
            i += 2;
            buf.append((char)n);
          } catch(NumberFormatException nfe) {
            buf.append(c);
          }
        }
        else buf.append(c);
      }
      return buf.toString();
    }

    /** Replace string with HTML entities for HTML output
     * replaces < with &lt;
     * replaces > with &gt;
     * etc.
     */
    public final static String HTMLEncode(String s) {
	StringBuffer buf = new StringBuffer(s.length() + 16);
	for (int i=0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (c == '<')
		buf.append("&lt;");
	    else if (c == '>')
		buf.append("&gt;");
	    else if (c == '&')
		buf.append("&amp;");
	    //else if (c == '"')
	    // buf.append("&quot;");
	    else
		buf.append(c);
	}
	return buf.toString();
    }

/**
 * Encode URL strings
 *
 * e.g. "foo bar.txt" -> "foo%20bar.txt"
 */
  public final static String urlPathEncode(String s)
  {
    /*
       // URL encode the string
       String tmpStr = java.net.URLEncoder.encode(s);
       // spaces are converted to +'s in CGI input but are valid as part of a
       // URL path in which case a '+' is treated as a '+' and not a space,
       // so must convert +'s to %20's.
     */
    StringBuffer buf = new StringBuffer(s.length());
    for(int i = 0; i < s.length(); i++)
    {
      char c = s.charAt(i);
      if(c == '_' || c == '.' || c == '-' || Character.isLetterOrDigit(c))
        buf.append(c);
      else
      {
        buf.append('%');
        char ch = Character.forDigit((c >> 4) & 0xF,16);
        // converting to use uppercase letter as part of
        // the hex value if ch is a letter.
        if(Character.isLetter(ch))
        {
          ch -= 32;
        // caseDiff 'a' -> 'A'
        }
        buf.append(ch);
        ch = Character.forDigit(c & 0xF,16);
        if(Character.isLetter(ch))
        {
          ch -= 32;
        }
        buf.append(ch);
      // buf.append(Integer.toHexString(c));
      }
    }
    // for
    return buf.toString();
  }

/** 
* Converts the latitude string into a numerical equivalent.
* Handles string as signed decimal degrees or DDD.MM.SSA format.
*
* Converts lat/lon DDD.MM.SSA (degrees/min/seconds) to double precison lat/lon values
* e.g. lat   34.10.32N  => 34.175555555555555
*      long 118.23.35W  => -118.39305555555556
*/
	public static double convertLatitudeStringToDegrees(String deg) throws NumberFormatException {
		return convertEarthCoordinateToDegrees(deg, true);
	}
/** 
* Converts the longitude string into a numerical equivalent.
* Handles string as signed decimal degrees or DDD.MM.SSA format.
*
* Converts lat/lon DDD.MM.SSA (degrees/min/seconds) to double precison lat/lon values
* e.g. lat   34.10.32N  => 34.175555555555555
*      long 118.23.35W  => -118.39305555555556
*/
	public static double convertLongitudeStringToDegrees(String deg) throws NumberFormatException {
		return convertEarthCoordinateToDegrees(deg, false);
	}
/** 
* Converts the latitude or longitude string into a numerical equivalent.
* Handles string as signed decimal degrees or DDD.MM.SSA format.
*
* Converts lat/lon DDD.MM.SSA (degrees/min/seconds) to double precison lat/lon values
* e.g. lat   34.10.32N  => 34.175555555555555
*      long 118.23.35W  => -118.39305555555556
*/
	public static double convertEarthCoordinateToDegrees(String deg, boolean isLatitude) throws NumberFormatException {
	
	  int degend = deg.indexOf('.');
	int minend = -1;
	if (degend > 0) {
		minend = deg.indexOf('.', degend+1);
	}
	
	// Check if value is already in degrees form; e.g. 40, 10.03, +70.123, -91
	if (minend < 0 || deg.startsWith("+") || deg.startsWith("-")) {
	  double degrees = Double.valueOf(deg).doubleValue();
	  double minDeg, maxDeg;
	  if (isLatitude) {
	    minDeg = -90;
	    maxDeg = +90;
	  } else {
	    minDeg = -180;
	    maxDeg = +180;
	  }
	  if (degrees < minDeg || degrees > maxDeg)
	    throw new NumberFormatException("degrees value not legal");
	  return degrees;
	}
	
	  char sign = Character.toUpperCase(deg.charAt(deg.length()-1));
	if (isLatitude) {
	  if (sign != 'N' && sign != 'S')
	    throw new NumberFormatException("sign is not legal: must be N or S");
	} else if (sign != 'E' && sign != 'W')
	  throw new NumberFormatException("sign is not legal: must be E or W");
	
	  int degs, mins, secs;
	degs = Integer.parseInt(deg.substring(0, degend));
	  if (degs < 0 || degs > 180)
	      throw new NumberFormatException("degrees decimal not legal");
	if (isLatitude && degs > 90)
	  throw new NumberFormatException("degrees decimal not legal for latitude");
	mins = Integer.parseInt(deg.substring(degend + 1, minend));
	  if (mins < 0 || mins > 59)
	      throw new NumberFormatException("minutes decimal not legal");
	secs = Integer.parseInt(deg.substring(minend + 1, deg.length() - 1));
	  if (secs < 0 || secs > 59)
	      throw new NumberFormatException("seconds decimal not legal");
	
	  double v = (double)degs + ((double)mins + (((double)secs) / 60)) / 60;
	
	if (sign == (isLatitude ? 'S': 'W')) return -v;
	
	return v;
	  }

}
