/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.StringTokenizer;
import java.util.Vector;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.StreamCopy;
import org.mitre.pam.interfaces.Reporter;

/**
 * This class allows the user to become an FTP client.
 *
 * This class has been written in compliance with RFC 959.
 * <pre>
 * The following commands have been implemented:
 *   - CWD   (String argument required)
 *   - DELE  (String argument required)
 *   - RMKD  (String argument required)
 *   - MKD   (String argument required)
 *   - RETR  (File argument required) (optional local file argument)
 *   - STOR  (File argument required) (optional local file argument)
 *   - LIST  (no argument)
 *   - PWD   (no argument)
 *   - QUIT  (no argument)
 *</pre>
 * The class is intended to be used in the following manner:
 *<br>
 *   - commands are sent using the method sendCommand which accepts two types of commands:
 *         (1) commands without arguments and
 *         (2) commands with arguments
 *<br>
 *   - the sendCommand method returns a String array which contains the lines of response returned from the
 *     server
 *<br>
 *   - the getData method will return a String array containing any data that was returned from the server and is
 *     not part of the response
 *<pre>
 *         ex:
 *               FTP myFTP = new FTP(host, user, password);
 *               String[] responses = null;
 *               String[] data = null;
 *               myFTP.connect();
 *               responses = myFTP.sendCommand(FTP.CWD, "/pub/java/jdk1.1.4");
 *               if(responses[0].charAt(0) == '2') {
 *                   responses = myFTP.sendCommand(FTP.LIST);
 *                   if(responses[0].charAt(0) == '2') {
 *                       data = myFTP.getData();
 *                   }
 *               }
 *</pre>
 *
 * @author   John Fantry, Technical Co-op (D510)
 * @author   Jason Mathews, many changes (D510)
 */
public class FTP 
{
	protected final static int DEFAULT_DATA_PORT = 20;

	// as defined in RFC 959
	protected final static int DEFAULT_CONTROL_PORT = 21;

	// as defined in RFC 959
	public final static String CWD = "CWD";

	// commands defined in RFC 959
	public final static String LIST = "LIST";
	public final static String MKD = "MKD";
	public final static String PWD = "PWD";
	public final static String QUIT = "QUIT";
	public final static String RETR = "RETR";
	public final static String STOR = "STOR";
	public final static String DELE = "DELE";
	public final static String RMKD = "RMKD";

	// pseudo FTP command implemented locally
	private static final String DEBUG_KEY = "FTP";

	// keyword for internal debugging
	protected BufferedReader serverInputStream = null;

	// this is the server command response stream
	protected BufferedWriter serverOutputStream = null;

	// commands are written to this stream
	protected String host;
	protected String user;
	protected String pass;
	protected int port;
	protected Socket server = null;
	protected String response = new String();
	protected String[] responses = null;
	protected String[] data = null;
	protected boolean connected = false;
	protected Reporter reporter;

	public FTP(String host, String user, String pass, Reporter reporter)
	{
		this.host = host;
		this.user = user;
		this.pass = pass;
		port = DEFAULT_CONTROL_PORT;
		// uses default control port
		this.reporter = reporter;
	}

	public FTP(String host, int port, String user, String pass, Reporter reporter)
	{
		this.host = host;
		this.port = port;
		this.user = user;
		this.pass = pass;
		this.reporter = reporter;
	}

	/**
	 * Attempts to establish a connection with an FTP server and complete the
	 * login sequence.
	 *
	 * @exception    IOException
	 *               Thrown if there is a problem communicating with the server
	 * @exception    UnknownHostException
	 *               Thrown if the host is unknown
	 */
	public void connect() throws IOException, UnknownHostException
	{
		if(connected)
			return;
		// already connected
		// Flag to close socket and streams if any exception is thrown while
		// establishing a connection
		boolean allowConnection = false;
		server = new Socket(host,port);
		try
		{
			serverInputStream = new BufferedReader(new InputStreamReader(server.getInputStream()));
			serverOutputStream = new BufferedWriter(new OutputStreamWriter(server.getOutputStream()));
			connected = true;
			reply();
			if(response.charAt(0) != '2')
			{
				throw new IOException("Error establishing connection with host " + host);
			}
			if(user != null)
			{
				send("USER " + user);
				if((response.charAt(0) == '4') || (response.charAt(0) == '5'))
				{
					throw new IOException(responses[0]);
				}
				if(response.charAt(0) == '3')
				{
					if(pass == null)
					{
						throw new IOException(responses[0]);
					}
					send("PASS " + pass);
					if(response.charAt(0) != '2')
					{
						throw new IOException(responses[0]);
					}
				}
			}
			allowConnection = true;
		// connection opened with no errors
		}
		finally
		{
			if(!allowConnection)
				closeResources();
		}
	}

	/**
	 * Closes the connection with the server by sending QUIT.
	 *
	 * @exception    IOException
	 *               Thrown if there is a problem communicating with the server
	 */
	public void close() throws IOException
	{
		try
		{
			if(connected)
			{
				send("QUIT");
				if(response.charAt(0) != '2')
					throw new IOException(responses[0]);
			}
		}
		finally
		{
			closeResources();
		}
	}

	/**
	 * Internal method to close all open resources (socket and i/o streams)
	 */
	private void closeResources()
	{
		if(server != null)
		{
			if(serverOutputStream != null)
			{
				try
				{
					serverOutputStream.close();
				}
				catch(Exception ignore)
				{
				}
			}
			if(serverInputStream != null)
			{
				try
				{
					serverInputStream.close();
				}
				catch(Exception ignore)
				{
				}
			}
			try
			{
				server.close();
			}
			catch(Exception ignore)
			{
			}
		}
		connected = false;
	}

	/**
	 * Reads the response from the server.
	 *
	 * @exception    IOException
	 *               Thrown if there is a problem communicating with the server
	 */
	protected void reply() throws IOException
	{
		if(!connected)
		{
			throw new IOException("Not connected");
		}
		Vector replies = new Vector();
		try
		{
			response = serverInputStream.readLine();
			reporter.debugMsg("FTP ",response,null,null,DEBUG_KEY);
			if(response.charAt(3) == '-')
			{
				// this implies a multi-line response
				String tag = response.substring(0,2);
				replies.addElement(response);
				while(!(response = serverInputStream.readLine()).startsWith(tag))
				{
					reporter.debugMsg("FTP ",response,null,null,DEBUG_KEY);
					replies.addElement(response);
				}
				replies.addElement(response);
			}
		}
		finally
		{
			if(replies.size() == 0)
			{
				responses = new String[1];
				responses[0] = response;
			}
			else
			{
				responses = new String[replies.size()];
				for(int i = 0;i < responses.length;i++)
				{
					responses[i] = (String)replies.elementAt(i);
				}
			}
		}
	}

	/**
	 * Used internally to send the port settings to the server.
	 */
	private void sendPort(ServerSocket serverSocket) throws IOException
	{
		int localPort = serverSocket.getLocalPort();
		//	InetAddress inetAddress = serverSocket.getInetAddress();
		InetAddress myAddress;
		try
		{
			URL url = new URL(ExecGlobal.getProtocolHostNoPort());
			myAddress = InetAddress.getByName(url.getHost());
		// catch UnknownHostException + MalformedURLException
		}
		catch(Exception e)
		{
			reporter.warning("FTP " + e.toString());
			myAddress = InetAddress.getLocalHost();
		}
		// System.out.println("inet address="+myAddress.getHostAddress());
		byte[] addrbytes = myAddress.getAddress();
		short addrshorts[] = new short[4];
		for(int i = 0;i <= 3;i++)
		{
			addrshorts[i] = addrbytes[i];
			if(addrshorts[i] < 0)
				addrshorts[i] += 256;
		}
		String commandSend = "port " + addrshorts[0] + "," + addrshorts[1] + "," + addrshorts[2] + "," + addrshorts[3] + "," + ((localPort & 0xff00) >> 8) + "," + (localPort & 0x00ff);
		send(commandSend);
		if(responses[0].charAt(0) != '2')
		{
			throw new IOException("Error creating data connection with host " + host + " " + responses[0] + " in response to " + commandSend);
		}
	}

	/**
	 * Used internally to send concatenated commands (CWD + arg).
	 *
	 * @exception    IOException
	 *               Thrown if there is a problem communicating with the server
	 */
	protected void send(String command) throws IOException
	{
		if(!connected)
		{
			throw new IOException("Not connected");
		}
		//System.out.println("FTP send: " + command);
		serverOutputStream.write(command);
		serverOutputStream.newLine();
		serverOutputStream.flush();
		reply();
		if(responses[0].startsWith("425"))
			// STOR | RETR | LIST
			sendDataCommand(command);
	}

	/**
	 * Used internally to send all commands to the server that will involve the data port (LIST).
	 *
	 * @exception    IOException
	 *               Thrown if there is a problem communicating with the server
	 */
	protected void sendDataCommand(String command) throws IOException
	{
		ServerSocket serverSocket = null;
		Socket socket = null;
		try
		{
			Vector dataVector = new Vector();
			serverSocket = new ServerSocket(0);
			sendPort(serverSocket);
			//System.out.println("FTP send data: " + command);
			serverOutputStream.write(command);
			serverOutputStream.newLine();
			serverOutputStream.flush();
			reply();
			if(responses[0].charAt(0) != '1')
			{
				throw new IOException(responses[0]);
			}
			socket = serverSocket.accept();
			BufferedReader br = new BufferedReader(new InputStreamReader(socket.getInputStream()));
			String s;
			while((s = br.readLine()) != null)
				dataVector.addElement(s);
			data = new String[dataVector.size()];
			for(int i = 0;i < data.length;i++)
				data[i] = (String)dataVector.elementAt(i);
			br.close();
			reply();
		}
		finally
		{
			if(socket != null)
			{
				try
				{
					socket.close();
				}
				catch(Exception ioe)
				{
				}
			}
			if(serverSocket != null)
			{
				try
				{
					serverSocket.close();
				}
				catch(Exception ioe)
				{
				}
			}
		}
	}

	/**
	 * Used internally to send all commands to the server that involve files (RETR, STOR).
	 *
	 * @param  command   The command to send to the server
	 * @param  file      The file that the method will work with
	 *
	 * @exception  IOException
	 *             Thrown if there is a problem communicating with the server
	 */
	protected void sendFileCommand(String command, File rmtFile, File localFile) throws IOException
	{
		ServerSocket serverSocket = null;
		Socket socket = null;
		try
		{
			serverSocket = new ServerSocket(0);
			sendPort(serverSocket);
			serverOutputStream.write("TYPE i");
			serverOutputStream.newLine();
			serverOutputStream.flush();
			reply();
			if(responses[0].charAt(0) != '2')
			{
				throw new IOException(responses[0]);
			}
			// System.out.println("FTP send file: " + command + ' ' + file.getName());
			serverOutputStream.write(command + ' ' + rmtFile.getName());
			serverOutputStream.newLine();
			serverOutputStream.flush();
			reply();
			if(responses[0].charAt(0) != '1')
			{
				throw new IOException(responses[0]);
			}
			socket = serverSocket.accept();
			if(command.startsWith("RETR"))
			{
				InputStream in = socket.getInputStream();
				FileOutputStream os = new FileOutputStream(localFile);
				StreamCopy.copyFully(in,os);
				reply();
			}
			else
				if(command.startsWith(STOR))
				{
					OutputStream out = socket.getOutputStream();
					BufferedOutputStream bout = new BufferedOutputStream(out);
					FileInputStream is = new FileInputStream(localFile);
					StreamCopy.copyFully(is,bout);
					reply();
				}
				else
				{
					throw new IllegalArgumentException("Unrecognized command: " + command);
				}
		}
		finally
		{
			if(socket != null)
			{
				try
				{
					socket.close();
				}
				catch(IOException ioe)
				{
				}
			}
			if(serverSocket != null)
			{
				try
				{
					serverSocket.close();
				}
				catch(IOException ioe)
				{
				}
			}
		}
	}

	/**
	 * Sends a command without argument to the server.
	 *
	 * @param  command   The command to send to the server
	 *
	 * @exception  IOException
	 *             Thrown if there is a problem communicating with the server
	 * @exception  IllegalArgumentException
	 *             Thrown if this method is called with a command that requires an argument
	 *
	 * @return   A String array containing the lines of response recevied from the server
	 */
	public String[] sendCommand(String command) throws IOException
	{
		if(!connected)
		{
			throw new IOException("Not connected");
		}
		responses = null;
		data = null;
		if(command.startsWith(LIST))
		{
			sendDataCommand(command);
			return responses;
		}
		else
			if(command.startsWith(RETR) || command.startsWith(STOR))
			{
				throw new IllegalArgumentException("Missing argument: " + command + " requires a File object as a secong argument");
			}
			else
			{
				send(command);
				return responses;
			}
	}

	/**
	 * Sends a command with argument to the server.
	 *
	 * @param  command   The command to send to the server
	 * @param  arg       The argument for the command
	 *
	 * @exception  IOException
	 *             Thrown if there is a problem communicating with the server
	 * @exception  IllegalArgumentException
	 *             Thrown if this method is called with a command that does not require an argument or
	 *             if the argument is not the correct type for the associated command
	 *
	 * @return   A String array containing the lines of response recevied from the server
	 */
	public String[] sendCommand(String command, Object arg) throws IOException
	{
		if(!connected)
		{
			throw new IOException("Not connected");
		}
		responses = null;
		data = null;
		if(command.startsWith(RETR) || command.startsWith(STOR))
		{
			if(!(arg instanceof File))
				throw new IllegalArgumentException("Command argument must be a file: " + arg);
			sendFileCommand(command,(File)arg,(File)arg);
			return responses;
		}
		else
			if(command.startsWith(CWD) || command.startsWith(MKD) || command.startsWith(DELE))
			{
				if(!(arg instanceof String))
					throw new IllegalArgumentException("Command argument must be a String: " + arg);
				send(command + ' ' + (String)arg);
				return responses;
			}
			else
				if(command.startsWith(RMKD))
				{
					if(!(arg instanceof String))
						throw new IllegalArgumentException("Command argument must be a String: " + arg);
					processRMKD((String)arg);
					return responses;
				}
				else
				{
					throw new IllegalArgumentException("Unrecognized command: " + command);
				}
	}

	/**
	 * Sends a command with file arguments to the server.
	 *
	 * @param  command    The command to send to the server
	 * @param  remoteFile The file on server to read/write
	 * @param  localFile  The file on local file system to read/write
	 *
	 * @exception  IOException
	 *             Thrown if there is a problem communicating with the server
	 * @exception  IllegalArgumentException
	 *             Thrown if this method is called with a command that does not require an argument or
	 *             if the argument is not the correct type for the associated command
	 *
	 * @return   A String array containing the lines of response recevied from the server
	 */
	public String[] sendCommand(String command, File remoteFile, File localFile) throws IOException
	{
		if(!connected)
		{
			throw new IOException("Not connected");
		}
		if(!command.startsWith(RETR) && !command.startsWith(STOR))
		{
			throw new IllegalArgumentException("Unrecognized command: " + command);
		}
		responses = null;
		data = null;
		sendFileCommand(command,remoteFile,localFile);
		return responses;
	}

	/**
	 * This function is responsible for traversing a path and creating any directories
	 * that do not already exist.
	 *
	 * @param path  This is the directory structure that will be created of it does not exist.
	 *
	 * @throws IOException  Thrown if there is an error communicating with the host.
	 */
	protected void processRMKD(String path) throws IOException
	{
		// System.out.println("FTP:processRMKD path="+path);//debug
		StringTokenizer st = new StringTokenizer(path,"/");
		if(!st.hasMoreElements())
		{
			reporter.assert(false,"processRMKD received empty path to create: ",path,null,null);
			return;
		}
		String dir = (String)st.nextElement();
		String responses[] = sendCommand(CWD,"/" + dir);
		while(responses[0].startsWith("2") && st.hasMoreElements())
		{
			dir = (String)st.nextElement();
			responses = sendCommand(CWD,dir);
		}
		if(!responses[0].startsWith("2"))
		{
			sendCommand(MKD,dir);
			responses = sendCommand(CWD,dir);
			if(!responses[0].startsWith("2"))
				throw new IOException("Can't create directory " + dir + " in FTP path " + path);
			while(st.hasMoreElements())
			{
				dir = (String)st.nextElement();
				sendCommand(MKD,dir);
				responses = sendCommand(CWD,dir);
				if(!responses[0].startsWith("2"))
					throw new IOException("Can't create directory " + dir + " in FTP path " + path);
			}
		}
	}

	/**
	 * @return   returns the line(s) of data received from the server
	 */
	public String[] getData()
	{
		return data;
	}

	/**
	 * @return   returns the line(s) of response received from the server
	 */
	public String[] getResponses()
	{
		return responses;
	}

	/**
	 * @return   returns the host
	 */
	public String getHost()
	{
		return host;
	}

	/**
	 * @return   returns the boolean connected
	 */
	public boolean isConnected()
	{
		return connected;
	}

	/**
	 * Closes the connection with the host if it is still open.
	 */
	protected void finalize()
	{
		try
		{
			close();
		}
		catch(IOException ioe)
		{
		}
	}

/*
   public static void main (String args[]) {
   new org.mitre.pam.executive.Exec();
   FTP myFTP = null;
   try {
   myFTP = new FTP("blackbird.mitre.org", "anonymous", "guest@");
   String[] responses = null;
   String[] data = null;
   myFTP.connect();
   responses = myFTP.sendCommand(FTP.CWD, "/incoming/test");
   if (responses[0].charAt(0) == '2') {
   
   // try to create a test file and send it across
   try {
   File file;
   if (args.length != 0) {
   file = new File(args[0]);
   } else {
   file = new File("foo.txt");
   java.io.PrintWriter out = new java.io.PrintWriter(new java.io.FileWriter(file), true);
   out.println(new java.util.Date());
   out.close();
   }
   
   System.out.println("Source file=" + file.getName() + ",size=" + file.length() + " bytes");
   
   responses = myFTP.sendCommand(FTP.STOR, file);
   
   // get file list
   responses = myFTP.sendCommand(FTP.LIST);
   if (responses[0].charAt(0) == '2') {
   data = myFTP.getData();
   for (int i=0; i < data.length; i++)
   System.out.println(data[i]);
   }
   
   // now try to retrieve it
   myFTP.sendCommand(FTP.RETR, file, new File("foo.copy"));
   
   // now delete the file
   myFTP.sendCommand(FTP.DELE, file.getName());
   
   } catch (IOException e) {
   System.out.println(e);
   }
   
   // get file list
   
   responses = myFTP.sendCommand(FTP.LIST);
   if (responses[0].charAt(0) == '2') {
   data = myFTP.getData();
   for (int i=0; i < data.length; i++)
   System.out.println(data[i]);
   }
   }
   
   } catch (IOException ioe) {
   ioe.printStackTrace();
   } finally {
   if (myFTP != null)
   try {
   myFTP.close();
   } catch (IOException ignore) {}
   }
   System.exit(0);
   }
 */
}

