/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util;

import java.io.*;
import java.io.StreamTokenizer;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ExecToController;
import org.mitre.pam.executive.ExecGlobal;

/**
* Copy the contents of one stream to another.  
*
* @author Nazario Irizarry and others
*/

public class StreamCopy {

  private static final int BUF_SIZE = 4096; //size of buffer, in bytes

/**
* Copy from intput stream to output stream until end of input stream.  
* Closes input and output stream when done.
*/
  public static void copyFully(InputStream in, OutputStream out) throws IOException {
    copyFully(in,out,true);
  }

/**
* Gives user ability to not close (close = false)
*/
  public static void copyFully(InputStream in, OutputStream out, boolean close) throws IOException {
    try {
      BufferedInputStream bin = new BufferedInputStream(in, BUF_SIZE);
    byte[] buf = new byte[BUF_SIZE];
    int bytesRead = 0;
      while (true) {
      bytesRead = bin.read(buf);
	  if (bytesRead <= 0) break;
	  out.write(buf, 0, bytesRead);
      }
    out.flush();
    } finally {
    if (close) {
	    try {
      out.close();
	    } catch (IOException ignore) {}
	    try {
		in.close();
	    } catch (IOException ignore) {}
	}
    }//finally
  }

/**
* Copy one file to another.
* This code should use the copyFully() method
*/
	public static void copyFile(File from, File to) throws IOException
	{
		int bufSize = 8192;
		if (!from.exists()) {
			throw new IOException("Attempt to copy nonexisting file: "+from.getAbsolutePath());
		}
		FileInputStream fin = null;
		FileOutputStream fout = null;
		try
		{
			fin = new FileInputStream(from);
			fout = new FileOutputStream(to);
			int bytesRead;
			byte[] buf = new byte[bufSize];
			do
			{
				bytesRead = fin.read(buf,0,bufSize);
				if(bytesRead > 0)
					fout.write(buf,0,bytesRead);
			}
			while(bytesRead > 0);
		}
		finally
		{
			if(fout != null)
				try
				{
					fout.close();
				}
				catch(Exception ignore)
				{
				}
			if(fin != null)
				try
				{
					fin.close();
				}
				catch(Exception ignore)
				{
				}
		}
	}

/**
* Get the first n tokens from the specified file.
* Designed so plugins could easily check to see
* the current "version" of their support directory
* by checking a few tokens in a well known file
* in their support directory.
*
* If the file does not exist none of the returnTokens are filled in.
* Uses java.io.StreamTokenizer.
*
* @param file to read
* @param returnTokens array of the first n tokens to return 
*/
	public static void getFirstTokensFromFile(File file, String[] returnTokens) {
		if (!file.exists()) {
			return;
		}
		FileReader reader = null;
		try {
			for (int i = 0; i<returnTokens.length; i++) {
				reader = new FileReader(file);
				StreamTokenizer st = new StreamTokenizer(reader);
				st.resetSyntax();
				st.wordChars('0','9');
				st.nextToken();
				returnTokens[i] = st.sval;
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		if (reader != null)
			try { reader.close(); } catch (IOException ignore){};
	}
/**
* Helps plugins create the necessary support file structure
* by copying an entire support directory from inside the .zip or .jar file.
* 
* @param exec a reference to the exec
* @param whichDir the name of the subdirectory under 0supportDir
* @param whichVersion a string which must match exactly with the first token in the file by versionFileName
* @param versionFileName a text file located in whichDir that contains the version of the support files in that subdirectory
* #param zipFile the zip file from which to extract support files should they need to be copied initially or on an update
*/
	public static void makeASupportDir(ExecToController exec, String whichDir, String whichVersion, String versionFileName, File zipFile){
		Reporter reporter = exec.getExecAsAgent().getReporter();
		File versionFile = new File(ExecGlobal.getSupportDir(whichDir), versionFileName);
		String[] token = new String[1];
		StreamCopy.getFirstTokensFromFile(versionFile, token);
		boolean needToCopy = (token[0] == null) || (!token[0].equals(whichVersion));
		if (needToCopy) {
			try {
				if (token[0] == null)
					reporter.log("Creating support directory "+whichDir);
				else
					reporter.log("Recreating support directory "+whichDir);
				StreamCopy.copySupportDirFromZipFile(whichDir, zipFile, reporter);
			}
			catch (Exception e) {
				reporter.reportException(e);
			}
		}
	}
/**
* Helps plugins created the necessary support file structure
* by copying an entire support directory from inside the .zip or .jar file
*/
	public static void copySupportDirFromZipFile(String supportSubdir, File zipFileExisting, Reporter reporter) throws IOException {
		
		if (!zipFileExisting.exists())
			throw new IOException("Required support file does not exist: "+zipFileExisting.getName());
		if (!zipFileExisting.isFile())
			throw new IOException("Specified support filename is not a file: "+zipFileExisting.getName());
			
		ZipFile zipFile;
		String expandedSubDirName = "0supportDir/"+supportSubdir;
		File supportDir = ExecGlobal.getSupportDir(supportSubdir);
		zipFile = new ZipFile(zipFileExisting);
		Enumeration entries = zipFile.entries();
		while (entries.hasMoreElements()) {
			ZipEntry entry = (ZipEntry) entries.nextElement();
			if (!entry.isDirectory()) {
				String name = entry.getName();
				if (name.startsWith(expandedSubDirName)) {
					copyEntryToSupportDir(zipFile, entry, supportSubdir, reporter);
				}
			}
		}
	}
	
	private static void copyEntryToSupportDir(ZipFile zipFile, ZipEntry entry, String supportSubdir, Reporter reporter) throws IOException {
		File subdir = ExecGlobal.getSupportDir();
		String justTheSub = entry.getName().substring("0supportDir/".length());
		if (File.separatorChar != '/') {
			justTheSub = justTheSub.replace('/',File.separatorChar);
		}
		int subdirIndex = justTheSub.lastIndexOf(File.separatorChar);
		if (subdirIndex >= 0) {
			File newSubdir = new File(subdir, justTheSub.substring(0,subdirIndex));
//			System.out.println("make dir "+newSubdir.getAbsolutePath());
			if (!newSubdir.exists())
				newSubdir.mkdirs();
		}
		File newFile = new File(subdir, justTheSub);

//		System.out.println(" --> copy to: "+newFile.getAbsolutePath());

		reporter.log("Creating support file "+justTheSub);
		StreamCopy.copyFully(zipFile.getInputStream(entry), new FileOutputStream(newFile));
	}

}
