package org.mitre.pam.util;

/**
 * URL Helper class
 * 
 * Creates url connections of the specified method type: GET, POST, HEAD, etc.
 * which is subject to protocol restrictions.
 * <P>
 * If POST method is used then the parameters after the "?" in the URL are stripped off
 * of the URL and passed via the input stream of the live connection. These parameters must
 * be encoded according to RFCs 1630 (URI) and 1945 (HTTP).
 * <P>
 * For example, the url "http://foo.com/test.cgi?x=1&y=2" passed as a POST request will create
 * a URL as "http://foo.com/test.cgi" and send as input "x=1&y=2".
 * 
 * @author Jason Mathews 12/01/99
 */
import java.net.*;
import java.io.*;
import java.util.Properties;
import java.util.Enumeration;

public final class URLHelper{
    private URL		       url;
    private String	       inputString;
    private String	       method;
    private Properties	       headers;
    public final static String GET = "GET";
    public final static String POST = "POST";
    public final static String HEAD = "HEAD";

    /**
     * Creates a URLHelper object with the default GET method
     * 
     * @param urlString   the String to parse as a URL.
     */
    public URLHelper(String urlString) throws MalformedURLException{
	this(urlString, GET);
    }

    /**
     * Creates a URLHelper object
     * 
     * @param urlString   the String to parse as a URL.
     * 
     * @param postMethod  request method, default="GET"
     */
    public URLHelper(String urlString, 
		     String method) throws MalformedURLException{
	if (method == null || method.length() == 0){
	    method = GET;// get by default
	}
	else if (method.equalsIgnoreCase(POST)){
	    method = POST;

	    int ind = urlString.indexOf('?');

	    if (ind != -1){
		inputString = urlString.substring(ind + 1);
		urlString = urlString.substring(0, ind);
	    }
	}

	this.method = method;
	url = new URL(urlString);
    }

    public URL getURL(){
	return url;
    }

    public String getMethod(){
	return method;
    }

    public String getInputString(){
	return inputString;
    }

    public Properties getRequestProperties(){
	return headers;
    }

    /**
     * Sets the general request property.
     * 
     * @param   key     the keyword by which the request is known
     * (e.g., "<code>accept</code>").
     * @param   value   the value associated with it.
     */
    public void setRequestProperty(String key, String value){
	if (key != null){
	    if (headers == null){
		headers = new Properties();
	    }

	    headers.put(key, value);
	}
    }

    /**
     * Creates url connections of the specified method type: GET, POST, HEAD, etc.
     * which is subject to protocol restrictions.
     * 
     */
    public URLConnection openConnection() throws IOException{
	URLConnection conn = url.openConnection();

	if (!method.equalsIgnoreCase(GET) 
		&& conn instanceof HttpURLConnection){
	    ((HttpURLConnection) conn).setRequestMethod(method);

	    // set the header info
	}

	if (headers != null){
	    Enumeration e = headers.keys();

	    while (e.hasMoreElements()){
		String key = (String) e.nextElement();
		String value = (String) headers.getProperty(key);

		conn.setRequestProperty(key, value);
	    }
	}

	if (method.equals(POST)){
	    if (inputString != null && inputString.length() != 0){
		conn.setDoOutput(true);

		PrintWriter pw = new PrintWriter(conn.getOutputStream());

		pw.print(inputString);
		pw.close();
	    }
	}

	return conn;
    }

    /*
     * public static void main(String argv[]) {
     * 
     * try {
     * String method = URLHelper.GET;
     * String urlString = "http://127.0.0.1/servlet/SnoopServlet?x=foo&y=bar";
     * String inputString = null;
     * 
     * for (int i=0; i < argv.length; i++)
     * if (argv[i].startsWith("-p"))
     * method = URLHelper.POST;
     * else if (argv[i].startsWith("-m"))
     * method = argv[++i];
     * else urlString = argv[i];
     * 
     * System.err.println(method + " " + urlString);
     * 
     * URLHelper url = new URLHelper(urlString, method);
     * 
     * URLConnection c = url.openConnection();
     * 
     * BufferedReader in = new BufferedReader(
     * new InputStreamReader(c.getInputStream()));
     * String line;
     * while ((line=in.readLine()) != null)
     * System.out.println(line);
     * } catch(Exception e) {
     * e.printStackTrace();
     * }
     * }
     */
}
