/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.alarm;

import org.mitre.pam.interfaces.Reporter;

/**
* A generic reporting class that may be created to report using
* a specific "context" string
*
* @author Tim Taylor
*/
public class AlarmReporter implements Reporter {
	String contextName;
	private volatile int debugKeyActivityNumber;
	private volatile String lastDebugKey;
	private volatile boolean lastDebugState;

	public AlarmReporter(String contextName) {
		if (contextName == null || contextName.length() == 0)
			contextName = "<>";
		this.contextName = contextName;
	}
	
/**
* Log events.  These are normal and are logged for information.
* @param msg is the message
*/
	public void log (String msg){
		Alarm.log(contextName, msg);
	}
/**
* Log an error message if an assertion fails.
* These are programming assumptions. If these occur
* the software support facility should be notified. Message parts are only concatenated
* if the assertion is false.
*
* @param cond expected to be true normally
* @param msg1 the message - part 1
* @param msg2 the message - part 2 - may be null
* @param msg3 the message - part 3 - may be null
* @param msg4 the message - part 4 - may be null
*/
	public void assert(boolean cond, Object msg1, Object msg2, Object msg3, Object msg4) {
		if (!cond) {
			Alarm.assert(contextName, msg1, msg2, msg3, msg4);
		}
	}
/**
* Report an exception in the context of the Agent.  An Exception is something
* that is not expected to occur and the stream of control
* should be aborted.  Unfortuneately the actual type is Throwable because
* some parts of JavaCC throw Errors instead of Exceptions
* <p>
* @param e the Exception or Throwable
*/
	public void reportException(Throwable e){
		Alarm.reportException(contextName, e);
	}
/**
* Issue a warning message in the context of the Agent .  A warning is something that somehow
* should not be but for which there is a recovery strategy, and
* operation may continue.
* @param msg the verbose message
*/
	public void warning (String msg) {
		Alarm.warn(contextName, msg);
	}

/**
* Issue an error message in the context of the Agent .  An error is something for
* which there is not recovery strategy and execution can not continue.
* However it is not a programming error.  It might be a protocol violation, for example.
* @param msg the  message
*/
	public void error (String msg){
		Alarm.error(contextName, msg);
	}

/**
* Report a verbose debugging message.
* @param msg1 the message - part 1
* @param msg2 the message - part 2 - may be null
* @param msg3 the message - part 3 - may be null
* @param msg4 the message - part 4 - may be null
* @param category is the category flag that may be used to 
* disable/enable entire sections of debug messages
*/
	public void debugMsg(Object msg1, Object msg2, Object msg3, Object msg4, String category){
		if (isDebugKey(category)) {
			if (category == null || category.length() == 0)
				category = "<>";
			Alarm.debugMsg(msg1, msg2, msg3, msg4, category);
		}
	}
/**
* Tell whether a certain key has been identified as a debug flag
* @return true if key is a debug flag that is set
*/
	final public boolean isDebugKey(String key){
		if ((Alarm.debugKeyActivityNumber != debugKeyActivityNumber) ||
				(key != lastDebugKey)) {
			synchronized (this) {
				debugKeyActivityNumber = Alarm.debugKeyActivityNumber;
				lastDebugKey = key;
				lastDebugState = Alarm.isDebugKey(key);
			}
		}
		return lastDebugState;
	}
	
}