/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.draw;

import java.awt.*;
import java.util.Vector;

import org.mitre.pam.interfaces.Agent;

public class Block {

	boolean hasInputs;
	boolean hasOutputs;
	String kind; // intentional package access only
	String name; // intentional package access only
	private String userSpecifiedInfoString;
	private String internalBuiltInfoString;
	private String userSpecifiedDisplayName;
		
	private boolean drawSelected;
	private static final int NOSE_DEPTH = 10; // half of the height; results in 45 degree angle
	protected Rectangle boundingRect = new Rectangle(50, 50, 50, NOSE_DEPTH*2+1);
	protected Rectangle outRect = new Rectangle();
	protected Rectangle inRect = new Rectangle();
	protected Rectangle bandArea = new Rectangle();
	private int[] xPoints, yPoints;
	protected DrawParent parent;
	int moveToX; // intentional package access only
	int moveToY; // intentional package access only
/*
	protected static Font enabledFont = new Font("Helvetica", Font.BOLD, 11);
	protected static Font disabledFont = new Font("Helvetica", Font.PLAIN, 10);
	protected static Color enabledColor = Color.black;
	protected static Color disabledColor = new Color((float)0.70,(float)0.48,(float)0.19); // brown;
*/
	protected Color displayFontColor;
	protected Font fontForName;
	protected Color outlineColor;
	
	protected Vector sources; 
	protected Vector sinks; 

	private int textLocOffestX;
	private int textLocOffestY;
	private static BlockGroup defaultBlockGroup = new BlockGroup();
	public BlockGroup blockGroup;
	
	int deltaXtoMark, deltaYtoMark;
	
	public Block() {
		blockGroup = defaultBlockGroup;
		hasInputs = true;
		hasOutputs = true;
		displayFontColor = blockGroup.enabledColor;
		fontForName = blockGroup.enabledFont;
		outlineColor = Color.black;
		sources = new Vector(); 
		sinks = new Vector(); 
	}
/**
* This should be called right after the block is created
* because we assume the block is close to its newly constructed
* state and not yet being displayed
*/
	public void setBlockGroup(BlockGroup bg) {
		this.blockGroup = bg;
		displayFontColor = blockGroup.enabledColor;
		fontForName = blockGroup.enabledFont;
	}
	
	void setParent(DrawParent parent) {
		this.parent = parent;
	}
	
	public void setHasInputs(boolean hasInputs) {
		this.hasInputs = hasInputs;
	}

	public void setHasOutputs(boolean hasOutputs) {
		this.hasOutputs = hasOutputs;
	}
	
	synchronized void select() {
		if (drawSelected)
			return;
		
		drawSelected = true;
		refresh();
	}

	synchronized void unselect() {
		if (!drawSelected)
			return;
		
		drawSelected = false;
		refresh();
	}
	
	public Point getLocation() {
		return new Point(boundingRect.x, boundingRect.y);
	}

	final protected void refresh() {
		if (parent != null)
			parent.needsRefresh(boundingRect);
	}
/**
* left of box
*/
	public int getX() {
		return boundingRect.x;
	}

/**
* right of box
*/
	public int getY() {
		return boundingRect.y;
	}
	

	protected void setOnce(int x, int y, String name) {
		String oldName = this.name;
		this.name = name;
		internalBuiltInfoString = null;
		boundingRect.x = x;
		boundingRect.y = y;
		if (parent != null) {
			refresh();
			adjustSizeForName();
			refresh();
			if (oldName != null && (oldName != name) && parent.blocks.contains(this)) {
//			System.out.println("rename from "+oldName+" to "+name);
				parent.blocks.remove(oldName);
				parent.blocks.put(name, this); // hash is keyed on name
			}
		}
		else {
			updateRectPos(x, y);
		}
	}
/**
* The name must be unique across all blocks in the drawParent.
* By default this name is displayed.  However the display name
* may be changed with the setDisplayName() method.
*/
	public void setName(String name) {
		setOnce(boundingRect.x, boundingRect.y, name);
	}
		
	public void adjustSizeForName() {
		if (parent == null)
			return;  // we will adjust later
		int stringW = parent.getFontMetrics(fontForName).stringWidth(getDisplayName());
		boundingRect.width = stringW + NOSE_DEPTH*2;
		updateRectPos(boundingRect.x, boundingRect.y); // update nubs
		textLocOffestX = (boundingRect.width - stringW)/2;
		textLocOffestY = (boundingRect.height + fontForName.getSize())/2;
		refresh();
	}

	private void updateRectPos(int x, int y) {
		/*
		* boundingRect outlines the drawable area
		*/
		boundingRect.x = x;
		boundingRect.y = y;
		/*
		* bandArea is the hotspot area for starting a rubberband
		* (output connection to another block)
		*/
		bandArea.x = boundingRect.x + boundingRect.width - NOSE_DEPTH-3;
		bandArea.y = boundingRect.y;
		bandArea.width = NOSE_DEPTH+6; // 3 to the right (6-3)
		bandArea.height = boundingRect.height;
		/*
		* inRect is the input rectangle connector box
		* (on the left)
		*/
		inRect.x = boundingRect.x;
		inRect.y = boundingRect.y + boundingRect.height/2 - 2;
		inRect.width = 4;
		inRect.height = 4;
		/*
		* outRect is the output rectangle connector box
		* (on the right)
		*/
		outRect.x = boundingRect.x + boundingRect.width-5;
		outRect.y = boundingRect.y + boundingRect.height/2 - 2;
		outRect.width = 4;
		outRect.height = 4;
//		System.out.println(name+" updatRectPos "+boundingRect);
		/*
		* xPoints and yPoints are the points of the polygon
		* that represent the drawn shape of this block
		*/
		if (hasInputs && hasOutputs) {
			if (xPoints == null || xPoints.length != 6) {
				xPoints = new int[6];
				yPoints = new int[6];
			}
			xPoints[0] = boundingRect.x + NOSE_DEPTH;
			yPoints[0] = boundingRect.y;
			
			xPoints[1] = boundingRect.x + boundingRect.width - NOSE_DEPTH; // make sure inside active area
			yPoints[1] = yPoints[0]; 

			xPoints[2] = xPoints[1] + NOSE_DEPTH;
			yPoints[2] = yPoints[1] + boundingRect.height/2;

			xPoints[3] = xPoints[1];
			yPoints[3] = boundingRect.y + boundingRect.height -1; // make sure inside active area

			xPoints[4] = xPoints[0];
			yPoints[4] = yPoints[3];

			xPoints[5] = boundingRect.x;
			yPoints[5] = yPoints[2];
		}
		else if (hasInputs) {
			if (xPoints == null || xPoints.length != 5) {
				xPoints = new int[5];
				yPoints = new int[5];
			}
			xPoints[0] = boundingRect.x + NOSE_DEPTH; 
			yPoints[0] = boundingRect.y;
			
			xPoints[1] = boundingRect.x + boundingRect.width -1; // make sure inside active area
			yPoints[1] = yPoints[0]; 

			xPoints[2] = xPoints[1];
			yPoints[2] = boundingRect.y + boundingRect.height -1; // make sure inside active area

			xPoints[3] = xPoints[0];
			yPoints[3] = yPoints[2];

			xPoints[4] = boundingRect.x;
			yPoints[4] = boundingRect.y + boundingRect.height/2;

		}
		else if (hasOutputs) {
			if (xPoints == null || xPoints.length != 5) {
				xPoints = new int[5];
				yPoints = new int[5];
			}
			xPoints[0] = boundingRect.x; 
			yPoints[0] = boundingRect.y;
			
			xPoints[1] = boundingRect.x + boundingRect.width - NOSE_DEPTH; // make sure inside active area
			yPoints[1] = yPoints[0]; 

			xPoints[2] = xPoints[1] + NOSE_DEPTH;
			yPoints[2] = yPoints[1] + boundingRect.height/2;

			xPoints[3] = xPoints[1];
			yPoints[3] = boundingRect.y + boundingRect.height -1; // make sure inside active area

			xPoints[4] = xPoints[0];
			yPoints[4] = yPoints[3];
		}
		else {
			xPoints = null;
			yPoints = null;
		}
	}

/**
* move box where x,y are the upper left
*/
	public void setPos(int x, int y) {
		moveTo(x+boundingRect.width/2, y+boundingRect.height/2);
	}
/**
* Set the color of the outline for hilighting
*/
	void setOutlineColor(Color color) {
		outlineColor = color;
	}

/**
* set enabled status
*/
	public void setEnabled(boolean enabled) {
		boolean nowEnabled = fontForName == blockGroup.enabledFont;
		
		if (nowEnabled != enabled) {
			fontForName = enabled ? blockGroup.enabledFont : blockGroup.disabledFont;
			displayFontColor = enabled ? blockGroup.enabledColor : blockGroup.disabledColor;
			refresh();
			adjustSizeForName();
		}
		
	}
/**
* is item enabled?
*/
	public boolean getEnabled() {
		return fontForName == blockGroup.enabledFont;
	}
	
/**
* the kind of block for information purposes, output on the status line
*/
	public void setKind(String kind) {
		this.kind = kind;
		internalBuiltInfoString = null;
	}
/**
* the kind of block
*/
	public String getKind() {
		return kind;
	}
/**
* move box so that its center is at the specified x,y 
*/
	public synchronized void moveTo(int x, int y) {
	/*
	* Refresh the old position and connecting lines
	*/
		if (parent != null) {
			refresh();
			for (int i = sources.size()-1; i>= 0; i--) {
				BlockConnector aLine = (BlockConnector) sources.elementAt(i);
				parent.needsRefresh(aLine.boundingRect);
			}
			for (int i = sinks.size()-1; i>= 0; i--) {
				BlockConnector aLine = (BlockConnector) sinks.elementAt(i);
				parent.needsRefresh(aLine.boundingRect);
			}
		}
	/*
	* Refresh the new position and connecting lines
	*/
		updateRectPos(x - boundingRect.width/2, y - boundingRect.height/2);
		unselect();
		if (parent != null) {
			for (int i = sources.size()-1; i>= 0; i--) {
				BlockConnector aLine = (BlockConnector) sources.elementAt(i);
				aLine.updateBounds();
				parent.needsRefresh(aLine.boundingRect);
			}
			for (int i = sinks.size()-1; i>= 0; i--) {
				BlockConnector aLine = (BlockConnector) sinks.elementAt(i);
				aLine.updateBounds();
				parent.needsRefresh(aLine.boundingRect);
			}
			refresh();
		}
	}
	
	
	public synchronized void paint(Graphics g) {
		if ( textLocOffestX == 0){
			textLocOffestX = boundingRect.width/2 - parent.getFontMetrics(fontForName).stringWidth(getDisplayName())/2;
			textLocOffestY = boundingRect.height/2 + fontForName.getSize()/2;
		}
		
	/*
	* fill main part of box
	*/
		g.setColor(getBackgroundColor());
		if (xPoints == null) {
			g.fillRect(boundingRect.x, boundingRect.y, boundingRect.width, boundingRect.height);
		}
		else {
			g.fillPolygon(xPoints, yPoints, xPoints.length);
		}

	/*
	* draw label
	* We used to draw this last but some weird NT or Windows bug (even with 1.1.8)
	* caused the label not to draw if it was the last thing drawn. This only
	* happened on the last block drawn to the offscreen image
	*/
		g.setFont(fontForName);
		g.setColor(displayFontColor);
		g.drawString(getDisplayName(), boundingRect.x+textLocOffestX , boundingRect.y + textLocOffestY);
/*
		g.drawLine(boundingRect.x, boundingRect.y, boundingRect.x+boundingRect.width, boundingRect.y+boundingRect.height);
		g.drawLine(boundingRect.x, boundingRect.y+boundingRect.height, boundingRect.x+boundingRect.width, boundingRect.y);
*/
	
	/*
	* select outline color
	*/
		Color color = displayFontColor;
		
		if (drawSelected) 
			color = Color.blue;
		else 
			color = outlineColor;
		g.setColor(color);
	/*
	* draw connector nubs if applicable
	*/
		if (hasOutputs) {
			g.fillRect(outRect.x, outRect.y, outRect.width, outRect.height);
		}
		if (hasInputs) {
			g.fillRect(inRect.x, inRect.y, inRect.width, inRect.height);
		}
	/*
	* draw outline
	*/
		if (xPoints == null) {
			g.drawRect(boundingRect.x, boundingRect.y, boundingRect.width-1, boundingRect.height-1);
		}
		else {
			g.drawPolygon(xPoints, yPoints, xPoints.length);
		}
	}
	
	public void editProperties() {
//		System.out.println("edit properties for block "+this.toString());
	}

	boolean isRubberbandLoc(int x, int y) {
		if (!hasOutputs)
			return false;

		return bandArea.contains(x, y);
	}
	
	public String getName() {
		return name;
	}
	
	public Rectangle getBoundingRect() {
		return boundingRect;
	}
	
	public String toString() {
		return "Block "+name;
	}

/**
* @param name null, or the information string to be displayed when
* user moves mouse over the block
*/
	public void setInfoString(String infoString) {
		userSpecifiedInfoString = infoString;
	}
/**
* If no info string has been explicitly set then the
* returned value depends on the kind and name properties of this block.
* If there is not explicit kind then the name is returned.  If
* there is a kind then the return string is
* of the form "<name> (<kind>)"
*/
	public final String getInfoString() {
		if (userSpecifiedInfoString != null)
			return userSpecifiedInfoString;
		if (internalBuiltInfoString != null)
			return internalBuiltInfoString;
		internalBuiltInfoString = (name == null)? "" : name;
		if (kind != null)
			internalBuiltInfoString += " ("+ kind + ")";
		return internalBuiltInfoString;
	}

/**
* The display name may be the same as the name or it may
* be overridden by the using app.  There reason to use this
* is that the name must be unique but the display name does
* not need to be unique.  
*/
	public void setDisplayName(String displayName) {
		userSpecifiedDisplayName = displayName;
	}
/**
* The display name may be the same as the name or it may
* be overridden by the using app
*/
	public final String getDisplayName() {
		return (userSpecifiedDisplayName) == null ? name : userSpecifiedDisplayName;
	}
/**
* Allows animation of background color
*/
	protected Color getBackgroundColor() {
		return blockGroup.normalBackgroundColor;
	}
/**
* Makes a note of the difference between the mark coordinates
* and the center of the object so that a future
* moveRelative() will move to the same relative position
* as the center is to the current mark.
*/
	final protected void markForMoveRelative(int markX, int markY) {
		deltaXtoMark = boundingRect.x + boundingRect.width/2  - markX; 
		deltaYtoMark = boundingRect.y + boundingRect.height/2 - markY;
		moveToX = markX;
		moveToY = markY;
	}
	
	final protected void moveRelative(int newX, int newY) {
		moveTo(newX + deltaXtoMark, newY + deltaYtoMark);
	}
	final void moveRelative() {
		moveRelative(moveToX, moveToY);
	}
}