package org.mitre.pam.util.nntp;

import java.io.*;
import java.util.*;
import java.net.*;
import java.text.*;

import org.mitre.pam.util.PrintCRLFWriter;
import org.mitre.pam.store.*;
//import org.mitre.pam.executive.Exec;
import org.mitre.pam.interfaces.Reporter;

/**
* The utility class implementing the NNTP Usenet news protocol. <p>
*
* <pre>
* Implements NNTP as specified in RFC 977 and RFC 850. 
* <a href=http://www.faqs.org/rfcs/rfc977.html>RFC 977</a> 
* <a href=http://www.faqs.org/rfcs/rfc850.html>RFC 850</a> 
*
* To use this NNTP class, you must execute an NNTP command and then 
* call getResponses() and getData() to figure out what responses and
* data were returned by the news server, if any.
*
* If there is an error response code (either 4xx or 5xx), then
* an exception will be thrown.
*
* At this time, the following NNTP commands are supported:
* ARTICLE
* GROUP
* NEWNEWS
* POST
* QUIT
* </pre>
*
* @author Jones Leung
*/
public class NNTPClient {
/**
* The default NNTP port
*/
    public static final int DEFAULT_PORT = 119;

/**
* NNTP Commands ...
*/  
    public static final String ARTICLE = "ARTICLE";
    public static final String GROUP = "GROUP";
    public static final String NEWNEWS = "NEWNEWS";
    public static final String POST = "POST";
    public static final String QUIT = "QUIT";
    
    // Unimplemented commands ...
/*    
    public static final String BODY = "BODY";
    public static final String HEAD = "HEAD":
    public static final String HELP = "HELP";
    public static final String IHAVE = "IHAVE";
    public static final String LAST = "LAST";
    public static final String LIST = "LIST";
    public static final String NEWGROUPS = "NEWGROUPS";
    public static final String NEXT = "NEXT";    
    public static final String SLAVE = "SLAVE";
    public static final String STAT = "STAT";
*/

/*
* NNTP Server Responses ...
*/    
    public static final int HELP = 100;
    public static final int DEBUG = 199;
    public static final int READY = 200;
    public static final int READ_ONLY = 201;
    public static final int SLAVE = 202;
    public static final int STREAMING_OK = 203;
    public static final int CLOSING = 205;
	public static final int GROUP_SELECTED = 211; 
	public static final int LISTING = 215; 
	public static final int ARTICLE_RETRIEVED_BOTH = 220;
	public static final int ARTICLE_RETRIEVED_HEAD = 221; 
	public static final int ARTICLE_RETRIEVED_BODY = 222;
	public static final int ARTICLE_RETRIEVED = 223;
	public static final int LISTING_OVERVIEW = 224;
	public static final int LISTING_ARTICLES = 230; 
	public static final int LISTING_NEW = 231; 
	public static final int ARTICLE_POSTED = 240; 
	public static final int SEND_ARTICLE = 340; 
    public static final int SERVICE_DISCONTINUED = 400;
	public static final int NO_SUCH_GROUP = 411;
	public static final int NO_GROUP_SELECTED = 412; 
	public static final int NO_ARTICLE_SELECTED = 420; 
	public static final int NO_SUCH_ARTICLE_IN_GROUP = 423; 
	public static final int NO_SUCH_ARTICLE = 430; 
	public static final int POSTING_NOT_ALLOWED = 440; 
	public static final int POSTING_FAILED = 441; 
    public static final int COMMAND_NOT_RECOGNIZED = 500;
    public static final int COMMAND_SYNTAX_ERROR = 501;
    public static final int PERMISSION_DENIED = 502;
    public static final int SERVER_ERROR = 503;
    
    public static final String DEBUG_KEY = "nntp";
    
    public static final String FILE_PREFIX = "news";
    
    protected Socket server = null;
    
    protected String host = null;
    protected int port = DEFAULT_PORT;    
    
    protected boolean connected = false;
    
    // status response lines
    protected Vector responses = new Vector();
    
    // data lines
    protected Vector data = new Vector();
    
    protected BufferedReader serverInputStream = null;
    protected BufferedWriter serverOutputStream = null;    

    protected File workingDir = null;       
    protected Reporter reporter; 

/**
* Constructor for NNTP client <p>
* @param    host        the hostname of the news server to connect to
* @param    workingDir  the directory where news articles will be saved
*/
    public NNTPClient (String host, File workingDir, Reporter reporter) {
        reporter.debugMsg("Creating NNTP client for '" , host , "'...",null, DEBUG_KEY);
        this.host = host;       
        this.workingDir = workingDir;
        this.reporter = reporter;
    }
/**
* Constructor for NNTP client <p>
* @param    host   the host name of the news server to connect to
* @param    port   the port number to connect on
*/
    public NNTPClient (String host, int port, Reporter reporter) {
        reporter.debugMsg("Creating NNTP client for '" , host , "' on port " + port , "...", DEBUG_KEY);
        this.host = host;
        this.port = port;
        this.reporter = reporter;
    }
/**
* Connect to specfied Usenet news server <p>
* @exception    IOException            if we can't connect through the socket
* @exception    UnknownHostException   if we can't lookup the IP address from the hostname
*/
    public void connect () throws IOException, UnknownHostException {
        server = new Socket (host, port);
        reporter.debugMsg("TIMEOUT: " , Integer.toString(server.getSoTimeout()),null,null, DEBUG_KEY);
        reporter.assert (server != null,  "Socket is null!", null, null, null);
        serverInputStream = new BufferedReader (new InputStreamReader (server.getInputStream()));
        serverOutputStream = new BufferedWriter (new OutputStreamWriter (server.getOutputStream()));
        connected = true;
        readResponse();
    }
/**
* Retrieve specific article by article number (must select a newsgroup first) <p>
* @param        article_number      the article number of the article to be retrieved
* @exception    IOException         if we are not connected to news server
*/
    public void article (int article_number) throws IOException {
        if (connected) {
            send (ARTICLE + " " + article_number);
        } else {
            throw new IOException ("not connected");
        }
        readResponse();
    }
/**
* Retrieve specific article by message-id (must select a newsgroup first) <p>
* Message-id's must be in angle brackets 
* (e.g. <tcl-faq/tk/windows_906899164@rtfm.mit.edu>)
* @param        message_id     the message id of the article to be retrieved
* @exception    IOException    if we are not connected to the news server
*/
    public void article (String message_id) throws IOException {
        if (connected) {
            send (ARTICLE + " " + message_id);
        } else {
            throw new IOException ("not connected");
        }
        readResponse();
    }    
    
/**
* Select a newsgroup <p>
* @param        newsgroup       the name of the newsgroup to select for reading or writing
* @exception    IOException     if we are not connected to the news server
*/
    public void group (String newsgroup) throws IOException {        
        if (connected) {
            send (GROUP + " " + newsgroup);
        } else {
            throw new IOException ("not connected");
        }        
        readResponse();
    }
/**
* Retrieve Usenet news articles posted after specific date <p>
* @param    newsgroup       a space-delimited string of newsgroup names
* @param    d               the date
*/
    public void newnews (String newsgroups, Date d) throws IOException {

        SimpleDateFormat formatter = new SimpleDateFormat ("yyMMdd HHmmss");
        
        if (connected) {
            String dateStr = formatter.format (d).toString();
            send (NEWNEWS + " " + newsgroups + " " + dateStr);            
        } else {
            throw new IOException ("not connected");
        }
        readResponse();
    }    
/**
* Post article to Usenet news server (must select a newsgroup first) <p>
* @param article    the article to post to the selected newsgroup
*/
    public void post (NewsWriter article) throws IOException {
        if (connected) {
            send (POST);
            readResponse();
            article.writeMessage (serverOutputStream);
        } else {
            throw new IOException ("not connected");
        }
        readResponse();
    }
/**
* Disconnect from news server <p>
* @exception    IOException     if there is a problem writing to the news server
*/
    public void quit () throws IOException {
        if (connected) {
            send (QUIT);
            readResponse();
            
            if (serverInputStream != null) {
                serverInputStream.close();
            }
            
            if (serverOutputStream != null) {
                serverOutputStream.close();
            }            
            connected = false;
        } 
    }
/**
* Send command to news server (used internally) <p>
* @param        command         the NNTP command string (with parameters) to send to the news server
* @exception    IOException     if there is a problem writing to the news server
*/
    protected void send (String command) throws IOException {
        reporter.debugMsg(command,null,null,null, DEBUG_KEY);
        serverOutputStream.write(command);
        serverOutputStream.newLine();
        serverOutputStream.flush();        
    }
/**
* Returns lines of responses from news server <p>
* @return   lines of responses from the news server
*/
    public Vector getResponses() {
        return responses;
    }
/**
* Whether or not this client is connected to a news server <p>
* @return    whether or not this client is connected to a news server
*/
    public boolean isConnected() {
        return connected;
    }
/**
* Read response from news server (used internally) <p>
*/
    protected void readResponse() throws IOException {        
        clearResponses();
                
        do {
            String line = serverInputStream.readLine();  
            reporter.debugMsg(line,null,null,null, DEBUG_KEY);            
            
            if (!line.equals ("") && line != null) {                
                char ch = line.charAt (0);                
                
                if (ch != '1' && ch != '2' && ch != '3' && ch != '4' && ch != '5') {
                    if (line.startsWith ("..")) {
                        reporter.debugMsg("Collapsing doubled period...",null,null,null, DEBUG_KEY);
                        responses.addElement (line.substring (1));
                    } else if (!line.trim().equals (".")) {
                        responses.addElement (line);
                    }
                }                
                
                if (ch == '4' || ch == '5') {
                    throw new IOException (line);
                }
            }
        } while (serverInputStream.ready());
    }
/**
* Clear the response queue (used internally)
*/ 
    protected void clearResponses() {
        responses = new Vector();
    }
/**
* Make sure that we close the connection to the news server. <p>
*/
    protected void finalize() {
        if (connected) {
            try {
                 quit();
            } catch(IOException ioe) {
                // ignore
            }
        }
    		//{{CLEANUP
		//}}
}   
/**
* Print out responses to most recent command -- for debugging purposes 
*/
    public void printResponses () {
        Enumeration enum = getResponses().elements();
        while (enum.hasMoreElements()) {
            reporter.debugMsg("response: '", (String)enum.nextElement().toString(), "'",null, DEBUG_KEY);
        }
    }
/**
* Returns an Enumeration of message ids of new articles.  Must call newnews() first.
*/
    public Enumeration getNewMessageIds() {
        Vector msgs = new Vector();
        Enumeration enum = responses.elements();
        while (enum.hasMoreElements()) {
            String line = (String)enum.nextElement();
            if (line.startsWith ("<")) {
                msgs.addElement (line);
            }
        }
        return msgs.elements();
    }
/**
* Get an Enumeration of news articles suitable for use with NewsDecomposers.<p>
* @param    an Enumeration of message-ids of news articles to be retrieved.
* @return   an Enumeration of File objects representing individual news articles.
*/
    public Enumeration getNewsArticles (Enumeration messageids) {
        Vector newArticles = new Vector();        
        
        while (messageids.hasMoreElements()) {
            try {
                File article = new File (workingDir, FileRef.getNewFileId (FILE_PREFIX));
                PrintCRLFWriter articleWriter = new PrintCRLFWriter(new BufferedOutputStream (new FileOutputStream (article)));
                article ((String)messageids.nextElement());
                
                Vector lines = getResponses();
                Enumeration enum = lines.elements();
                while (enum.hasMoreElements()) {
                    articleWriter.println((String)enum.nextElement());
                }
                articleWriter.flush();
                articleWriter.close();  
                newArticles.addElement (article);
            } catch (Exception e) {
                reporter.reportException ( e);
            }
        }
        
        return newArticles.elements();
    }
/**
* Main Program for testing NNTP utility class
*/
/*
    public static void main (String[] args) {
        
        try {
            // turn on debugging messages
            Alarm.setTerseDebug (DEBUG_KEY, true);
            
            // display log window
            AlarmFrame alarmFrame = new AlarmFrame ("NNTPClient Log");
            alarmFrame.show();
            
            // instantiate a new NNTP client
            NNTPClient client = new NNTPClient(args[0], null);
            
            // connect to NNTP server
            client.connect();
            
            // print responses        
            reporter.debugMsg("Responses: ",null,null,null DEBUG_KEY);
            client.printResponses();
                        
            // join the specified newsgroup
            client.group (args[1]);
            
            // print responses (response code from server)
            client.printResponses();            
            
            // get a list of articles posted since last week (7 days ago)
            Date lastweek = new Date (System.currentTimeMillis() - 7 * 60 * 60 * 24 * 1000);
            client.newnews (args[1], lastweek);
            
            // print responses (response code -- single line only)
            client.printResponses();            
                        
            // retrieve articles
            Vector message_id = client.getResponses();
            Enumeration enum = message_id.elements();
            while (enum.hasMoreElements()) {
                String id = (String)enum.nextElement();
                reporter.debugMsg("Retrieving article: " , id,null,null, DEBUG_KEY);
                client.article (id);
                client.printResponses();
            }
            
            // post an article
            NewsWriter message = new NewsWriter();
            message.setFrom("jleung@mitre.org");
            message.setBody("This is a test");
            message.setSubject("PutNews Test");
            message.addNewsgroup ("mitre.test");                
                        
            client.post (message);
            client.printResponses();
            
            // terminate NNTP session
            client.quit();
            client.printResponses();
        } catch (Exception e) {
            Alarm.reportException ("NNTP", e);
        }
    }
   */
}