package org.mitre.pam.util.nntp;

import java.util.*;
import java.io.*;
import java.text.FieldPosition;
import java.text.SimpleDateFormat;

import org.mitre.pam.util.PrintCRLFWriter;
import org.mitre.pam.util.smtp.*;
import org.mitre.pam.executive.Exec;

/**
* This utility class is used to create Usenet news articles. <p>
*
* @author Jones Leung
*/
public class NewsWriter extends MIMEc implements NewsComposer {
/**
* The newsgroups that the news article is to be posted to
*/    
    protected Vector newsGroups = new Vector();
/**
* Date format for news articles
*/
    public static final String DATE_FORMAT = "EEE, dd MMM yyyy hh:mm:ss zz";
/**
* Date formatter for news articles (overrides MIMEc)
*/
    protected SimpleDateFormat dateFormatter = new SimpleDateFormat(DATE_FORMAT);    
/**
* Constructor for NewsWriter.
*/
    public NewsWriter () {
        super();    
    }
    
/**
* This function will write a properly formatted news article to the provided
* output stream.  <p>
*
* @param out  The output stream to write the message to.
* @throws IOException  Thrown if there is an error writing to the stream, or
*                      if there is an error reading the file attachments.
*/
    public void writeMessage(BufferedWriter out) throws IOException {
        super.writeMessage (out);
    }
    
/**
* This function will write a properly formatted news article to the provided
* output stream.  <p>
*
* @param out  The output stream to write the message to.
* @throws IOException  Thrown if there is an error writing to the stream, or
*                      if there is an error reading the file attachments.
*/    
    public void writeMessage(PrintCRLFWriter out) throws IOException {
        super.writeMessage (out);
    }
    
/**
* Add a single newsgroup to the newsgroup line of the news article being composed
*/
    public void addNewsgroup(String newsgroup) {
        newsGroups.addElement (newsgroup);
    }
    
/**
* Set the poster of the news article being composed<p>
* <pre>
* Acceptable formats:
* user@hostname
* <user@hostname>
* username <user@hostname>
* </pre>
*/
    public void setFrom(String from) {
        this.from = from;
    }
    
/**
* Set the subject of the news article being composed
*/
    public void setSubject(String subject) {
        super.setSubject (subject);
    }
    
/**
* Set the body of the news article being composed
*/
    public void setBody(String body) {
        super.setBody (body);
    }
    
/**
* News articles may contain attachments, also known as enclosures.  This function
* represents the minimum amount of support that any news composer must have for
* adding attachments to a news article.
*/
    public void addAttachment(File attachment) {
        super.addAttachment (attachment);
    }

/**
* @return An Enumeration of the "newsgroups" field of the news article.
*/
    public Enumeration getNewsgroups() {        
        return newsGroups.elements();
    }

/**
* @return The "from" field of the news article.
*/
    public String getFrom() {
        return super.getFrom();
    }
/**
* This method is not applicable to Usenet news articles.<p>
* It is only here to implement the MailComposer interface.
* @throws   RuntimeException    if this method is called
*/
    public void setTo (String to) throws RuntimeException {
        throw new RuntimeException ("To: header field not used in Usenet news articles!");
    }
/**
* This method is not applicable to Usenet news articles.<p>
* It is only here to implement the MailComposer interface.
* @throws   RuntimeException    if this method is called
*/
    public String getTo () throws RuntimeException {
        throw new RuntimeException ("To: header field not used in Usenet news articles!");
    }
/**
* Writes the header of th email message to the output stream. <p>
* (overrides writeHeader in MIMEc)
* @param    out             The output stream to write to.
* @throws   IOException     if there is an error writing to the stream.
*/
    protected void writeHeader(BufferedWriter out) throws IOException {
        out.write("From: " + from);
        out.newLine();
//        out.write("To: " + to);
//        out.newLine();
        out.write ("Newsgroups: " + getNewsGroupString());
        out.newLine();
        out.write("Subject: " + subject);
        out.newLine();
        out.write("Date: " + dateFormatter.format(new Date(), new StringBuffer(), new FieldPosition(0)));
        out.newLine();
        out.write("MIME-Version: 1.0");
        out.newLine();

        if(attachments.isEmpty()) {
            out.write("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
            out.newLine();
            out.write("Content-Transfer-Encoding: 7bit");
            out.newLine();
        } else {
            out.write("Content-Type: " + MULTIPART_MIXED + "; boundary=\"" + BOUNDARY + "\"");
            out.newLine();
        }

        out.flush();
    }
/**
* Writes the header of the email message to the output stream. <p>
* (overrides writeHeader in MIMEc)
* @param    out             The output stream to write to.
* @throws   IOException     if there is an error writing to the stream.
*/
    protected void writeHeader(PrintCRLFWriter out) throws IOException {
        
        out.println("From: " + from);
//       out.println("To: " + to);    
        out.println ("Newsgroups: " + getNewsGroupString());
        out.println("Subject: " + subject);    
        out.println("Date: " + dateFormatter.format(new Date(), new StringBuffer(), new FieldPosition(0)));
        out.println("MIME-Version: 1.0");

        if(attachments.isEmpty()) {        
            out.println("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
            out.println("Content-Transfer-Encoding: 7bit");
        } else {
            out.println("Content-Type: " + MULTIPART_MIXED + "; boundary=\"" + BOUNDARY + "\"");
        }
        out.flush();
    }    
/**
* @return a list of newsgroups in format specified in RFC 850 (comma-separated).
*/
    private String getNewsGroupString() {
        String newsgroups = "";
        
        Enumeration enum = newsGroups.elements();
        
        if (enum.hasMoreElements()) {
            newsgroups = (String)enum.nextElement();
        }
        
        while (enum.hasMoreElements()) {
            newsgroups = newsgroups + "," + (String)enum.nextElement();
        }
        
        return newsgroups;
    }
}