/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */

package org.mitre.pam.util.pdifHashtables;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.interfaces.HasProperties;

/**
 * A node used by a PDIFRead object.  Stores the node's properties
 * in a hash table, with their names as a key.  References to the
 * connections associated with the node are stored using themselves
 * as a key.
 * 
 * @version 1, 3 December 97
 * @version 1.1, 30 December 97 
 * @version 2, 27 January 1999
 * @author Nazario Irizarry
 */
public class PDIFNode implements HasProperties, Cloneable {
/*
* A dictionary so that references to the same attribute or kind will resolve
* to the same string reference
*/
//	static private Hashtable dictionary = new Hashtable(1000);
/**
* The peer for a node can be anything
*/
	Object peer;
/**
 * the name (ID) of the node
*/
	protected String name;
/**
 * the type (kind) of the node
*/
	protected String type;
/** 
 * a reference to an associated object (not used in this package-
 * intended for application specific reference)
*/
	protected boolean changed;
/**
* attributes
*/
	protected HasProperties attributes;
/**
* connections
*/
	protected Vector connections = new Vector(5);
	

/**
* name and type become the dictionary version of the String
*/
	public PDIFNode ( String type, String name, HasProperties attributes) {
		super(); 
		this.attributes = attributes;
		this.name = name;
		this.type = type;
	}
	
	public void removeConn(PDIFConn conn) {
		connections.removeElement(conn);
	}

	public void putConn(PDIFConn value) {
		connections.addElement( value);
	}

	public Enumeration connections() {
		return connections.elements();
	}

/**
* Will copy the attributes of another node. 
* Does not copy name or type.
*/
	public void copyAttributesNotConnections(PDIFNode otherNode) {
		Enumeration keys = otherNode.attributes.getProperties();
		while (keys.hasMoreElements()) {
			String key = (String)keys.nextElement();
			attributes.copyPropertyFromAs(key, otherNode, key);
		}
	}

	
	public void setChanged(boolean f) {
		changed = f;
	}
	public boolean isChanged() {
		return changed;
	}

/**
* Set the name of this node without regard
* for any impact that may have on any existing
* connections.
* The name used is the dictionary version of the String
*/
	public void setName(String name) {
		this.name = name.intern();
	}
/**
* On a rename also set the changed flag
* on any connections since the PDIF
* for those connections will need to change
* The name used is the dictionary version of the String
*/
	public void rename(String name) {
		this.name = name.intern();
		changed = true;
	}

	public String getName() {
		return name;
	}
	public void setType(String type) {
		this.type = type.intern();
	}

	public void setPeer(Object peer) {
		this.peer = peer;
	}

	public Object getPeer() {
		return this.peer;
	}

	public String getType() {
		return type;
	}

	public HasProperties getAttributes() {
		return attributes;
	}

	public String getProperty(String propertyName){
		return attributes.getProperty(propertyName);
	}
/**
* Get a generic object that holds the value of the key.
* @return an object of String, Numeric, or ValueList type, or null
*/
	public Object getPropertyAsObject(String propertyName){
		return attributes.getPropertyAsObject(propertyName);
	}
/**
* @return 0 if property does not exist or is an empty valuelist
* @return 1 if string or valuelist with 1 object
* @return >1 if valuelist with more than one value
*/
	public int getCurrentDimension(String propertyName){
		return attributes.getCurrentDimension(propertyName);
	}
/**
* Copy the property and value potentially with a different name
*/
	public void copyPropertyFromAs(String fromName, HasProperties fromOther, String newName){
		if (fromOther.getCurrentDimension(fromName) > 1)
			setPropertyValues(newName, fromOther.getPropertyValues(fromName));
		else
			setProperty(newName, fromOther.getProperty(fromName));
	}
/**
* Stores the dictionary version of the propertyName
*/
	public void setProperty(String propertyName, String value){
		String currentValue = attributes.getProperty(propertyName);
		if (currentValue == null || !currentValue.equals(value)) {
			changed = true;
			attributes.setProperty( propertyName,  value);
		}
	}
	public void removeProperty(String property){
//		changed = true;
		attributes.removeProperty( property);
	}
/**
* Set a boolean-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(String property, boolean value){
		setProperty(property, value?"true":"false");
	}
/**
* Set a single integer-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, int value){
		setProperty(property, Integer.toString(value));
	}
/**
* Set a single double-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, double value){
		setProperty(property, Double.toString(value));
	}
/**
* Set a single long-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, long value){
		setProperty(property, Long.toString(value));
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*
*/
	public void removePropertyAndRemoteCopies(String property){
		changed = true;
		attributes.removePropertyAndRemoteCopies(property);
	}
	public ValueList getPropertyValues(String propertyName){
		return attributes.getPropertyValues( propertyName);
	}
/**
* Stores the dictionary version of the propertyName
*/
	public void setPropertyValues(String propertyName, ValueList values){
		changed = true;
		attributes.setPropertyValues( propertyName,  values);
	}
	public Enumeration getProperties(){
		return attributes.getProperties();
	}
	
	public boolean getBooleanProperty(String property, boolean defaultValue){
		return attributes.getBooleanProperty(property, defaultValue);
	}
	
	public String getStringProperty(String property, String defaultValue){
		return attributes.getStringProperty(property, defaultValue);
	}

	public boolean isProperty(String propertyName){
		return attributes.isProperty(propertyName);
	}
	
	public int getIntegerProperty(String property, int defaultValue){
		return attributes.getIntegerProperty(property, defaultValue);
	}

	public long getLongProperty(String property, long defaultValue){
		return attributes.getLongProperty(property, defaultValue);
	}

	public double getDoubleProperty(String property, double defaultValue){
		return attributes.getDoubleProperty(property, defaultValue);
	}
}

