/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.popmail;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;

import java.net.SocketException;

import java.util.StringTokenizer;

import org.mitre.pam.store.FileRef;

/**
* This class provides access to mail messages from a POP3
* mail server.  This version supports multiple attachments.
*
* @author John R. Fantry
*/
public class RecvMail {
  protected static final String filePrefix = new String("mm");

  protected File workingDir;

  protected String host;
  protected String user;
  protected String password;

  protected boolean connectedToPOP;
  protected Pop3 pop;
  protected int mailIndex;

  protected final static int MAX_ATTEMPTS = 5;

  /**
  * Creates a new <tt>RecvMail</tt> object.
  *
  * @param host        The internet address of the mail server
  * @param user        Username
  * @param password    User password
  * @param workingDir  The directory where mail messages will be saved
  */
  public RecvMail(String host, String user, String password, File workingDir) {
    this.host = host;
    this.user = user;
    this.password = password;
    this.workingDir = workingDir;

    connectedToPOP = false;
    pop = new Pop3(this.host, this.user, this.password);

    mailIndex = 1;
  }

  /**
  * Establishes a POP3 connection with the host.
  * If there is an error while attempting to open the socket, there will
  * be a delay of 60 seconds and then another connection attempt will
  * be made.  A maximum of 5 connection attempts will be made before an
  * exception is thrown.
  *
  * @exception  SocketException
  *             Occurs when a socket connection can not be established
  *             with the host.
  */
  protected void connectToPOP() throws SocketException {
    int tries = 1;
    PopStatus status = pop.connect();   // Establish connection with POP server
    if (status.OK()) {
      status = pop.login();
    }

    if (status.OK()) {
      connectedToPOP = true;
      reset();
    } else {                            // wait and try again
      while ((!(status.OK())) && (tries < MAX_ATTEMPTS)) {
        try {
          Thread.sleep(60000);
        } catch(Exception e) {
        }

        status = pop.connect();

        if (status.OK()) {
          status = pop.login();
        }

        if (status.OK()) {
          connectedToPOP = true;
          reset();
        }
        tries++;
      }
    }
    if (!(connectedToPOP)) {
      throw new SocketException("Can't establish connection with " + host);
    }
  }

  /**
  * Checks to see if there are any messages available from the server.
  *
  * @exception  SocketException
  *             Occurs if there is an error while trying to open a socket
  *             connection with the host.
  *
  * @return     True if mail exists, false otherwise.
  */
  public boolean isMailExisting() throws SocketException {
    boolean mail = false;

    if (!connectedToPOP) {
      connectToPOP();
    }

    if (connectedToPOP) {
      PopStatus status = pop.stat();
      String response = status.Response();
      StringTokenizer st = new StringTokenizer(response, " ");
      response = st.nextToken();
      response = st.nextToken();
      if (Integer.parseInt(response) > 0) {  // Check for messages
        mail = true;
      }
    }
    return(mail);
  }

  /**
  * Retrieves the next piece of mail from the host. Can extract a null content.
  *
  * @exception  FileNotFoundException
  *             Occurs when an attempt is made to retrieve mail
  *             that does not exist.
  *
  * @exception  SocketException
  *             Occurs if a socket connection can not be established
  *             with the host
  *
  * @exception IOException
  *            Thrown if there is a problem saving the mail message to the
  *            working directory.
  */
  public File getNextPost() throws IOException, FileNotFoundException, SocketException {
    if(!connectedToPOP) {
      connectToPOP();
    }

    if(connectedToPOP) {
      PopStatus status = pop.retr(mailIndex);
      if(!status.OK()) {
        throw new FileNotFoundException();
      }
      else {
        File mailMessage = new File(workingDir, FileRef.getNewFileId(filePrefix));
        FileWriter mailMessageWriter = new FileWriter(mailMessage);
        String[] lines = status.Responses();
        for(int i=0; i<lines.length; i++) {
          mailMessageWriter.write(lines[i] + "\n", 0, lines[i].length() + 1);
        }
        mailMessageWriter.flush();
        mailMessageWriter.close();
        mailIndex++;
        return mailMessage;
      }
    }
    return null;
  }

  /**
  * Marks a piece of mail for deletion.
  *
  * @exception  SocketException
  *             Occurs if a socket connection can not be established
  *             with the host.
  */
  public void deleteMail() throws SocketException {
    if (connectedToPOP) {
      if (mailIndex == 1) {               // In case someone tries to delete mail before reading it
        PopStatus status = pop.dele(mailIndex);
      } else {
        PopStatus status = pop.dele(mailIndex - 1);
      }
    }
  }

  /**
  * Closes the POP3 connection.
  */
  public void closePOP() {
    if (connectedToPOP) {
      PopStatus status = pop.quit();

      if (status.OK()) {
        pop.close();
        connectedToPOP = false;
        mailIndex = 1;
      }
    }
  }

  /**
  * Closes the connection with the host if it is still open.
  */
  protected void finalize() {
    closePOP();
  }

  /**
  * Removes all tags from messages that were marked for deletion.
  *
  * @exception  SocketException
  *             Occurs if a socket connection can not be created
  *             with the host.
  */
  public void undeleteMail() throws SocketException {
    if (!connectedToPOP) {
      connectToPOP();
    }
    PopStatus status = pop.rset();
  }

  /**
  * Resets the mail index so that it points to the front of the list.
  */
  public void reset() {
    mailIndex = 1;
  }
}