/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtp;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;

import java.text.FieldPosition;
import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.mitre.pam.util.Codecs;
import org.mitre.pam.util.PrintCRLFWriter;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.executive.ExecGlobal;


/**
* This class composes MIME formatted email messages.
*
* Jason Mathews - iPAM 2.3.64 6/2/99 - fixed date code used in header
*/
public class MIMEc implements MailComposer {

  // MIME content types
  public static final String APPLICATION = "application/";
  public static final String IMAGE       = "image/";
  public static final String MULTIPART   = "multipart/";
  public static final String TEXT        = "text/";

  // MIME content types + subtypes
  public static final String GIF_IMAGE = IMAGE + "gif";
  public static final String MULTIPART_MIXED = MULTIPART + "mixed";
  public static final String OCTET_STREAM = APPLICATION + "octet-stream";
  public static final String TEXT_HTML = TEXT + "html";
  public static final String TEXT_PLAIN = TEXT + "plain";
  public static final String TEXT_XML = TEXT + "xml";

  public static final String BOUNDARY = "----=_NextPart_000_01BC7FBC.7F1086E0";
  public static final String DATE_FORMAT = "EEE, dd MMM yyyy HH:mm:ss zzzzz";

  // extension -> mimetype mappings
  protected static Hashtable MIMEtypes = new Hashtable();
  
  protected static Reporter reporter;

  // mimetype -> extension mappings
  protected static Hashtable MIMEextensions = new Hashtable();

  protected Hashtable attachments = new Hashtable();

  protected static SimpleDateFormat dateFormatter = new SimpleDateFormat(DATE_FORMAT);

  protected String body = new String();
  protected String from = new String();
  protected String subject = new String();
  protected String to = new String();

  static {
    initMIMEtable();
  }


  public MIMEc() {
  }

  /**
  * Initializes our MIMEtypes Hashtable with all the default file/MIMEtype pairs
  * that we should be aware of by default.
  */
  protected static void initMIMEtable() {

		reporter = ExecGlobal.getExec().getExecAsAgent().getReporter();
      	BufferedReader rdr = null;

      try {
        File mimeFile = new File(ExecGlobal.getSupportDir("smtp"), "mime.types");
        rdr = new BufferedReader(new FileReader(mimeFile));
        String s;
        for (int line=0; (s = rdr.readLine()) != null; line++) {
          // ignore comments and blank lines
          if (s.length() == 0 || s.startsWith("#")) continue;
          int pos = s.indexOf(':');
          if (pos <= 1) {
            reporter.warning( "invalid mimeType definition at line " + line);
            continue;
          }
          String MIMEtype = s.substring(0,pos).trim();
          boolean useType = true;
          /* Mimetypes prefixed with '*' indicates a one-way mapping from
          * a file extension to a mime type but not vice versa (e.g. .exe implies
          * octet-stream but not the other away around.
          */
          if (MIMEtype.startsWith("*")) {
            useType = false;
            MIMEtype = MIMEtype.substring(1);
          }
          // System.out.println(MIMEtype + ": " + s.substring(pos+1).trim());
          StringTokenizer MIMEcontentType = new StringTokenizer(s.substring(pos+1), ", ");
          for (int i=0; MIMEcontentType.hasMoreElements(); i++) {
            String ext = (String) MIMEcontentType.nextElement();
            // Store only the first occurence of an extension -> mimetype mapping
            if (!MIMEtypes.containsKey(ext))
              MIMEtypes.put(ext, MIMEtype);
              // else exec.warning("smtp", "Extension (" + ext + ") already mapped to a mimetype: ignoring " + MIMEtype);
            if (useType) {
              MIMEextensions.put(MIMEtype, ext);
              useType = false;                // only use the first extension for lookup
            }
          }
        } // for

      } catch(java.io.IOException ioe) {
        reporter.reportException( ioe);
      } finally {
        if (rdr != null)
          try {
            rdr.close();
          } catch (IOException ignore) {}
    }
  }

  /**
  * This function sets the "to" field for this MIME message.
  *
  * @param to  The "to" address.  The address should be a user@host pair.
  */
  public void setTo(String to) {
    if (to.indexOf("<") == -1) {
      to = "<" + to + ">";
    }

    this.to = to;
  }

  /**
  * This function sets the "from" field for this MIME message.
  *
  * @param from  The "from" address.  The address should be a user@host pair.
  */
  public void setFrom(String from) {
    if (from.indexOf("<") == -1) {
      from = "<" + from + ">";
    }

    this.from = from;
  }

  /**
  * This function sets the "subject" field for this MIME message.
  *
  * @param subject  The "subject" of the message.
  */
  public void setSubject(String subject) {
    this.subject = subject;
  }

  /**
  * This function sets the "body" of this MIME message.
  *
  * @param subject  The "body" of the message.
  */
  public void setBody(String body) {
    this.body = body;
  }

  /**
  * This function will add a file to this message as an attachment.  Since the
  * MIMEtype is unknown, we'll try and look it up.
  *
  * @param attachment  The file to attach.
  */
  public void addAttachment(File attachment) {
    String MIMEtype = getMIMEtype(attachment);

    attachments.put(attachment, MIMEtype);
  }

  /**
  * This function will add a file to this message as an attachment.
  *
  * The first argument to this function is supposed to be the MIME type of the
  * attachment.  This class maintains a Hashtable of all file/MIMEtype pairs that it is
  * aware of.  If the MIMEtype argument is null, then the function will look in
  * the Hashtable to see if it knows what the proper type should be for the file
  * argument.  If the type is unknown than the default OCTET_STREAM will be used.
  *
  * If the MIMEtype argument is not null, then the function will see if it has an
  * entry in its Hashtable for this file/MIMEtype pair.  If it doesn't then an
  * entry will be added.  This way the class will "remember" all file/MIMEtype pairs that
  * it comes into contact with and will therefore have a better chance of matching
  * a MIME type with a file when the MIMEtype argument is null.
  *
  * @param MIMEtype  The MIME type of the attachment (may be null).
  * @param attachment  The file to be included as an attachment.
  */
  public void addAttachment(String MIMEtype, File attachment) {
    if (MIMEtype == null || MIMEtype.length() == 0) {
      MIMEtype = getMIMEtype(attachment);
    }
    else {
      String filename = attachment.getName();
      String extension = filename.substring(filename.lastIndexOf(".") + 1);

      if (!MIMEtypes.containsKey(extension)) {
        MIMEtypes.put(extension, MIMEtype);
      }
    }

    attachments.put(attachment, MIMEtype);
  }

/** Find a file extension given a MIME type
 */
  public static String findExtension(String MIMEtype) {
        if (MIMEtype == null || MIMEtype.length()==0) return null;
	return (String) MIMEextensions.get(MIMEtype);
  }

  public void addAttachment(String MIMEtype, File attachment, String attachmentName) {
    if (MIMEtype == null || MIMEtype.length() == 0) {
      MIMEtype = getMIMEtype(attachment);
    }
    else {
      //String filename = attachment.getName();
      String extension = attachmentName.substring(attachmentName.lastIndexOf(".") + 1);

      if (!MIMEtypes.containsKey(extension)) {
        MIMEtypes.put(extension, MIMEtype);
      }
    }

    if (isValidFileName(attachmentName)) {
      Vector list = new Vector();
      list.addElement(MIMEtype);
      list.addElement(attachmentName);
      attachments.put(attachment, list);
    }
    else {
      attachments.put(attachment, MIMEtype);
    }
  }

  protected boolean isValidFileName(String fileName) {
    if (fileName.startsWith("http") || fileName.startsWith("ftp")) {
      return false;
    }
    return true;
  }

  /**
  * This function will write the properly formatted MIME message, that is represented
  * by this object, to the stream provided.
  *
  * @param out  The stream to write the message to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  public void writeMessage(BufferedWriter out) throws IOException {

    // write header
    writeHeader(out);

    // write line to separate header from body
    out.newLine();
    out.flush();

    // write body
    if(attachments.isEmpty()) {
      out.write(body);
      out.newLine();
      out.write(".");
      out.newLine();
    }
    else {
      out.write("This is a multi-part message in MIME format.");
      out.newLine();
      out.write("--" + BOUNDARY);
      out.newLine();
      out.write("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      out.newLine();
      out.write("Content-Transfer-Encoding: 7bit");
      out.newLine();

      out.newLine();

      out.write(body);
      out.newLine();

      out.newLine();

      writeAttachments(out);

      out.write(".");
      out.newLine();
    }

    out.flush();
  }

  /**
  * Writes the header of the mail message to the output stream.
  *
  * @param out  The output stream to write to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  protected void writeHeader(BufferedWriter out) throws IOException {
    out.write("From: " + from);
    out.newLine();
    out.write("To: " + to);
    out.newLine();
    out.write("Subject: " + subject);
    out.newLine();
    out.write("Date: " + dateFormatter.format(new Date()));
    out.newLine();
    out.write("MIME-Version: 1.0");
    out.newLine();

    if (attachments.isEmpty()) {
      out.write("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      out.newLine();
      out.write("Content-Transfer-Encoding: 7bit");
      out.newLine();
    }
    else {
      out.write("Content-Type: " + MULTIPART_MIXED + "; boundary=\"" + BOUNDARY + "\"");
      out.newLine();
    }

    out.flush();
  }

  /**
  * Writes the attachments of the mail message to the output stream.  Any attachments
  * that are not of the TEXT MIME type will be Base64 encoded.
  *
  * @param out  The output stream to write to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  protected void writeAttachments(BufferedWriter out) throws IOException {
    Enumeration keys = attachments.keys();

    while (keys.hasMoreElements()) {
      File attachment = (File)keys.nextElement();
      //String MIMEtype = (String)attachments.get(attachment);
      String MIMEtype = new String();
      String fileName = new String();
      Object obj = attachments.get(attachment);

      if (obj instanceof String) {
        MIMEtype = (String)obj;
        fileName = attachment.getName();
      }
      else {
        MIMEtype = (String)((Vector)obj).elementAt(0);
        fileName = (String)((Vector)obj).elementAt(1);
      }

      out.write("--" + BOUNDARY);
      out.newLine();
      out.write("Content-Type: " + MIMEtype + "; name=\"" + fileName + "\"");
      out.newLine();
      out.write("Content-Disposition: inline; filename=\"" + fileName + "\"");
      out.newLine();

      if(MIMEtype.startsWith(TEXT)) {
        out.write("Content-Transfer-Encoding: " + MIMEtype + "; charset=us-ascii");
        out.newLine();

        out.newLine();

        BufferedReader fileIn = new BufferedReader(new FileReader(attachment));
        String line = fileIn.readLine();
        while(line != null) {
          out.write(line, 0, line.length());
          out.newLine();
          line = fileIn.readLine();
        }
        fileIn.close();

        //out.newLine();
      }
      else {
        out.write("Content-Transfer-Encoding: base64");
        out.newLine();

        out.newLine();

        BufferedInputStream fileIn = new BufferedInputStream(new FileInputStream(attachment), 54);
        byte inbuf[] = new byte[54];
        int bytesRead = fileIn.read(inbuf, 0, inbuf.length);
        while(bytesRead != -1) {
          byte outbuf[];
          if(bytesRead == 54) {
            outbuf = Codecs.base64Encode(inbuf);
          } else {
            byte tempbuf[] = new byte[bytesRead];
            System.arraycopy(inbuf, 0, tempbuf, 0, bytesRead);
            outbuf = Codecs.base64Encode(tempbuf);
          }
          out.write(new String(outbuf, 0, outbuf.length));
          out.newLine();
          bytesRead = fileIn.read(inbuf, 0, inbuf.length);
        }
        fileIn.close();
      }

      if (!keys.hasMoreElements()) {
        out.write("--" + BOUNDARY + "--");
        out.newLine();
      }
    }

    out.flush();
  }

  /**
  * Returns the "to" field of this mail message.
  *
  * @return  The "to" field.
  */
  public String getTo() {
    return to;
  }

  /**
  * Returns the "from" field of this mail message.
  *
  * @return  The "from" field.
  */
  public String getFrom() {
    return from;
  }

  /**
  * This function attempts to determine the MIME type of the file argument.  If the
  * MIME type can not be determined then the default of OCTET_STREAM will be returned.
  *
  * @return  A string representing the MIME type of the file argument.
  *
  * @param file  The file to determine the MIME type of.
  */
  public static String getMIMEtype(File file) {
    String filename = file.getName();
    int pos = filename.lastIndexOf(".");
    if (pos > 0) {
     String extension = filename.substring(pos + 1);
     if (extension.length() > 0) {
      if (MIMEtypes.containsKey(extension)) {
        return (String) MIMEtypes.get(extension);
      }
      }
    }

    return OCTET_STREAM;
  }

  /**
  * This function will write the properly formatted MIME message, that is represented
  * by this object, to the stream provided.
  *
  * @param out  The stream to write the message to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  public void writeMessage(PrintCRLFWriter out) throws IOException {

    // write header
    writeHeader(out);

    // write line to separate header from body
    //out.newLine();
    out.println();
    out.flush();

    // write body
    if (attachments.isEmpty()) {
      //out.write(body);
      out.println(body);
      //out.newLine();
      //out.write(".");
      out.println(".");
      //out.newLine();
    }
    else {
      //out.write("This is a multi-part message in MIME format.");
      out.println("This is a multi-part message in MIME format.");
      //out.newLine();
      //out.write("--" + BOUNDARY);
      out.println("--" + BOUNDARY);
      //out.newLine();
      //out.write("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      out.println("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      //out.newLine();
      //out.write("Content-Transfer-Encoding: 7bit");
      out.println("Content-Transfer-Encoding: 7bit");
      //out.newLine();

      //out.newLine();
      out.println();

      //out.write(body);
      out.println(body);
      //out.newLine();

      //out.newLine();
      out.println();

      writeAttachments(out);

      //out.write(".");
      out.println(".");
      //out.newLine();
    }

    out.flush();
  }

  /**
  * Writes the header of the mail message to the output stream.
  *
  * @param out  The output stream to write to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  protected void writeHeader(PrintCRLFWriter out) throws IOException {
    //out.write("From: " + from);
    out.println("From: " + from);
    //out.newLine();
    //out.write("To: " + to);
    out.println("To: " + to);
    //out.newLine();
    //out.write("Subject: " + subject);
    out.println("Subject: " + subject);
    //out.newLine();
    //out.write("Date: " + dateFormatter.format(new Date(), new StringBuffer(), new FieldPosition(0)));
    out.println("Date: " + dateFormatter.format(new Date(), new StringBuffer(), new FieldPosition(0)));
    //out.newLine();
    //out.write("MIME-Version: 1.0");
    out.println("MIME-Version: 1.0");
    //out.newLine();

    if(attachments.isEmpty()) {
      //out.write("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      out.println("Content-Type: " + TEXT_PLAIN + "; charset=us-ascii");
      //out.newLine();
      //out.write("Content-Transfer-Encoding: 7bit");
      out.println("Content-Transfer-Encoding: 7bit");
      //out.newLine();
    }
    else {
      //out.write("Content-Type: " + MULTIPART_MIXED + "; boundary=\"" + BOUNDARY + "\"");
      out.println("Content-Type: " + MULTIPART_MIXED + "; boundary=\"" + BOUNDARY + "\"");
      //out.newLine();
    }

    out.flush();
  }

  /**
  * Writes the attachments of the mail message to the output stream.  Any attachments
  * that are not of the TEXT MIME type will be Base64 encoded.
  *
  * @param out  The output stream to write to.
  *
  * @throws IOException  Thrown if there is an error writing to the stream.
  */
  protected void writeAttachments(PrintCRLFWriter out) throws IOException {
    Enumeration keys = attachments.keys();

    while(keys.hasMoreElements()) {
      File attachment = (File)keys.nextElement();
      //String MIMEtype = (String)attachments.get(attachment);
      String MIMEtype = new String();
      String fileName = new String();
      Object obj = attachments.get(attachment);

      if (obj instanceof String) {
        MIMEtype = (String)obj;
        fileName = attachment.getName();
      }
      else {
        MIMEtype = (String)((Vector)obj).elementAt(0);
        fileName = (String)((Vector)obj).elementAt(1);
      }

      //out.write("--" + BOUNDARY);
      out.println("--" + BOUNDARY);
      //out.newLine();
      //out.write("Content-Type: " + MIMEtype + "; name=\"" + fileName + "\"");
      out.println("Content-Type: " + MIMEtype + "; name=\"" + fileName + "\"");
      //out.newLine();
      //out.write("Content-Disposition: inline; filename=\"" + fileName + "\"");
      out.println("Content-Disposition: inline; filename=\"" + fileName + "\"");
      //out.newLine();

      if (MIMEtype.startsWith(TEXT)) {
        //out.write("Content-Transfer-Encoding: " + MIMEtype + "; charset=us-ascii");
        out.println("Content-Transfer-Encoding: " + MIMEtype + "; charset=us-ascii");
        //out.newLine();

        //out.newLine();
        out.println();

        BufferedReader fileIn = new BufferedReader(new FileReader(attachment));
        String line = fileIn.readLine();
        while(line != null) {
          //out.write(line, 0, line.length());
          out.println(line);
          //out.newLine();
          line = fileIn.readLine();
        }
        fileIn.close();

        //out.newLine();
      }
      else {
        //out.write("Content-Transfer-Encoding: base64");
        out.println("Content-Transfer-Encoding: base64");
        //out.newLine();

        //out.newLine();
        out.println();

        BufferedInputStream fileIn = new BufferedInputStream(new FileInputStream(attachment), 54);
        byte inbuf[] = new byte[54];
        int bytesRead = fileIn.read(inbuf, 0, inbuf.length);
        while(bytesRead != -1) {
          byte outbuf[];
          if(bytesRead == 54) {
            outbuf = Codecs.base64Encode(inbuf);
          } else {
            byte tempbuf[] = new byte[bytesRead];
            System.arraycopy(inbuf, 0, tempbuf, 0, bytesRead);
            outbuf = Codecs.base64Encode(tempbuf);
          }
          //out.write(new String(outbuf, 0, outbuf.length));
          out.println(new String(outbuf, 0, outbuf.length));
          //out.newLine();
          bytesRead = fileIn.read(inbuf, 0, inbuf.length);
        }
        fileIn.close();
      }

      if(!keys.hasMoreElements()) {
        //out.write("--" + BOUNDARY + "--");
        out.println("--" + BOUNDARY + "--");
        //out.newLine();
      }
    }

    out.flush();
  }
}