/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtpIn;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.OutputStreamWriter;

import java.net.ServerSocket;
import java.net.Socket;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import org.mitre.pam.executive.ExecGlobal;


/**
* The main purpose of this class is to process incoming SMTP connections by
* opening streams with a client, receiving/saving new mail messages, and
* delivering new mail to the proper recipiants.
*
* Incoming messages are only accepted if the addressee is registered in the
* SMTPregistry.  All incoming messages are individually saved directly to the working
* directory.  After a session is complete and all new mail has been received,
* the mail is delivered by creating an SMTPevent object for each mail message
* and sending it to the proper recipiants.  After delivery is complete the mail
* file is delted.
*<P>
* The delivery mechanism is modeled after the java bean event listener model.
* New mail messages are wrapped in SMTPevent objects which are then passed to
* registered classes which implement the SMTPlistener interface.  One major
* difference between this implementation and the java bean model is that
* registered classes will only receive certain SMTPevent objects.  Part of
* the registration process requires all registrants to supply mailing lists.
* Registered classes will only receive SMTPevent objects if the corresponding
* mail message is addressed to a user in their mailing list.
*<P>
* If you are unfamiliar with the Simple Mail Transfer Protocol (SMTP) please
* see RFC 821.  This document can be found at "ftp://ds.internic.net/rfc/rfc821.txt".
*
* @author John R. Fantry
* @author Jason Mathews, 11/1/99
*     Report exceptions to exec reporter not System.err if possible.
*     Add finally block to ensure that socket is closed.
*/
public class SMTPconnection extends Thread {

  protected static final String filePrefix = "mm";

  /**
  * All connection threads will stop when this flag is true
  */
  protected static boolean stop = false;

  protected BufferedReader in;
  protected BufferedWriter out;

  protected File workingDir;

  protected Hashtable sessionMessages;

  /**
  * Used to create new file names
  */
  protected static Random randFileId = new Random(System.currentTimeMillis());

/**
* socket server is running on
*/
  protected ServerSocket server;
  protected Socket client = null;
  protected String hostname;

  /**
  * This SMTPconnection constructor requires three parameters.  The hostname
  * is used to generate human readable responses during an SMTPsession.
  * The workingDir is used to save all incoming mail messages.  These files
  * are deleted after they are delivered.  The server represents the
  * SocketServer object upon which the accept() method will be called.
  *
  * @param hostname    hostname of the machine that the server is running on.
  * @param workingDir  Directory used to store temporary mail files.
  * @param server      The connection thread will listen on this server.
  */
  public SMTPconnection(String hostname, File workingDir, ServerSocket server) {
    this.server = server;
    this.hostname = hostname;

    if(workingDir.isDirectory())
      this.workingDir = workingDir;
    else
      this.workingDir = new File(System.getProperty("user.dir"));
  }

  public void run() {
    try {

      // loop runs until the stop flag is set to true
      while(!stop) {

        /**
        * Loop runs until the stop flag is set to true, or we have a new client.
        * When the stop flag is set to true we will exit from the inner loop and
        * also exit from the outer loop, thus ending the function.
        */
        while((!stop) && (client == null)) {
          try {
            client = server.accept();
          } catch(InterruptedIOException iioe) {
            // thrown by the accept() method when it times out
          }
        }
        if(client != null) {
          buildStreams();
          prepareConnection();
          enterDialog();
          destroyConnection();
          deliverMail();
        }
      }
    } catch(IOException ioe) {
      reportException(ioe);
    } finally {
      shutdownConnection();
    }
  }

  /**
  * This function creates input and output streams with a new client.
  *
  * @exception IOException  Thrown if there is an error creating the streams.
  */
  protected void buildStreams() throws IOException {
    in = new BufferedReader(new InputStreamReader(client.getInputStream()));
    out = new BufferedWriter(new OutputStreamWriter(client.getOutputStream()));
  }

  /**
  * This function prepares an SMTP session by telling the client that it
  * is ready.
  *
  * @exception IOException  Thrown if there is an error communicating with the
  *                         client.
  */
  protected void prepareConnection() throws IOException {

    String message = new String("220 " + hostname + " Simple Mail Transfer Service Ready");
    send(message);

    while(!in.ready()) {
      try {
        Thread.currentThread().sleep(1000);
      } catch(InterruptedException ie) {
      }
    }
  }

  /**
  * This function sends a string to the client.
  *
  * @param message  The message to send to the client.
  *
  * @exception IOException  Thrown if there is an error communicating with the
  *                         client.
  */
  protected void send(String message) throws IOException {
    out.write(message, 0, message.length());
    out.newLine();
    out.flush();
  }

  /**
  * This function is responsible for carrying out an SMTP session.  All SMTP
  * commands as specified in RFC 821 are here, but only some are implemented.
  * The commands that are not implemented are mostly out dated.
  *
  * Incoming messages are only accepted if the addressee is registered in the
  * SMTPregistry.  All incoming messages are saved directly to the working
  * directory.
  *
  * @exception IOException  Thrown if there is an error communicating with the
  *                         client, or if there is an error saving mail messages
  *                         to the local file system.
  */
  protected void enterDialog() throws IOException {
    String line;
    sessionMessages = new Hashtable();
    Vector mailingList = new Vector();

    while((line = in.readLine()) != null && ! (line.startsWith("QUIT"))) {
      if(line.startsWith("DATA")) {
        send("354 Start mail input");
        File mailMessage = saveMailMessage();
        sessionMessages.put(mailMessage, mailingList);
        send("250 Mail action completed");
      }
      else if(line.startsWith("EXPN")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("HELO")) {
        send("250 OK");
      }
      else if(line.startsWith("HELP")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("MAIL")) {
        send("250 OK");
      }
      else if(line.startsWith("NOOP")) {
        send("250 OK");
      }
      else if(line.startsWith("RCPT")) {
        String recipiant = line.substring(line.indexOf("<") + 1, line.length() - 1);
        if(SMTPserver.registered(recipiant)) {
          mailingList.addElement(recipiant);
          send("250 OK");
        } else {
          send("550 No such user here");
        }
      }
      else if(line.startsWith("RSET")) {
        sessionMessages = new Hashtable();
        mailingList = new Vector();
        send("250 Requested mail ation okay, completed");
      }
      else if(line.startsWith("SAML")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("SEND")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("SOML")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("TURN")) {
        send("502 Command not implemented");
      }
      else if(line.startsWith("VRFY")) {
        String recipiant = line.substring(line.indexOf("<")+1, line.length()-1);
        if(SMTPserver.registered(recipiant)) {
          send("250 OK");
        } else {
          send("550 No such user here");
        }
      }
      else {
        send("500 Syntax error, command unrecognized");
      }
    }
  }

  /**
  * This function saves an incoming mail message to the local file system.
  * A new file name is created using the random number generator.
  *
  * @return A File object that points to the new mail message on the local
  *         file system.
  *
  * @exception IOException  Thrown if there is an error saving mail messages
  *                         to the local file system.
  */
  protected File saveMailMessage() throws IOException {
    String line;

    long id = randFileId.nextLong();
    if(id < 0) {
      id = - id;
    }
    id %= 1000000;

    File mailMessage = new File(workingDir, filePrefix + '-' + Long.toString(id));
    FileWriter mailMessageWriter = new FileWriter(mailMessage);

    try {
    /**
    * A line consisting of a single "." character indicates the end of a
    * message.
    */
    while((line = in.readLine()) != null && ! line.equals(".")) {
      if(line.startsWith(".")) {

        /**
        * The SMTP protocol uses dot stuffing if the first character in a
        * line of text is the "." character.  If a line begins with the "."
        * character then we want to ignore it.
        */
        mailMessageWriter.write(line + "\n", 1, line.length());
      }
      else {
        mailMessageWriter.write(line + "\n", 0, line.length() + 1);
      }
    }
    mailMessageWriter.flush();
   } finally {
     try {
    mailMessageWriter.close();
     } catch (IOException ioe) {}
   }

    return mailMessage;
  }

  /**
  * This function closes an SMTP session by telling the client that it is
  * closing the connection and then closing the input and output streams.
  *
  * @exception IOException  Thrown if there is an error communicating with the
  *                         client, or if there is an error closing the input
  *                         and output streams.
  */
  protected void destroyConnection() throws IOException {
    send("221 " + hostname + " closing channel");
    shutdownConnection();
  }

  /**
  * This function closes the I/O streams and socket of an SMTP session.
  */
  private void shutdownConnection() {
    if (client != null) {
      if (in != null)
	try {
    in.close();
	} catch(IOException e){}
      if (out != null)
	try {
    out.close();
	} catch(IOException e){}
      try {
    client.close();
      } catch(IOException e){}
    client = null;
  }
  }

  /**
  * This function delivers all new mail messages by creating a new SMTPevent
  * object for each message and sending it out. Once a mail message
  * has been delivered it is deleted from the local file system.
  */
  protected void deliverMail() {
    Enumeration mailMessages = sessionMessages.keys();
    while(mailMessages.hasMoreElements()) {
      File mailMessage = (File)mailMessages.nextElement();
      Vector mailingList = (Vector)sessionMessages.get(mailMessage);
      try {
        SMTPevent event = new SMTPevent(mailMessage, mailingList);
        SMTPserver.fireSMTPevent(event);
        mailMessage.delete();

        /**
        * These exceptions would be thrown by the SMTPevent constructor.
        * I don't know why this would ever happen, but if it does I guess all
        * I can do is ignore it, consider the message dead, and move on.
        */
      } catch(FileNotFoundException fnfe) {
	reportException(fnfe);
      } catch(IOException ioe) {
	reportException(ioe);
      }
    }
  }

  /**
  * This function sets the stop flag to true, telling all the SMTPconnection
  * threads to stop.
  */
  public static void stopConnection() {
    stop = true;
  }

/**
 * reports exceptions to exec report if possible
 * otherwise prints stacktrace to System.err
 */
  public static void reportException(Throwable t) {
    try {
      ExecGlobal.getExec().getExecAsAgent().getReporter().reportException(t);
    } catch (Exception e) {
      t.printStackTrace(System.err);
    }
  }

}
