/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtpIn;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Date;
import java.util.Vector;


/**
* This class is essentially a wrapper for a mail message.  When a message needs
* to be delivered an SMTPevent object is created to hold it.  This wrapper class
* provides the functions necessary to extract information about the message
* like subject, body, and date.  This class also provides the functions
* necessary to extract attachments.
*
* Internally this class uses a second class (i.e. MIMEd) to process the mail
* message.
*
* It should be noted that this class contains a customized mailing list.  Sometimes
* a mail message might be addressed to several recipiants.  If an SMTPlistener
* is registered to receive mail from only JohnDoe@host and a message has multiple
* recipiants, JohnDoe@host and JaneDoe@host, then the customized mailing list
* will only contain the user JohnDoe@host.  This will save the SMTPlistener which
* receives this object from having to go through the list returned by getTo() to find
* out who the message is supposed to go to.  By calling getMailingList() instead,
* an SMTPlistener will receive a customized mailing list that only contains
* users that they registered for.
*
* Note that by the time an SMTPlistener receives the event the mailing
* list is customized.  Prior to this time the mailing list actually contains
* all the users that the message is addressed to.
*
* The majority of the functions in this class are very easy to understand.  They
* are not commented.
*
* @author John R. Fantry
*/
public class SMTPevent implements MailDecomposer {

  protected File mailMessage;

  protected MailDecomposer decomposer;

  protected Vector mailingList;

  /**
  * This SMTPevent constructor requires two parameters.  mailMessage represents
  * the message to be delivered.  mailingList represents the users to whom
  * the message is addressed.  This list is customized before event delivery.
  *
  * @param mailMessage  Mail message to be delivered.
  * @param mailingList  List of addressees.
  *
  * @exception FileNotFoundException  Thrown if there is an error trying to
  *                                   access the mail message.
  * @exception IOException            Thrown if there is an error trying to
  *                                   access the mail message.
  */
  public SMTPevent(File mailMessage, Vector mailingList) throws FileNotFoundException,
                                                                IOException {
    if(mailMessage.canRead())
      this.mailMessage = mailMessage;
    else
      throw new FileNotFoundException("File " + mailMessage.getPath() +
                                      mailMessage.getName() + " not found");

    this.mailingList = mailingList;

    if(MIMEd.isMIMEencoded(this.mailMessage)) {
      decomposer = new MIMEd(mailMessage);
    }
    else {
      decomposer = new maild(mailMessage);
    }
  }

  public Vector getMailingList() {
    return mailingList;
  }

  public File getMailMessage() {
    return mailMessage;
  }

  public Date getDate() {
    return decomposer.getDate();
  }

  public String getFrom() {
    return decomposer.getFrom();
  }

  public String getSubject() {
    return decomposer.getSubject();
  }

  public Vector getTo() {
    return decomposer.getTo();
  }

  public boolean attachmentsExist() {
    return decomposer.attachmentsExist();
  }

  public Vector getAttachmentNames() {
    return decomposer.getAttachmentNames();
  }

  public int numOfAttachments() {
    return decomposer.numOfAttachments();
  }

  public String getBody() throws IOException {
    return decomposer.getBody();
  }

  /**
  * This function attempts to extract an attachment with the given name.
  * The attachment is saved into the dst file.
  *
  * @return A String representing the MIME type of the attachment that
  * is extracted.
  *
  * @param attachmentName  The name of the attachment to extract.
  * @param dst             The destination file for the attachment.
  *
  * @exception IOException            Thrown if there is an error reading the file.
  */
  public String extractFileAttachment(String attachmentName, File dst) throws IOException {
    return decomposer.extractFileAttachment(attachmentName, dst);
  }

  /**
  * This function attempts to extract an attachment that is identified within
  * the mail message by index.  The index is zero based.
  * The attachment is saved into the dst file.
  *
  * @return A String representing the MIME type of the attachment that
  * is extracted.
  *
  * @param attachmentNum  The index of the attachment to extract.
  * @param dst            The destination file for the attachment.
  *
  * @exception IOException            Thrown if there is an error reading the file.
  */
  public String extractFileAttachment(int attachmentNum, File dst) throws IOException {
    return decomposer.extractFileAttachment(attachmentNum, dst);
  }
}