/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtpIn;

import org.mitre.pam.util.Codecs;
import java.io.BufferedReader;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;

import java.text.ParsePosition;

import java.util.Date;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;


/**
* This class decomposes uuencoded mail messages (non MIME formatted messages).
*
* If you are unfamiliar with the format of electronic mail messages please
* see RFC 822.
*
* An example of the format of a mail message:
*<pre>
* Received: from jleung-pc (jleung-pc.mitre.org [129.83.142.48])
* by mbunix.mitre.org (8.8.8/8.8.8/mitre.0) with ESMTP id MAA03518
* for <test1@jleung-pc.mitre.org>; Mon, 13 Apr 1998 12:43:53 -0400 (EDT)
* Message-Id: <199804131643.MAA03518@mbunix.mitre.org>
* From: "John Fantry" <jfantry@mitre.org>
* To: <test1@jleung-pc.mitre.org>
* Subject: uuencode
* Date: Mon, 13 Apr 1998 12:45:33 -0400
* X-MSMail-Priority: Normal
* X-Priority: 3
* X-Mailer: Microsoft Internet Mail 4.70.1161
*
* message body
* begin 600 Notepad.exe
* M35J0``,````$````__\``+@`````````0 ``````````````````````````
* M````````````````````@ ````X?N@X`M G-(;@!3,TA5&AI<R!P<F]G<F%M
* M(&-A;FYO="!B92!R=6X@:6X@1$]3(&UO9&4N#0T*) ````````!010``3 $%
* M`!$N[3$`````>@(``. ``@,+`0,*`$@```!J````````@%0````0````8 ``
* M``"T`0`0`````@``! ````0````$``````````#@````! ``1H@!``(`````
* M````````````````````````````````````````````````````````````
* M````````````````````````````````````````````````````````````
* M````````````````````````````````````````````````````````````
* M````````````````````````````````````````````````````````````
* -````````````````````
* `
* end
*
* begin 600 clock.avi
* M4DE&1FI"`0!!5DD@3$E35(P%``!H9')L879I:#@```! 0@\`O!L````````0
* M" ``# `````````"`````#(``$$!``!!`0```````````````````````$Q)
* M4U20! ``<W1R;'-T<F@X````=FED<U),12 ````````````````!`````0``
* M```````,````?!<``````````````````$$!00%S=')F* 0``"@```!!`0``
* M00$```$`" `!`````B4#``````````````$```````````````" ``" `( `
* M@( `@ ``@( `@ " @ " P,# `,#<P,#PRJ; ^_L#IOO[^P/[`P/[`P,#`WM[
* M`P-[>WL#`P/Z>ZNK`_K[>P,#4U,#`[N[`P,K*P,#.SL#`]/3`P/3T],#*RLK
* M````````````````````````````````````````````````````````````
* M````````````````````````````````````````````````````````````
* )````````````
* `
* end
*</pre>
*
* @author John R. Fantry
* @author Jones Leung, fix null header field exceptions
* @author Jason Mathews, 11/1/99
*     Report exceptions to exec reporter not System.err if possible.
*     Add checks for readLine() != null to avoid null ptr exceptions.
*     Add finally blocks to ensure that I/O streams are closed.
*/
public class maild extends GenericMaild {

  protected static final String BEGIN_TOKEN = "begin";

  protected BufferedReader in;

  protected File mailMessage;

  protected String attachmentName;

  /**
  * Constructor.
  *
  * @param mailMessage  The mail message to decompose.
  *
  * @exception IOException  Thrown if there is an error reading the mail message.
  */
  public maild(File mailMessage) throws IOException {
    this.mailMessage = mailMessage;
    processHeader();
    processAttachments();
  }

  /**
  * This function extracts fields out of the header of the message.
  *
  * @exception IOException  Thrown if there is an error reading the mail message.
  */
  protected void processHeader() throws IOException {
   in = new BufferedReader(new FileReader(mailMessage));

   try {
    String line;

    // First blank line represents end of header
    while((line = in.readLine()) != null && line.length() > 0) {
      if(containsField(line)) {
        if(field.equalsIgnoreCase(DATE_FIELD)) {
// JL: Set the date to current date in case it is null             
          date = dateFormatter.parse(value, new ParsePosition(0));
          if (date == null) {
            date = new Date();
          }
        }
// JL: Set the from field to "" in case it is null        
        else if(field.equalsIgnoreCase(FROM_FIELD)) {
          from = value;
          if (from == null) {
            from = "";
          }
        }
// JL: Set the subject field to "" in case it is null        
        else if(field.equalsIgnoreCase(SUBJECT_FIELD)) {
          subject = value;
          if (subject == null) {
            subject = "";
          }
        }
// JL: Set the to field to "" in case it is null        
        else if(field.equalsIgnoreCase(TO_FIELD)) {
          if (value == null) {
            to.addElement("");  
            continue;
          }
          
          if(value.indexOf(',') < 0) {
            to.addElement(value);
          }
          else {

            // Recipiants are separated by commas
            StringTokenizer st = new StringTokenizer(value, ",");

            while(st.hasMoreElements()) {
              to.addElement(trim((String)st.nextElement()));
            }
            in.mark(256);
            line = in.readLine();
            while(line != null && line.indexOf(':') < 0) {
              st = new StringTokenizer(line, ",");
              while(st.hasMoreElements()) {
                to.addElement(trim((String)st.nextElement()));
              }
              in.mark(256);
              line = in.readLine();
            }
            in.reset();
          }
        }
      }
    }

   } finally {
     try {
       in.close();
     } catch (IOException io) {}
   }
  }

  /**
  * This function checks to see if a string matches the pattern of the
  * begin line in a uuencoded message.
  *
  * An example of a begin line is "begin 600 filename.exe".
  *
  * @return  True if the line matches the pattern, false otherwise.
  *
  * @param line  The string to check.
  */
  protected boolean isBeginPattern(String line) {
    StringTokenizer tokenizer = new StringTokenizer(line, " ");
    String firstToken;
    String secondToken;
    String thirdToken;

    try {
      firstToken = tokenizer.nextToken();
      secondToken = tokenizer.nextToken();
      thirdToken = tokenizer.nextToken();
    } catch(NoSuchElementException nsee) {
      // If there aren't at least three tokens then the pattern doesn't match
      return false;
    }

    if(tokenizer.hasMoreElements()) {
      // If there are more than three tokens then the pattern doesn't match
      return false;
    }

    try {
      if(firstToken.equalsIgnoreCase(BEGIN_TOKEN)) {

        // test to verify that secondToken is a parsable Integer
        int i = Integer.parseInt(secondToken);

        /**
        * At this point we know that there are exactly three tokens.
        * We also know that the first token equals BEGIN_TOKEN, and that
        * the secondToken is a parsable Integer.
        * I can't think of any test for the thirdToken so it looks like
        * a good match.
        */
        attachmentName = thirdToken;
        return true;
      }
    } catch(NumberFormatException nfe) {
      // If secondToken is not an integer then the pattern doesn't match
      return false;
    }

    // won't compile without this return statement but it should never be executed
    return false;
  }

  /**
  * This function attempts to determine whether or not a string is uuencoded.
  *
  * @return  True if the line appears to be uuencoded, false otherwise.
  *
  * @param line  The string to check.
  */
  protected boolean isUUENCODEDline(String line) {
    if (line == null || line.length()==0) return false;
    try {
      /**
      * If it is possible to uudecode this string without error then hopefully
      * that means it is indeed uuencoded.
      */
      byte test[] = Codecs.uudecode(line.getBytes());
    } catch(Exception e) {

      // If anything goes wrong during decoding then it's not uuencoded
      return false;
    }

    return true;
  }

  /**
  * This function parses through a mail message extracting the names of any
  * attachments.
  *
  * @exception IOException  Thrown if there is an error reading the mail message.
  */
  protected void processAttachments() throws IOException {
   in = new BufferedReader(new FileReader(mailMessage));

   try {
    String line;
    // read until end of file
    while((line = in.readLine()) != null) {
      if(isBeginPattern(line)) {
        if((line = in.readLine()) != null) {
          if(isUUENCODEDline(line)) {
            attachmentNames.addElement(attachmentName);
          }
        }
      }
    }
   } finally {
     try {
       in.close();
     } catch (IOException io) {}
   }
  }

  /**
  * This function extracts the message body out of the mail message.
  *
  * @return  The body of the mail message.
  *
  * @exception IOException  Thrown if there is an error reading the mail message.
  */
  public String getBody() throws IOException {
    String line;
    StringBuffer body = new StringBuffer();
    in = new BufferedReader(new FileReader(mailMessage));
   try {
    // Advance to the start of the message body
    while((line = in.readLine()) != null && line.length() > 0) {
    }

    /**
    * If there are no attachments then read until the end of the file, else
    * read until the first attachment starts.
    */
    if(attachmentNames.size() == 0) {
      while((line = in.readLine()) != null) {
        body.append(line + "\n");
      }
    }
    else {
      String temp;
// JL: prevent NullPointerException when encountering empty message body
      while((line = in.readLine()) != null) {
        
        if(isBeginPattern(line)) {
          temp = line;
          in.mark(256);
          line = in.readLine();
          if(isUUENCODEDline(line)) {
            break;
          }
          else {
            body.append(temp + "\n");
            in.reset();
          }
        }
        else {
          body.append(line + "\n");
        }
      }
    }
   } finally {
     try {
       in.close();
     } catch (IOException io) {}
   }
    return body.toString();
  }

  /**
  * This function attempts to extract an attachment that is identified within
  * the mail message by index.  The index is zero based.
  * The attachment is saved into the dst file.
  *
  * @return A String representing the MIME type of the attachment that
  * is extracted.
  *
  * @param attachmentNum  The index of the attachment to extract.
  * @param dst            The destination file for the attachment.
  *
  * @exception IOException  Thrown if there is an error reading the file.
  */
  public String extractFileAttachment(int attachmentNum, File dst) throws IOException {
    String MIMEtype = "none";
    in = new BufferedReader(new FileReader(mailMessage));
    BufferedOutputStream out = null;
    try {
    advanceToAttachment(attachmentNum);
      String line = in.readLine();
      out = new BufferedOutputStream(new FileOutputStream(dst));
      while((line = in.readLine()) != null && !line.equals("`")) {
      byte data[] = Codecs.uudecode(line.getBytes());
      try {
        out.write(data);
      } catch(IOException ioe) {
	  SMTPconnection.reportException(ioe);
      }
    }
    out.flush();
    } finally {
      if (out != null)
	try {
    out.close();
	} catch (IOException io) {}
      try {
	in.close();
      } catch (IOException io) {}
    }
    return MIMEtype;
  }

  /**
  * Advances through the message body to the given attachment index.  Assumes
  * that we are reading from the start of the mail message.
  *
  * @param index  The index of the attachment to advance to.
  *
  * @exception IOException  Thrown if there is an error reading the mail message.
  */
  protected void advanceToAttachment(int index) throws IOException {
    while(index > -1) {
      in.mark(256);
      String line = in.readLine();
      if (line == null) break;
      if(isBeginPattern(line)) {
        line = in.readLine();
        if(isUUENCODEDline(line)) {
          index--;
        }
      }
    }
    in.reset();
  }
}