/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pdif.eventParser;

import java.io.*;

/** 
* A parser that deliver higher level PDIF parse events in an "event" loop.
*
*<p> The ideal PDIF parser can be decomposed several levels.
*
*
* <p> Level 3 recognizes higher level events.  metaEvents are
*     reserved for future use.  nodeEvents are generated when a parenthesis level
*     is entered or exited in the node section of the pdif stream.
*     connectorEvent is generated only for the connector portion of the pdif
*     stream.  connectorEvents are generated every time a complete "relation" is identified.
*     A complete relation consists of the two items that are related, their types,
*     and their ids.  Level 3 reports the types for the relation in alphabetical order based on
*     "kind.role".
*     (This allows different systems to transmit the relation with either item first without
*     imparing the ability of the receiver code to recognize the relation).  The fourth
*     Level 3 event is some kind of exitEvent.  As with Level 2, this should be tied
*     to a return to zero parenthesis rather than an end-of-file so that multiple
*     PDIF blocks can be sent in one stream. <p>
*
*
* This PDIF parser generates several parse events: 
* HEADER_ATTRIBUTE, NODE_ENTER, NODE_LEAVE, NODE_ATTRIBUTE, 
* CONNECTOR, NON_NODE, NON_CONNECTOR, EOF, END_OF_PDIF_BLOCK. <p>
*
*
* As the parser scans the following:
*<pre>
*(pdif (header (domain foo))
*    (node
*       (point 39a (latitude 45))
*       (leg eg9887)
*    )
*    (connector
*        (leg.container eg9887 (point.contained 39a)(point.contained 503b))
*    (nonconnector
*        (point.contained 39a (leg.container eg9887))
*    )
*)
*</pre>
*<p>
* It generates HEADER_ATTRIBUTE after parsing "domain foo"  <p>
* It generates NODE_ENTER events on entry to point, leg.  <p>
* It generates NODE_ATTRIBUTE events on entry to latitude <p>
* It generates NODE_LEAVE events at the closing parenthesis for point and leg  <p>
* It generates CONNECTOR events (two of them) after it has parsed the first 
* occurance of "point.contained 39a" and  "point.contained 503b"  <p>
* It generates NON_CONNECTOR after parsing the second occurance of "leg.container eg9887". 
* note that this NON_CONNECTOR event is intended to "undo" a previous CONNECTOR
* even though the order of the connection endpoints is different.  <p>
* It generates END_OF_PDIF_BLOCK at the closing parenthis for the entire block <p>
* It will finally generate EOF at the end of the stream<p>
*
*
* @version  4, 12 Dec 1997  objectized return types, support for multivalues
* @version  3, 12 Aug 1997  partial support for "nonnode" and "nonconnector"
* @version  2, 17 Sept 1996
* @author   Nazario Irizarry, Jr.
*/
public class ParseEvent3   {

/** 
 * Returned as the parse event type at end of file.
 */
	public static final int EOF = 19;
/** 
 * Returned as the parse event type when we enter the next parenthesized item.
 * In the node section of PDIF. 
 */
	public static final int NODE_ENTER = 20;
/** 
 * Returned as the parse event type when we leave a parenthesized level in the node section.
 */
	public static final int NODE_LEAVE = 21;
/** 
 * Returned as the parse event type identifying a node attribute.
 */
	public static final int NODE_ATTRIBUTE = 22;
/**
 * Returned as the parse event type identifying a connector. 
 */
	public static final int CONNECTOR = 23;
/** 
 * Returned as the parse event when we find a nonnode
 */
	public static final int NON_NODE = 24;
/** 
 * Returned as the parse event when we find a nonnode
 */
	public static final int NON_CONNECTOR = 25;
/** 
 * Returned as the parse event for a header attribute
 */
	public static final int HEADER_ATTRIBUTE = 26;
/**
* Returned as parse event when at end of a PDIF block
*/
	public static final int END_OF_PDIF_BLOCK = 27;
/**
* When eventType == CONNECTOR or NON_CONNECTOR
* this array contains the two ends of the connector. They are sorted
* by kind.role such that connectorEnds[1].kindRole > connectorEnds[0].kindRole
*/
	public ConnectorEnd connectorEnds[];
/**
* When eventType == NODE_ENTER, NODE_LEAVE, NON_NODE this is the node
*/
	public Node node;
/**
* When eventType == NODE_ATTRIBUTE or HEADER_ATTRIBUTE this is the attribute
*/
	public Attribute attribute;
	
/** 
 * This is the same as the returned event type from the last
 * call to next().  Valid for any return type;
 */
	public int eventType;
/*
* Constants to track what portion of stream we are in
*/
	private static final int IN_PDIF = 9;
	private static final int IN_HEADER = 9;
	private static final int IN_NODE = 10;
	private static final int IN_CONNECTOR = 12;
	private static final int IN_NON_NODE = 13;
	private static final int IN_NON_CONNECTOR = 14;
	private static final int IN_META = 15;
	private static final int NOWHERE = 0;

	private int inWhere = NOWHERE;
	private ConnectorEnd outerMost=null;
	
	ParseEvent2 parse2;
/** 
 * Attaches to an input stream to parse PDIF streams.  
 * Ignores all tokens prior to first parenthesis.
 */
	public ParseEvent3 (Reader inStream) {
		parse2 = new ParseEvent2(inStream);
		inWhere = IN_PDIF;
	}
	
/** 
 * Gets the next PDIF token. 
 * @return HEADER_ATTRIBUTE, NODE_ENTER, NODE_LEAVE, NODE_ATTRIBUTE, 
 * CONNECTOR, NON_NODE, NON_CONNECTOR, END_OF_PDIF_BLOCK, EOF
 */
	public int next() throws  BadPDIFException,IOException {
		int commentLevel=0;
		
		while ( true ) {
			
			int level3Type = parse2.next();

			switch (level3Type) {
				case ParseEvent2.EOF:
					return (eventType = EOF);
					
				case ParseEvent2.LEAVE:
					switch(parse2.depth) {
						case 1:
							if (inWhere == IN_PDIF) {
								inWhere = NOWHERE;
								return (eventType = END_OF_PDIF_BLOCK);
								}
							inWhere = NOWHERE;
							continue;
							
						case 2:
							inWhere = IN_PDIF;
							continue;
							
						case 3:
							if (inWhere == IN_NODE) {
								attribute = null;
								return (eventType = NODE_LEAVE);
							}
							continue;

						default:
							continue;
					}
					
				case ParseEvent2.ENTER:
					switch(parse2.depth) {
				/*
				* enter level 1
				*/
						case 1:
							if (parse2.arg1.equals("pdif")) {
								inWhere = IN_PDIF;
								continue;
							}
							skipThisLevel();
							break;

				/*
				* enter level 2
				*/
						case 2:
							if (parse2.arg1.equals("node")) {
								attribute = null;
								inWhere = IN_NODE;
							}
							else if (parse2.arg1.equals("connector")) {
								inWhere = IN_CONNECTOR;
							}
							else if (parse2.arg1.equals("meta")) {
								inWhere = IN_META;
							}
							else if (parse2.arg1.equals("nonnode")) {
								inWhere = IN_NON_NODE;
							}
							else if (parse2.arg1.equals("nonconnector")) {
								inWhere = IN_NON_CONNECTOR;
							}
							else if (parse2.arg1.equals("header")) {
								node = new Node("header", "");
								inWhere = IN_HEADER;
							}
							continue;
				/*
				* enter level 3
				*/
						case 3:
							switch(inWhere) {
								case IN_NODE:
									node = new Node(parse2.arg1, parse2.arg2);
									return (eventType = NODE_ENTER);

								case IN_NON_NODE:
									node = new Node(parse2.arg1, parse2.arg2);
									return (eventType = NON_NODE);

								case IN_HEADER:
									attribute = new Attribute(parse2.arg1,  parse2.multiValues);
									return (eventType = HEADER_ATTRIBUTE);

								case IN_NON_CONNECTOR:
								case IN_CONNECTOR:
									connectorEnds = new ConnectorEnd[2];
									outerMost = new ConnectorEnd(parse2.arg1, parse2.arg2, "container");
									continue;


								default:
									skipThisLevel();
							}
							break;

				/*
				* enter level 4
				*/
						case 4:
							switch(inWhere) {
								case IN_NODE:
									attribute = new Attribute(parse2.arg1, parse2.multiValues);
									return (eventType = NODE_ATTRIBUTE);
									
								case IN_CONNECTOR:
									ConnectorEnd ce = new ConnectorEnd(parse2.arg1, parse2.arg2, "contained");
									if (ce.kindRole.compareTo(outerMost.kindRole) > 0) {
										connectorEnds[0] = outerMost;
										connectorEnds[1] = ce;
									}
									else {
										connectorEnds[1] = outerMost;
										connectorEnds[0] = ce;
									}
									return (eventType = CONNECTOR);
								
								case IN_NON_CONNECTOR:
									ConnectorEnd cd = new ConnectorEnd(parse2.arg1, parse2.arg2, "contained");
									if (cd.kindRole.compareTo(outerMost.kindRole) > 0) {
										connectorEnds[0] = outerMost;
										connectorEnds[1] = cd;
									}
									else {
										connectorEnds[1] = outerMost;
										connectorEnds[0] = cd;
									}
									return (eventType = NON_CONNECTOR);
								
								default:
									skipThisLevel();
							}
							break;

				/*
				* enter level 5
				*/
						case 5:
							skipThisLevel();
					}
			}
		}
	}

/*
*
*/
	private void throwError(String s) throws BadPDIFException {
		throw new BadPDIFException(s);
	}

/*
* utility to skip through a level as if commented
*/
	private void skipThisLevel() {
		parse2.commentLevel = 1;
		parse2.depth--;
		parse2.currentDepth--;
	}

}