/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon;

import java.util.Enumeration;
import java.util.Vector;

import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.product.view.BasicProduct;

/**
* This class creates a timer Daemon that sets a user_alarm property
* when the system time reaches the given alarm time for any number of alarms.
* When fired, the <code>userAlarm.name</code> property is created, which can
* be watched by other agents/modules (i.e. AssertAgent) to take the
* appropriate actions.
*
* @author Jason Mathews
* @version 1.2, 08/07/98
*/

public class PropertyTimer extends DaemonBase {

    OrderedAlarmList alarms = new OrderedAlarmList();
    private static final String DEBUG_KEY = "PropertyTimer";

  class OrderedAlarmList {
	Vector list = new Vector();

	OrderedAlarmList() {}

	synchronized void insert(PropertyAlarm alarm) {
	  int i = list.size();
	  PropertyAlarm found = null;
	  for (int pos = 0; pos < list.size(); pos++) {
		PropertyAlarm a = (PropertyAlarm) (list.elementAt(pos));
		if (alarm.name.intern() == a.name.intern()) {
			found = a;
			// System.out.println("found at " + pos);
			if (i < pos) break;
		}
		else if (pos < i && alarm.time < a.time) {
			i = pos;
			if (found != null) break;
		}
	   }
	   // System.out.println("insert at " + i);
           list.insertElementAt(alarm, i);
	   if (found != null)
		list.removeElement(found);
	   //dump();
	}

	final synchronized void remove(PropertyAlarm alarm) {
		list.removeElement(alarm);
	}

	final int size() {
		return list.size();
	}

	void dump() {
	      Enumeration e = list.elements();
	      while (e.hasMoreElements()) {
		PropertyAlarm a = (PropertyAlarm) e.nextElement();
		System.out.println( " time=" + a.time + " " + a.values);
	      }
	}

	final PropertyAlarm first() {
		return (PropertyAlarm) list.firstElement();
	}

	final Enumeration elements() {
		return list.elements();
	}
  }

  class PropertyAlarm {
     String name;
     long time;
     Vector values;

     PropertyAlarm(String name, long time, Vector values) {
	this.name = name;
        this.time = time;
        this.values = values;
     }

     PropertyAlarm(String name, long time, String value) {
	this.name = name;
	this.time = time;
	values = new Vector(1);
	values.addElement(value);
     }

     public String toString() {
       return "PropertyAlarm[name=" + name + ",time=" + time + ",values=" + values+"]";
     }
   }

  public PropertyTimer() {
    super();
  }

 /**
  * Add new timer event to list of alarms
  */
  public void addTimer(String name, long time, Vector values) {
      PropertyAlarm alarm = new PropertyAlarm(name, time, values);
      reporter.debugMsg("add timer: " , alarm, null,null, DEBUG_KEY);
      alarms.insert(alarm);
      // if (enabled==false) setEnabled(true);
  }

/**
* Upon request we check the alarms and activate the alarm (i.e. set the alarmed properties)
* when the alarm time has elapsed. Stop checking when all alarms are fired or an alarm time
* has not yet been triggered.
*/

  public void doDaemonTask() {

    long current = System.currentTimeMillis();

    if (org.mitre.pam.util.alarm.Alarm.isDebugKey(DEBUG_KEY)) {
	    System.out.println(getId() + ": time=" + current);
    if (alarms.size() == 0) System.out.println( " no alarms defined");
    else alarms.dump();
    }

    while (alarms.size() != 0) {
      PropertyAlarm alarm = alarms.first();
      if (alarm.time > current) return; // has alarm triggered yet ?
      alarms.remove(alarm); // remove named alarm from the list

      // System.out.println("Fire alarm " + alarm); // DEBUG
      reporter.debugMsg( " fire alarm: " , alarm,null,null, DEBUG_KEY);

      if (alarm.values.size() > 1)
	setPropertyValues("user_alarm."+alarm.name, new ValueList(alarm.values));
      else
	setProperty("user_alarm."+alarm.name, (String)alarm.values.elementAt(0));
    }
    // If no alarms are left then turn off daemon ??
  }

  /**
  * Identifies the persistent product for this agent, this
  * is the method used by the exec to restore product during
  * system restart
  *
  * @param product The saved product to restore.
  */

  public void setPersistentProduct(Product p) {
    // log(getId() + " setPersistentProduct " + p);
    if (p instanceof VectorProduct) {
    Enumeration e = (Enumeration) ((VectorProduct)p).getContent();
    while (e.hasMoreElements()) {
	BasicProduct bp = (BasicProduct) e.nextElement();
        PropertyAlarm alarm = new PropertyAlarm(bp.getId(),
				Long.parseLong(bp.getDescription()),
				bp.getPropertyValues("alarmfact").getVectorOfElements());
	alarms.insert(alarm);
/*
	log("Product " + bp); log("Alarm " + alarm);
	System.out.println("Product " + bp);System.out.println("Alarm " + alarm);
	System.out.println("  VV " + bp.getPropertyValues("alarmfact"));
*/
    }
   }
  }

  /**
  * Identifies the persistent product for this agent, this
  * is the method used by the exec to find out what to save
  * on a checkpoint and/or system shutdown.
  *
  * @return a vector of alarms to save
  */

  public Product getPersistentProduct() {

    VectorProduct container = new VectorProduct(getId(), getDescription());
    Enumeration e = alarms.elements();
    while (e.hasMoreElements()) {
        PropertyAlarm alarm = (PropertyAlarm) e.nextElement();
	BasicProduct bp = new BasicProduct(alarm.name, Long.toString(alarm.time));
	ValueList list = new ValueList(alarm.values);
	bp.setPropertyValues("alarmfact", list);
	container.addElement(bp);
    }
    return (container.size() == 0) ? null : container;
  }

/*

  // DEBUG method + code to test this class

  public void addTimer(String name, long time, String value) {
        Vector v = new Vector(1);
	v.addElement(value);
	addTimer(name, time, v);
  }

  public static void main (String argv[]) {
    new org.mitre.pam.executive.Exec();

    PropertyTimer prop = new PropertyTimer();

    long current = System.currentTimeMillis();

    prop.addTimer("300",  current + 300,  "300");
    prop.addTimer("250",  current + 250,  "250");
    prop.addTimer("500",  current + 500,  "500");
    prop.addTimer("300",  current + 355,  "350x");
    prop.addTimer("1500", current + 1500, "1500");
    prop.addTimer("500",  current + 550,  "550x");
    prop.addTimer("1000", current + 1000, "1000");
    prop.addTimer("-500", current - 500,  "-500");

    while (prop.alarms.size() != 0) {
      prop.doDaemonTask();
      try {
	Thread.sleep(100);
      } catch (Exception e) {}
    }

    System.exit(0);
  }
*/

}
