/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon.rules;

import org.mitre.pam.daemon.rules.PropertiesTable;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.ExecToEditor;
import org.mitre.pam.interfaces.Sink;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.interfaces.ConfigurationListener;
import org.mitre.pam.interfaces.ConfigurationEvent;
import org.mitre.pam.interfaces.MementoEvent;
import org.mitre.pam.interfaces.MementoListener;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pdif.eventParser.ValueList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import jess.Clearable;
import jess.Context;
import jess.Fact;
import jess.Funcall;
import jess.RU;
import jess.Resetable;
import jess.Rete;
import jess.ReteException;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;

/**
* Assert the properties of a PAM agent. Part are asserted as a single unordered facts. The
* rest of the properties are asserted as individual ordered facts, which can be named by
* a given string (default name is "pamAgent3").
*
* Usage:
*  (assertAgent)
*  (assertAgent <agent-id> [ordered-fact-label])
*
* No parameters to function call will define the fact template if it has not already been defined
* and list all asserted agents to the console.
* Calling assertAgent with an agent-Id will add that agent to the assert list.
*
* @author Nazario Irizarry, Jr.
* @author Jason Mathews Added ordered-fact label + list of asserted agents for no params
*/

class AssertAgent implements Userfunction, MementoListener,
         Clearable, Resetable, ConfigurationListener
{
	SyncSemaphore triggerObject;
	boolean templateIsBuilt;
	Reporter reporter;
	ExecToEditor exec;
	/**
	* if true means that we accept all new connections of specified kind
	*/
	Hashtable acceptableKinds = new Hashtable(101);

	private boolean acceptAll;
		
        private static final String DEFAULT_ORDERED_FACT_NAME = "pamAgent3";
	
	/*
	* Stores references to FactTables indexed on Agent
	*/
	private Hashtable masterFactTable = new Hashtable(101);

	/* 
	* these are the properties that are part of the defTamplate
	*/
	private static final String[] ignoreList =
		{"mementoKind","mementoUid","javaClass","kind","id"};
	private static final String[] alwaysList =
		{"kind","id","enabled","description","creator","creatorEmailAddress"};

	public static Hashtable ignore;
	public static Hashtable always;
	
	Rete m_engine;

	AssertAgent(ExecToEditor exec, Reporter reporter, SyncSemaphore triggerObject) {
		this.triggerObject = triggerObject;
		this.reporter = reporter;
		this.exec = exec;
		
		if (ignore == null) {
			ignore = buildHashTable(ignoreList);
		}
		if (always == null) {
			always = buildHashTable(alwaysList);
		}
		exec.addConfigurationListener(this);

	}

	public String name() { 
		return "assertAgent"; 
	}
	
	void buildTemplate(Context context) throws ReteException{
		if (templateIsBuilt)
			return;
			
	    StringBuffer dt = new StringBuffer("(deftemplate pamAgent ");
	    dt.append(" \n\"$PAM-AGENT$ " + "\"\n ");

	    // Make all the readable memento properties into slots
	    Enumeration st = always.elements();
	    
	    while (st.hasMoreElements()) {
	        dt.append("(slot ");
	        dt.append((String)st.nextElement());
	        dt.append(")\n");
		}
	    dt.append(" )");
		context.engine().executeCommand(dt.toString());
		templateIsBuilt = true;
	}

	public void clear() { 
		Enumeration enum = masterFactTable.keys();
		while (enum.hasMoreElements()) {
			Agent agent = (Agent) enum.nextElement();
			agent.getMemento().removeMementoListener(this);
		}
		masterFactTable.clear(); 
		templateIsBuilt = false;
		acceptAll = false;
		acceptableKinds.clear();
	}
	
	public Value retractAgent(Agent mem) throws ReteException {
		PropertiesTable factTable = (PropertiesTable) masterFactTable.remove(mem);
		Enumeration props = factTable.elements();
		while (props.hasMoreElements()) {
			Fact f = (Fact) props.nextElement();
			m_engine.retract(f.factData());
		}
		m_engine.retract(factTable.getRootFact().factData());
		mem.getMemento().removeMementoListener(this);
		return Funcall.TRUE();
	}
	
	public PropertiesTable getPropertiesTable(Agent agent) {
		return (PropertiesTable) masterFactTable.get(agent);
	}

	public boolean containsAgent(Agent agent) {
		return ( masterFactTable.get(agent) != null);
	}
	
	public void reset() throws ReteException
	{
		Enumeration e = masterFactTable.keys();
		while (e.hasMoreElements())
		{
			Agent agent = (Agent) e.nextElement();

			PropertiesTable factTable = getPropertiesTable(agent);

			factTable.setRootFact(null);  // so we don't get an integrity error
			createFact(agent, factTable);
			changeOrderedFacts(agent, factTable);
		}
	}

	Value assertOneAgent(Agent agent, Context context) throws ReteException{
		return assertOneAgent(agent, null, context);
	}

	Value assertOneAgent(Agent agent, String orderedFactName, Context context) throws ReteException{

  		if (!templateIsBuilt && context != null)
  			buildTemplate(context);
  			
  		return assertOneAgent(agent, orderedFactName, false);
  	}

	Value assertOneAgent(Agent agent, boolean specialFlag) throws ReteException {
		return assertOneAgent(agent, null, specialFlag);
	}

	Value assertOneAgent(Agent agent, String orderedFactName, boolean specialFlag) throws ReteException {

        if (masterFactTable.get(agent) != null)
          return Funcall.FALSE();

		PropertiesTable factTable = new PropertiesTable(31, agent);
		factTable.setSpecialAttributes(specialFlag);

	  if (orderedFactName != null && orderedFactName.length() > 0)
	    factTable.setOrderedFactName(orderedFactName);

		masterFactTable.put(agent, factTable);
		
        createFact(agent, factTable);
        changeOrderedFacts(agent, factTable);
     /*
     * iterate looking for connections we care about
     */
        if (masterFactTable.size() > 1) {
        	if (agent instanceof Sink) {
        		Enumeration sources = ((ConnectedAgent)agent).getSourceList().elements();
        		while(sources.hasMoreElements()) {
        			Agent aSource = (Agent) sources.nextElement();
        			if (containsAgent(aSource)) {
        				assertConnection(aSource, agent);
        			}
        		}
        	}
        	if (agent instanceof Source) {
        		Enumeration sinks = ((ConnectedAgent)agent).getSinkList().elements();
        		while(sinks.hasMoreElements()) {
        			Agent aSink = (Agent) sinks.nextElement();
        			if (containsAgent(aSink)) {
        				assertConnection(agent, aSink);
        			}
        		}
        	}
        }

        // Register ourselves to receive change events

		agent.getMemento().addMementoListener(this);
		
        return Funcall.TRUE();
	}

/**
* Assert facts for one connection
*/
	void assertConnection(Agent from, Agent to) throws ReteException{
        Fact fact = new Fact("pamConnection", RU.ORDERED_FACT, m_engine);
        fact.addValue(new Value(from.getId(),RU.STRING));
        fact.addValue(new Value(to.getId(),RU.STRING));
        m_engine.assert(fact.factData());
	}

/**
* Assert facts for one connection
*/
	void retractConnection(Agent from, Agent to) throws ReteException{
        Fact fact = new Fact("pamConnection", RU.ORDERED_FACT, m_engine);
        fact.addValue(new Value(from.getId(),RU.STRING));
        fact.addValue(new Value(to.getId(),RU.STRING));
        m_engine.retract(fact.factData());
	}

/*
* SYNTAX: (assertAgent <pam-agent-id> ) - aserts the agent
* SYNTAX: (assertAgent)  - builds the template only
*/
  public Value call(ValueVector vv, Context context) throws ReteException
  {
		return callSpecial(vv, context, false);
  }
  
  Value callSpecial(ValueVector vv, Context context, boolean specialFlag) throws ReteException
  {
      if (vv.size() <= 1) {
	  		if (!templateIsBuilt)
	  			buildTemplate(context);

	  Enumeration enum = masterFactTable.keys();
	  while (enum.hasMoreElements()) {
	      Agent agent = (Agent) enum.nextElement();
	      System.out.println("assertAgent " + agent);
	  }

	  return Funcall.TRUE();
      }
 
	        String agentId = vv.get(1).stringValue();
	
      // optional argument: factName defines the atom header of ordered facts
      String factName;
      if (vv.size() == 3)
	  factName = vv.get(2).stringValue();
      else
	  factName = null; // use default factname
	        
	        Agent agent =  exec.getActiveAgent(agentId);
	
	        // Make sure agent exists
	        if (agent == null)
	          return Funcall.FALSE();
	
      return assertOneAgent(agent, factName, context);
  }
  
  private synchronized void createFact(
  							Agent agent, 
  							PropertiesTable factTable) 
       throws ReteException
	{
        Fact fact = factTable.getRootFact();
        if (fact != null) {
        	throw new ReteException("createFact","fact already exists","");
	    }

        fact = new Fact("pamAgent", RU.UNORDERED_FACT, m_engine);
        factTable.setRootFact(fact);
        
        Enumeration st = always.elements();
        
        while (st.hasMoreElements()){
            String name = (String) st.nextElement();

            Value newV;
            
            if (name.equals("kind")) {
//            	newV = new Value(RU.putAtom(agent.getKind()), RU.ATOM); // Jess 4.0
            	newV = new Value(agent.getKind(), RU.ATOM);
            }
            else if (name.equals("id")) {
            	newV = new Value(agent.getId(), RU.STRING);
            }
            else {
				String val = agent.getProperty(name);
				if (val == null || val.length()==0)
					val = "";
				newV = new Value(val, RU.STRING);
	        }
            fact.addValue(name, newV);
        }
        m_engine.assert(fact.factData());
	}

/*
* @param mem the memento
* @param changedName the name of the single property to change
*/
  synchronized void changeTemplateFact(
  							Agent agent, 
  							String name,
  							String newV,
  							PropertiesTable factTable) 
       throws ReteException
	{
        Fact fact = factTable.getRootFact();
        if (fact == null) {
        	throw new ReteException("changeTemplateFact","missing root fact","");
	    }
        
        if (newV == null || newV.length() == 0) {
	        Value oldV = fact.findValue(name);
	        if (!oldV.equals("")) {
	            fact.addValue(name, new Value("", RU.STRING));
				m_engine.assert(fact.factData());
			}
        }
        else {
	        Value oldV = fact.findValue(name);
	        if (!oldV.equals(newV)) {
	            fact.addValue(name, new Value(newV, RU.STRING));
				m_engine.assert(fact.factData());
			}
        }

	}

	void changeOrderedFacts( Agent agent, PropertiesTable factTable) throws ReteException {
        Enumeration enum = agent.getProperties();
        
        while (enum.hasMoreElements()){
            String property = (String)enum.nextElement();

	    if (containsWord(ignore, property)  ||
		containsWord(always, property) ||
		// ignore don't care attributes by prefix
		// ignore meta_ properties unless prefix is meta_jess
		(property.startsWith("meta_") && !property.startsWith("meta_jess")) ||
		// ignore bert_ properties
		property.startsWith("bert_")) {
              continue;
	    }            
            
			ValueList val = agent.getPropertyValues(property);
			
			changeOrderedFact(agent, property, val, factTable);
        }
	}
	
	static boolean containsWord(Hashtable list, String word) {
		return list.containsKey(word);
	}
	
	private void addValueToFact(Fact f, String value) throws ReteException {
		try {
			f.addValue(Integer.parseInt(value), RU.INTEGER);
		}
		catch (NumberFormatException nfe) {
			try {
				double rval = Long.parseLong(value);
				f.addValue(rval, RU.FLOAT);
			}
			catch (NumberFormatException nfe2) {
				f.addValue(value, RU.STRING);
			}						
		}
	}
	
	void changeOrderedFact  (
			Agent agent, 
			String property, 
			Object newValue,
			PropertiesTable factTable) throws ReteException
	{
		changeOrderedFact(agent, property, newValue, factTable, false);
	}

	void changeOrderedFact(
			Agent agent, 
			String property, 
			Object newValue,
			PropertiesTable factTable,
			boolean ignoreFirstInMultiValued) throws ReteException
	{
		Fact existingFact = (Fact) factTable.get(property);
		if (existingFact != null) {
			m_engine.retract(existingFact.factData());
		}

		if (newValue == null)
			return;

		String factName = factTable.getOrderedFactName();
		if (factName == null || factName.length()==0)
			factName = DEFAULT_ORDERED_FACT_NAME;

		Fact f = new Fact(factName, RU.ORDERED_FACT, m_engine);
		f.addValue(agent.getId(), RU.STRING);
	//	f.addValue(RU.putAtom(property), RU.ATOM); // Jess 4.0
		f.addValue(property, RU.ATOM);

	/*
	* Single string value
	*/
		if (newValue instanceof String) {
			addValueToFact(f, (String)newValue);
			m_engine.assert(f.factData());
			factTable.put(property, f);
		}
	/*
	* Potentially multiple values
	*/
		else {
			ValueList valueList = (ValueList) newValue;
			if (valueList.size() >= 1) {
			/*
			* Single valued attribute
			*/
				if (valueList.size() == 1) {
					addValueToFact(f, valueList.elementAt(0));
				}
			/*
			* Multi-valued attribute
			*/
				else {
					boolean ignore = ignoreFirstInMultiValued && property.equals("user_");
					for (int i = ignore?1:0; i < valueList.size(); i++) {
						addValueToFact(f, valueList.elementAt(i));										
					}
				}
	
				m_engine.assert(f.factData());
				factTable.put(property, f);
			}
		}
	}

	public synchronized void propertyChange(MementoEvent pce)
	{
		Agent agent = (Agent) exec.getActiveAgent(((Memento) pce.getSource()).getParentId());

		try {
		  	String property = pce.getPropertyName();
		  	if (containsWord(ignore, property))
		  		return;

			PropertiesTable factTable = getPropertiesTable(agent);

		  	if (containsWord(always, property) ) {
		  		changeTemplateFact(agent, property, (String) pce.getNewValue(), factTable);
		  	}
		  	else {
		  		changeOrderedFact(agent, property, pce.getNewValue(), factTable, factTable.hasSpecialAttributes());
		  	}
	  		triggerObject.trigger();  // used to cause jess to run indirectly
		}
		catch (ReteException re) {
		    System.out.println("Async Error: " + re);
		}
	}
/**
* A node has been created 
*/
	public void nodeCreated(ConfigurationEvent evt) {
		Agent agent =  evt.getAgent();
		if (containsAgent(agent))
			return;
		if (acceptAll || acceptableKinds.containsKey(agent.getKind())){
			try {
				assertOneAgent(agent, null);
			}
			catch (ReteException re) {
				reporter.reportException(re);
			}
		}
	}
/**
* Add acceptable kind
*/
	void addAcceptableKind(String kind) {
		acceptableKinds.put(kind, kind);
	}
/**
* Accept all
*/
	void acceptAllNewNodes() {
		acceptAll = true;
	}

/**
* A node deleted
*/
	public void nodeDeleted(ConfigurationEvent evt) {
		Agent agent =  evt.getAgent();
		if (!containsAgent( agent))
			return;
		try {
			retractAgent(agent);
		}
		catch (ReteException e) {
			reporter.reportException(e);
		}
	}

/**
* A connection made
*/
	public void connectionMade(ConfigurationEvent evt) {
		Agent source = evt.getAgentFrom();
		if (!containsAgent(source))
			return;
		Agent sink = evt.getAgentTo();
		if (!containsAgent(sink))
			return;
		try {
			assertConnection(source, sink);
		}
		catch (ReteException e) {
			reporter.reportException(e);
		}
	}

/**
* A connection broken
*/
	public void connectionBroken(ConfigurationEvent evt) {
		Agent source = evt.getAgentFrom();
		if (!containsAgent(source))
			return;
		Agent sink = evt.getAgentTo();
		if (!containsAgent(sink))
			return;
		try {
			retractConnection(source, sink);
		}
		catch (ReteException e) {
			reporter.reportException(e);
		}
	}
/**
* Build a hashtable that contains specified keys
*/
	public Hashtable buildHashTable(String[] items) {
		Hashtable tab = new Hashtable(3*items.length);
		for (int i = 0; i < items.length; i++ ) {
			tab.put(items[i], items[i]);
		}
		return tab;
	}
/**
* Set the rete engine we are connected to
*/
	public void setEngine(Rete engine) {
		m_engine = engine;
	}
}

