/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.*;
import java.text.SimpleDateFormat;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.ProductShell;
import org.mitre.pam.executive.Module;

/**
* Makes a clone of its incoming product and changes the 
* id of the cloned product to a specified format. <p>
*
* Relies on the ProductIDFilter to modify the id's of the cloned product.<p>
* <pre>
* There are two possible formats for the new id: 
* 1) the old id comes first followed by the date in user-specified format
* 2) the date in user-specified format comes first followed by the old id
* </pre>
*
* @author Jones Leung
* @version 1 Jones Leung
* @version 1.1 Jason Mathews Change default date to 4-digit year and
* store this as a static string member.
*/
public class AppendDateToId extends NoSchedBase {	
    public static final String DEBUG_KEY = "appendID";
/**
* If true, old id is added to front of the date to create new id. <br>
* Othwerwise, old id is added to the end of the date to create new id. 
*/
    public static final String ID_AT_FRONT = "idAtFront";
/**
* The SimpleDateFormat pattern string to used to format the date.<br>
* Default is:  " M/d/yyyy h:mm a "
*/
    public static final String DATE_PATTERN = "datePattern";

    public static final String DEFAULT_DATE = " M/d/yyyy h:mm a ";

/**
* This channel is not scheduled.
*/
	public AppendDateToId () {
		setDescription("Prepends or appends a product's id");
//        setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
        setProperty(ID_AT_FRONT, "true");
        setProperty(DATE_PATTERN, DEFAULT_DATE);
		setProperty(Module.LOG_ACTIVITY, "false");

        metaBooleanInfo(ID_AT_FRONT, true,"True if the id should be appended to front, otherwise it is appended to the back");
        metaStringInfo(DATE_PATTERN, DEFAULT_DATE,"The date format, such as: \"M/d/yyyy h:mm a\" ");
	}
/**
* Create a clone of the incoming product and change the id's
*/
	protected Product buildProduct(Product inProduct) {
	    if (inProduct == null) {     
	        return null;
        }
        
        reporter.debugMsg("incoming product",inProduct,null,null, DEBUG_KEY);
        
        Product deepCopy = inProduct.cloneDeeper();

        ProductIDFilter pif = new ProductIDFilter (
        		getStringProperty(DATE_PATTERN, DEFAULT_DATE),
        		getBooleanProperty(ID_AT_FRONT, true));
        
        deepCopy.apply (pif);
        
        return deepCopy;
	}
}

/**
* This utility class allows users to change the id's on Products so that they
* conform to a certain format.<p>
*
* This filter should receive a deep copy of the Product so that the original
* product is not modified. <p>
*
* @author Jones Leung
*/
class ProductIDFilter implements ProductFilter {
    protected boolean idAtFront;
    protected SimpleDateFormat formatter = null;
/**
* Constructor for ProductIDFilter.<p>
* @param pattern    the SimpleDateFormat pattern string to use
* @param id         whether to add the original id to the front or end of the new id
*/
    public ProductIDFilter (String pattern, boolean idAtFront) {
        this.idAtFront = idAtFront;        
        formatter = new SimpleDateFormat (pattern);
    }    
/**
* Changes id's on Products to make them match a specified format.<p>
*/
    public boolean filter (ContainerProduct parent, Product productToTest) {       
        String oldId = productToTest.getId();
        Date modificationDate = productToTest.getModificationDate();
                
        String newId = oldId;
        if (idAtFront) {
            // add the old id string to the front
            newId = oldId + formatter.format(modificationDate);
        } else {
            // add the old id string to the end
            newId = formatter.format(modificationDate) + oldId;
        }
        
        // change the product's id
        ((ProductShell)productToTest).setId (newId);
        
        // do not stop iteration until we have gone through the entire product
        return true;  
    }
}