/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.StringTokenizer;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pdif.eventParser.ValueList;

/* ExtractProperties walks the incoming product and examines
 * any basic products with string content for name-value pairs
 * from which to add to the product's properties.
 * <P>
 * The output of this filter can be further filtered and processed
 * by other agents accordingly (e.g. passIfProductAttributesMatch).
 * <P>
 * For example, if the product has its content as "location: boston" then the
 * property "location" is added to the product with the value "boston".
 * <P>
 * Changing the key-value separator to "=" and line deliminator to ";" with the input of
 * "location=boston;foo=bar" would extract two name-value pairs: 1) location=boston, 2) foo=bar.
 *
 * @author Jason Mathews
 * @since iPAM 2.4.90 (11/01/99)
 * @see passIfProductAttributesMatch
 */

public class ExtractProperties extends NoSchedBase {

  protected static String DEF_KEYVALUE_SEPARATOR = ":";
  protected static String DEF_LINE_DELIM = "\r\n";
  protected static String DEF_VALUE_DELIM = "|";

  public static final String KEYVALUE_SEPARATOR = "keyValueSeparator";
  public static final String LINE_DELIM = "lineSeparator";
  public static final String VALUE_DELIM = "valueSeparator";

  private static final String DEBUG_KEY = "ExtractProperties";

  public ExtractProperties() {
    setDescription( "Extracts name/value pairs from a product's string content and sets them to its properties" );
    setProperty(LINE_DELIM, DEF_LINE_DELIM);
    setProperty(KEYVALUE_SEPARATOR, DEF_KEYVALUE_SEPARATOR);
    setProperty(VALUE_DELIM, DEF_VALUE_DELIM);
    metaStringInfo(LINE_DELIM, DEF_LINE_DELIM, "The characters in the delim string are the delimiters for separating lines.");
    metaStringInfo(KEYVALUE_SEPARATOR, DEF_KEYVALUE_SEPARATOR, "String that the separates key name from its value.");
    metaStringInfo(VALUE_DELIM, DEF_VALUE_DELIM, "The characters in the delim string are the delimiters for separating multiple values.");
  }

  protected Product buildProduct(Product inProduct) {

    if (inProduct == null)
      return null;

    Product p = inProduct.cloneDeeper();

    Filter filter = new Filter(getStringProperty(LINE_DELIM, DEF_LINE_DELIM),
			getStringProperty(KEYVALUE_SEPARATOR, DEF_KEYVALUE_SEPARATOR),
			getStringProperty(VALUE_DELIM, DEF_VALUE_DELIM), reporter);
    p.apply(filter);

    return p;
  }

  class Filter implements ProductFilter {
    Reporter reporter;
    String lineDelim;
    String keyValueSeparator;
    String valueDelim;
		
    Filter(String lineDelim, String keyValueSeparator, String valueDelim, Reporter reporter ) {
      this.lineDelim = lineDelim;
      this.keyValueSeparator = keyValueSeparator;
      this.valueDelim = valueDelim;
      this.reporter = reporter;
    }

    private void processLine(String line, Product p) {
      StringTokenizer st = new StringTokenizer(line, lineDelim);
      while (st.hasMoreElements()) {
	/*  
	 * Key/value pairs should be separated by a separator string (default=':'),
	 * but there may also be some extra white space thrown in. If a key has multiple
	 * values then those values may be separated by a value separator (default='|').
	 */
	String tok = st.nextToken();
	int pos = tok.indexOf(keyValueSeparator);
	if (pos < 1) continue; // not a key value pair - ignore this line
	String key = tok.substring(0,pos).trim();
	String value = tok.substring(pos+1).trim();
	reporter.debugMsg("found key=", key, ", value=", value, DEBUG_KEY);
	if (value.length() != 0) {
	  StringTokenizer keyValue = new StringTokenizer(value, valueDelim);
	  if (keyValue.hasMoreElements()) {
	    ValueList vl = new ValueList(keyValue);
	    p.setPropertyValues(key, vl);
	    reporter.debugMsg("set property: key=", key, " value=", vl, DEBUG_KEY);
	  }
	}
      }
    }

    public boolean filter ( ContainerProduct parent, Product p ) {
      if ( p instanceof BasicProduct ) {
	Object content = p.getContent();
	if (content instanceof String)
	  processLine((String) content, p);
      }

      return true; // continue search
    }
  }

}
