/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.*;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.product.view.BasicProduct;
//import org.mitre.pam.util.alarm.*;

/**
* This Juggler: Passes through a product with everything but the headline stripped off.
* Optionally, the user may choose to leave the description up to a maximum length
* and/or leave url content.
*
* @author A. Kazura
*/
public class HeadlinesOnly extends NoSchedBase {

  public static String KEEPING_URL = "keepURLs";
  public static String MAX_DESCRIPTION_LENGTH = "maxDescriptionLength";

  /**
  * A ProductFilter that removes the content and descriptions from the product
  * as specified by the user
	*/
   class FilterAllButHeadline implements ProductFilter{
      int maxDescriptionLength;
      boolean keepingURL;

      FilterAllButHeadline ( int maxDescriptionLength, boolean keepingURL ) {
         this.maxDescriptionLength = maxDescriptionLength;
         this.keepingURL = keepingURL;
      }

      public boolean filter ( ContainerProduct parent, Product p ) {
         if ( p instanceof ContainerProduct ) {
            return true; // continue search
         }
         if ( p instanceof BasicProduct ) {
            BasicProduct bp = (BasicProduct) p;
            if ( bp.getDescription().length() > maxDescriptionLength ) {//
               bp.setDescription( "" );
            }
            if ( !( bp.getContent() instanceof java.net.URL && keepingURL ) ) {
               bp.setContent( "" );
            }
         }
			return true; // continue search
		}

   }    // end of class   FilterAllButHeadline

/**
* Removes duplicate headlines and notes on the first instance of the repeated
* headline that more than one occurrence was found
*/
   class FilterOutDuplicates implements ProductFilter {

      boolean keepingURL;
      Hashtable unique = new Hashtable();

      FilterOutDuplicates ( boolean keepingURL ) {
         this.keepingURL = keepingURL;
      }

      public boolean filter ( ContainerProduct parent, Product p ) {
         if ( p instanceof ContainerProduct ) {
            return true; // continue search
         }
         if ( p instanceof BasicProduct ) {
            BasicProduct bp = (BasicProduct) p;
            String productId = p.getId();
           // check if we already have an instance of this headline
           // if not, save it
            if ( !unique.containsKey( productId ) )   {
               unique.put( productId, bp );
            }
            // else change description to reflect there are multiples
            else {
               BasicProduct uniqueProduct = (BasicProduct) unique.get( productId );
               parent.removeElement( p );
               String d = keepingURL ?
                     "* Multiple instances of this product id exist, first URL is:" :
                     "* Multiple instances of this product id exist *";
               uniqueProduct.setDescription( d );
            }
         }
         return true; // continue search
		}

   }    // end of class FilterOutDuplicates


/**
* This channel is not scheduled.
*/
  public HeadlinesOnly () {
		setDescription( "Removes everything but the product's headline" );
		setProperty(KEEPING_URL, "false");
		setProperty(MAX_DESCRIPTION_LENGTH, "0" );

		metaBooleanInfo(KEEPING_URL, false, "Include URLs in the output");
		metaIntegerInfo(MAX_DESCRIPTION_LENGTH, 0,"Maximum size of the output" );
  }

/**
* The product is built by stripping off each product's description and content (default),
* or optionally leaving the description up to a maximum length specified by the user
* and/or optionally leaving url content.
*/
	protected Product buildProduct(Product jugglersProduct) {

      // if product is null, do nothing
      if ( jugglersProduct == null ) {
         return null;
      }

      Product p = jugglersProduct.cloneDeeper();
      boolean keepingURL = false;
      keepingURL = getBooleanProperty( KEEPING_URL, false );
      p.apply( new FilterAllButHeadline( getMaxDescriptionLength(), keepingURL ) );
      p.apply( new FilterOutDuplicates( keepingURL ) );
      return p;
   }


/**
* Returns the maximum length for the description specified by user (or zero as default).
*/
   public int getMaxDescriptionLength () {
      int maxDescriptionLength = 0;
      String max = getProperty( MAX_DESCRIPTION_LENGTH );
      if ( max != null ) {
         try {
            maxDescriptionLength = new Integer( max ).intValue();
         }
         catch ( NumberFormatException ex ) {
         }
      }
      return maxDescriptionLength;
   }

 /*
*/

}