/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Vector;

import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.interfaces.ContainerProduct;

/**
* A Juggler that builds an aggregate product and passes it on only
* if any of the String product content matches any/all of the given keywords, which
* are delimited by whitespace or some specified string.
*
* keys field specifies a list of keywords (e.g. "foo bar test")
* method specifies the match method ("ANY" -> must match any keyword [default]
*				     "ALL" -> must match all keywords)
* delimiter specifies the separation string between keywords (default=whitespace)
*
* 1998 The MITRE Corporation
*
* @author Jason Mathews
*/
public class KeyMatch extends NoSchedBase {

  private String delimiter = " ";
  private int method = MATCH_ANY;
  private String keyList[];
  private boolean found[];

  private static final int MATCH_ANY = 0;
  private static final int MATCH_ALL = 1;

  private static final String KEYS = "keys";
  private static final String KEY_WORDS = "keyWords";
  private static final String METHOD = "method";
  private static final String DELIMITER = "delimiter";
  private static final String[] DEFAULT_KEYS = {"keyWord1", "keyWord2", "keyWordn"};
  private static final String[] METHODS = {"any", "all"};
  private static final String DEBUG_KEY = "KeyMatch";

  public KeyMatch () {
    setDescription("Passes aggregate product if it matches a list of keywords");
    
    ValueList vl = new ValueList(DEFAULT_KEYS);
//    setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
    setPropertyValues(KEY_WORDS, vl);
    setProperty(METHOD, "any" );

    metaMultiStringInfo(KEY_WORDS, vl, "keys separated by spaces");
    metaEnumerationInfo(METHOD, "any", "how key words are used", METHODS );
  }


  public Product buildProduct(Product prod) {

    if (prod == null) 
      return null;

	return isMatched(prod) ? prod : null;
  }

  // Test whether a given product matches any or all of the keywords
  // Returns true immediately if any match is required otherwise continues
  // to check all products against all keywords

	protected boolean isMatched (Product productToTest) {
	
	/*
	* catch old style of keywords with space delimter
	*/
	    String keys = getStringProperty(KEYS,"");
	    if (keys.length() > 0) {
			
		    String delimeter = getStringProperty(DELIMITER," ");
	
		    StringTokenizer st = new StringTokenizer(keys, delimiter);
		    int n = st.countTokens();
		
		    if (n == 0) 
		    	return false;
		
		    Vector vKeys = new Vector(n);
		    for (int i = 0; i < n && st.hasMoreElements(); i++) {
				vKeys.addElement( st.nextElement());
		    }
		    removeProperty(KEYS);
		    removeProperty(DELIMITER);
		    setPropertyValues(KEY_WORDS, new ValueList(vKeys));
		}

	/*
	* the newer style has keywords in a multvalued list
	*/
		Vector keyWords = getPropertyValues(KEY_WORDS).getVectorOfElements();
		int nWords = keyWords.size();
		if (nWords == 0)
			return false;
	
	    keyList = new String[nWords];
	    found = new boolean[nWords];
	    for (int i = 0; i < nWords; i++) {
			keyList[i] = (String) keyWords.elementAt(i);
	    }
	
		String methProp = getStringProperty(METHOD,"any");
		int method;
		if (methProp.equalsIgnoreCase("any"))
			method = MATCH_ANY; // default match method
		else
			method = MATCH_ALL;
		
		return visit(productToTest, (method == MATCH_ANY) ? 1 : keyList.length);
  }


  /**
   * Recursively visit product and inner products to determine if they match given keywords
   */

	private boolean visit (Product productToTest, int matchesNeeded) {
		
		reporter.debugMsg("looking for matches ",null,null,null, DEBUG_KEY); 
		if (productToTest instanceof ContainerProduct) {
			// get the enumeration of products inside VectorProduct
			ContainerProduct vp = (ContainerProduct) productToTest;
			Enumeration enum = (Enumeration) vp.getContent();
			
			// check content on each of the top-level inner products
			while (enum.hasMoreElements()) {	
				Product element = (Product) enum.nextElement();
				if (visit(element, matchesNeeded))
					return true; // match found -> done
			}
		} else if (productToTest instanceof BasicProduct) {
			Object content = productToTest.getContent();
			if (!(content instanceof String)) {
				return false;
			}
			int n = keyList.length;
			String s = (String)content;
			for (int i=0; i < n; i++) {
				if (!found[i] && s.indexOf(keyList[i]) >= 0) {
					found[i] = true;
					reporter.debugMsg(" matched '",keyList[i],"'",null, DEBUG_KEY); // DEBUG
					if (--matchesNeeded == 0) {
						return true;
					}
				}
			}
		} else {
			// don't know how to handle this product type
			reporter.assert(false, "Unknown Product type encountered.",productToTest,null,null);
		}
		
		return false;
	}


}
