/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.Enumeration;
import java.util.Hashtable;
import java.io.StringWriter;
import java.io.PrintWriter;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.productVisitor.AsTree;
import org.mitre.pam.productVisitor.AsPlainText;
import org.mitre.pam.util.PrintLFWriter;
import org.mitre.pdif.eventParser.ValueList;

/**
 * A Juggler that builds a notification product from its inputs.
 *
 * @author Jason Mathews
 */
public class NotifyOfProduct extends NoSchedBase {
  /** Id of product for notification of available product */
  protected static final String NOTIFY_TITLE = "notificationId";

  /** Size limit for product content (listing or summary) in the number of bytes */
  protected static final String PRODUCT_LIMIT = "maxProductContentSize";

  /** Flag to show content type and/or element size of product */
  protected static final String SHOW_PRODUCT_SUMMARY = "showProductSummary";

  /** Flag to show product with asPlainText viewer and store this list in the juggler's
   *  product content. */
  protected static final String SHOW_PRODUCT_LIST = "showProductList";

  /** List of product properties to show (e.g. id, description) **/
  protected static final String PROPERTY_LIST = "propertyList";

	private static final String[] INIT_PROPERTIES = {"id"};
	
  public NotifyOfProduct () {
    setDescription("Passes notification of a product if available");
    setProperty(NOTIFY_TITLE, "");
    setProperty(PRODUCT_LIMIT, "80");
    setProperty(SHOW_PRODUCT_SUMMARY, "false");
    setProperty(SHOW_PRODUCT_LIST, "false");
    
    ValueList props = new ValueList(INIT_PROPERTIES);
    setPropertyValues(PROPERTY_LIST, props);

    metaStringInfo(NOTIFY_TITLE, "","id to use on the delivered product which may override the input products id");
    metaIntegerInfo(PRODUCT_LIMIT, 80, "maximum number of lines in the notification that is generated");
    metaBooleanInfo(SHOW_PRODUCT_SUMMARY, false,"show/append original product summary?");
    metaBooleanInfo(SHOW_PRODUCT_LIST, false,"");
    metaMultiStringInfo(PROPERTY_LIST, props,"list of product properties or meta-data to include in the output, such as id");
  }

  protected Product buildProduct(Product jugglersProduct) {

    // if product is null then we're done
    if (jugglersProduct == null) {
      return null;
    }

    Hashtable hash = new Hashtable();
    ValueList list = getPropertyValues(PROPERTY_LIST);
    String key;
    boolean showProductId = false;
    if (list != null && list.size() != 0) {
      for (Enumeration e=list.elements(); e.hasMoreElements(); ) {
	key = (String) e.nextElement();
	if (key.equals("id"))
		showProductId = true;
	else hash.put(key, key);
      }
    }

    String id = getStringProperty(NOTIFY_TITLE,"");
    if (id.length()==0) 
    	id = jugglersProduct.getId();
    else if (showProductId)
      id = id + " - " + jugglersProduct.getId();

    BasicProduct bp = new BasicProduct(id, getDescription());
    StringBuffer buf = new StringBuffer();

    // check product's properties and show any requested attributes
    if (hash.size() != 0) {
      for (Enumeration e=jugglersProduct.getProperties();
	   e.hasMoreElements();) {
	key = (String) e.nextElement();
	if (hash.containsKey(key)) {
	  buf.append(key);
	  buf.append(": ");
	  buf.append(jugglersProduct.getStringProperty(key,""));
	  buf.append('\n');
	}
      }
    }

    int limit = getIntegerProperty(PRODUCT_LIMIT,-1);

    if (getBooleanProperty(SHOW_PRODUCT_LIST,false)) {
      AsPlainText apt = new AsPlainText();
      StringWriter sw = new StringWriter();
      apt.visit(jugglersProduct);
      apt.writeResults(new PrintLFWriter(sw));
      buf.append(sw.toString());
    }
    else if (getBooleanProperty(SHOW_PRODUCT_SUMMARY,false)) {
      if (jugglersProduct instanceof ContainerProduct) {
	AsTree asTree = new AsTree();
	asTree.visit(jugglersProduct);
	buf.append("Elements: ");
	buf.append( asTree.getLeafCount() );
      } else {
	buf.append("Content: ");
	buf.append(jugglersProduct.getContent());
      }
      buf.append('\n');
    }
    
    if ( limit != 0 ) {
      if (buf.length() > limit && limit > 0) {
         if ( limit > 3 ) {
          buf.setLength(limit - 3);
          buf.append("...");
        }  else
          buf.setLength(limit);
      }
    bp.setContent(buf.toString());
    }

    return bp;
  }

}
