/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;
import java.util.Enumeration;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pdif.eventParser.ValueList;

/**
 * This juggler passes products if the product has attributes that
 * match certain specified attributes.
 * 
 * @author Michael DeCorpo 7/21/98
 * @author Jones Leung, changed description and commented out default property,
 * updated so that it uses a ProductFilter to check inner-products instead of
 * just the outer-product.
 * @author Jason Mathews allow filter to handle HashtableProducts,
 * combine filter match tests into a single function productMatch().
 * 1/21/00 Added or-attribute test to compare a string property against a list of possible values.
 * 1/30/00 Added not-attribute test to reject products having a given property value.
 */
public class PassProductAttributes extends NoSchedBase {
	public final static String      ATTRIBUTE_PREFIX = "productAttribute_";
	public final static String      ATTRIBUTE_OR_PREFIX = "productOrAttribute_";
	public final static String      ATTRIBUTE_NOT_PREFIX = "productNotAttribute_";

	public PassProductAttributes() {
		setDescription("Passes products with matching attributes of the form: product[Or|Not]Attribute_ + attrname, attrvalue.");
		// setProperty(ATTRIBUTE_PREFIX+"xxx","yyy");
		// setProperty(ATTRIBUTE_OR_PREFIX+"xxx","yyy");
		// setProperty(ATTRIBUTE_NOT_PREFIX+"xxx","yyy");
		// setProperty(META_GUI_GROUP, "New Filter");
		setProperty(META_ATTRIBUTE_ADD_PREFIX + ATTRIBUTE_PREFIX + "newAttribute", "Product Equals Attribute");
		setProperty(META_ATTRIBUTE_ADD_PREFIX + ATTRIBUTE_OR_PREFIX + "newAttribute", "Product OR Attribute");
		setProperty(META_ATTRIBUTE_ADD_PREFIX + ATTRIBUTE_NOT_PREFIX + "newAttribute", 
			    "Product Not equals Attribute");
	}

	/**
	 * Go through the juggler's properties one by one looking for ones that
	 * begin with the value of ATTRIBUTE_PREFIX, ATTRIBUTE_OR_PREFIX, or
	 * ATTRIBUTE_NOT_PREFIX. If we find one then the product's corresponding
	 * attribute and values are checked against that in the filter.
	 * A positive match returns true and a negative match return false.
	 */
	protected boolean productMatch(Product productToTest) {
		Enumeration     e = getProperties();

		while (e.hasMoreElements()) {
			String  prop = (String) e.nextElement();

			if (prop.startsWith(ATTRIBUTE_PREFIX)) {
				// remove the prefix and check the jugglers product
				String  property = prop.substring(ATTRIBUTE_PREFIX.length());
				String  val = getProperty(prop);
				// check if product matches
				String  productsval = productToTest.getProperty(property);

				reporter.debugMsg(property, val, "-->", productsval, getId());
				if (productsval == null ||!productsval.equals(val)) {
					return false;
				}
			}
			else if (prop.startsWith(ATTRIBUTE_OR_PREFIX)) {
				// remove the prefix and check the jugglers product
				String		property = prop.substring(ATTRIBUTE_OR_PREFIX.length());
				ValueList       vals = getPropertyValues(prop);
				// check if product matches
				String		productsval = productToTest.getProperty(property);

				reporter.debugMsg(property, vals, "-->", productsval, getId());
				if (productsval == null) {
					return false;
				}
				boolean		match = false;
				// Check value against each of the possible values and stop when it matches
				// otherwise there is no match.
				Enumeration     props = vals.elements();

				while (props.hasMoreElements()) {
					String  val = (String) props.nextElement();

					if (productsval.equals(val)) {
						match = true;
						break;
					}
				}
				if (!match) {
					return false;
				}
			}
			else if (prop.startsWith(ATTRIBUTE_NOT_PREFIX)) {
				// remove the prefix and check the jugglers product
				String  property = prop.substring(ATTRIBUTE_NOT_PREFIX.length());
				String  val = getProperty(prop);
				// check if product matches
				String  productsval = productToTest.getProperty(property);

				reporter.debugMsg(property, val, "-->", productsval, getId());
				if (productsval != null && productsval.equals(val)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * This filter builds a composite product from the product that its sources
	 * provide and removes all subproducts that do not match any of the
	 * attribute filters.
	 */
	protected Product buildProduct(Product productIn) {
		if (productIn == null) {
			return null;
		}
		Product jugglersProduct = productIn.cloneDeeper();

		// filter out a BasicProduct
		if (jugglersProduct instanceof BasicProduct) {
			if (!productMatch(jugglersProduct)) {
				return null;
			}
			return jugglersProduct;
		}

		/*
		 * Do container products
		 */
		ProductFilter   filter = new filterProducts();

		jugglersProduct.apply(filter);
		if (jugglersProduct instanceof ContainerProduct) {
			ContainerProduct	cp = (ContainerProduct) jugglersProduct;

			if (cp.size() == 0) {
				return null;
			}
			if (cp.size() == 1) {
				Enumeration     e = (Enumeration) cp.getContent();

				return (Product) e.nextElement();
			}
		}
		else {
			assert(false, "Unknown product type", jugglersProduct, null, null);
		}
		return jugglersProduct;
	}

	/**
	 * Filters out products which don't have the correct attribute-value pair.
	 * Only a deep-copy of a Product should be filtered, not the original Product.
	 */
	public class filterProducts implements ProductFilter {
		public boolean filter(ContainerProduct parent, Product productToTest) {
			// this is to get by the test in apply() so that mapply() will be called
			if (parent == null) {
				return true;    // continue
			}
			// go through the jugglers properties 1by1 looking for ones that
			// begin with the value of ATTRIBUTE_PREFIX or ATTRIBUTE_OR_PREFIX.
			// If we find one, the product must have an attribute with this
			// name and the value must be the same.
			if (!productMatch(productToTest)) {
				parent.removeElement(productToTest);
				// do not stop iteration until we have gone through the entire product
			}
			return true;
		}

	}
}
