/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.Hashtable;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pdif.eventParser.ValueList;

/** SubstitutionFilter walks the incoming product and examines
 * any basic products with string content that match target strings
 * in a substitution table and if present will then replace the
 * string content with the replacement string value.
 *<P>
 * This can be used to replace long repeated phrases (such as traffic
 * reports) with terse ones to output to pager or other compact devices.
 *
 * @author Jason Mathews
 * @since iPAM 2.4.90 (11/02/99)
 */

public class ReplaceWords extends NoSchedBase {

  /**
   * The string replace list of target and replacement word or phrase pairs.
   */
  public static final String REPLACE_LIST = "replacementList";
  public static final String IGNORE_WHITESPACE = "ignoreWhitespace";
  public static final String IGNORE_CASE = "ignoreCase";

  private static final String[] DEFAULT_LIST = {
    "There are no problems or delays to report at this time.", "No problems"
  };

  public ReplaceWords() {
    setDescription("Filter to replace known string content text with alternative text");
    ValueList replaceList = new ValueList(DEFAULT_LIST);
    setProperty(IGNORE_WHITESPACE, "false");
    setProperty(IGNORE_CASE, "false");
    setPropertyValues(REPLACE_LIST, replaceList);
    metaBooleanInfo(IGNORE_WHITESPACE, false, "Set to ignore leading and trailing whitespace in string content");
    metaBooleanInfo(IGNORE_CASE, false, "Set to ignore case of string content for matching");
    metaMultiStringInfo(REPLACE_LIST, replaceList,"The string replace list of target and replacement word or phrase pairs.");
  }

  protected Product buildProduct(Product inProduct) {

    if (inProduct == null)
      return null;

    ValueList replaceList = getPropertyValues(REPLACE_LIST);
    if (replaceList.size() == 0 || (replaceList.size() % 2) != 0) {
      reporter.warning("Cannot do any filtering: invalid or missing value for " + REPLACE_LIST);
      return inProduct;
    }

    boolean ignoreWhitespace = getBooleanProperty(IGNORE_WHITESPACE, false);
    boolean ignoreCase = getBooleanProperty(IGNORE_CASE, false);

    // create replacement mapping table
    Hashtable replaceTable = new Hashtable();
    for (int i=0; i < replaceList.size(); i+=2) {
      String target = replaceList.elementAt(i);
      if (ignoreWhitespace) target = target.trim();
      if (ignoreCase) target = target.toLowerCase();
      replaceTable.put(target, replaceList.elementAt(i+1));
    }

    StringReplaceFilter filter = new StringReplaceFilter(replaceTable,
					 ignoreWhitespace, ignoreCase);

    Product p = inProduct.cloneDeeper();
    p.apply(filter);

    return p;
  }

  class StringReplaceFilter implements ProductFilter {

    Hashtable replaceTable = new Hashtable();
    boolean ignoreWhitespace;
    boolean ignoreCase;

    StringReplaceFilter(Hashtable replaceTable, boolean ignoreWhitespace,
			boolean ignoreCase) {
      this.replaceTable = replaceTable;
      this.ignoreWhitespace = ignoreWhitespace;
      this.ignoreCase = ignoreCase;
    }
 		
    public boolean filter ( ContainerProduct parent, Product p ) {
      if ( p instanceof BasicProduct ) {
	Object content = p.getContent();
	if (content instanceof String) {
	  String stringContent = (String) content;
	  if (ignoreWhitespace) stringContent = stringContent.trim();
	  if (ignoreCase) stringContent = stringContent.toLowerCase();
	  Object newContent = replaceTable.get(stringContent); 
	  if (newContent != null) {
	    BasicProduct bp = (BasicProduct)p;
	    bp.setContent((String) newContent);
	  }
	}
      }

      return true; // continue search
    }
  }
}  
