/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator.detectChange;
	
import java.util.*;
import java.io.*;
import java.net.URL;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import org.mitre.pam.aggregator.ChanSchedBase;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.PAMDate;
import org.mitre.pdif.eventParser.ValueList;

/**
 * This agent checks its input content
 * to see if it is different than last time using message digests (i.e. digital signatures)
 *
 * @author Nazario Irizarry, Jr.
 * @author Jason Mathews, rewritten to support String and URL product types
 *			with MessageDigests to verify true content changes
 *			independent of source or product meta data,
 *                      Added clearCache trigger,
 *                      Added support for other ContainerProducts not just VectorProducts.
 */
public class NotifyOfChange2 extends ChanSchedBase 
{
	protected static final String SOMETHING_CHANGED = "somethingChangedMessage";

	protected static final String NOTHING_CHANGED = "nothingChangedMessage";

	protected static final String FETCH_FTP_CONTENTS = "fetchFtpContents";

	protected static final String TIMEOUT = "timeout";

	protected static final String CLEAR_CACHE = "clearCache";

	protected static final String DIGEST_METHOD = "SHA-1";

	protected static final String DEFAULT_TIMEOUT = "5";

	private Product interimProduct;

	private int changeCount;

	private Vector deleteNodes;

	private MessageDigest md = null;

	private String timeout;

	public NotifyOfChange2()
	{
		setPeriod(3600);
		setDescription("Detect changes based on content of FILES, STRINGS, or URL'S.");
		setProperty(SOMETHING_CHANGED,"New or updated content");
		setProperty(NOTHING_CHANGED,"");
		// setProperty(FETCH_FTP_CONTENTS,"false");
		setProperty(CLEAR_CACHE,"false");
		metaStringInfo(SOMETHING_CHANGED,"New or updated content","Text to be used when something has changed");
		metaStringInfo(NOTHING_CHANGED,"","Text to be used when there are no changes");
		metaBooleanInfo(CLEAR_CACHE,false,"Clears cache upon setting to true");
	}

	class DeletePair 
	{
		ContainerProduct container;
		Product product;

		DeletePair(ContainerProduct c, Product p)
		{
			container = c;
			product = p;
		}
	}

	/**
	 * A channel can build and change the product that it's pims provide
	 */
	public Product buildProduct(Product product)
	{
		if(product == null)
			return null;

		if(md == null) {
			try {
			  md = MessageDigest.getInstance(DIGEST_METHOD);
			}
			catch(NoSuchAlgorithmException e) {
			  reportException(e);
			  return null;
			}
		}
		interimProduct = product.cloneDeeper();
		timeout = getStringProperty(TIMEOUT, DEFAULT_TIMEOUT);
		/*
		 * Parse inner products and make list of those products
		 * with unchanged content to delete later.
		 */
		deleteNodes = new Vector();
		changeCount = 0;
		checkProduct(null, interimProduct);
		/*
		 * Remove unused crc entries
		 */
		for (Enumeration e = getProperties(); e.hasMoreElements(); )
		{
			String key = (String) (e.nextElement());
			if(key.startsWith("cache."))
			{
				String value = getProperty(key);
				int usedCount = 0;
				if(value != null && value.length() != 0)
				{
					try
					{
						usedCount = Integer.parseInt(value);
					}
					catch(NumberFormatException nfe)
					{
						reportException(nfe);
					}
				}
				if(--usedCount < 0)
					removeProperty(key);
				else
					setProperty(key, Integer.toString(usedCount));
				}
			}
		if(interimProduct != null && interimProduct instanceof ContainerProduct)
		{
			if(changeCount == 0)
			{
				// all inner products are unchanged so discard the whole thing
				interimProduct = null;
			}
			else
			{
				// remove nodes marked for deletion from the product
				Enumeration enum = deleteNodes.elements();
				while(enum.hasMoreElements())
				{
					DeletePair pair = (DeletePair)(enum.nextElement());
					pair.container.removeElement(pair.product);
				}
				// cleanup vector product removing dead branches
				ContainerProduct vp = (ContainerProduct)interimProduct;
				// System.out.print("ContainerProduct " + vp.size());
				vp.prune();
				// System.out.println(" -> " + vp.size());
				if(vp.size() == 0)
				{
					// System.out.println("Null out empty product");
					interimProduct = null;
				}
				else
					if(vp.size() == 1)
					{
						// System.out.println("Make basic out of single vector element");
					    // interimProduct = vp.elementAt(0);
					    Enumeration e = (Enumeration) vp.getContent();
					    interimProduct = (Product) e.nextElement();
					}
			}
			deleteNodes = null; // discard this list
		}
	/*
	* Nothing changed message
	*/
		if(interimProduct == null)
		{
			reporter.assert(changeCount == 0,"Changed products have been discarded",null,null,null);
			String nothingChanged = getProperty(NOTHING_CHANGED);
			if(nothingChanged != null && nothingChanged.length() != 0)
			{
				return new BasicProduct(nothingChanged,"Checked on " + (new PAMDate()).toString());
			}
			return null;
		}
	/*
	* Something changed message
	*/
		else
		{
			reporter.assert(changeCount != 0,"Composite product has inconsitent change count",null,null,null);
			String somethingChanged = getStringProperty(SOMETHING_CHANGED,"");
			if (somethingChanged.length()==0) {
				return interimProduct;
			}
			else {
				VectorProduct prod = new VectorProduct(somethingChanged,changeCount + " items, checked on " + (new PAMDate()).toString());
				prod.addElement(interimProduct);
				return prod;
			}
		}
	}

	private void checkProduct(ContainerProduct parent, Product p)
	{
		if(p instanceof ContainerProduct)
		{
			Enumeration enum = (Enumeration)p.getContent();
			while(enum.hasMoreElements())
			{
				checkProduct((ContainerProduct)p,(Product)enum.nextElement());
			}
		}
		else
			if(p instanceof BasicProduct)
			{
				if(contentsChanged(p))
				{
					changeCount++;
				}
				else
				{
					if(parent == null)
						interimProduct = null;
					else
						deleteNodes.addElement(new DeletePair(parent,p));
				}
			}
			else
			{
				// don't know how to handle this product type
				reporter.assert(false,"NotifyOfChange: Unknown Product type encountered ",p.getClass(),null,null);
			}
	}

	/**
	 * run a signature on file contents to see
	 * if they have changed
	 */
	protected boolean contentsChanged(Product p)
	{
		String lookupName;
		Object content = p.getContent();
		if(content instanceof FileRef)
		{
			FileRef fileRef = (FileRef)content;
			// System.out.println("FileRef '" + p.getId() + "' : " + fileRef.getSourceName());
			lookupName = computeCRC(fileRef);
		}
		else
			if(content instanceof String)
			{
				// System.out.println("String '" + p.getId() + "' : " + (String)content);
				lookupName = computeCRC((String)content);
			}
			else
				if(content instanceof URL)
				{
					URL uri = (URL)content;
					String url = uri.toString();
					lookupName = computeCRC(url);
				}
				else
				{
					reporter.assert(false,"NotifyOfChange: Unknown content type encountered ",content.getClass(),null,null);
					return true;
				// must assume that contents have changed - no way to tell
				}
		lookupName = "cache." + lookupName;
		/*
		 * check if crc is different
		 */
		String usedCount = getProperty(lookupName);
		setProperty(lookupName, timeout);

		return (usedCount == null);
	}

	/**
	 * Return message digest bytes as a hex string
	 */
	private static String hexDigits(byte[] bytes)
	{
		char c;
		StringBuffer buf = new StringBuffer(40);
		// hold 40-hex digits
		for(int i = 0; i < bytes.length; i++)
		{
			c = (char)((bytes[i] >> 4) & 0xf);
			if(c > 9)
			{
				c = (char)((c - 10) + 'a');
			}
			else
			{
				c = (char)(c + '0');
			}
			buf.append(c);
			// high-order bits
			c = (char)(bytes[i] & 0xf);
			if(c > 9)
			{
				c = (char)((c - 10) + 'a');
			}
			else
			{
				c = (char)(c + '0');
			}
			buf.append(c);
		// low-order bits
		}
		return buf.toString();
	}

	/**
	 * Compute a signature on file
	 */
	protected String computeCRC(FileRef fileRef)
	{
		try
		{
			md.reset();
			byte[] b = new byte[4096];
			FileInputStream in = new FileInputStream(fileRef.getFile());
			while(true)
			{
				int len = in.read(b,0,4096);
				if(len <= 0)
					break;
				md.update(b,0,len);
			}
			in.close();
		}
		catch(IOException e)
		{
			reportException(e);
		}
		return hexDigits(md.digest());
	}

	protected String computeCRC(String s)
	{
		md.reset();
		byte[] b = s.getBytes();
		int len = b.length;
		md.update(b,0,len);
		return hexDigits(md.digest());
	}

/**
* override setProperty and handle cache settings
*/
	public void setProperty(String attribute, String value) {
	  if (attribute.equals(CLEAR_CACHE) &&
	      "true".equalsIgnoreCase(value)) {
	    // clearCache does not actually store the store of true
	    // but triggers the following process of clearing the
	    // cache entries.
	    for (Enumeration e=getProperties(); e.hasMoreElements(); ) {
	      String key = (String) (e.nextElement());
	      if (key.startsWith("cache."))
		removeProperty(key);
	    }
	  }
	  else
	    super.setProperty(attribute, value);
	}
}

