/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator.detectChange;

import java.util.Enumeration;
import java.util.Date;
import java.io.*;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.PAMDate;
import org.mitre.pam.aggregator.ChanSchedBase;
import org.mitre.pam.store.JoinMemento;

/**
* This filter checks its input content
* to see if it is different than last time.
* The modification dates of products are checked.
*
* @author John Fantry, Technical Co-op (D510)
* @author Jason Mathews, Added clearCache trigger + many other changes,
*                        Added support for non-file product types.
*/
public class NotifyOfDateChange extends ChanSchedBase  {

	protected static final String SOMETHING_CHANGED = "somethingChangedMessage";
	protected static final String NOTHING_CHANGED = "nothingChangedMessage";
	protected static final String CLEAR_CACHE = "clearCache";

	private int changeCount;

/*
* Unique channel attributes
*/
	protected SignatureTable signatureCache = new SignatureTable();

	public NotifyOfDateChange () {
		super(false, false);
		stateMemento = new JoinMemento("generic","notifier", signatureCache, reporter);
		initChanSchedBase(true);
		setDescription("Detect changes based on modification date of products.");
		setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
		setProperty(SOMETHING_CHANGED,"New or updated content");
		setProperty(NOTHING_CHANGED,"");
		setProperty(CLEAR_CACHE,"false");

		metaStringInfo(SOMETHING_CHANGED,"New or updated content","Text to be used when something has changed");
		metaStringInfo(NOTHING_CHANGED,"","Text to be used when there are no changes");
		metaBooleanInfo(CLEAR_CACHE,false,"Clears cache upon setting to true");
	}

/**
* A channel can build and change the product that it's pims provide
*/
	protected Product buildProduct(Product productIn) {
		
	    if (productIn == null) {
		return null;
	    }

	    Product jugglersProduct = productIn.cloneDeeper(); /* working here */

	    changeCount = 0; // reset count of changed products

	    // filter out a BasicProduct
	    if (jugglersProduct instanceof BasicProduct) {
		if (!productChanged(jugglersProduct))
		    return null;
		return jugglersProduct;
	    }
	    
	    /*
	     * Do container products
	     */
	    ProductFilter filter = new filterProducts();
	    jugglersProduct.apply (filter);
	    if (jugglersProduct instanceof ContainerProduct) {
	        ContainerProduct cp = (ContainerProduct) jugglersProduct;
	        if (cp.size() == 0) {
	            jugglersProduct = null;
	        }
	        else if (cp.size() == 1) {
		    Enumeration e = (Enumeration) cp.getContent();
		    jugglersProduct = (Product) e.nextElement();
	        }
	    }
	    else {
	    	assert(false, "Unknown product type",jugglersProduct,null,null);
	    }

	    /*
	     * Remove unused history entries
		*/
		Enumeration er = signatureCache.keys();
		while (er.hasMoreElements()) {
			Object key = er.nextElement();
			SignatureEntry entry = signatureCache.getCRCEntry(key);
		// log("entry " + key + " " + entry.usedCount);
			if (--entry.usedCount < 0) {
				signatureCache.remove(key);
			}
		}
	/*
	* Nothing changed message
	*/
	    if ( jugglersProduct == null ) {
			String nothingChanged = getStringProperty(NOTHING_CHANGED,"");
		if ( nothingChanged.length() != 0) {
				BasicProduct prod = new BasicProduct(
							nothingChanged,
							"Checked on "+(new PAMDate()).toString());
				return prod;
			}
			return null;
		}
	/*
	* Something changed message
	*/
		else {
			String somethingChanged = getStringProperty(SOMETHING_CHANGED,"");
			if (somethingChanged.length()==0) {
		    return jugglersProduct;
		}
		else {
		    VectorProduct prod = new VectorProduct(somethingChanged,changeCount + " items, checked on " + (new PAMDate()).toString());
		    prod.addElement(jugglersProduct);
		    return prod;
		}
	    }
	}

    protected boolean productChanged(Product productToTest) {

	Object content = productToTest.getContent();
	String lookupName;
	long crc;

	if (content instanceof FileRef) {
	    FileRef fileRef = (FileRef)content;
	    crc = computeCRC(fileRef);
	    // use fileref source Name to keep track of FileRef products
	    lookupName = "cache."+fileRef.getSourceName();
	} else {
	    Date date = productToTest.getModificationDate();
	    crc = date.getTime();
	    // use product Id to keep track of all other products
	    lookupName = "cache." + productToTest.getId().replace(' ', '_');
	}

		SignatureEntry oldCRC = signatureCache.getCRCEntry(lookupName);
	if (oldCRC == null) {
	    changeCount++;
	    // log(s + " [NEW]");
			SignatureEntry entry = new SignatureEntry();
			entry.crc = crc;
			signatureCache.put(lookupName, entry);
			return true;
		}
		else {
	    // reset time to live counter for this entry
			oldCRC.usedCount = 5;
	    if (oldCRC.crc != crc) {
		//log(s + " [CHANGED] " + oldCRC + " -> " + crc);
		changeCount++;
		oldCRC.crc = crc;
		return true;
	    }
	    //log(s + " ** DUPLICATE ** " + crc);
	    return false;
		}		
	}

/**
* override setProperty and handle cache settings
*/
	public void setProperty(String attribute, String value) {
	
		if (attribute.startsWith("cache.")) {
			SignatureEntry entry = new SignatureEntry();
			try {
				entry.crc = Long.parseLong(value);
			}
			catch (NumberFormatException e) {
				return;
			}
			signatureCache.put(attribute, entry);
		} else if (attribute.equals(CLEAR_CACHE) &&
			   "true".equalsIgnoreCase(value)) {
		  signatureCache.clear();
		}
		else
			super.setProperty(attribute, value);
	}

  protected long computeCRC(FileRef fileRef) {
      File file = fileRef.getFile();
      return (file != null) ? file.lastModified() : 0;
  }

/**
* Filters out products that have not changed since last visited
* Only a deep-copy of a Product should be filtered, not the original Product.
*/
    public class filterProducts implements ProductFilter {
        public boolean filter (ContainerProduct parent, Product productToTest) {
	    
            // this is to get by the test in apply() so that mapply() will be called
            if (parent == null) {
                return true;
            }
	    
	    if (productToTest instanceof BasicProduct) {
		if (!productChanged(productToTest))
                    parent.removeElement(productToTest);
	    }
	    // do not check container products only their contents...
            
            return true; // continue through the entire product
        }    
    }
}
