/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator.detectChange;
	
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.FTP;
import org.mitre.pam.aggregator.ChanSchedBase;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.productVisitor.AsFlatURLList;
import org.mitre.pam.store.JoinMemento;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.PAMDate;

/**
 * This channel checks its input content
 * to see if it is different than last time. Only
 * the modification date of the products are checked.
 *
 * @author   John Fantry, Technical Co-op (D510)
 * @author   Nazario Irizarry, Jr.
 * @author   Jason Mathews, Added clearCache trigger + many other changes.
 */
public class NotifyOfDateChangeFTP extends ChanSchedBase 
{
	protected static final String SOMETHING_CHANGED = "somethingChangedMessage";

	protected static final String NOTHING_CHANGED = "nothingChangedMessage";

	protected static final String USER_HOST_AND_PASSWORD = "userHostAndPassword";

	protected static final String CLEAR_CACHE = "clearCache";

	private static final String DEBUG_KEY = "notifyOfDateChangeFTP";

	/*
	 * Unique channel attributes
	 */
	protected SignatureTable signatureCache = new SignatureTable();

	/*
	 *
	 */
	public NotifyOfDateChangeFTP()
	{
		super(false,false);
		stateMemento = new JoinMemento("generic","notifier",signatureCache,getReporter());
		initChanSchedBase(true);
		setPeriod(600);
		setDescription("Detect file changes based on content.");

		setProperty(SOMETHING_CHANGED,"New or updated content");
		setProperty(NOTHING_CHANGED,"");
		setProperty(USER_HOST_AND_PASSWORD,"user@host{password}");
		setProperty(CLEAR_CACHE,"false");

		metaStringInfo(SOMETHING_CHANGED,"New or updated content","Text to be used when something has changed");
		metaStringInfo(NOTHING_CHANGED,"","Text to be used when there are no changes");
		metaStringInfo(USER_HOST_AND_PASSWORD,"user@host{password}","user host and password in the form user@host{password}");
		metaBooleanInfo(CLEAR_CACHE,false,"Clears cache upon setting to true");
	}

	/**
	 * A channel can build and change the product that it's pims provide
	 */
	protected Product buildProduct(Product productUnderInspection)
	{
		if(productUnderInspection == null)
		{
			return null;
		}
		// Use creator email as anonymous password if available
		String anonPassword = getProperty(CREATOR_EMAIL);
		if(anonPassword == null || anonPassword.indexOf('@') == -1)
			anonPassword = "anonymous@";
		Vector changedURLList = new Vector(30);
		/*
		 * Scan product
		 */
		try
		{
			AsFlatURLList URLProductVisitor = new AsFlatURLList();
			URLProductVisitor.visit(productUnderInspection);
			Enumeration URLEnumeration = URLProductVisitor.getURLProductList();
			/*
			 * Check each file that is a product's content
			 */
			while(URLEnumeration.hasMoreElements())
			{
				Product aURLProduct = (Product)URLEnumeration.nextElement();
				URL aURLRef = (URL)aURLProduct.getContent();
				if(URLHasNewContents(aURLRef,aURLProduct.getModificationDate()))
				{
					reporter.debugMsg("URL changed for ",name," --> ",aURLRef,DEBUG_KEY);
					changedURLList.addElement(aURLProduct);
				}
				else
					reporter.debugMsg("URL has NOT changed for ",name," --> ",aURLRef,DEBUG_KEY);
			}
		}
		catch(ClassCastException ex)
		{
			reportException(ex);
		}
		/*
		 * Remove unused crc entries
		 */
		Enumeration er = signatureCache.keys();
		while(er.hasMoreElements())
		{
			Object key = er.nextElement();
			SignatureEntry entry = signatureCache.getCRCEntry(key);
			if(--entry.usedCount < 0)
			{
				signatureCache.remove(key);
			}
		}
		/*
		 * Nothing changed message
		 */
		if(changedURLList.size() == 0)
		{
			String nothingChanged = getStringProperty(NOTHING_CHANGED,"");
			if(nothingChanged.length() != 0)
			{
				BasicProduct prod = new BasicProduct(nothingChanged,"Checked on " + (new Date()).toString());
				prod.setTopic(getProperty(TOPIC));
				return prod;
			}
			return null;
		}
		/*
		 * Report on something new
		 */
		Hashtable hostLookup = new Hashtable();
		StringTokenizer st = new StringTokenizer((String)getProperty(USER_HOST_AND_PASSWORD),"}");
		while(st.hasMoreElements())
		{
			String s = (String)st.nextElement();
			int hpos = s.indexOf('@');
			int ppos = s.indexOf('{',hpos);
			String user = s.substring(0,hpos);
			String host = s.substring(hpos + 1,ppos);
			String password = s.substring(ppos + 1);
			hostLookup.put(host,user + '$' + password);
		}
		File workingDir = ExecGlobal.g_getWorkingDir();
		// this is where new files will go
		if(!workingDir.exists())
			workingDir.mkdirs();
		FTP ftp = null;
		String currentPath = new String();
		boolean retrieved = false;
		Vector changedProducts = new Vector();
		for(int i = 0;i < changedURLList.size();i++)
		{
			// retrieved = false;
			Product p = (Product)changedURLList.elementAt(i);
			URL url = (URL)p.getContent();
			String host = url.getHost();
			if((ftp == null) || (!ftp.getHost().equalsIgnoreCase(host)))
			{
				if(ftp != null)
				{
					try
					{
						ftp.close();
					// close the old connection before the handle is lost
					}
					catch(IOException ioe)
					{
						reportException(ioe);
					}
				}
				String s = (String)hostLookup.get(host);
				String user, password;
				// if no login then try anonymous
				if(s == null)
				{
					user = "anonymous";
					password = anonPassword;
				}
				else
				{
					int pos = s.indexOf('$');
					user = s.substring(0,pos);
					password = s.substring(pos + 1);
				}
				ftp = new FTP(host,user,password,reporter);
			}
			String path = url.getFile();
			int endpos = path.lastIndexOf('/');
			String fileName = path.substring(endpos + 1);
			path = path.substring(0,endpos);
			File localFile = new File(workingDir,fileName);
			// these if tests exist in an effort to minimize network traffic
			try
			{
				if(!ftp.isConnected())
				{
					ftp.connect();
					currentPath = "/";
				}
				if(!currentPath.equalsIgnoreCase(path))
				{
					String[] responses = ftp.sendCommand(FTP.CWD,path);
					if(responses[0].charAt(0) != '2')
					{
						// directory does not exist
						throw new IllegalArgumentException(responses[0]);
					}
					currentPath = path;
				}
				ftp.sendCommand(FTP.RETR,localFile);
				// build a new product preserving the old id, description and modification date
				BasicProduct bp = new BasicProduct(p.getId(),p.getDescription());
				bp.setModificationDate(p.getModificationDate());
				// the modification date of this FileRef is set to the original file date obtained from the remote system
				FileRef f = new FileRef(localFile,p.getModificationDate().getTime(),localFile.getName());
				bp.setContent(f);
				changedProducts.addElement(bp);
			}
			catch(Exception e)
			{
				reportException(e);
			}
		}
		//for
		try
		{
			ftp.close();
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		String prodId = getStringProperty(SOMETHING_CHANGED,"");
		if(prodId.length() == 0)
		{
			if(changedProducts.size() == 1)
			{
				return (Product)changedProducts.elementAt(0);
			}
			// otherwise use agent ID for vector product ID
			prodId = getId();
		}
		VectorProduct prod = new VectorProduct(prodId,changedProducts.size() + " items, checked on " + (new PAMDate()).toString());
		prod.setContent(changedProducts);
		return prod;
	}

	/**
	 * run a signature on file contents to see
	 * if they have changed
	 */
	protected boolean URLHasNewContents(URL URLRef, Date modificationDate)
	{
		long crc = computeCRC(modificationDate);
		/*
		 * check if crc is different
		 */
		String lookupName = "cache." + URLRef.getHost() + URLRef.getFile();
		SignatureEntry oldCRC = signatureCache.getCRCEntry(lookupName);
		if(oldCRC == null || (oldCRC.crc != crc))
		{
			SignatureEntry entry = new SignatureEntry();
			entry.crc = crc;
			signatureCache.put(lookupName,entry);
			return true;
		}
		else
		{
			oldCRC.usedCount = 5;
			return (oldCRC.crc != crc);
		}
	}

	/**
	 * Compute a signature on file
	 */
	protected long computeCRC(Date modificationDate)
	{
		return modificationDate.getTime();
	}

	/**
	 * override setProperty to handle cache settings
	 */
	public void setProperty(String attribute, String value)
	{
		if(attribute.startsWith("cache."))
		{
			SignatureEntry entry = new SignatureEntry();
			try
			{
				entry.crc = Long.parseLong(value);
			}
			catch(NumberFormatException e)
			{
				return;
			}
			signatureCache.put(attribute,entry);
		} else if (attribute.equals(CLEAR_CACHE) &&
			   "true".equalsIgnoreCase(value)) {
		  signatureCache.clear();
		}
		else
			super.setProperty(attribute,value);
	}
}

