/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon;

import java.util.*;
import java.io.*;

import org.mitre.pam.interfaces.ExecToMonitor;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pam.interfaces.DIM;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.interfaces.Juggler;


import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.putter.BackWebLocalDIM;
import org.mitre.pam.monitor.SrvBackWebSubLive;

/**
* This class creates a Daemon that generates a static web page
* from which uses can subscribe to BackWeb channels defined by BackWebLocalDIM
* instances.
*
* @author Nazario Irizarry, Jr.
*/
public class BackWebSubscriber extends DaemonBase implements HasPrintLine {

	public static final String SUB_PAGE_DIR ="BackWebSubscribe";
	public static final String SUB_PAGE_NAME= "channels.html";

	public static final String USE_PEERWEBSVR = "usePeerWebServer";
	
	private transient PrintStream printer;

	public BackWebSubscriber() {
		setPeriod(15*60*60+57);	// every 5 minutes
		metaBooleanInfo(USE_PEERWEBSVR, false,"Flag to use peer web server to serve BackWeb subscription servlet otherwise uses iPAM internal web server");
	}
	
	public void doDaemonTask() {
	
		File dir = ExecGlobal.getSupportDir(SUB_PAGE_DIR);
		File file = new File(dir, SUB_PAGE_NAME);
		
		try {
			FileOutputStream fout = new FileOutputStream(file);
			printer = new PrintStream(fout);
			Enumeration chans = ((ExecToMonitor)exec).getListOfActiveModules(ExecToMonitor.JUGGLER);
			String postURL = SrvBackWebSubLive.postURL;
			if (getBooleanProperty(USE_PEERWEBSVR, false))
				postURL = ExecGlobal.getPeerProtocolHostAndPort() + postURL;
			makeOutput(this, chans, postURL);
		}
		catch (IOException e) {
			reportException(e);
		} finally {
			if (printer != null) {
				printer.close();
				printer = null;
			}
		}
	}
	
	public void println(String s) throws IOException {
		printer.println(s);
	}

	public void print(String s) throws IOException{
		printer.print(s);
	}

	public void flush() throws IOException{
		printer.flush();
	}

	public static void makeOutput(HasPrintLine p, Enumeration chans, String action) throws IOException {

		String backWebHost=null;
		String backWebTCPPort=null;
		String backWebUDPPort=null;
		String backWebExposure=null;

		Hashtable topicList = new Hashtable();

		p.println( "<HTML><HEAD>" );
		p.println( "<TITLE>BackWeb/iPAM Channels</TITLE>" );
		p.println( "</HEAD><BODY>" );
		p.print("<FORM");
		if (action != null) {
			p.print(" action=\"");
			p.print(action);
			p.print("\"");
		}
		p.print(" method=\"POST\">");
		p.println( "<H2>Channel Subscription</H2>" );

//		Enumeration chans = ((ExecToMonitor)exec).getActiveJugglerList();
	/*
	* Get all channels available
	*/
		
		while ( chans.hasMoreElements() ) {
			ConnectedAgent chan = (ConnectedAgent) chans.nextElement();
			DIM bwSink = channelConnectedToBackweb(chan);
			if (bwSink != null) {
			
			/*
			* lookup back web host and port for first channel
			*/
				if (backWebHost == null) {					
					Memento bw = bwSink.getMemento();
					backWebHost = bw.getStringProperty(BackWebLocalDIM.BW_HOST_DIM_PROP,"");
					backWebTCPPort = bw.getStringProperty(BackWebLocalDIM.BW_TCP_SERVER_PORT_NUMBER,"");
					backWebUDPPort = bw.getStringProperty(BackWebLocalDIM.BW_UDP_SERVER_PORT_NUMBER,"");
					backWebExposure = bw.getStringProperty(BackWebLocalDIM.EXPOSURE_GROUP_PROP,"");
				}
			/*
			* identify channel topic, and current contributer/member list
			*/
				String topic = buildProductTopic(chan);
				Vector topicMembers = (Vector) topicList.get(topic);
			/*
			* Check to see if the topic exists, and add if required
			* to the topic list.  TopicMembers is a Vector or
			* jugglers that contribute to the topic
			*/
				if (topicMembers == null){
					topicMembers = new Vector(10);
					topicList.put(topic, topicMembers);
				}
				topicMembers.addElement(chan);
			}
		}
	/*
	* Build table
	*/
		if (topicList.size() == 0) {
			p.println("Currently there are no channels available for subscription.");
		}
		else {
			buildPageWithContents(p, topicList, 
				backWebHost, backWebTCPPort, 
				backWebUDPPort,  backWebExposure);
		}

		p.println( "</BODY></HTML>" );
		
		p.flush();
	}

	private static void buildPageWithContents(HasPrintLine p, Hashtable topicList, 
		String backWebHost, String backWebTCPPort, String backWebUDPPort, 
		String backWebExposure) throws IOException {

		p.println("<BLOCKQUOTE><TABLE BORDER=0>");
		p.println("<TR> <TD COLSPAN=2>");
		p.println("<P>Mark all channels you would like to subscribe to:<P>");
		p.println("</TD></TR>");
        
      
		
		int columnCount=0;
		Enumeration topics = topicList.keys();
		while ( topics.hasMoreElements() ) {
			String topic = (String) topics.nextElement();
			if (columnCount == 0) 
				p.print("<TR><TD>");
			else 
				p.print("<TD>");
			
/*
			p.print("<INPUT TYPE=\"checkbox\" NAME=\"");
			p.print(channelId);
			p.print("\" VALUE=\"true\" >");
*/
		/*
		* a check box name="channel" value="<channelid>"
		*/
			p.print("<INPUT TYPE=\"checkbox\" NAME=\"channel\" VALUE=\"");
			p.print("pam_");
			p.print(topic);
			p.print("\" >");

			
			p.print("<A HREF=\"#");
			p.print(topic);
			p.print("\">");
			p.print(topic);
			p.print("</A>");

			if (columnCount++ >= 2) {
				p.print("</TD></TR>");
				columnCount = 0;
			}
			else {
				p.print("</TD>");
			}

		}
	/*
	* The last thing to do is the submit button
	*/
		if (columnCount == 0) 
			p.print("<TR><TD>");
		else 
			p.print("<TD>");

		p.print("<P><INPUT TYPE=\"submit\" NAME=\"Subscribe\" VALUE=\"Subscribe\">");
		
		if (columnCount++ >= 2) {
			p.print("</TD></TR>");
			columnCount = 0;
		}
		else {
			p.print("</TD>");
		}
	
		p.println("</TABLE>");

		p.println( "<P><P>Note: Selections will completely replace any current subscription.</BLOCKQUOTE><HR>" );
	/*
	* Now build descriptions
	*/
		p.println("<H2>Channel Descriptions</H2>");
		

		topics = topicList.keys();
		while ( topics.hasMoreElements() ) {
			String topic = (String) topics.nextElement();
			
		/*
		* Write out the topic and an anchor for the topic
		*/
			p.print("<A NAME=\""); p.print(topic); p.println("\"></A>");
			p.print("<H3>"); p.print(topic); p.println("</H3>");
			
			Enumeration chans = ((Vector) topicList.get(topic)).elements();

		
			p.println( "<P><BLOCKQUOTE><DL>");
			
		/*
		* Write out the topic's contributing jugglers
		*/
			while ( chans.hasMoreElements() ) {
				Agent chan = (Agent) chans.nextElement();
				p.print("<dt>");
				p.print(chan.getId());
				p.print("</dt>");

				p.print("<dd>");
				p.print(chan.getDescription());
				Memento mem=chan.getMemento();
				p.print("</dd>");
				String schedule = buildScheduleString(mem);
				if (schedule != null) {
					p.print("<DD><FONT SIZE=\"-1\"> ");
					p.print(schedule);
					p.println("</FONT></DD> ");
				}
			}
			p.println( "</DL></BLOCKQUOTE>" );
		}
		
	/*
	* identify the back web host and port in the form
	*/
		p.print("<INPUT TYPE=\"hidden\" NAME=\"bwHost\" VALUE=\"");
		p.print(backWebHost);
		p.print("\" >");

		p.print("<INPUT TYPE=\"hidden\" NAME=\"bwTCPPort\" VALUE=\"");
		p.print(backWebTCPPort);
		p.print("\" >");

		p.print("<INPUT TYPE=\"hidden\" NAME=\"bwUDPPort\" VALUE=\"");
		p.print(backWebUDPPort);
		p.print("\" >");

		p.print("<INPUT TYPE=\"hidden\" NAME=\"bwExposure\" VALUE=\"");
		p.print(backWebExposure);
		p.print("\" >");
	}

	
	
	private static String buildScheduleString(Memento mem) {
		String schedule = mem.getStringProperty(ScheduleAttributes.SCHEDULE_VERB,"");
		if (schedule.length() == 0)
			return null;
			
		StringBuffer buf = new StringBuffer(80);
		buf.append("updates ");
		if (schedule == ScheduleAttributes.SCHEDULE_ONCE) {
			buf.append("once");
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_NONE) {
			buf.append("never");
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_DAILY) {
			buf.append("daily at ");
			buf.append(mem.getStringProperty(ScheduleAttributes.HOUR_OF_DAY,""));
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_HOURLY) {
			buf.append("hourly at ");
			buf.append(mem.getStringProperty(ScheduleAttributes.MINUTE_OF_HOUR,""));
			buf.append(" minutes after the hour");
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_WEEKLY) {
			buf.append("weekly on ");
			buf.append(mem.getStringProperty(ScheduleAttributes.DAY_OF_WEEK,""));
			buf.append(" at ");
			buf.append(mem.getStringProperty(ScheduleAttributes.HOUR_OF_DAY,""));
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_PERIODIC) {
			int howOften = 0;
			try {
				howOften = mem.getIntegerProperty(ScheduleAttributes.PERIOD,0);
			}
			catch (Exception ignore){}
			if (howOften < 180) {
				buf.append("every ");
				buf.append(mem.getStringProperty(ScheduleAttributes.PERIOD,"0"));
				buf.append(" seconds");
			}
			else {
				buf.append("every ");
				int minutes = howOften/60;
				buf.append(Integer.toString(minutes));
				if (minutes == 1)
					buf.append(" minute");
				else
					buf.append(" minutes");
			}
		}
		else if (schedule == ScheduleAttributes.SCHEDULE_NONE) {
			buf.append("depend on other processing schedules ");
		}
		else {
			buf.append("unknown");
		}
		return buf.toString();
	}

	private static DIM channelConnectedToBackweb(ConnectedAgent chan) {
		
		Enumeration e = chan.getSinkList().elements();
		while(e.hasMoreElements()) {
			Object sink = e.nextElement();
			if (sink instanceof BackWebLocalDIM)
				return (DIM) sink;
		}
		return null;
	}

	private static String buildProductTopic(ConnectedAgent chan) {
		String topic = chan.getStringProperty(Juggler.TOPIC,"");
		if (topic.length() == 0)
			return "default";
		
		return BackWebLocalDIM.sanitizeTopicName(topic);
	}

}