package org.mitre.pam.getter;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;

import org.mitre.pam.product.view.DictionaryProduct;

import org.mitre.pam.productVisitor.Serializer;

import COM.TIBCO.rv.RvDataCallback;
import COM.TIBCO.rv.RvException;
import COM.TIBCO.rv.RvListener;
import COM.TIBCO.rv.RvOpaque;
import COM.TIBCO.rv.RvSender;
import COM.TIBCO.rv.RvSession;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import java.util.zip.InflaterInputStream;



/**
* This class is a PIM that will listen for incoming TIBCO messages. <p>
*
* Its incoming product is a byte array inside an RvOpaque
* object containing a Product serialized by various methods in
* in org.mitre.pam.util.Serializer. <p>
*
* The byte array is unserialized by varous methods in org.mitre.pam.util.Serializer
* which convert it from a byte array back into a Product. <p>
*
* <pre>
* Config vars:
* daemonHost - the IP Address or hostname of the machine on which
*              your TIB/Rendezvous daemon (RVD) is running (default
*              localhost or 127.0.0.1)
*
* daemonPort - the UDP port number where this PIM should listen for its messages
*              by communicating with the TIB/Rendezvous daemon (RVD) listening on that
*              port. (default 7500)
*
* subjectName - the static or dynamic "subject name" under which messages
*               of interest to this PIM will appear; analogous to a radio
*               broadcast frequency.  This subject name may contain wildcards
*               to allow retrieval of messages on multiple subjects.
*               (no default value)
*
* pollDaemonInterval - the number of milliseconds between each check that verifies that
*                      the 'rva' session we are using is valid.  If the 'rva' session is
*                      invalid, then we won't get callbacks when new messages arrive.
*                      So, it is necessary to periodically check to make sure that our 'rva'
*                      session is still valid.
*                      (default 60000) (i.e. 1 minute)
*
* Sample PDIF:
* (daemonHost jleung-pc.mitre.org)
* (daemonPort 7501)
* (subjectName NEWS.BOSNIA)
* (pollDaemonInterval 60000)
* </pre>
*
* To prevent RvMsgDispatch:Invalid RVMSG_MAGIC error message when trying to specify
* host and port number for RvSession just use RVA (TIB/Rendezvous agent) instead of
* RVD (TIB/Rendezvous daemon). <p>
*
* The RVA will find and connect to an existing RVD with the specified host
* and port number or it will start an appropriate RVD and connect to it.  <p>
*
* The RVA cannot start RVDs on remote machines unless there is already an
* RVA running on that remote machine. <p>
*
* By default, RVA listens on port 7600.  There should be an RVA running on the
* same machine as iPAM in order for this PIM to connect to an RVD through the RVA. <p>
*
* @author Jones Leung
* @author John R. Fantry - modified class so that it uses DictionaryBasePIM
*
* @see org.mitre.pam.interfaces.AsynchronousAgent
* @see org.mitre.pam.putter.PutTIBCO
* @see com.TIBCO.rv.RvDataCallback
* @see org.mitre.pam.util.Serializer
*/
public class GetAsyncTIBCO extends DictionaryBasePIM implements AsynchronousAgent, RvDataCallback {
  /** The IP Address or hostname of the machine on which the RVD is running */
  protected final static String DAEMON_HOST = "daemonHost";
  protected final static String DEFAULT_HOST = "localhost";

  /** The UDP port where this PIM expects to communicate with the RVD to get messages */
  protected final static String DAEMON_PORT = "daemonPort";
  protected final static String DEFAULT_PORT = "7600";

  /** The subject this PIM is interested in */
  protected final static String SUBJECT_NAME = "subjectName";
  protected final static String DEFAULT_SUBJECT = "";

  /** The number of milliseconds between 'rva' session validity checks */
  protected final static String POLL_DAEMON_INTERVAL = "pollDaemonInterval";
  protected final static String DEFAULT_INTERVAL = "60";

  // The number of seconds to wait until we poll the RvSession to make sure it's still valid
  long pollDaemonInterval = 60; // default is 1 min.

  /**
  * If true, this PIM will continue to queue all new messages, but these queued messages
  * won't become the pimsProduct until the flag becomes true.
  */
  protected boolean paused = false;

  // This RvSession is used for all communications between this PIM and the RVD
  // N.B. Starting RvSessions is expensive.
  RvSession session = null;

  // This RvListener is used to receive all messages for this PIM
  RvListener listener = null;

  // This thread is used to poll the RvSession to make sure that it is functional
  IsSessionValidThread it = null;

  String daemonHost = null;
  String daemonPort = null;
  String subjectName = null;

/**
* Creates a new GetAsyncTIBCO object.
*/
  public GetAsyncTIBCO() {
    setDescription ("Asynchronously retrieve TIBCO messages");

    metaStringInfo(DAEMON_HOST, DEFAULT_HOST, "the host to retrieve messages from");
    metaStringInfo(DAEMON_PORT, DEFAULT_PORT, "the port to talk to");
    metaStringInfo(SUBJECT_NAME, DEFAULT_SUBJECT, "the TIBCO subject");
    metaStringInfo(POLL_DAEMON_INTERVAL, DEFAULT_INTERVAL, "how often (in seconds) to check the server connection");
  }


/**
* Begin queueing TIBCO messages.
*/
  public void startService() {
    // get config var values
    daemonHost = getProperty (DAEMON_HOST);
    daemonPort = getProperty (DAEMON_PORT);
    subjectName = getProperty (SUBJECT_NAME);
    pollDaemonInterval = Long.parseLong (getProperty (POLL_DAEMON_INTERVAL)) * 1000;

    CreateRvSessionAndListener();

    // create thread which checks to make sure that session is valid
    it = new IsSessionValidThread (pollDaemonInterval, this, session, reporter);
    it.start();
  }

/**
* Create a new RvSession and RvListener
*/

  public RvSession CreateRvSessionAndListener () {
    // Create an RvSession
    try {
//        System.out.println ("Create RvSession...");
        session = new RvSession(daemonHost, (new Integer (daemonPort)).intValue());
//        session = new RvSession();  // Connect to local RVA.
    } catch (RvException rve) {
        log (getId() + ": create RvSession:  " + rve.getMessage());
        reportException (rve);
        return null;
    }

    // Create an RvListener
    try {
//        System.out.println ("Create RvListener...");
        listener = session.newListener (subjectName, this);
    } catch (RvException rve) {
        log (getId() + ": create RvListener: " + rve.getMessage());
        reportException (rve);
        return null;
    }

    return session;
  }


/**
* Shutdown PIM.
*/
  public void stopService() {
        // Terminate connection with RVA.

        // stop thread which is polling the RvSession to see if it is still active
        it.stop();
        it = null;

        try {
            if (session != null) {
                session.term();
            }
        } catch (RvException rve) {
            log (getId() + ": terminate RvSession:  " + rve.getMessage());
            reportException (rve);
        }
        session = null;
  }

/**
* Tell this PIM to stop injecting asynchronous events into PAM.
*/
  public void pauseService() {
        paused = true;
  }

/**
* Tell this PIM to continue injecting asynchronous events into PAM.
*/
  public void resumeService() {
        paused = false;
  }

/**
* This PIM is totally asynchronous.  All incoming Prodcuts are injeced into the system
* immediately using pushProductToAllClients().  doFetch() will always return null.
*/
  public Product doFetch() {
    return null;
  }

/**
* Callback to receive TIBCO messages.  This method should not block. <p>
* @param subject - This parameter receives the subject of the arriving message.
* @param replySender  - If the arriving message carries a reply subject, this
*  parameter receives a newly created sender activity that sends to
*  that reply subject. Otherwise, this parameter receives null.
* @param data - This parameter receives the data object.
* @param listener - Each listener appoints a handler object to process arriving
*  data on its behalf. This parameter receives the listener that appointed
*  this handler.
*/
  public void onData (String subject, RvSender replySender, Object data, RvListener listener) {
        if (data instanceof RvOpaque) {
            byte[] b = ((RvOpaque) data).contents;
            processByteArray(b);
        } else {
            // ignore
//            System.out.println ("Message is not RvOpaque...");
        }
  }

  // restore the contents of the byte array
  public void processByteArray(byte[] bbuf) {
    Object obj = null;

    ByteArrayInputStream bis = new ByteArrayInputStream(bbuf);
    InflaterInputStream iis = new InflaterInputStream(bis);

    try {
      ObjectInputStream in = new ObjectInputStream(iis);
      obj = Serializer.unSerialize(in, getWorkingDir());
      in.close();
    } catch(IOException ioe) {
      reporter.reportException(ioe);
    } catch(ClassNotFoundException cnfe) {
      reporter.reportException(cnfe);
    }

    processObject(obj);
  }


/**
* This class is a Thread which polls to make sure that our
* RvSession is valid -- meaning that we still have a functional
* connection to the RVA.
*/
	public class IsSessionValidThread extends Thread {
	    long sleepTime = 60000;  // default is to poll every 60000 milliseconds.
      Reporter reporter;
      RvSession session;
      GetAsyncTIBCO agent;

      public IsSessionValidThread (long sleepTime, GetAsyncTIBCO agent, RvSession session, Reporter reporter) {
	        this.sleepTime = sleepTime;
          this.reporter = reporter;
          this.session = session;
          this.agent = agent;
	    }

	    public void run() {
            // Re-create RvSession and RvListener, if necessary.
            // Sometimes, the session is disconnected due to RVA crash or RVA license expiration.
            // When the session is invalid, all its listeners and senders also
            // become invalid.
            // We may lose some messages if the session becomes invalid.
            while (true) {
                if (session == null || !session.isValid()) {
//                    System.out.println ("Session is invalid.  Re-creating session and listener.");
                    reporter.log (agent.getId() + ": RvSession is invalid.  Re-creating session and listener.");
                    session = agent.CreateRvSessionAndListener();
                } else {
//                    System.out.println ("Session is valid.");
                }
                try {
                     this.sleep (sleepTime);
                } catch (Exception e) {
                    reporter.reportException (e);
                }
            }
	    }
  } // end of inner class IsSessionValidThread


}