/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.productVisitor.Serializer;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.DatagramSocket;
import java.util.zip.InflaterInputStream;
import java.net.DatagramPacket;

/**
* This class is a PIM that will listen for incoming UDP datagrams. <p>
*
* <pre>
* Config vars:
*
* listenPort - the UDP port number where this PIM should listen for the datagram
*              (default 7502)
*
* Sample PDIF:
* (listenPort 7502)
* </pre>
*
* NOTE: The maximum reliable length of a UDP datagram is 64900 bytes.
*       If the byte array size exceeds 64900 an AssertionException will occur
*       in the DIM which tries to send the oversized datagram to this PIM. <p>
*
* The largest possible UDP datagram is 64K, but some space is taken up
* by the UDP header and IP header. <p>
*
* @author Jones Leung
* @author John R. Fantry - modified class so that it uses DictionaryBasePIM
*/
public class GetAsyncUDP extends DictionaryBasePIM implements AsynchronousAgent {
  /** The UDP port where this PIM will listen for incoming datagrams */
  protected final static String LISTEN_PORT = "listenPort";
  public final static String DEFAULT_PORT = "7502";

  public final static int MAX_PACKET_SIZE = 64900;

  // this feature is not actually implemented, pausing this agent has no affect
  protected boolean paused;

  // The DatagramSocket used for receiving/sending UDP datagrams
  protected transient DatagramSocket ds;

  // The thread which is used to listen for incoming datagrams
  protected transient ReceiveDatagramThread listener;

/**
* Creates a new GetAsyncUDP object.
*/
  public GetAsyncUDP() {
    setDescription ("Asynchronously retrieve UDP datagrams");
    metaStringInfo(LISTEN_PORT, DEFAULT_PORT, "the port to listen on");
  }

/**
* Begin queueing UDP datagrams
*/
  public void enableService() {
    try {
    // get config var values
	String listenPort = getProperty(LISTEN_PORT);

    // Create a DatagramSocket
        ds = new DatagramSocket (Integer.parseInt (listenPort));
	// paused = false;

    // ...and start listening for datagrams
        listener = new ReceiveDatagramThread(ds, reporter, MAX_PACKET_SIZE);
	listener.setDaemon(true);
        listener.start();
    } catch (Exception e) {
        e.printStackTrace();
        reportException (e);
    }
  }


/**
* Shutdown PIM.
*/
  public void disableService() {
        // Stop listener thread
        if (listener != null) {
            listener.running = false;
            listener = null;
        }

        // close DatagramSocket
        if (ds != null) {
            ds.close();
            ds = null;
        }
  }

/**
* Tell this PIM to stop injecting asynchronous events into PAM.
*
* Not implemented.
*/
  public void pauseService() {
        //paused = true;
  }

/**
* Tell this PIM to continue injecting asynchronous events into PAM.
*
* Not implemented.
*/
  public void resumeService() {
        //paused = false;
  }

/**
* This PIM is totally asynchronous.  All incoming Prodcuts are injeced into the system
* immediately using pushProductToAllClients().  doFetch() will always return null.
*/
  public Product doFetch() {
    return null;
  }

  // restore the contents of the byte array
  public void processByteArray(byte[] bbuf) {
    Object obj = null;

    ByteArrayInputStream bis = new ByteArrayInputStream(bbuf);
    InflaterInputStream iis = new InflaterInputStream(bis);

    try {
      ObjectInputStream in = new ObjectInputStream(iis);
      obj = Serializer.unSerialize(in, getWorkingDir());
      in.close();
    } catch(IOException ioe) {
      reporter.reportException(ioe);
    } catch(ClassNotFoundException cnfe) {
      reporter.reportException(cnfe);
    }

    processObject(obj);
  }


/**
* Listen for incoming UDP packets and pass them on to the agent reference
*/
final class ReceiveDatagramThread extends Thread {
  byte[] b;
  DatagramSocket ds;
  Reporter reporter;
  int maxDatagramSize;
  boolean running = true;

  public ReceiveDatagramThread (DatagramSocket ds, Reporter reporter, int maxDatagramSize) {

    this.ds = ds;
    this.reporter = reporter;
    this.maxDatagramSize = maxDatagramSize;

    // this listening socket should never time out
    try {
      ds.setSoTimeout(0);
    } catch (Exception e) {
      reporter.reportException(e);
    }
  }

  /**
  * Listen for incoming packets, and pass the contents to agent.processByteArray()
  */
  public void run() {
    while (running) {
      DatagramPacket dp = new DatagramPacket(new byte[maxDatagramSize], maxDatagramSize);

      if (ds == null) {
        reporter.assert(false, "Datagram listener socket is null!  Not receiving product asynchronously via UDP.", null, null, null);
        break;
      }

      if (dp == null) {
        reporter.assert(false, "DatagramPacket is null! Not receiving product asynchronously via UDP.", null, null, null);
        break;
      }

      try {
        ds.receive(dp);
        b = dp.getData();

        // get "real" data without the padding
        byte[] temp = new byte[dp.getLength()];
        System.arraycopy (b, 0, temp, 0, dp.getLength());
        b = temp;
        processByteArray(b);

      } catch (Exception e) {
        reporter.reportException(e);
      }
    }
  }
}

}
