/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import java.util.*;
import java.net.*;
import java.io.*;

import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Product;


import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.store.FileRef;

import org.mitre.pam.getter.PIMShell1;

/**
* Gets one directory, one level deep. <P>
*
* @author Nazario Irizarry, Jr.
* @author Michael E. Los  Enhanced to support copying of subdirectories
*/
public class GetDirectory extends PIMShell1 {

  //--------------------------------------------
  // PUBLIC STATIC DATA MEMBERS
  //--------------------------------------------
  public static final String RECURSE_DIRS = "recurseDirectories";
  public static final String KEEP_SOURCE_FILES = "keepSourceFiles";
  public static final String LAST_UPDATE_LONG = "lastUpdatedLong";
  public static final String USE_FILE_MODIFY_TIME = "useSourceFileTimeStamp";
  public static final String SOURCE_DIRECTORY = "sourceDirectory";

  private String zRootDirPath;
  private boolean keepSourceFiles;
  private boolean useSourceFileTimeStamp;
  private long lastUpdateLong;

  public GetDirectory () {
	setDescription("Gets the contents of a directory");
	
	metaBooleanInfo(RECURSE_DIRS, false, "Get directory for nested directories and their subdirectories");
	metaBooleanInfo(KEEP_SOURCE_FILES, true, "If set then source files will not be deleted once they have been scanned");
	metaIntegerInfo(LAST_UPDATE_LONG, 1, "Time in milliseconds when the directory was last scanned (for info only)");
	metaBooleanInfo(USE_FILE_MODIFY_TIME, false, "If set then new files are identified by their last-modified time stamp");
	metaStringInfo(SOURCE_DIRECTORY, File.separator, "If set then new files are identified by their last-modified time stamp");
	}

/**
* Get directory contents in reponse to a doFetch command.
* Since the files were put there by someone else, mark the
* files as keepers.
*/
	public synchronized Product doFetch(){
		Product directoryProduct=null;
		
		keepSourceFiles = getBooleanProperty(KEEP_SOURCE_FILES, true);
		useSourceFileTimeStamp = getBooleanProperty(USE_FILE_MODIFY_TIME, false);
		lastUpdateLong = Long.parseLong(getStringProperty(LAST_UPDATE_LONG,"1"));
		if (!useSourceFileTimeStamp)
			lastUpdateLong = 1;
		ContainerProduct list;
		try {
			String rootName = getStringProperty(SOURCE_DIRECTORY,File.separator);
			if (rootName.endsWith(File.separator)) {
				rootName = rootName.substring(0, rootName.length()-1);
			}
			File root = new File (rootName);
			//System.out.println("getDir: RootDir: " + root.getPath());
	
			boolean bRecurseDirs = getBooleanProperty(RECURSE_DIRS, false);
			// if recursing dirs, store path of the directory at which we are starting
			zRootDirPath = root.getPath();
			list = getList(root, bRecurseDirs);
	
			if ( (list == null) || (list.size() == 0) ) {
				directoryProduct = null;
		    }
		    else {
		        // getList sets a default Id and Description; we want specific values
		        list.setId(getId());
		        list.setDescription(getDescription());
		        directoryProduct = list;
			}
		}
		catch (Exception e) {
			directoryProduct = null;
			reportException(e);
		}

    	setProperty(LAST_UPDATE_LONG, Long.toString((new Date()).getTime()));
    	return directoryProduct;
	}

  /**
   * Get a listing of a directory, mark the files as keepers.
   * If the flag to recurse directories is true, then the relative
   * path to the file is included as part of the name stored in the
   * FileRef for it.
   */
	protected ContainerProduct getList(File vDir, boolean bRecurseDirs) {
		if (!vDir.isDirectory())
			return null;
	
	
		/*
		* use the name of directory (not the full path) 
		* as the id of the VectorProduct
		*/
		String sDirName = vDir.getName();
		if (sDirName.length() == 0) // this happened on windows when root name ended with "\"
			return null;
		
		//    System.out.println("getDir:getList: sDirName " + sDirName);
		VectorProduct result = new VectorProduct(sDirName, "directory listing");
		try {
			// Retrieve a listing of the directory's contents
			String [] fileList = vDir.list();
			if (fileList == null) {
				reporter.warning("Directory "+vDir.getAbsolutePath()+" is empty ");
			}
			else {
				for (int i = 0; i < fileList.length; i++) {
					File oneFile = new File(vDir, fileList[i]);
					if (oneFile.isDirectory() && bRecurseDirs) {
						ContainerProduct subDirProd = getList(oneFile, bRecurseDirs);
						result.addElement(subDirProd);
					}
					else if (oneFile.isFile() && oneFile.lastModified() > lastUpdateLong) {
						FileRef f = new FileRef(oneFile, oneFile.lastModified(), oneFile.getName());
						
						f.setKeepFile(keepSourceFiles);
						
						BasicProduct p = new BasicProduct(oneFile.getName(),"");
						p.setContent(f);
						if (result == null)
							reporter.assert(false,"reporter null! file:",oneFile.getAbsolutePath(),null,null);
						else
							result.addElement(p);
					}
				}//for
			}
		}
		catch (Exception e) {
			reportException(e);
		}
		return result;
	}//getList()
}