/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import java.io.IOException;

import java.net.MalformedURLException;
import java.net.URL;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.getter.PIMShell1;

import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.util.FTP;


/**
* This class creates a PIM for retrieving a file listing from a remote
* directory on a remote machine using the FTP protocol.
*
* @author   John Fantry, Technical Co-op (D510)
* @author   Jones Leung, added WINNT_FORMAT2 to handle alternative WINNT FTP servers
*/
public class GetFTPDirectoryList extends PIMShell1 {
  /**
  * The UNIX_FORMAT and WINNT_FORMAT arrays represent patterns that are compared to the responses
  * returned by the FTP.LIST command in order to determine what type of host machine the responses
  * came from.  Each number represents an index into each line of the response where a space character
  * should be.  Every line of a given response must conform to the defined pattern or else it is
  * rejected.
  *
  * NOTE: The first line of response from a UNIX machine is generally the string "total xxx".  This line does
  * not conform to the pattern, so it is discarded.
  *
  * The MIN_UNIX_LENGTH and MIN_WINNT_LENGTH integers represent the minimum character length that every line
  * of a response must have in order to pass.
  */
  protected final static int[] UNIX_FORMAT = {10, 40, 44, 47, 53};
  protected final static int[] WINNT_FORMAT = {8, 17, 29, 38};  // 29 is iffy, may not apply to all servers
  protected final static int[] WINNT_FORMAT2 = {12, 14, 23, 45, 58};
  protected final static int MIN_UNIX_LENGTH = 54;
  protected final static int MIN_WINNT_LENGTH = 39;

  protected final static String PASSWORD = "password";
  protected final static String USER = "user";
  protected final static String FTP_URL = "ftpUrl";
  protected final static String INITIAL_URL = "ftp://x.y.z/rootDir";

  private String host;
  private String path;

  public GetFTPDirectoryList() {
    setDescription("Get an FTP directory listing");
    setProperty(PASSWORD, "<ftp password here>");

    metaStringInfo(FTP_URL, INITIAL_URL,"ftp source address in URL form, such as: ftp://x.y.z/rootDir");
    metaStringInfo(USER, "anonymous", "ftp user name");
    metaSetIsPassword(PASSWORD, "ftp password");
  }

  public Product doFetch() {
      String url = getStringProperty(FTP_URL,INITIAL_URL);

	  if ( !url.startsWith("ftp://")) {
		reportException(new IllegalArgumentException("Invalid FTP URL: " + url));
		return null;
	  }

      int pos = url.indexOf('/', 6);
      if (pos == -1) {
	  host = url.substring(6);
	  path = "/";
      }
      else {
	  host = url.substring(6, pos);
	  path = url.substring(pos);  // extract path
      }

      String user = getStringProperty(USER,"anonymous");
      String password = getStringProperty(PASSWORD,"");
      FTP ftp = new FTP(host, user, password, reporter);
      String[] fileList = null;

    try {
      ftp.connect();
	String[] responses = ftp.sendCommand(FTP.CWD, path);
/*      
      System.out.println ("FTP.CWD: ");
      for (i = 0; i < responses.length; i++) {
        System.out.println (responses[i]);
      }
*/      
      
      if(responses[0].charAt(0) != '2') {  // directory does not exist
/*      
        System.out.println ("FTP.DIR_NOT_EXIST: ");
        for (i = 0; i < responses.length; i++) {
            System.out.println (responses[i]);
        }       
*/        
        
	    reporter.log("Failed to change to destination directory: " + path);
	    reporter.reportException(new IOException(responses[0]));
        return null;
      }
      responses = ftp.sendCommand(FTP.LIST);
/*      
      System.out.println ("FTP.LIST: ");
      for (i = 0; i < responses.length; i++) {
        System.out.println (responses[i]);
      }
*/      

	if (responses[0].charAt(0) == '2') {  // indicates success
        fileList = ftp.getData();
/*        
        System.out.println ("FTP.GET_DATA: ");
        for (i = 0; i < fileList.length; i++) {
            System.out.println (fileList[i]);
        }
*/        
	} else {
	    throw new IOException("Failed to get directory listing from " +
				  url +  ": " + responses[0]);
	}
      } catch (IOException ioe) {
	  	reporter.reportException(ioe);
	  	return null;
      } finally {
	  if (ftp.isConnected()) {
	      try {
			ftp.close();
	      } catch (IOException ignore) {}
	  }
      }

      if (fileList == null || fileList.length == 0) {
		reportException(new IllegalArgumentException("Unable to get directory list from " + url));
		return null;
      }

		if (isUnixFormat(fileList)) {
			return createProductFromUnix(fileList);
		} else if(isWinNTFormat(fileList)) {
			return createProductFromWinNT(fileList);
		} else if (isWinNTFormat2 (fileList)) {
			return createProductFromWinNT2(fileList);
		} else {
			reporter.assert(false, "Directory format from ", host, " is unknown,\n Send 0checkpts/log.txt file to software maintainer(s)",null);
			reportUnknownFormat(fileList);
			return null;
		}
  }

  /**
  * Compares the String[] argument to the UNIX_FORMAT pattern and the
  * MIN_UNIX_LENGTH spec to determine if there is a match.
  *
  * @param    list    An FTP directory listing
  *
  * @return   true if there is a pattern match
  */
  protected boolean isUnixFormat(String[] list) {
    if(list[0].startsWith("total")) {
      for(int x=1; x<list.length; x++) {
        if(list[x].length() < MIN_UNIX_LENGTH) {
          return false;
        }
        for(int y=0; y<UNIX_FORMAT.length; y++) {
          if(list[x].charAt(UNIX_FORMAT[y]) != ' ') {
            return false;
          }
        }
      }
    } else {
      for(int x=0; x<list.length; x++) {
        if(list[x].length() < MIN_UNIX_LENGTH) {
          return false;
        }
        for(int y=0; y<UNIX_FORMAT.length; y++) {
          if(list[x].charAt(UNIX_FORMAT[y]) != ' ') {
            return false;
          }
        }
      }
    }
    return true;
  }

  /**
  * Compares the String[] argument to the WINNT_FORMAT pattern and the
  * MIN_WINNT_LENGTH spec to determine if there is a match.
  *
  * @param    list    An FTP directory listing
  *
  * @return   true if there is a pattern match
  */
  protected boolean isWinNTFormat(String[] list) {
    for(int x=0; x<list.length; x++) {
      if(list[x].length() < MIN_WINNT_LENGTH) {
        return false;
      }
      for(int y=0; y<WINNT_FORMAT.length; y++) {
        if(list[x].charAt(WINNT_FORMAT[y]) != ' ') {
          return false;
        }
      }
    }
    return true;
  }

  /**
  * Compares the String[] argument to the WINNT_FORMAT2 pattern and the
  * MIN_WINNT_LENGTH spec to determine if there is a match.
  *
  * @param    list    An FTP directory listing
  *
  * @return   true if there is a pattern match
  */
  protected boolean isWinNTFormat2(String[] list) {
    for(int x=0; x<list.length; x++) {
      if(list[x].length() < MIN_WINNT_LENGTH) {
//        System.out.println ("WINNT_FORMAT2[y]: " + "< MIN_WINNT_LENGTH");
        return false;
      }
      for(int y=0; y<WINNT_FORMAT2.length; y++) {        
        if(list[x].charAt(WINNT_FORMAT2[y]) != ' ') {
/*            
          System.out.println ("list[x]: " + list[x]);  
          System.out.println ("WINNT_FORMAT2[y]: " + WINNT_FORMAT2[y]);  
*/          
          return false;
        }
      }
    }
    return true;
  }

  /**
  * Builds a vector product containing URL's that point to the file on
  * the remote machine.
  *
  * @param    list    An FTP directory listing
  */
  protected Product createProductFromUnix(String[] list) {
    SimpleDateFormat dateAndTime = new SimpleDateFormat("MMM d HH:mm yyyy");
    SimpleDateFormat dateAndTime1 = new SimpleDateFormat("MMM  d HH:mm yyyy");
    SimpleDateFormat dateAndYear = new SimpleDateFormat("MMM d  yyyy");
    SimpleDateFormat dateAndYear1 = new SimpleDateFormat("MMM  d  yyyy");
    dateAndTime.setTimeZone(TimeZone.getDefault());
    dateAndTime1.setTimeZone(TimeZone.getDefault());
    dateAndYear.setTimeZone(TimeZone.getDefault());
    dateAndYear1.setTimeZone(TimeZone.getDefault());

    String fileName;
    Date fileDate = new Date();
    Calendar rightNow = Calendar.getInstance();

    VectorProduct container = new VectorProduct(getId(), getDescription());

    if(list[0].startsWith("total")) {
      for(int i=1; i<list.length; i++) {
        if(list[i].charAt(0) != 'd') {  // don't want to process any directories
          fileName = list[i].substring(54);
          try {
            fileDate = dateAndTime.parse(list[i].substring(41, 53) + ' ' + rightNow.get(Calendar.YEAR));
          } catch(ParseException pe) {
            try {
              fileDate = dateAndTime1.parse(list[i].substring(41, 53) + ' ' + rightNow.get(Calendar.YEAR));
            } catch(ParseException pe1) {
              try {
                fileDate = dateAndYear.parse(list[i].substring(41, 53));
              } catch(ParseException pe2) {
                try {
                  fileDate = dateAndYear1.parse(list[i].substring(41, 53));
                } catch(ParseException pe3) {
                  reportException(new IllegalArgumentException("Can not create Date for: " + list[i].substring(41, 53)));
                }
              }
            }
          }

          BasicProduct bp = new BasicProduct(getId(), getDescription());
          bp.setModificationDate(fileDate);

          try {
            if(path.endsWith("/")) {
              URL url = new URL("ftp://" + host + path + fileName);
              if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
              }
            } else {
              URL url = new URL("ftp://" + host + path + '/' + fileName);
              if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
              }
            }
          } catch(MalformedURLException mue) {
            reportException(mue);
          }

          if(bp.getContent() instanceof URL) {
            container.addElement(bp);
          }
        }
      }
    } else {
      for(int i=0; i<list.length; i++) {
        if(list[i].charAt(0) != 'd') {  // don't want to process directories
          fileName = list[i].substring(54);

          try {
            fileDate = dateAndTime.parse(list[i].substring(41, 53) + ' ' + rightNow.get(Calendar.YEAR));
          } catch(ParseException pe) {
            try {
              fileDate = dateAndYear.parse(list[i].substring(41, 53));
            } catch(ParseException pe1) {
              reportException(new IllegalArgumentException("Can not create Date for: " + list[i].substring(41, 53)));
            }
          }

          BasicProduct bp = new BasicProduct(getId(), getDescription());
          bp.setModificationDate(fileDate);

          try {
            if(path.endsWith("/")) {
              URL url = new URL("ftp://" + host + path + fileName);
              if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
              }
            } else {
              URL url = new URL("ftp://" + host + path + '/' + fileName);
              if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
              }
            }
          } catch(MalformedURLException mue) {
            mue.printStackTrace();
          }

          if(bp.getContent() instanceof URL) {
            container.addElement(bp);
          }
        }
      }
    }
    return container;
  }

  /**
  * Builds a vector product containing URL's that point to the file on
  * the remote machine.
  *
  * @param    list    An FTP directory listing
  */
  protected Product createProductFromWinNT(String[] list) {
    SimpleDateFormat dateAndTime = new SimpleDateFormat("MM-d-yy  hh:mma");
    dateAndTime.setTimeZone(TimeZone.getDefault());

    String fileName;
    Date fileDate = new Date();

    VectorProduct container = new VectorProduct(getId(), getDescription());

    for(int i=0; i<list.length; i++) {
      if(!list[i].substring(24, 29).equalsIgnoreCase("<DIR>")) {  // don't want to process directories
        fileName = list[i].substring(39);

        try {
          fileDate = dateAndTime.parse(list[i].substring(0, 17));
        } catch(ParseException pe) {
          reportException(new IllegalArgumentException("Can not create Date for: " + list[i].substring(0, 17)));
        }

        BasicProduct bp = new BasicProduct(getId(), getDescription());
        bp.setModificationDate(fileDate);

        try {
          if(path.endsWith("/")) {
            URL url = new URL("ftp://" + host + path + fileName);
            if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
            }
          } else {
            URL url = new URL("ftp://" + host + path + '/' + fileName);
            if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
            }
          }
        } catch(MalformedURLException mue) {
          mue.printStackTrace();
        }

        if(bp.getContent() instanceof URL) {
          container.addElement(bp);
        }
      }
    }

    return container;
  }


  /**
  * Builds a vector product containing URL's that point to the file on
  * the remote machine.
  *
  * @param    list    An FTP directory listing
  */
  protected Product createProductFromWinNT2(String[] list) {
    SimpleDateFormat dateAndTime = new SimpleDateFormat("MMM d HH:mm yyyy");
    SimpleDateFormat dateAndTime1 = new SimpleDateFormat("MMM  d HH:mm yyyy");
    SimpleDateFormat dateAndYear = new SimpleDateFormat("MMM d  yyyy");
    SimpleDateFormat dateAndYear1 = new SimpleDateFormat("MMM  d  yyyy");
    dateAndTime.setTimeZone(TimeZone.getDefault());
    dateAndTime1.setTimeZone(TimeZone.getDefault());
    dateAndYear.setTimeZone(TimeZone.getDefault());
    dateAndYear1.setTimeZone(TimeZone.getDefault());
    dateAndTime.setTimeZone(TimeZone.getDefault());
    Calendar rightNow = Calendar.getInstance();

    String fileName;
    Date fileDate = new Date();

    VectorProduct container = new VectorProduct(getId(), getDescription());

    for(int i=0; i<list.length; i++) {
      if(!list[i].substring(0, 1).equalsIgnoreCase("d")) {  // don't want to process directories
        fileName = list[i].substring(59);

        try {
            fileDate = dateAndTime.parse(list[i].substring(46, 58) + ' ' + rightNow.get(Calendar.YEAR));
          } catch(ParseException pe) {
            try {
              fileDate = dateAndTime1.parse(list[i].substring(46, 58) + ' ' + rightNow.get(Calendar.YEAR));
            } catch(ParseException pe1) {
              try {
                fileDate = dateAndYear.parse(list[i].substring(46, 58));
              } catch(ParseException pe2) {
                try {
                  fileDate = dateAndYear1.parse(list[i].substring(46, 58));
                } catch(ParseException pe3) {
                  reportException(new IllegalArgumentException("Can not create Date for: " + list[i].substring(46, 58)));
                }
              }
            }
          }
          

        BasicProduct bp = new BasicProduct(getId(), getDescription());
        bp.setModificationDate(fileDate);
        
//        System.out.println ("fileDate: " + fileDate);

        try {
          if(path.endsWith("/")) {
            URL url = new URL("ftp://" + host + path + fileName);
            if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
            }
          } else {
            URL url = new URL("ftp://" + host + path + '/' + fileName);
            if(!url.getFile().endsWith("/")) {
                bp.setContent(url);
            }
          }
        } catch(MalformedURLException mue) {
          mue.printStackTrace();
        }

        if(bp.getContent() instanceof URL) {
          container.addElement(bp);
        }
      }
    }

    return container;
  }

	private void reportUnknownFormat(String[] list){
		int lines = list.length;
		if (lines > 10)
			lines = 10;
		StringBuffer str = new StringBuffer(2000);
		str.append(" begin FTP fragment dump");
		str.append('\n');
		for (int i = 0; i< lines; i++ ) {
			str.append(" >");
			str.append(list[i]);
			str.append(" <");
			str.append('\n');
		}
		str.append(" end FTP fragment dump");
		str.append('\n');
		reporter.assert(false,str.toString(),null,null,null);
	}
}
