/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.URL;
import java.net.URLEncoder;
import java.net.MalformedURLException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import org.mitre.pam.getter.search.parse.LoggingHTMLParser;
import org.mitre.pam.getter.search.parse.ParseException;
import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pdif.eventParser.ValueList;

/**
* This PIM fetches quotes (for any mutual fund, market index, or corporate bond)
* and reports a given quote and optionally if it is below a min price
* or above a max price.
*
* Input example:
*
* Property = symbols, value = { "FMAGX:100:150", "JAGIX:18", "IBM" }
*
* Symbols property takes single or multiple tick symbols each optionally with target prices.
* This example creates a product when the Fidelity Magellen Fund falls below $100 or rises above $150,
* and if Janus G&I falls below $18, and always includes the IBM stock price regardless of price.
*
* The ticker names are case insensitive.
*
* This search would build the following URL: http://quote.yahoo.com/q?d=v1&s=fmagx+jagix+ibm
* Note that the yahoo script and query syntax is subject to change as well as the
* format of the query result.
*
* @author Jason Mathews
*/

public class GetQuotePIM extends PIMShell1 {

    private static final String SYMBOLS = "symbols";	 // property name

    private static final String DEBUG_KEY = "logQuotes"; // debugKey

    private static final String[] labels = new String[] {
	"Symbol", "Last trade date", "Last trade price", "Min change", "Change %", "Volume"
    };
    
    private static final String SOURCE_URL = "http://quote.yahoo.com/q?d=v1&s=";

  /**
  * Creates a new <tt>GetQuotePIM</tt> object.
  */
    public GetQuotePIM() {
		setDescription("Stock quotes from yahoo");
		Vector v = new Vector();
		v.addElement("FMAGX:100:150");
		v.addElement("JAGIX:18:");
		v.addElement("IBM");
		ValueList vl = new ValueList(v);
		
		setPropertyValues(SYMBOLS, vl);
		metaMultiStringInfo(SYMBOLS, vl, "list of ticker symbols and their limits such as \"FMAGX:100:150\", \"JAGIX:18:\", \"IBM\"");
    }

    private void getQuote(VectorProduct product, Vector symbols, Hashtable props) throws MalformedURLException, IOException, ParseException {
	
     StringBuffer urlbuf = new StringBuffer(SOURCE_URL);
     urlbuf.append(URLEncoder.encode( (String) symbols.elementAt(0)));
     for (int i =1; i < symbols.size(); i++) {
	 urlbuf.append('+');
	 urlbuf.append(URLEncoder.encode((String)symbols.elementAt(i)));
     }
	
     URL url = new URL(urlbuf.toString());
     reporter.debugMsg("yahoo query for ",getId()," --> ",url, LOG_QUERIES);
	
     BufferedInputStream urlStream = new BufferedInputStream(url.openStream());

      try {
	LoggingHTMLParser parser = new LoggingHTMLParser(urlStream, reporter);

	// Stop after the following HTML line:
	// </td></tr></table><p><table border=0 width="100%" cellpadding=4 cellspacing=0><tr bgcolor="#6699cc"><td><font face=arial><b><big>Quotes</big></b></font></td>
	    
	// System.out.println("Parser - looking for <big>");
	if (!parser.SkipToAfterTag("<big>") || !parser.SkipToString("Quotes") || !parser.SkipToAfterOpenTag(LoggingHTMLParser.TABLE)) {
	    throw new ParseException("Unable to parse " + url.getHost() + " content");
	}

	// System.out.println("\nParser - looking for <TD>");
	    
	for (int tick=0; parser.SkipToAfterOpenTag(LoggingHTMLParser.TR) && tick < symbols.size(); tick++) {

	  reporter.debugMsg("Parse quote for ", symbols.elementAt(tick),null,null, DEBUG_KEY);
	  Vector v = new Vector(6);

/*
Parse out quotes from the following HTML table:

<table border=1 cellpadding=1 cellspacing=0>
<tr bgcolor="#dcdcdc">
<th nowrap>Symbol</th>
<th nowrap colspan=2>Last Trade</th>
<th nowrap colspan=2>Change</th>
<th nowrap>Volume</th>
<th nowrap>More Info</th>

</tr>
<tr align=right>
<td nowrap align=left><a href="/q?s=PBHGX&d=t">PBHGX</a></td>
<td nowrap align=center>Jul 29</td>
<td nowrap><b>24.92</b></td>
<td nowrap><font color="#ff0020">-0.43</font></td>
<td nowrap><font color="#ff0020">-1.70%</font></td>
<td nowrap>N/A</td>
<td nowrap align=center><small>
<a href="/q?s=pbhgx&d=b">Chart</a>
, 
<a href="http://biz.yahoo.com/n/p/pbhgx.html">News</a>
, 
<a href="http://biz.yahoo.com/p/p/pbhgx.html">Profile</a>
</small>
</td>
</tr></table>
*/

	  while (parser.SkipToAfterOpenTag(LoggingHTMLParser.TD) && v.size() < 6) {
	      String s = LoggingHTMLParser.stripHTML(parser.GetUntilClosedTagOfType(LoggingHTMLParser.TD));
	      v.addElement(s);
	  }

	if (v.size() != labels.length) {
	   // could be an unknown symbol
	   reporter.warning("GetQuotePIM unable to parse input: " + v + " for " + url);
	   continue; // try next tick symbol
	}

	double price; // last price

	String s = (String) v.elementAt(2);
	int pos;
	if ((pos = s.indexOf(' ')) != -1) {
		price = Double.valueOf(s.substring(0,pos)).doubleValue();
		// check and add fractional amount (e.g. 14 7/16 -> 14.4375) for stock prices
		int div = s.indexOf('/', pos+1);
		if (div != -1) {
			try {
			 int num = Integer.parseInt(s.substring(pos+1, div));
			 int denum = Integer.parseInt(s.substring(div+1));
			 if (num > 0 && denum > 0) {
				// System.out.println("ADD fraction: " + num + " / " + denum);
				price += (double) num / denum;
			 }
			} catch (NumberFormatException nfe) {
			  reportException(nfe);
			}
		}
	} else price = Double.valueOf(s).doubleValue();

	// debugMsg("last price=\""+s + "\" -> " +price, DEBUG_KEY);

	  String symbol = (String) symbols.elementAt(tick);

	  Double min = (Double) props.get(symbol + "_min");

	  if (min != null) {
	      double minval = min.doubleValue();
	      reporter.debugMsg("target min="+minval+",price="+price,null,null,null, DEBUG_KEY);
	      if (minval != 0.0 && price < minval) {
		  BasicProduct prod = new BasicProduct(symbol, "quote below min price");
		  StringBuffer buf = new StringBuffer();
		  buf.append("Min price: " + min + "\n");
		  for (int i= 0; i < v.size(); i++) {
		      buf.append(labels[i]);
		      buf.append(": ");
		      buf.append(v.elementAt(i));
		      buf.append('\n');
		  }
		  prod.setContent(buf.toString());
		  product.addElement(prod);
		  continue;
	      }
	}//if
	  
	  Double max = (Double) props.get(symbol + "_max");
	  if (max != null) {
	      double maxval = max.doubleValue();
	      reporter.debugMsg("target max="+maxval+",price=" + price,null,null,null, DEBUG_KEY);
	      if (maxval != 0.0 && price > maxval) {
		  BasicProduct prod = new BasicProduct(symbol, "quote above max price");
		  StringBuffer buf = new StringBuffer();
		  buf.append("Max price: " + max + "\n");
		  for (int i= 0; i < v.size(); i++) {
		      buf.append(labels[i]);
		      buf.append(": ");
		      buf.append(v.elementAt(i));
		      buf.append('\n');
		  }
		  prod.setContent(buf.toString());
		  product.addElement(prod);
		  continue;
	      }
	} //if
	else if (min == null) {
	    // no target min or max prices -> default show
	    BasicProduct prod = new BasicProduct(symbol, "quote at $" + gfloat(price));
	    StringBuffer buf = new StringBuffer();
	    for (int i= 0; i < v.size(); i++) {
		buf.append(labels[i]);
		buf.append(": ");
		buf.append(v.elementAt(i));
		buf.append('\n');
	    }
	    prod.setContent(buf.toString());
	    product.addElement(prod);
	  }
        } //for
      } //try
      catch (ParseException e) {
	  reporter.assert(false, "The following error occured on this url - ", url,null,null);
	  reporter.reportException(e);
	  throw e;
      }
      catch (Exception e) {
	 	 reporter.assert(false, "The following error occured on this url - ", url,null,null);
	  	reporter.reportException( e);
	  throw new ParseException(e.getMessage());
      }
      finally {
	  try {
	      urlStream.close();
	  }
	  catch (IOException ignore) {}
      }
    }

  private static String gfloat(double f) {
    f = (int) ((f + 0.005) * 100.)/100.;
    String s = Double.toString(f);
    if (s.endsWith(".0")) return s.substring(0, s.length()-2);
    return s;
  }

  /**
   * Callback to receive a product.
   */
    public Product doFetch() {

	ValueList values = getPropertyValues(SYMBOLS);

	if (values == null) {
	    reporter.error("GetQuotePIM requires at least one ticker symbol");
	    return null;
	}

	Vector symbols = new Vector();
	Hashtable props = new Hashtable();

	reporter.debugMsg("Values: ", values," n="+values.size(),null, DEBUG_KEY);

	for (Enumeration e=values.elements(); e.hasMoreElements(); ) {
	  String s = (String) e.nextElement();

	  Vector tokens = getTokens(s);

	  String symbol = null;
	  if (tokens.size() != 0) {
	    symbol = (String) tokens.elementAt(0);
	    if (symbol.length() == 0) symbol = null;
	  }
	  if (symbol == null) {
	    log("invalid ticker symbol: " + s);
	    continue;
	  }

	  symbols.addElement(symbol);

	  boolean always = true;
	  if (tokens.size() >= 2) {
	    s = (String) tokens.elementAt(1);
	    if (s.length() != 0) {
	      props.put(symbol + "_min", Double.valueOf(s));
	      always = false;
	    }
	    if (tokens.size() == 3) {
	      s = (String) tokens.elementAt(2);
	      if (s.length() != 0) {
		props.put(symbol + "_max",  Double.valueOf(s));
		always = false;
	      }
	    }
	  }

	  if (always) {
	    // symbol only - no min/max targets
	    props.put(symbol + "_always",  new Boolean(true));
	  }
	}// for

	reporter.debugMsg("Query symbols: " + symbols, null,null,null, DEBUG_KEY);
	reporter.debugMsg("Query targets: " + props, null,null,null,DEBUG_KEY);
	
	if (symbols.size() == 0 || props.size() == 0) {
	    reporter.assert(false, "GetQuotePIM requires at least one ticker symbol with min/max prices",null,null,null);
	    return null;
	}

        VectorProduct product = new VectorProduct(getId(), getDescription());

	try {
	    setProgress("Trying GetQuotePIM");
	    getQuote(product, symbols, props);
	}
	catch (IOException ignore) {
	    try {
		Thread.sleep(60000);
		setProgress("Retrying GetQuotePIM (2)");
		getQuote(product, symbols, props);
	    }
	    catch (IOException ignore2) {
		try {
		    Thread.sleep(120000);
		    setProgress("Retrying getQuotePIM (3)");
		    getQuote(product, symbols, props);
		}
		catch (IOException ignore3) {
		    log("Unable to get GetQuotePIM on third try");
		}
		catch (Exception ie3) {}
	    }
	    catch (Exception ie2) {}
	}
	catch (Exception ignoreParseException) {
	}

	if (product.size() == 0) {
	    return null;
	} else if (product.size() == 1) {
	    return product.elementAt(0);
	} else {
	    return product;
	}
  }

  private static Vector getTokens(String token) {
    int pos = -1;
    int curpos;
    Vector tokens = new Vector();
    while (true) {
      curpos = token.indexOf(':', pos+1);
      if (curpos == -1) {
	tokens.addElement(token.substring(pos+1).trim());
	break;
      }
      tokens.addElement(token.substring(pos+1, curpos).trim());
      pos = curpos;
    }
    return tokens;
  }

// /*
    public static void main (String argv[]) throws Exception {
	new org.mitre.pam.executive.Exec();

      // enable debugging
      org.mitre.pam.util.alarm.Alarm.setDebug(DEBUG_KEY, true);
      org.mitre.pam.util.alarm.Alarm.setDebug(LOG_QUERIES, true);

	GetQuotePIM pim = new GetQuotePIM();
		Vector v = new Vector();
		v.addElement("");
		v.addElement("IBM");
		v.addElement("BA:");
		v.addElement("LU::");
		v.addElement("DELL:18");
		v.addElement("JAGIX:18:");
		v.addElement("GE::150");
		v.addElement("FMAGX:100:150");
      pim.setKind("getQuotePIM");
      pim.setId("test");
 	pim.setPropertyValues("symbols", new ValueList(v));
      Product p = pim.doFetch();
      if (p != null) {
	org.mitre.pam.productVisitor.AsPlainText apt =
	  new org.mitre.pam.productVisitor.AsPlainText();
	java.io.StringWriter sw = new java.io.StringWriter();
	apt.visit(p);
	apt.writeResults(new org.mitre.pam.util.PrintLFWriter(sw));
	System.out.println(sw.toString());
      }

	System.exit(0);
    }
// */

}
