/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import java.util.Enumeration;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.HttpURLConnection;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.StreamCopy;
import org.mitre.pam.util.smtp.MIMEc;
import org.mitre.pam.util.URLHelper;
import org.mitre.pdif.eventParser.ValueList;

/**
* Gets a set of files which are defined by a list of URLs and saves each to a
* local file in its working directory.
*
* @author Ayal Spitz
* @author Jason Mathews (many changes/added basic authentication + smart file naming
*                       Add checking for http response codes to not store failed requests +
*                       Close streams in finally clause)
*                       Added URLHelper to get URLConnection as GET or POST request method.
* @author Nazario Irizarry removed calls to setProperty that were no longer needed
*/
public class GetURL extends PIMShell1 {

    private static final String DEBUG_KEY = "GetURL";

    public static final String AUTH_USER_NAME = "authUser";
    public static final String AUTH_PASSWORD = "authPassword";
    public static final String SOURCE_URL = "URLsToFetch";
    public static final String POST_METHOD = "usePostMethod";

    private String authEncoding;
	private static final String[] DEF_URLS = {"http://x.y.z/url1", "http://x.y.z/urln"};

    public GetURL() {
      setDescription("Get list of URLs and save each as a local file");
      
      ValueList vl = new ValueList(DEF_URLS);

      metaStringInfo(AUTH_USER_NAME,"","user name if required to access url");
      metaStringInfo(AUTH_PASSWORD,"","password if required to access url");
      metaBooleanInfo(POST_METHOD,false,"flag to get URL as POST method (default=GET)");
      metaMultiStringInfo(SOURCE_URL,vl,"url(s) to access");
	  metaSetIsPassword(AUTH_PASSWORD, "password to access url [if required]");
    }

    public Product doFetch() {
      ValueList urlList = getPropertyValues(SOURCE_URL);

      String user = getStringProperty(AUTH_USER_NAME,"");
      String pass = getStringProperty(AUTH_PASSWORD,"");
      if (user.length() != 0) {
        reporter.debugMsg("Build auth string: " , user , ":" , pass, DEBUG_KEY);
        authEncoding = "Basic " + org.mitre.pam.util.Codecs.base64Encode(user+":"+pass);
      } else {
        authEncoding = null;
      }

      return retrieveURLs(urlList.elements(), urlList.size());
    }

    protected Product retrieveURLs(Enumeration enum, int numURLs) {

      if (numURLs == 0) return null;

      String method = getBooleanProperty(POST_METHOD, false) ? URLHelper.POST : URLHelper.GET;

      VectorProduct vp = new VectorProduct(getId(), getDescription());
      while (enum.hasMoreElements()) {
        BasicProduct bp = getBP((String) enum.nextElement(), null, null, method);
        if (bp != null)
          vp.addElement(bp);
      }

      if (vp.size() == 0)
        return null;
      else if (vp.size() == 1) {
        BasicProduct bp = (BasicProduct) vp.elementAt(0);
        bp.setDescription(getDescription());
       return bp;
      }
      else
        return vp;
    }

    /**
     * Create a file in the working directory given the base file name from the URL.
     * Uses the extension mapped to the mime type if available otherwise uses
     * the extension from the URL.
     */
    private File getTempFile(String file, String defExtension) {
	File workingDir = getWorkingDir();
	String tempFile;
	String prefix = null;
	String ext = defExtension;

	if (file != null) {
	    int pos = file.lastIndexOf('.');
	    if (pos > 0) {
		prefix = file.substring(0, pos);
		if (ext == null || ext.length() == 0)
		    ext = file.substring(pos+1); // use extension of file
	    }
	    else if (pos == 0) {
		prefix = file.substring(1); // i.e. ".cshrc"
	    }
	    else prefix = file; // no extension (use default)
	} else prefix = "index";

	if (ext == null) {
		tempFile = prefix;  // extension cannot be determined
	} else {
		tempFile = prefix + "." + ext;
	}

	return new File(workingDir, tempFile);
    }

    protected BasicProduct getBP(String strURL, String id, String Dis, String method) {
      BasicProduct  bp = null;
      FileOutputStream tempFileStream = null;
      InputStream urlStream = null;

      try {
	    URLHelper urlHelper = new URLHelper(strURL, method);

	    // check if authentication credentials are provided
	    if (authEncoding != null) {
		reporter.debugMsg("send auth: ", authEncoding, null,null, DEBUG_KEY);
		urlHelper.setRequestProperty("Authorization", authEncoding);
	    }

	    URL tempURL = urlHelper.getURL();
            URLConnection urlc = urlHelper.openConnection();

	    if (urlc instanceof java.net.HttpURLConnection) {
	      java.net.HttpURLConnection httpConn = (java.net.HttpURLConnection) urlc;
	      int response = httpConn.getResponseCode();
	      // catch 4xx client errors and 5xx server errors
	      if (response >= HttpURLConnection.HTTP_BAD_REQUEST) {
		warning("Unable to get URL " + strURL + " -- " + response + " " +httpConn.getResponseMessage());
		return null;
	      }
	    }

	    urlStream = urlc.getInputStream();

	    /*
	     * Get mime type then build filename using basename
	     * from the URL and the mimetype for its extension
	     */
	    String mimeType = urlc.getContentType();
	    reporter.debugMsg("Url=", strURL, ", MIMEtype=", mimeType, DEBUG_KEY);

	    String sourceFileName = null;
	    String tempName = org.mitre.pam.util.Codecs.URLdecode(tempURL.getFile());
	    String prodName;
	    if (tempName != null) {
              int slashPos = tempName.lastIndexOf('/');
              if (slashPos != -1 && slashPos + 1 < tempName.length()) {
		// truncate any query parameters off the URL to get only the filename.
                int paramPos = tempName.indexOf('?', slashPos);
                if (paramPos != -1)
                  sourceFileName = tempName.substring(slashPos+1, paramPos);
                else
                  sourceFileName = tempName.substring(slashPos+1);
              }
	    }
	    File tempFile;
	    if (sourceFileName == null || sourceFileName.length() == 0) {
              prodName = getId();
              tempFile = getTempFile(null, MIMEc.findExtension(mimeType));
	      // use the source file name that was generated
              sourceFileName = tempFile.getName();
	    } else {
              prodName = sourceFileName;
              tempFile = getTempFile(sourceFileName, MIMEc.findExtension(mimeType));
	    }

            long fileDate = urlc.getLastModified();
	    if (fileDate == 0)
	      fileDate = System.currentTimeMillis();

	    FileRef tempFileRef = new FileRef(tempFile, fileDate, sourceFileName);

	    tempFileStream = new FileOutputStream(tempFile);

	    StreamCopy.copyFully(urlStream, tempFileStream, false);

	    if (mimeType != null)
		tempFileRef.setMIMEType(mimeType);

	    bp = new BasicProduct(prodName, strURL);
	    bp.setContent(tempFileRef);

	} catch (MalformedURLException e) {
	    reporter.assert(false, "Error: Malformed URL <" , strURL , ">",null);
	} catch (IOException e) {
	    reportException(e);
	} finally {
	  if (tempFileStream != null)
	    try {
	    tempFileStream.close();
	  } catch (IOException ignore) {}
	  if (urlStream != null)
	    try {
	    urlStream.close();
	  } catch (IOException ignore) {}
	}

	return bp;
    }
}
