/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.ProductPool;
import org.mitre.pam.interfaces.ProductFactory;

import org.mitre.pam.interfaces.Product;
import org.mitre.pdif.eventParser.ValueList;


/**
* Test PIM designed to help evaluate
* throughput of a near-realtime iPAM thread. <P>
*
* @author Nazario Irizarry, Jr.
*/
public class TestAsyncSource extends TestPIMByteArray implements AsynchronousAgent, Runnable {

	private final static String DELAY = "sendToSendDelayMillis";
	private final static String CYCLE_TIME = "computedSendCycleTimeSecs";
	private final static String SEND_TIME = "computedSendTransferTimeSecs";
	private final static String PRIORITY = "priority";
	private final static String REPETITIONS = "repetitions";
	private final static String AVERAGE_ITER_TIME = "longTermCycleTimeSecs";
	private final static String PRODUCT_POOL_SIZE = "sizeOfInternalProductPool";

	long sendTime;
	long delayTime;
	Thread thread;
	Product cannedProduct;
	boolean running;
	
	public TestAsyncSource () {
		setDescription("Computes throughput time over 16 transfers");

		setProperty(META_GUI_GROUP, TestPIM1.TEST_CATEGORY);

		metaIntegerInfo(REPETITIONS, 25, "Number of times to send product (-1 means unending)");
		metaDoubleInfo(DELAY, 0.0, "Amount (secs) to delay between each send");
		metaDoubleInfo(CYCLE_TIME, 16, "Computed total time in seconds based on throughput");
		metaDoubleInfo(SEND_TIME, 16, "Computed transfer time ( = delay + transfer) in seconds");
		metaIntegerInfo(PRIORITY, 5, "Java thread priority");
		metaIntegerInfo(PRODUCT_POOL_SIZE, 0, "Number of items in the product pool");
	}

	
	public void setProperty(String name, String value) {
		super.setProperty(name,value);
		if (name.equals(DELAY)) {
			delayTime = getIntegerProperty(DELAY,0);
		}
		else if (name.equals(PRIORITY)) {
			if (thread != null) {
				thread.setPriority(getIntegerProperty(PRIORITY,5));
			}
		}
	}
	
	public void enableService() {
		if (thread == null) {
			thread = new Thread(this);
			thread.setPriority(getIntegerProperty(PRIORITY,5));
			running = true;
			thread.start();
		}
	}

	public void disableService() {
		running = false;
	}

	public void run() {
		if (logActivity) {
			reporter.log("starting");
		}
		Product cannedProduct = super.doFetch();
		int iters = getIntegerProperty(REPETITIONS, 25);
		double iterdiv = iters; // do this up here in case the user changes mid stream
		boolean forever = iters < 0;
		long iterCount = 0;
		
		if (iters == 0) {
			thread = null;
			return;
		}
		
		int productPoolSize = getIntegerProperty(PRODUCT_POOL_SIZE, 25);
		ProductPool productPool = null;
		if (productPoolSize > 0) {
			productPool = makeProductPool(productPoolSize);
		}
		
		setProperty(AVERAGE_ITER_TIME,"Computed at end of iterations");
		long iterStart = System.currentTimeMillis();

		while((forever || --iters >= 0) && running) {
			iterCount++;
			long startTime = System.currentTimeMillis();
			if (productPool == null) {
				cannedProduct.setModificationDate(startTime);
				pushProductToAllClients(cannedProduct);
			}
			else {
				try {
					Product p = productPool.getBlocking(this);
					pushProductToAllClients(p);
					productPool.doneWithProduct(this, p);
				}
				catch (InterruptedException e) {
					reporter.warning("interrupted while getting next product from pool");
				}
			}
			sendTime = ((System.currentTimeMillis() - startTime) + 15*sendTime)/16;
			try {
				Thread.sleep(delayTime);
			}
			catch (InterruptedException e) {};	
		}
		if (iterCount > 0) {
			double avgIter = 0.001*((double)(System.currentTimeMillis() - iterStart) / iterCount);
			setProperty(AVERAGE_ITER_TIME, Double.toString(avgIter));
		}
		else {
			setProperty(AVERAGE_ITER_TIME, "Nothing sent");
		}
		thread = null;
		if (productPool != null) {
			productPool.dispose(); // must explicitly dispose of the pool
			productPool = null;
		}
		if (logActivity) {
			reporter.log("done");
		}
	}
	
	private void recomputeTimes() {
		double cycleTime = (sendTime + delayTime)/1000.0;
		setProperty(CYCLE_TIME, Double.toString(cycleTime));
		setProperty(SEND_TIME, Double.toString(((double)sendTime)/1000.0));
	}
	
	public Memento getMemento() {
		recomputeTimes();
		return super.getMemento();
	}
	public String getProperty(String name) {
		if (name.equals(CYCLE_TIME) || name.equals(SEND_TIME)) {
			recomputeTimes();
		}
		return super.getProperty(name);
	}
	public ValueList getPropertyValues(String name) {
		if (name.equals(CYCLE_TIME) || name.equals(SEND_TIME)) {
			recomputeTimes();
		}
		return super.getPropertyValues(name);
	}
/*
* Make a product factory with the specified number of products in the pool.
* The product pool is generic.  It is up to us to populate it with
* the right kinds of basic products.
*/
	private ProductPool makeProductPool(int productPoolSize) {
		ProductFactory pfab = exec.getProductFactory();
		Product poolMembers[] = new Product[productPoolSize];

		for (int i = 0; i< productPoolSize; i++) {
			poolMembers[i] = pfab.makeProduct("member "+i,"","");
		}
		
		return pfab.makeProductPool(this, poolMembers);
	}
}
