/* Copyright 2000 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter.remote;

import org.mitre.pam.getter.PIMShell1;
import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.productVisitor.Serializer;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.zip.InflaterInputStream;

/**
 * An input agent that will listen for incoming serialized products
 * from another iPAM's putSerializedProduct agent.
 *
 * This agent is Asynchronous -- it creates a product on demand when it is received.
 *
 * This has the following configurable parameters (defaults in parens):
 *   listenPort       - port number where this agent should listen (19099)
 *
 * This is a sample configuration to help new users of this agent set it up:
 *   (listenPort 19099)
 *
 * NOTE: More than one GetSerializedProduct agent can run on the same PAM, but
 * each must listen on a different server port.  An AcceptSerializedProduct agents
 * can, however, receive products from multiple PutSerializedProduct agents
 * on different hosts.
 *
 * @author Jason Mathews   Based on GetAsyncUDP by Jones Leung.
 *
 * @since iPAM 2.5.101 2-Mar-2000
 *
 * @see org.mitre.pam.puter.remote.PutSerializedProduct
 * @see org.mitre.pam.util.Serializer
 * @see org.mitre.pam.interfaces.AsynchronousAgent
 */

public class AcceptSerializedProduct extends PIMShell1 implements AsynchronousAgent, SerializedConstants {

  protected final static String LISTEN_PORT = "listenPort";

  private transient ReceiveThread listener;
  private transient ServerSocket ss;
  private static final String DEBUG_KEY = "GetSerializedProduct";

  public AcceptSerializedProduct() {
    setDescription("Asynchronously retrieve serialized products");
    metaIntegerInfo(LISTEN_PORT, DEFAULT_PORT, "the port to listen on");
  }

  /**
   * This is where the agent will register with the registry and will start the
   * SMTPserver if it is not already running.
   */
  public void enableService()
  {
    int listenPort = getIntegerProperty(LISTEN_PORT, -1);
    if (listenPort < 0) {
      reporter.error("Illegal " + LISTEN_PORT + " value: " + getProperty(LISTEN_PORT));
      return;
    }

    try {
      // Create a ServerSocket
      ss = new ServerSocket(listenPort);

      // specify 0 uses any free port 
      if (listenPort == 0)
	log("Listening on serverSocket port=" + ss.getLocalPort());

      // ... and start listening
      listener = new ReceiveThread(ss, reporter);
      listener.setDaemon(true);
      listener.start();
    } catch (Exception e) {
      reportException (e);
    }
  } /* enableService */

  /**
   * This is where the listener thread will be shut down if it is still running
   */
  public void disableService()
  {
    if (listener != null) {
      listener.running = false;
      listener = null;
    }

    // close ServerSocket
    if (ss != null) {
      try {
	ss.close();
      } catch (IOException ioe) {}
      ss = null;
    }
  }

  /**
   * This agent is completely asynchronous.  All incoming Prodcuts are injeced
   * into the system immediately using pushProductToAllClients().
   * doFetch() will always return null.
   */
  public Product doFetch() {
    return null;
  }

  private void processSerializedStream(InputStream is) throws IOException {
    try {
      reporter.debugMsg("Reading serialized object from stream",null,null,null,DEBUG_KEY);
      InflaterInputStream iis = new InflaterInputStream(is);
      ObjectInputStream ois = new ObjectInputStream(iis);
      Object o = Serializer.unSerialize(ois, getWorkingDir());
      reporter.debugMsg("Unserialized object",o,null,null,DEBUG_KEY);
      ois.close();
      if (o instanceof Product)
	pushProductToAllClients((Product)o);
      else
	reporter.error("Serialized object not a product: " +
		       (o == null ? o : o.getClass()));
    }
    catch (ClassNotFoundException nfe) {
      reporter.reportException(nfe);
    }
  }

  final class ReceiveThread extends Thread {

    ServerSocket ss;
    Reporter reporter;
    boolean running = true;

    public ReceiveThread (ServerSocket ss, Reporter reporter) {
      this.ss = ss;
      this.reporter = reporter;

      // this listening socket should never time out
      try {
	ss.setSoTimeout(0);
      } catch (Exception e) {
	reporter.reportException(e);
      }
    }

    /**
     * Listen for incoming properties, and pass contents to agent.process()
     */
    public void run() {
      Socket socket = null;
      while (running) {
	try {
	  socket = ss.accept();
	  processSerializedStream(socket.getInputStream());
	} catch (Exception e) {
	  reporter.reportException(e);
	} finally {
	  if (socket != null)
	    try {
		socket.close();
	    } catch (IOException ioe) {}
	}
      }
    } /* run */

  } /* ReceiveThread */

}
