/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter.search.bnn;

import java.util.Hashtable;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;
import java.util.StringTokenizer;
import java.net.URL;
import java.io.IOException;
import java.io.File;

import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.getter.PIMShell1;
import org.mitre.pdif.eventParser.ValueList;

/**
 * This PIM fetches a vector of stories from the 
 * Mitre Broadcast News Navigator (BNN).
 *
 * The stories returned are all the matching stories in the BNN database from
 * the specified number of days back.<p>
 *
 * 8/10/98: Added passthru of metadata such as named entities<p>
 *
 * 4/22/98: Can now take advantage of BNN features allowing topic, person,
 * organization, and location searches.  These keywords are, however, limited
 * to the list provided by BNN in their user interface (they are case
 * sensitive)  <p>
 *
 * 3/12/98: If a search string consisting of a number of words, separated
 * by spaces is specified, then stories matching all of these words are
 * returned.  <p>
 *
 *
 * @author Michael DeCorpo
 * @author Srinivas Inguva (added search capability)
 * @author Jason Mathews (added multi value searches)
 */

public class BNNPIM3 extends PIMShell1 {

	private static final String TIME_BACK = "nDaysToSearchBack";
	private static final String SEARCH_STRING = "searchWords";
	private static final String TOPIC_STRING = "findTopic";
	private static final String PERSON_STRING = "findPerson";
	private static final String ORG_STRING = "findOrg";
	private static final String LOC_STRING = "findLocation";
	private static String DEFAULT_DESCRIPTION = "Stories from the MITRE Broadcast News Navigator.";
	private static final String BNN_HOST_QUERY = "bnnHostQuery";

  //GeoNODE properties
  public static final String GeoNODE_SourceProgram = "sourceProgram";
  public static final String GeoNODE_DATE = "date";
  public static final String GeoNODE_LOC = "namedEntityLocation";
  public static final String GeoNODE_ORG = "namedEntityOrganization";
  public static final String GeoNODE_PERSON = "namedEntityPerson";
  public static final String SOURCE_SPEC = "sourceSpec";

  public BNNPIM3 () { 
    setDescription(DEFAULT_DESCRIPTION);
    setProperty(TIME_BACK, "1");
    setProperty(SEARCH_STRING, "");
    setProperty(TOPIC_STRING, "");
    setProperty(PERSON_STRING, "");
    setProperty(ORG_STRING, "");
    setProperty(LOC_STRING, "");
    setProperty(BNN_HOST_QUERY,"http://videoserver3.mitre.org/tr/owa/BNN_revh.BnnFindStories");
    setProperty(SOURCE_SPEC,"CNN_Prime_News");

    metaIntegerInfo(TIME_BACK, 1,"Number of days to search");
    metaStringInfo(SEARCH_STRING, "","Search string");
    metaStringInfo(TOPIC_STRING, "", "News topic");
    metaMultiStringInfo(PERSON_STRING, new ValueList(""),"People (can be more than one) to search for");
    metaStringInfo(ORG_STRING, "","Organization to search for");
    metaStringInfo(LOC_STRING, "","Location to search for");
    metaStringInfo(BNN_HOST_QUERY,"http://videoserver3.mitre.org/tr/owa/BNN_revh.BnnFindStories","URL for BNN server");
    metaStringInfo(SOURCE_SPEC,"CNN_Prime_News","BNN news source");
  }

/**
* Callback to receive a command.  We ignore the
* command for now since we can only do one thing - fetch.
*/
  public Product doFetch() {
    setProgress("beginning");
    
    String host = getStringProperty(BNN_HOST_QUERY,"");
    if (host == null || host.length()==0)
      throw new IllegalArgumentException("Attribute "+BNN_HOST_QUERY+" must be supplied ");

    long nDays = 1;
    try {
      nDays = Long.parseLong(getStringProperty(TIME_BACK,"1"));
    }
    catch (NumberFormatException e) {
      reporter.warning( TIME_BACK + "value = " + getStringProperty(TIME_BACK,"1") + "; is not a valid integer");
    }
    long date = System.currentTimeMillis() - nDays*1000*60*60*24;
    if (date <= 0) {
      throw new IllegalArgumentException("Value for " + TIME_BACK + " is out of range");
    }
    Date startDate = new Date(date);

    ValueList sourceList = getPropertyValues(SOURCE_SPEC);

    GetBNN getbnn = new GetBNN(host, sourceList, reporter);

    String searchString = getStringProperty(SEARCH_STRING,"");
    if (searchString.length() != 0)
      getbnn.setSearchString(searchString);
    ValueList topics = getPropertyValues(TOPIC_STRING);
    if (topics.size() != 0)
      getbnn.setTopics(topics);
    ValueList people = getPropertyValues(PERSON_STRING);
    if (people.size() != 0)
      getbnn.setPerson(people);
    String orgString = getStringProperty(ORG_STRING,"");
    if (orgString.length() != 0)
      getbnn.setOrg(orgString);
    String locString = getStringProperty(LOC_STRING,"");
    if (locString.length() != 0)
      getbnn.setLoc(locString);

    try {
      setProgress("Trying bnn");
      getbnn.getNow(startDate);
    }
    catch (IOException ignore) {
      try {
	Thread.sleep(60000);
	setProgress("Retrying bnn (2)");
	getbnn.getNow(startDate);
      }
      catch (IOException ignore2) {
	try {
	  Thread.sleep(120000);
	  setProgress("Retrying bnn (3)");
	  getbnn.getNow(startDate);
	}
	catch (IOException ignore3) {
	  log("Unable to get bnn on third try");
          // return;
	}
	catch (Exception ie3) {}
      }
      catch (Exception ie2) {}
    }
    catch (Exception ignoreParseException) { 
    }    
    
    Vector list = getbnn.getStories();

    if (list.size() == 0) {
      return null;
    }

//    int ndays = Integer.parseInt(getProperty(TIME_BACK));

    if (searchString.length() != 0) {
      setDescription( " Stories with keyword(s): " + searchString );
    }

    VectorProduct container = new VectorProduct(getId(), getDescription());

      Enumeration e = list.elements();
      // add summaries
      while (e.hasMoreElements()) {
	BNNStory it = (BNNStory) e.nextElement();

	// Make each story a BasicProduct with id='Story # - date',
	// description=the story summary, content=the URL of the search.
	BasicProduct prod = new BasicProduct(it.id , it.date + " (Summary) " +  it.summary);
      addProperties(prod, it);
      String url = getbnn.getURLfromStory(it);
      try {
        prod.setContent(new URL(url));
      }
      catch (Exception exc) {
        prod.setContent(url);
      }
      container.addElement(prod);
    }

      // add stories
    for (e = list.elements(); e.hasMoreElements(); ) {
	BNNStory it = (BNNStory) e.nextElement();

	// Make each story a BasicProduct with id='Story # - date',
	// description=the story summary, content=the URL of the search.
	BasicProduct prod = new BasicProduct(it.id, it.date + " (Story) " +  it.story);
      addProperties(prod, it);
      String url = getbnn.getURLfromStory(it);
      try {
        prod.setContent(new URL(url));
      }
      catch (Exception exc) {
        prod.setContent(url);
      }
      container.addElement(prod);
    }

      return container;
    
  }//end of doFetch()

  private void addProperties(Product p, BNNStory story) {
    p.setProperty(GeoNODE_SourceProgram, story.source);
    p.setProperty(GeoNODE_DATE, story.date);
//    p.setProperty(GeoNODE_ID, story.id); // already done when product was created

    Vector v = new Vector();
    StringTokenizer st = new StringTokenizer(story.locations, ",");
    while (st.hasMoreElements()) {
      v.addElement(st.nextElement());
    }
    ValueList vl = new ValueList(v);
    p.setPropertyValues(GeoNODE_LOC, vl);

    v = new Vector();
    st = new StringTokenizer(story.people, ",");
    while (st.hasMoreElements()) {
      v.addElement(st.nextElement());
    }
    vl = new ValueList(v);
    p.setPropertyValues(GeoNODE_PERSON, vl);

    v = new Vector();
    st = new StringTokenizer(story.organizations, ",");
    while (st.hasMoreElements()) {
      v.addElement(st.nextElement());
    }
    vl = new ValueList(v);
    p.setPropertyValues(GeoNODE_ORG, vl);
  }
}
