/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter.search.smarTraveler;

import java.util.Hashtable;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;
import java.net.URL;
import java.io.IOException;
import java.text.SimpleDateFormat;

import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.getter.PIMShell1;

/**
* This PIM fetches a vector of traffic reports from the SmarTraveler website
* whose URL is specified in the souceSpec attribute. All information is
* returned unless a search key (e.g. "128 NB") is defined in which case the
* defined method (exact or substring matches) is performed on each paragraph
* and only those matching are part of the product.
*
* Input example:
*
*  souceSpec: "http://www.smartraveler.com/scripts/bostraffic.asp?index=6&city=bos&cityname=Boston"
*  key:       "128 NB"		(optional search string; default = no filter)
*  method:    "substring"	(default = "exact" if key is specified)
*
* @author Jason Mathews
*/

public class TravelerPIM extends PIMShell1 {

  // cities that use new SmarTraveler format as of 4/20/1999
  protected final static String newCities[] = new String[] {
    "bos", // Boston
    "wdc", // Washington DC
    "nyc"  // New York
  };
  // Philadelphia, Cincinnati still use older smarTraveler format

  private static final String DEF_URL = "http://www.smartraveler.com/scripts/bostraffic.asp?index=6&city=bos&cityname=Boston";

  private static final String KEY = "key";
  private static final String METHOD = "method"; // for backward compatibility
  private static final String SOURCE_URL = "sourceURL";
  protected static final String SHORT_NAMES = "useShortRouteNames";

  public TravelerPIM() {
    setDescription("Travel reports from www.SmarTraveler.com");
    setProperty(KEY, "");
    setProperty(SOURCE_URL, DEF_URL);
    // setProperty(SHORT_NAMES, "false");
    metaStringInfo(KEY, "","Key to parse on part of the traffic report; e.g. West");
    metaStringInfo(SOURCE_URL, DEF_URL,"Full url for your SmarTraveler query");
    // metaBooleanInfo(SHORT_NAMES, false, "Abbreviate route names where appropriate; e.g. 128 North -> 128N");
  }

  private BasicProduct buildNewStory (TravelerStory ts, SimpleDateFormat formatter) {
//    System.out.println("Creating Basic product: " + ts.title + "|" + ts.description + "|" + ts.story);
    String id = ts.title;
    if (id == null || id.equals("")) id = getId();
    String description = ts.description;
    BasicProduct prod = new BasicProduct(id, description == null ? "" : description);
    prod.setContent(ts.story);
    if (ts.date != null) {
      try {
	Date date = formatter.parse(ts.date);
	prod.setModificationDate(date);
      }
      catch (java.text.ParseException ignoreException) {}
    }
    return prod;
  }


/**
* Callback to receive a command.  We ignore the
* command for now since we can only do one thing - fetch.
*/
  public Product doFetch() {

    String urlString = getStringProperty(SOURCE_URL, "");
    if (urlString.length() == 0) {
	reporter.assert(false, "SmarTravelerPIM requires a ", SOURCE_URL, " property", null);
	return null;
    }

    reporter.debugMsg("Traveler query for ",name," --> ",urlString, LOG_QUERIES);

    String key = getStringProperty(KEY, "");

    GetTraveler gettraveler;

    if (useNewFormat(urlString)) {
      gettraveler = new GetTravelerBos(urlString, reporter, key, getId());
    } else {
      // otherwise use the old SmarTraveler format
      String methProp = getStringProperty(METHOD,"");
      int method;
      if (key == null || key.length() == 0) method = TravelerStory.NONE;
      else if (methProp == null || methProp.equalsIgnoreCase("EXACT"))
	method = TravelerStory.EXACT;
      else
	method = TravelerStory.SUBSTRING;
      gettraveler = new GetTravelerWash(urlString, reporter, key, method, getId());
    }

    if (getBooleanProperty(SHORT_NAMES, false))
	gettraveler.setShortLabels(true);

    try {
      setProgress("Trying SmarTraveler");
      gettraveler.getNow();
    }
    catch (IOException e) {
      if (urlString.startsWith("file:")) {
		reporter.error("Unable to get SmarTraveler info from " + urlString);
		reportException(e);
		return null;
	}
      try {
	Thread.sleep(60000);
	setProgress("Retrying SmarTraveler (2)");
	gettraveler.getNow();
      }
      catch (IOException ignore2) {
	try {
	  Thread.sleep(120000);
	  setProgress("Retrying SmarTraveler (3)");
	  gettraveler.getNow();
	}
	catch (IOException e3) {
	  log("Unable to get SmarTraveler on third try");
	  reportException(e3);
	}
	catch (Exception ie3) {}
      }
      catch (Exception ie2) {}
    }
    catch (Exception ignoreParseException) {}

    Vector list = gettraveler.getStories();

    if (list.size() == 0) {
		return null;
    }
    else {
      // If there is only one product then create a single BasicProduct
      // otherwise create a vectorProduct container for each product.
      // use date format e.g. 6/9/98 09:09 am
		SimpleDateFormat formatter = new SimpleDateFormat("M/d/yy hh:mm a", java.util.Locale.US);
		if (list.size() == 1) {
			return buildNewStory( (TravelerStory) list.firstElement(), formatter);
		} else {
			Enumeration e = list.elements();
			// gettraveler.getTitle()
			VectorProduct container = new VectorProduct(getId(), "");
			while (e.hasMoreElements()) {
			container.addElement( buildNewStory( (TravelerStory) e.nextElement(), formatter));
			}
			return container;
		}
    }
  }
  
/* Check if URL uses the new SmarTraveler format which is determined by the
 * city parameter in the URL and matched against the list of new cities:
 * e.g. http://www.smartraveler.com/scripts/bostraffic.asp?index=6&city=bos&cityname=Boston
 */
private boolean useNewFormat(String urlString) {
  int ind = urlString.indexOf("city=");
  if (ind < 0) return false;
  int endpos = urlString.indexOf('&', ind);
  if (endpos < 0) endpos = urlString.length();
  String city = urlString.substring(ind+5, endpos);
  for (int i=0; i < newCities.length; i++)
    if (newCities[i].equals(city)) return true;
  return false;
}

 /*
 * for backward compatibility
 */
	public void setProperty(String attribute, String value) {
		if (attribute.equalsIgnoreCase("sourceSpec")) {
			attribute = SOURCE_URL;
		}
		super.setProperty(attribute, value);
	}

}

