/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.monitor.watcher;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.awt.Point;
import java.awt.Dimension;
import java.util.Random;

import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.interfaces.DIM;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pam.interfaces.Reporter;


/**
* A cluster is a group of agents that are connected to each other.
* Two clusters have no agents in common.
* There may only be one cluster if everything is connected to everything.
*
* @author Nazario Irizarry, Jr.
*/
class Cluster extends Island {
	static int clusterSeq;
	private String clusterId = "cluster" + clusterSeq++;
	int verticalBlockSpacing = 0;
	int horizontalBlockSpacing = 100;
	int x;
	int y;
	int width;
	int height;
	int minX;
	int minY;
	int maxX;
	int maxY;
	static Random random = new Random(System.currentTimeMillis());
	Vector verticalSlots;
	private static final NoSuchAgentException NSE = new NoSuchAgentException();
	private static final int CHANGE_THRESHOLD = 1;
	protected int changeCount;
	Reporter reporter;
	private boolean hasDied;
	private Dimension mass;

	Cluster(Hashtable allNodes, Reporter reporter) {
		super(allNodes);
		this.reporter = reporter;
	}

/**
* Allows us to promote an Island into a cluster.
* Once this has been done the Island can no longer be used
* cause we reference its contents and don't expect them to change
* in the Island
*/
	Cluster(Island island, Reporter reporter) {
		this(island.allNodes, reporter);

		Enumeration e = island.items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			addNode(node);
		}
		this.items = island.items;
//		markMembers();
//		layout();
	}
/**
* Once we kill a cluster it can have children for a while
* but we can never grow this cluster
* we only keep the cluster around so that its children
* have a position on the screen for a while
*/
	void kill() {
		hasDied = true;
	}
	
/**
* Is it dead?
*/
	boolean isDead() {
		return hasDied;
	}
/**
* Set our origin to a goal.
* If the new goal is not too far away
* don't move
*/
	void setOrigin(int targetX, int targetY) {
		boolean moved=false;
		if (Math.abs(targetX-x) > 20) {
			x = targetX;
			moved = true;
		}
		if (Math.abs(targetY-y) > 20) {
			y = targetY;
			moved = true;
		}
		if (moved) {
			tellChildrenWeMoved();
		}
	}
	
	int getOriginX() {
		return x;
	}
	
	int getOriginY() {
		return y;
	}
/**
* Tell all member nodes that we moved
*/
	void tellChildrenWeMoved() {
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			node.transferPos();
		}
	}
/**
* Tell the cluster that it needs a new layout
*/
	void layoutNeeded() {
		changeCount++;
	}
/**
* When we add node we mark ourselves as the parent cluster
* and if we have been ordered then we need to add new node
* into the order
*/
	void addNode(NodeData node) {
		if (verticalBlockSpacing == 0) {
			verticalBlockSpacing = node.block.getBoundingRect().height + 5;
		}
		if (items.containsKey(node.agent)) {
			reporter.debugMsg("nothing to add ",node," already in cluster ",clusterId, DEBUG_NAME);
			return;
		}
		changeCount++;
		items.put(node.agent, node);
		node.setParentCluster(this);
	}
/**
* When we remove a node we need to do some housecleaning
* but we don't redo the order of anything
*/
	void removeNode(NodeData node) {
		if (items.remove(node.agent) == null)
			return;
		node.loseParent();
		changeCount++;
	}
/**
* A child node says that they are leaving the cluster.
*/
	void nodeSaysBye(NodeData node) {
		items.remove(node.agent);
		changeCount++;
	}
/**
* Look to see if the node can be be given a horizontal order
* in between the current allocation of source and sink positions.
*
*/	
	private void findPositionInBetween(NodeData node) {
		int insertHPosition = 0;
		int insertVPosition = 0;
		
		ConnectedAgent agent = node.agent;

	/*
	* Look within a rectangle bound by the clusters current extent
	*/
		int maximumLeftPos = minX;
		int minimumRightPos = maxX;
		int minimumTopPos = minY;
		int maximumBottomPos = maxY;
		
	/*
	* Narrow that down based on the rectangle defined by the extent
	* of inputs and outputs
	*/
		if (!(agent instanceof PIM)) {
			Enumeration e = node.agent.getSourceList().elements();
			while (e.hasMoreElements()) {
				try {
					NodeData source = nodeForAgent((ConnectedAgent) e.nextElement());
					if (source.clusterOrderX > maximumLeftPos)
						maximumLeftPos = source.clusterOrderX;
					if (source.clusterOrderY < minimumTopPos)
						minimumTopPos = source.clusterOrderY;
					if (source.clusterOrderY > maximumBottomPos)
						maximumBottomPos = source.clusterOrderY;
				} catch (NoSuchAgentException ignore) {}
			}
		}
		if (!(agent instanceof DIM)) {
			Enumeration e = node.agent.getSinkList().elements();
			while (e.hasMoreElements()) {
				try {
					NodeData sink = nodeForAgent((ConnectedAgent) e.nextElement());
					if (sink.clusterOrderX < minimumRightPos)
						minimumRightPos = sink.clusterOrderX;
					if (sink.clusterOrderY < minimumTopPos)
						minimumTopPos = sink.clusterOrderY;
					if (sink.clusterOrderY > maximumBottomPos)
						maximumBottomPos = sink.clusterOrderY;
				} catch (NoSuchAgentException ignore) {}
			}
		}

	/*
	* PIMs go all the way to the left
	*/
		if (agent instanceof PIM) {
			insertHPosition = 0;
			return;		
		}

	/*
	* DIMs go all the way to the right
	*/
		if (agent instanceof DIM) {
			insertHPosition = maxX;
			return;		
		}
	}
/**
* Add the members of a second cluster as our own
* We expect that the second cluster will cease to be used
* after this.  Nodes cannot belong to two clusters simultaneously.
*/	
	void mergeWith(Cluster other) {
		Enumeration e = other.items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			addNode(node); // also sets layout to invalid
		}
	}
/*
* Get the "center of mass"
*/
	Point getCentroid() {
		int cumx = 0;
		int cumy = 0;
		int count = 0;
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData nodeData = (NodeData) e.nextElement();
			cumx += nodeData.getOffsetInClusterX();
			cumy += nodeData.getOffsetInClusterY();
			count++;
		}
		if (count > 0) {
			return new Point(cumx/count, cumy/count);
		}
		else {
			return new Point(50,50);
		}
	}
/*
* Push each node in the cluster to its goal position
*/
	protected void assignEachNodeAPosition() {
		horizontalBlockSpacing = determineHSpacing();

/*
		if (reporter.isDebugKey(DEBUG_NAME)) {
			reporter.log(" Assign each node a position");
			reporter.log("   horizontal space ="+horizontalBlockSpacing);
			reporter.log("   vertical space   ="+verticalBlockSpacing);
			reporter.log("   x ="+x+",  y ="+y);
		}
*/
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			node.setOffsetInCluster(node.clusterOrderX*horizontalBlockSpacing, node.clusterOrderY*verticalBlockSpacing);
		}
		width = horizontalBlockSpacing * (maxX - minX + 1);
		height = verticalBlockSpacing * (maxY - minY + 1);
		
		adjustForHorizontalLineCase1();
	}
/*
* This case is one where a horizontal line between two nodes goes directly behind
* a third agent that is in between.  We don't want it to look like the third
* agent is also connected as an artifact of the way it is drawn
*/
	protected void adjustForHorizontalLineCase1() {
		for (int column = 0; column < (verticalSlots.size()-1); column++) {
			Vector v = (Vector) verticalSlots.elementAt(column);
			for (int row = 0; row < v.size(); row++) {
				NodeData node = (NodeData)v.elementAt(row);
				if (node == null)
					continue;
				Enumeration enum = outputNodesOn(node);
				while (enum.hasMoreElements()) {
					NodeData outNode = (NodeData) enum.nextElement();
					int xSpan = Math.abs(node.clusterOrderX - outNode.clusterOrderX);
					if (xSpan > 1) {
						int ySpan = Math.abs(node.getOffsetInClusterY() - outNode.getOffsetInClusterY());
						while (ySpan < 3) {
							outNode.setOffsetInClusterY( outNode.getOffsetInClusterY()+ random.nextInt() % 5);
							ySpan = Math.abs(node.getOffsetInClusterY() - outNode.getOffsetInClusterY());
						}
					}
				}
			}
		}
	}
/*
* Push each node in the cluster to its goal position
*/
	private int determineHSpacing() {
		int spacing = 100; // minimum
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			int width = node.block.getBoundingRect().width;
			if (width > spacing)
				spacing = width;
		}
		spacing = ((spacing+9)/10)*10;
//		reporter.debugMsg(this," new hspacing= ",Integer.toString(spacing),null,DEBUG_NAME);
		return spacing;
	}

/*
* Get the width and height
*/
	Dimension getMass() {
		if (mass == null)
			mass = new Dimension(maxX - minX, maxY - minY);
		return mass;
	}
/*
* Empty the cluster
*/
	void clear() {
		items.clear();
	}
/*
* Assign a layout order left-to-right and top-to-bottom to
* every cluster member.  As this currently stands it gets a sequencing
* which may unnecessarily have a distance of more than one between
* neighbors.  This is because an iteration may push a neighbor without
* moving its community, but this is ok for now.
*/
	void layout() {
		reporter.debugMsg("** Layout cluster :",this,null,null,DEBUG_NAME);
		Enumeration e = items.elements();
		NodeData startNode = null;
		minX = minY = maxX = maxY = 0;
	/*
	* clear order
	*/
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			if (startNode == null)
				startNode = node;
			node.clusterOrderX = 0;
			node.clusterOrderY = 0;
			node.updateConnectionInfo();
		}
		if (startNode == null) {
			changeCount = 0;
			return;
		}
	/*
	* propagate order to the right
	*/
		layoutToRight(startNode, startNode);
	/*
	* propagate order to the left
	*/
		layoutToLeft(startNode, startNode);
	/*
	* check all nodes (except for start node)
	* and ensure proper sequencing
	*/
		boolean changesMade;
		do {
			changesMade = false;
			e = items.elements();
			while (e.hasMoreElements()) {
				NodeData node = (NodeData) e.nextElement();
				if (node.clusterOrderX == 0) {
					changesMade |= layoutToRight(node, startNode);
					changesMade |= layoutToLeft(node, startNode);
				}
			}
		} while (changesMade);
	/*
	* "Normalize" such that x and y >= 0
	*/
		e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			node.clusterOrderX -= minX;
			node.clusterOrderY -= minY;
		}

		maxX -= minX;
		maxY -= minY;
		minX = 0;
		minY = 0;
	/*
		if (reporter.isDebugKey(DEBUG_NAME)) {
			reporter.debugMsg(clusterId,null,null,null,DEBUG_NAME);
			reporter.debugMsg("  minX=",Integer.toString(minX),"  maxX=",Integer.toString(maxX),DEBUG_NAME);
			reporter.debugMsg("  minY=",Integer.toString(minY),"  maxY=",Integer.toString(maxY),DEBUG_NAME);
		}
	*/
	/*
	* Allocate space for vertical ordering
	*/
		verticalSlots = new Vector(maxX+3); // room for growth
		for (int i = 0; i<= maxX; i++) {
			verticalSlots.addElement( new Vector(10));
		}
	/*
	* Allocate vertical order
	*/
		assignVerticalOrder();
	
		assignEachNodeAPosition();
		changeCount = 0;
		debugListMembers();
	}
/*
* Assign vertical order to the nodes.  Each node
* is initially given a unique position so that 
* no two nodes layout on top of each other.  The
* initial ordering has no aesthetics.  Many
* silly layouts result at this point.
*
* @return true if a change was made in a nodes order
*/
	private void assignVerticalOrder() {
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			Vector v = (Vector)verticalSlots.elementAt(node.clusterOrderX);
			int index = v.size();
			v.addElement(node);
			node.clusterOrderY = index;
			if (maxY < index) 
				maxY = index;
		}
	}
/*
* Assign layout to output nodes.
* Every connected node to the right must have an ordinal
* value greater than the argument node.
*
* @return true if a change was made in a nodes order
*/
	private boolean layoutToRight(NodeData node, Object ignoreNode) {
		boolean changeMade = false;
		if (node.agent instanceof DIM)
			return false;
		Enumeration e = node.agent.getSinkList().elements();
		while (e.hasMoreElements()) {
			try {
				NodeData relative = nodeForAgent((ConnectedAgent) e.nextElement());
				if (relative == ignoreNode)
					continue;
				if (relative.clusterOrderX <= node.clusterOrderX) {
					relative.clusterOrderX = node.clusterOrderX +1;
					layoutToRight(relative, ignoreNode);
					changeMade = true;
					if (maxX < relative.clusterOrderX) 
						maxX = relative.clusterOrderX;
				}
			} catch (NoSuchAgentException ignore) {}
		}
		return changeMade;
	}
/*
* Assign layout to input nodes.
* Every connected node to the left must have an ordinal
* value less than the argument node.
*
* @return true if a change was made in a nodes order
*/
	private boolean layoutToLeft(NodeData node, Object ignoreNode) {
		boolean changeMade = false;
		if (node.agent instanceof PIM)
			return false;
		Enumeration e = node.agent.getSourceList().elements();
		while (e.hasMoreElements()) {
			try {
				NodeData relative = nodeForAgent((ConnectedAgent) e.nextElement());
				if (relative == null || relative == ignoreNode)
					continue;
				if (relative.clusterOrderX >= node.clusterOrderX) {
					relative.clusterOrderX = node.clusterOrderX -1;
					layoutToLeft(relative, ignoreNode);
					changeMade = true;
					if (minX > relative.clusterOrderX) 
						minX = relative.clusterOrderX;
				}
			} catch (NoSuchAgentException ignore) {}
		}
		return changeMade;
	}

/**
* Marks each node as belonging to this cluster uniquely
*/
/*
	private void markMembers() {
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			Cluster itsParentCluster = node.getParentCluster();
			if (itsParentCluster != null && itsParentCluster != this) {
				itsParentCluster.removeNode(node);
			}
			node.setParentCluster(this);
		}
	}
*/
/*
* Grow a cluster that contains a specified node
*/
/*
	void growAround(NodeData node) {
		if (verticalBlockSpacing == 0) {
			verticalBlockSpacing = node.block.getBoundingRect().height + 5;
		}
		growAroundR(node);
		markMembers();
		changeCount++;
	}
*/
	
	public String toString() {
		return clusterId;
	}

	void debugListMembers() {
/*
		if (!reporter.isDebugKey(DEBUG_NAME))
			return;
		reporter.debugMsg("List of ",this, null,null, DEBUG_NAME);
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			reporter.debugMsg("  -",Integer.toString(node.clusterOrderX),Integer.toString(node.clusterOrderY),node,DEBUG_NAME);
		}
*/
	}
	
	/**
	* Identifies if the cluster has a current layout
	*/
	boolean hasGoodLayout() {
		return (changeCount == 0);
	}
/**
* Cluster should not be displayed if it only has one agent
*/
/*
	public boolean shouldBeDisplayed() {
		return items.size() > 1;
	}
*/
/**
* Tell cluster to slide its children into position
*/
	public void slide() {
		Enumeration nodeList = items.elements();
		while (nodeList.hasMoreElements()) {
			NodeData node = (NodeData) nodeList.nextElement();
			node.slide();
		}
	}
/**
* Scan all memebers updating their connection info
*/
	public void scanMembersConnections() {
		Enumeration e = items.elements();
		while (e.hasMoreElements()) {
			NodeData node = (NodeData) e.nextElement();
			node.updateConnectionInfo();
//			if (reporter.isDebugKey(DEBUG_NAME)) {
//				reporter.debugMsg(" cons:",node, Integer.toString(node.nInputs),Integer.toString(node.nOutputs),DEBUG_NAME);
//			}
		}
	}

}