/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.productVisitor;

import java.util.Vector;
import java.util.Enumeration;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.PrintWriter;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.ProductVisitor;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.PrintCRLFStream;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.product.view.EnumeratedProduct;

/*
 * A ProductVisitor which is able to create a BackWeb Infopak.
 * Assumes the existence of an infopak template (.bis, .bif,... files)
 * which AsBWInfopak has access to.
 *
 * @author Srinivas Inguva
 *
 * modified by Michael DeCorpo
 * modified by Jason Mathews - don't generate summary page when product is a root element (i.e. index page)
 */

/*
  !!__To Do__!!
  1) In addSummary(): Use AsSummary visitor instead of AsPlainText visitor
  2) in addProductFiles(): check for the existence of an "index[*].html"
     file. If it doesn't exist, then make a copy of summary.html called
     "index.html" 
     -->> NO better to use new Enumerated product interfaces getIndexItems or getRoot!!!!!
 */

public class AsBWInfopak implements ProductVisitor {

/**
* name of the subdirectory that this visitor uses for its files
*/
  public static final String identifier = "BackWeb";

/**
* Vector which will be used to temporarily hold all of the infopak files:
*/
	Vector infopakFiles;

/**
* working directory
*/
	protected File workingDir;
	protected Reporter reporter;

	public AsBWInfopak(Reporter reporter, File wdir) {
		if ( !wdir.isDirectory()) {
			throw new IllegalArgumentException("Constructor needs directory as arg");
		}
		workingDir = wdir;
		this.reporter = reporter;
	}

	public void visit(Product product)  {

		infopakFiles = new Vector(50);
		
		File infopakDir = new File(ExecGlobal.getSupportDir(),identifier);
		try {
			addProductFiles(product,infopakFiles);
				
			addInfopakTemplate(infopakDir, infopakFiles);

			Product root;
			if (product instanceof EnumeratedProduct)
			  root = ((EnumeratedProduct) product).getRoot();
			else if (product.getProperty(Product.PROPERTY_ROOT) != null)
			  root = product;
			else root = null;

			String summaryFileName = null;

			if (root != null) {
			  Object content = root.getContent();
			  if (content instanceof FileRef) {
			    File file = ((FileRef)content).getFile();
			    summaryFileName = makeSummaryFile(file, infopakFiles);
			  }
			}

			if (summaryFileName == null)
			  summaryFileName = addSummaryFile(product,infopakFiles);

			// reporter.log("product summaryFileName=" + summaryFileName);
			
			addInfoFlashScriptFile(product, infopakFiles, summaryFileName);
		}
		catch (IOException e) {
			reporter.reportException(e);
		}
	}

  public Enumeration getInfopakEnumeration() {
    return infopakFiles.elements();
  }

  public Vector getInfopakFileRefVector() {
    return infopakFiles;
  }


/* 
* Get prebuilt infopak pieces.  For now, we are
* getting them from fixed directory.  At the minimum,
* this directory must contain: 
*      - CDAFlash.bif file (BW Frame file)
*/
	private void addInfopakTemplate(File infopakDir, Vector v) throws IOException {
		FileRef fr;
		String[] files = infopakDir.list();
		int copied = 0;
		
		try {
			for(int i=0;i<files.length;i++) {
				File afile = new File(infopakDir,files[i]);
				fr = new FileRef(afile,System.currentTimeMillis(),afile.getName());
				fr.setKeepFile(true);	// do not let garbage collection get these
				
				v.addElement(makeLocalCopy(fr));
				copied++;
			}
		}
		finally {
			if (copied <= 2) {
				reporter.warning("There were only "+files.length+" backweb template files, was expecting more");
			}
		}
	}

/*
* Get files from the product we have been passed in.
* Currently using the AsFlatFileList visitor to get *all*
* files from the product.  Can just as easily use some 
* other (filter) visitor.
*/
	private void addProductFiles(Product product, Vector v) throws IOException {
		AsFlatFileList myList = new AsFlatFileList();
		myList.visit(product);
		Enumeration fileProdList = myList.getFileProductList(); // enum of filerefs
		
		while (fileProdList.hasMoreElements()) {
			FileRef fr = (FileRef) ((Product)fileProdList.nextElement()).getContent();
			v.addElement(makeLocalCopy(fr));
		}
	}

/*
* Create a summary file from product and add it to v.
* 
* (Should look through product for a file named "root"+<whatever>+".html", 
*  if we find one then we can just make a copy of it called 
*  "summary.html")
*      -->> NO better to use new Enumerated product interfaces getIndexItems or getRoot!!!!!
*/
	private String addSummaryFile(Product product, Vector v) {
	
		AsHTMLIndex summarizer = new AsHTMLIndex(false, null, null, reporter);
		
		summarizer.setFlatFile(true);
		summarizer.setBackground("ringbinder.jpg");
		summarizer.visit(product);
		
		String summaryName = FileRef.getNewFileId("summary","html");
		File f = new File(workingDir, summaryName);
		FileRef fr = new FileRef(f,System.currentTimeMillis(),summaryName);
		fr.setKeepFile(true); // this could potentially be gc'd
		
		PrintStream ps = null;
		try {
			FileOutputStream fout = new FileOutputStream(f);
			ps = new PrintStream(fout,true);
			summarizer.writeResults(ps);
		} catch (IOException e) {
			reporter.reportException(e);
		} finally {
			if (ps != null) ps.close();
		}
		v.addElement(fr);
		
		return summaryName;
	}

    /** Make a summary file from a given "root" index page
     * duplicate index file and name it summary-???.html if a summary.html page is not
     * already present.
     */
    private String makeSummaryFile(File src, Vector v) {
	String name = src.getName();
	// check if root product is already a summary page
	if (name.startsWith("summary") && name.endsWith(".html"))
	    return name;

	String summaryName = FileRef.getNewFileId("summary","html");
	File f = new File(workingDir, summaryName);
	try {
	    FileRef.copyFile(src, f);
	    FileRef fr = new FileRef(f, System.currentTimeMillis(), summaryName);
	    fr.setKeepFile(true); // this could potentially be gc'd
	    v.addElement(fr);
	    return summaryName;
	} catch (IOException ioe) {
	    reporter.reportException(ioe);
	}
	return null;
	}


/*
* Create a the InfoFlash Script and add it to v.
*/
	private void addInfoFlashScriptFile(Product product, Vector v, String summaryFileName) throws IOException{
		AsBWInfoFlash summarizer = new AsBWInfoFlash(summaryFileName, reporter);
		summarizer.visit(product);
		
		String flashName = FileRef.getNewFileId("CDAFlash","bis");
		File f = new File(workingDir, flashName); 
		FileRef fr = new FileRef(f, System.currentTimeMillis(), flashName);
		fr.setKeepFile(true); // this could potentially be gc'd
		
		FileOutputStream fout = new FileOutputStream(f);
		PrintWriter ps = new PrintCRLFStream(fout,true);
		String prefix = new String("    ");
		ps.println("Start {\n");
		ps.println(prefix + "set frameFile=\"CDAFlash.bif\"");
		ps.println(prefix + "set reference=midTop");
		ps.println("\n");
		ps.println(prefix + "onClick {");
		ps.println(prefix + prefix + "rect (386,5,399,18) does { end all }");
		ps.println(prefix + "}\n\n");
		ps.println(prefix + "animate from (500,1000) toPixelOffset (0, -106) in 1 seconds\n");
		
		summarizer.writeResults(ps);
		
		ps.println(prefix + "pause for 1 seconds");
		ps.println(prefix + "animate to (500,1000) in 1 seconds");
		ps.println("}");
		ps.flush();
		fout.close();

		v.addElement(fr);
	
	}

	private FileRef makeLocalCopy(FileRef fr) throws IOException {
		File fileRemote = fr.getFile();
		
		File localCopy = new File(workingDir,fileRemote.getName());
		
		FileRef.copyFile(fileRemote, localCopy);
		
		FileRef theCopy = new FileRef(localCopy, 1, localCopy.getName());
		theCopy.setKeepFile(true); // this could potentially be gc'd
		
		return theCopy;
	}

}

