/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.productVisitor;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.PrintWriter;
import java.net.URL;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductWriter;
import org.mitre.pam.store.FileRef;
import org.mitre.pdif.eventParser.ValueList;

/**
* A basic product can be anything almost.<p>
*
* @author Nazario Irizarry, Jr.
* @author Jason Mathews made this implement ProductWriter interface and added various suppression options
*                       fixed processByteArray to properly print out bytes.
*/
public class AsPlainText implements ProductWriter
{
	protected int depthOfProduct;
	protected Product inputProduct = null;
	protected char separatorChar = ' ';

	// suppresses id from being displayed (duplicated for email also in subject)
	protected boolean rootIdSuppress;

    // suppresses root description from being displayed if part of container
    protected boolean rootDescSuppress;
    
    // suppress verbose labeling for comptact display
    protected boolean compact;

    // show metadata
    protected boolean showMetaData;
    
    public static final String NO_ROOT_ID = "norootid";
    public static final String NO_ROOT_DESC = "norootdesc";
    public static final String COMPACT = "compact";
    public static final String SHOW_METADATA = "showmetadata";

    public AsPlainText() {
    }

    public AsPlainText(Vector options) {
	for (int i=0; i < options.size(); i++) {
	    String opt = ((String) options.elementAt(i)).toLowerCase();
	    if (opt.equals(NO_ROOT_ID))
			rootIdSuppress = true;
	    else if (opt.equals(NO_ROOT_DESC))
			rootDescSuppress = true;
	    else if (opt.equals(COMPACT))
			compact = true;
	    else if (opt.equals(SHOW_METADATA))
			showMetaData = true;
	    else
			throw new IllegalArgumentException("AsPlainText: unknown option " + opt);
		// should log assert/warning not throw an exception...
	}
    }

	public void writeResults(PrintWriter out) {

		process(out, inputProduct, 0);
		out.flush();
	}

	public void setRootIdSuppress(boolean suppress) {
		rootIdSuppress = suppress;
	}
	
	public void setSeparatorChar(char separator) {
		separatorChar = separator;
	}

	public void visit(Product inputProduct) {
		this.inputProduct = inputProduct;
		AsTree asTree = new AsTree();
		asTree.visit(inputProduct);
		depthOfProduct = asTree.getDepth();
	}

	protected void process(PrintWriter out, Product product) {
		process(out, product, 0);			
	}

	protected void process(PrintWriter out, Product product, int level) {
//		System.out.println(" level= "+level+" product= "+product);
		if (product == null)
			return;
	/*
	* "name"
	*/
		if (product.getId().length() > 0) {
			if (level == 0 && depthOfProduct > 1) {
			    if (!rootIdSuppress ) {
					int nChars = product.getId().length() + 4;
					if (!compact) {
					  for (int i = 0; i<nChars; i++)
						out.print("*");
					  out.println();
					}
					out.print("* ");
					out.print(product.getId());
					out.println(" *");
					if (!compact) {
					  for (int i = 0; i<nChars; i++)
						out.print("*");
					  out.println();
					}
			    }
			}
			else if (level == 1 && depthOfProduct > 1) {
				int nChars = product.getId().length() + 4;
				out.println();
				if (!compact) {
				  for (int i = 0; i<nChars; i++)
					out.print("-");
				  out.println();
				}
				out.print("+ ");
				out.print(product.getId());
				out.println(" +");
				if (!compact) {
				  for (int i = 0; i<nChars; i++)
					out.print("-");
				  out.println();
				}
			}
			else if (level > 0) {
				out.println();
	//			generatePrefix(out, level);
				out.println(product.getId());
			}
			else if (!rootIdSuppress) {
	//			generatePrefix(out, level);
				out.println(product.getId());
			}
		}

	/*
	* " - description"
	*/
		if (product.getDescription().length() > 0) {
			if (!rootDescSuppress || !(level == 0 && depthOfProduct > 1)) {
				StringTokenizer st = new StringTokenizer(product.getDescription(), "\n");
				while (st.hasMoreElements()) {
					String sub = (String) st.nextElement();
		//			generatePrefix(out, level);
					out.println(sub);
				}
			}
		}

	/*
	* Metadata
	*/
		if (showMetaData) {
			outputMetaData(product, out);
		}
	/*
	* Content
	*/
		Object content = product.getContent();
		/*
		* vector contents
		*/
		if (content instanceof Enumeration) {
     		    Product root = ((ContainerProduct)product).getRoot();
     		    if (root == null) {
				Enumeration enum = (Enumeration) content;
				while (enum.hasMoreElements()) {
					process(out, (Product)enum.nextElement(), level+1);
				}
		    }
		    else {
				process(out, root, level);
		    }
		}
		/*
		* String contents
		*/
		else if (content instanceof String) {
			String s = (String) content;
//			System.out.println(" AsPlainText, string content "+s);
			StringTokenizer std = new StringTokenizer(s, "\n");
			while (std.hasMoreElements()) {
				String sub = (String) std.nextElement();
//				generatePrefix(out, level);
				out.println(sub);
			}
		}
		/*
		* Byte Array contents
		*/
		else if (content instanceof byte[]) {
			processByteArray(out, (byte[])content);
			out.println();
		}
		/*
		* FileRef contents
		*/
		else if (content instanceof FileRef) {
//			generatePrefix(out, level);
			out.print("File name: ");
			String name = ((FileRef) content).getFile().getName();
			out.println(name);
		}
		/*
		* URL contents
		*/
		else if (content instanceof URL) {
			String s = ((URL)content).toString();
//			generatePrefix(out, level);
			out.print("URL - ");
			out.println(s);
		}
		/*
		* unknown contents
		*/
		else {
			throw new IllegalArgumentException(" AsPlainText, unrecognized content "+content);
		}
	}
/**
* do some output for byte arrays
*/
	static void processByteArray(PrintWriter out, byte[] bytes) {
		int limit = bytes.length;
		if (limit > 30) 
			limit = 30;
		for (int i = 0; i< limit; i++) {
			int b = bytes[i];
			if (i != 0) out.print(' ');
			String hexString = Integer.toHexString(b & 0xff);
			if (hexString.length() < 2)
				out.print('0');
			out.print(hexString);
		}
		if (limit != bytes.length) {
			out.println(" ... ["+limit+" of "+bytes.length+"]");
		}
		out.flush();
	}
	
/**
* prepend a certain number of tabs
*/
	private void generatePrefix(PrintWriter out, int level){
		for (int i = 0; i< level; i++)
			out.print(separatorChar);
	}
/**
* show a products metadata, for debugging
*/
	private void outputMetaData(Product p, PrintWriter out){
		out.println(" Metadata");
		out.println("  Size = "+p.getSizeInBytesAsDouble()+" bytes");
		Enumeration props = p.getProperties();
		while(props.hasMoreElements()) {
			String name = (String) props.nextElement();
			ValueList vl = p.getPropertyValues(name);
			out.print("  ");
			out.print(name);
			out.print(" = ");
			out.println(vl.toString());
		}
		
		Object content = p.getContent();
		if (content instanceof FileRef) {
			FileRef fr = (FileRef) content;
			out.println(" File metadata");
			out.print("  keep: ");
			out.println(fr.getKeepFile()?"true":"false");
			out.print("  mime type: ");
			out.println(fr.getMIMEType());
			out.print("  mod date: ");
			out.println(fr.getModificationDate().toString());
			out.print("  source name: ");
			out.println(fr.getSourceName());
		}
	}	
}
