/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.productVisitor;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductWriter;
import org.mitre.pam.store.FileRef;
import java.io.PrintWriter;
import java.net.URL;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

/**
* Converts an arbitrary Product structure into an XML file.
*
* Below is an example of the XML that this filter generates.  This XML
* was created by piping two getPIM1 agents through the filter.
*
* <?xml version="1.0" ?>
* <Container>
*   <Property name="knownContentProperties">
*    <Value>description</Value>
*    <Value>id</Value>
*    <Value>lifeTimeLong</Value>
*  </Property>
*  <Property name="description">
*    <Value />
*  </Property>
*  <Property name="id">
*    <Value>Multiple products</Value>
*  </Property>
*  <Property name="lifeTimeLong">
*    <Value>259200</Value>
*  </Property>
*  <Content>
*    <Property name="description">
*      <Value>Creates a string product</Value>
*    </Property>
*    <Property name="id">
*      <Value>testPIM1</Value>
*    </Property>
*    <Property name="lifeTimeLong">
*      <Value>259200</Value>
*    </Property>
*    <StringValue>content for test1</StringValue>
*  </Content>
*  <Content>
*    <Property name="description">
*      <Value>Creates a string product</Value>
*    </Property>
*    <Property name="id">
*      <Value>testPIM2</Value>
*    </Property>
*    <Property name="lifeTimeLong">
*      <Value>259200</Value>
*    </Property>
*    <StringValue>content for test2</StringValue>
*  </Content>
* </Container>
*
* BasicProducts are represented in the XML as <Content> and ContainerProducts
* are represented as <Container>.  It is perfectly legal to nest containers
* inside containers.
*
* Each <Content> tag will include entries for all the products attributes
* (except for those defined in the exclusion list) along with an entry
* for the products content.
*
* Each <Container> tag will include a field called "knownContentProperties".
* This list defines all the properties that are used by the <Content> tags
* defined later.
*
* The DTD for the XML document is the following:
*
* <!ELEMENT Container (Property+,(Content|Container)*)>
* <!ELEMENT Content (Property+,(StringValue|A|Filename|ByteValue)?)>
* <!ELEMENT Property (Value*)>
* <!ATTLIST Property name CDATA #IMPLIED>
* <!ELEMENT A (#PCDATA)>
* <!ATTLIST A HREF CDATA #IMPLIED>
* <!ELEMENT ByteValue (#PCDATA)>
* <!ELEMENT Filename (#PCDATA)>
* <!ELEMENT StringValue (#PCDATA)>
* <!ELEMENT Value (#PCDATA)>
*
* @author John R. Fantry
* @author Jason Mathews - added Byte Array handling
*/
public class AsXML implements ProductWriter {
	protected static final String   XML_VERSION = "1.0";
	protected static final String   CONTAINER_PRODUCT_OPEN_TAG = "<Container>";
	protected static final String   CONTAINER_PRODUCT_CLOSE_TAG = "</Container>";
	protected static final String   PRODUCT_OPEN_TAG = "<Content>";
	protected static final String   PRODUCT_CLOSE_TAG = "</Content>";
	protected static final String   COLUMN_NAMES = "knownContentProperties";
	protected Product		product = null;
	protected Vector		exclusionList = null;

	public AsXML() {}

	/**
	 * Write a Product as an XML entry.  All output will be prefixed with the prefix
	 * arg which contains some number of tab chars to help preserve the hierarchical structure.
	 */
	protected void process(PrintWriter out, Product product, String prefix, Vector columnList) {
		if (product == null) {
			out.println(prefix + PRODUCT_OPEN_TAG);
			out.println(prefix + PRODUCT_CLOSE_TAG);
			out.flush();
			return;
		}
		Object  content = product.getContent();

		/*
		 * vector contents
		 */
		if (content instanceof Enumeration) {
			Vector  localList = new Vector();

			out.println(prefix + CONTAINER_PRODUCT_OPEN_TAG);
			Enumeration     enum = (Enumeration) content;

			while (enum.hasMoreElements()) {
				// build knownContentProperties list
				Product p = (Product) enum.nextElement();
				Object  obj = p.getContent();

				if (!(obj instanceof Enumeration)) {
					Enumeration     e = p.getProperties();

					while (e.hasMoreElements()) {
						String  s = (String) e.nextElement();

						if (!localList.contains(s)) {
							localList.addElement(s);
						}
					}
				}
			}
			out.println(prefix + "\t" + "<Property name=\"" + COLUMN_NAMES + "\">");
			for (int n = localList.size(), i = 0; i < n; i++) {
				String  columnName = (String) localList.elementAt(i);

				if (exclusionList == null ||!exclusionList.contains(columnName)) {
					out.println(prefix + "\t\t" + "<Value>" + columnName + "</Value>");
				}
			}
			out.println(prefix + "\t" + "</Property>");
			Enumeration     props = product.getProperties();

			while (props.hasMoreElements()) {
				String  prop = (String) props.nextElement();

				out.println(prefix + "\t" + "<Property name=\"" + prop + "\">");
				// treats multi-valued properties as single value
				out.println(prefix + "\t\t" + "<Value>" + product.getProperty(prop) + "</Value>");
				out.println(prefix + "\t" + "</Property>");
			}
			Enumeration     prods = (Enumeration) product.getContent();

			while (prods.hasMoreElements()) {
				process(out, (Product) prods.nextElement(), prefix + "\t", localList);
			}
			out.println(prefix + CONTAINER_PRODUCT_CLOSE_TAG);
		}
		else {
			Enumeration     props = product.getProperties();

			out.println(prefix + PRODUCT_OPEN_TAG);
			props = product.getProperties();
			while (props.hasMoreElements()) {
				String  prop = (String) props.nextElement();

				if (exclusionList == null ||!exclusionList.contains(prop)) {
					out.println(prefix + "\t" + "<Property name=\"" + prop + "\">");
					// treats multi-valued properties as single value
					out.println(prefix + "\t\t" + "<Value>" + product.getProperty(prop) 
						    + "</Value>");
					out.println(prefix + "\t" + "</Property>");
				}
			}
			if (content != null) {

				/*
				 * String contents
				 */
				if (content instanceof String) {
					StringBuffer    parsedContent = new StringBuffer();
					StringTokenizer st = new StringTokenizer((String) content, "<", true);

					while (st.hasMoreTokens()) {
						String  s = st.nextToken();

						if (s.equals("<")) {
							parsedContent.append("&lt;");
						}
						else {
							parsedContent.append(s);
						}
					}
					out.println(prefix + "\t" + "<StringValue>" + parsedContent + "</StringValue>");
				}

				/*
				 * FileRef contents
				 */
				else if (content instanceof FileRef) {
					out.println(prefix + "\t" + "<Filename>" + ((FileRef) content).getSourceName() 
						    + "</Filename>");
				}

				/*
				 * URL contents
				 */
				else if (content instanceof URL) {
					String  url = content.toString();

					out.println(prefix + "\t" + "<A HREF=\"" + url + "\">" + url + "</A>");
				}

				/*
				 * Byte Array contents
				 */
				else if (content instanceof byte[]) {
					out.print(prefix + "\t<ByteValue>");
					processByteArray(out, (byte[]) content);
					out.println("</ByteValue>");
				}

				/*
				 * unknown contents
				 */
				else {
					throw new IllegalArgumentException(" AsXML, unrecognized content " + content);
				}
			}
			out.println(prefix + PRODUCT_CLOSE_TAG);
		}
	}

	public void visit(Product product) {
		visit(product, null);
	}

	public void visit(Product product, Vector exclusionList) {
		this.product = product;
		this.exclusionList = exclusionList;
	}

	protected void writeHeader(PrintWriter out) {
		out.println("<?xml version=\"" + XML_VERSION + "\"?>");
	}

	public void writeResults(PrintWriter out) {
		writeHeader(out);
		process(out, product, "", new Vector());
		out.flush();
	}

	/**
	 * do output for byte arrays
	 */
	static void processByteArray(PrintWriter out, byte[] bytes) {
		for (int size = bytes.length, i = 0; i < size; i++) {
			int     b = bytes[ i ];

			if (i != 0) {
				out.print(' ');
			}
			String  hexString = Integer.toHexString(b & 0xff);

			if (hexString.length() < 2) {
				out.print('0');
			}
			out.print(hexString);
		}
		out.flush();
	}

}
