/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
		
package org.mitre.pam.putter;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.FileReader;
import java.net.InetAddress;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.productVisitor.AsBWInfopak;
import org.mitre.pam.util.DownloadNumber;
import org.mitre.pam.util.PrintCRLFWriter;
import org.mitre.pam.util.PAMDate;

/**
 * A DIM that takes its product, creates a BackWeb infopak,
 * and publishes this infopak onto a BackWeb server running
 * on the same machine as the DIM. If the received iPAM product has
 * the bw_urgent property set to "true" then the backweb infopak
 * will be marked as urgent (display_asap=1). Likewise, if the product
 * has the bw_mustSee property set to "true" then the infopak will be
 * marked not to be deleted while it is new (must_see=1).
 *
 * @author Srinivas Inguva
 * Modified by Jason Mathews 5/13/99 [iPAM 2.2.63]
 *    - use native BackWeb directory (not "C:\BackWeb\..." as the default)
 *    - use BufferedReader instead of RandomAccessdFile stream
 * Modified by Jason Mathews 2/15/00 [iPAM 2.5.99]
 *    - Added ability to flag BackWeb infopaks as urgent and/or must-see
 *      if the received product has the appropriate properties.
 **/
public class BackWebLocalDIM extends DIMShell1 
{
	private Hashtable segTable;

	private Enumeration filelist = null;

	private DownloadNumber dn;

	private String scriptFile = null;

	private boolean initialized = false;

	/**
	 * The property name for the directory that the backweb server is in
	 */
	public static final String BW_SERVER_DIR_PROP = "backWebServerDir";

	private static final String SEG_FIELDS_FILE_PROP = "backWebSegfields";

	private static final String SEG_RULES_FILE_PROP = "backWebSegrules";

	private static final String SEG_FIELDS_FILE_VALUE = "segfields";

	private static final String SEG_RULES_FILE_VALUE = "segrules";

	private static final String PAM_DEFAULT = "pam_default";

	//The block of segmentation rules we have been allocated in the BW Server segrules file
	/**
	 * The property name for the starting rule in the block of segmentation rules
	 * that have been allocated for our use
	 */
	public static final String BLOCK_START_PROP = "backWebSegrulesStart";

	/**
	 * The property name for the ending rule in the block of segmentation rules
	 * that have been allocated for our use
	 */
	public static final String BLOCK_END_PROP = "backWebSegrulesEnd";

	public static final String EXPOSURE_GROUP_PROP = "backWebExposureGroup";

	private static final String DB_INSERT_PROGRAM_PROP = "backWebInsertProgramName";

	private static final String DB_INSERT_PROGRAM_VALUE = "insert_prs";

	public static final String BW_HOST_DIM_PROP = "backWebServerHost";

	public static final String BW_TCP_SERVER_PORT_NUMBER = "backWebTCPPort";

	public static final String BW_UDP_SERVER_PORT_NUMBER = "backWebUDPPort";

	public static final String BW_CONSOLE_PORT_NUMBER_PROP = "backWebConsolePort";

	public static final String BW_CHANNEL_NAME_PROP = "backWebChannelName";

	private static final String BW_PASSWORD_PROP = "backWebPasswordFile";

	private static final String BW_PASSWORD_VALUE = "passwd";

	private static final String BW_TOPIC_NAME = "backWebTopicName";

	private static final String DEBUG_KEY = "BackWebLocalDIM";
	
	public static final String BW_URGENT = "bw_urgent";

	public static final String BW_MUSTSEE = "bw_mustSee";
	
	private Vector currentInfoPakFileRefs;

	public BackWebLocalDIM()
	{
		setDescription("BackWeb DIM");
		//    setProperty(BW_PASSWORD_PROP, "passwd");
		setProperty(BW_CHANNEL_NAME_PROP,"CDA");
		setProperty(BW_TCP_SERVER_PORT_NUMBER,"81");
		setProperty(BW_UDP_SERVER_PORT_NUMBER,"370");
		String hostAddress = null; 
		try
		{
			hostAddress = InetAddress.getLocalHost().getHostName();
		}
		catch(Exception e)
		{
			hostAddress ="xxx.yyy.zzz";
		}
		setProperty(BW_HOST_DIM_PROP,hostAddress);
		//    setProperty(DB_INSERT_PROGRAM_PROP, DB_INSERT_PROGRAM_VALUE);
		setProperty(EXPOSURE_GROUP_PROP,"PAM");
		setProperty(BLOCK_START_PROP,"1");
		setProperty(BLOCK_END_PROP,"10");
		setProperty(BW_CONSOLE_PORT_NUMBER_PROP,"6789");
		//    setProperty(SEG_RULES_FILE_PROP, SEG_RULES_FILE_VALUE);
		//    setProperty(SEG_FIELDS_FILE_PROP, SEG_FIELDS_FILE_VALUE);
		setProperty(BW_TOPIC_NAME,"");
		String bwRootDir;
		if (File.separatorChar == '/')
		  bwRootDir = "/usr/local/BackWeb/BackWebServer/";
		else
		  bwRootDir = "c:\\program files\\BackWeb\\BackWeb Server\\";
		setProperty(BW_SERVER_DIR_PROP,bwRootDir);
		dn = new DownloadNumber();

		metaStringInfo(BW_CHANNEL_NAME_PROP,"CDA","BackWeb channel name (as seen by the console tool)");
		metaStringInfo(BW_TCP_SERVER_PORT_NUMBER,"81","TCP/IP port as specified for your BackWeb server configuration");
		metaStringInfo(BW_UDP_SERVER_PORT_NUMBER,"370","UDP/IP port as specified for your BackWeb server configuration");
		metaStringInfo(BW_HOST_DIM_PROP,hostAddress,"Host name for the machine running the BackWeb server");
		//    metaStringInfo(DB_INSERT_PROGRAM_PROP, DB_INSERT_PROGRAM_VALUE);
		metaStringInfo(EXPOSURE_GROUP_PROP,"PAM","BackWeb exposure group name (as seen by the console tool)");
		metaIntegerInfo(BLOCK_START_PROP,1,"First rule to use in looking for iPAM topics in the BackWeb segrules");
		metaIntegerInfo(BLOCK_END_PROP,10,"First rule to use in looking for iPAM topics in the BackWeb segrules");
		metaStringInfo(BW_CONSOLE_PORT_NUMBER_PROP,"6789","The UDP/IP port used by the BackWeb console program for your BackWeb server");
		//    metaStringInfo(SEG_RULES_FILE_PROP, SEG_RULES_FILE_VALUE);
		//    metaStringInfo(SEG_FIELDS_FILE_PROP, SEG_FIELDS_FILE_VALUE);
		metaStringInfo(BW_TOPIC_NAME,"","Optional topic to use when publishing to BackWeb, used in subscriptions.  A topic foo will match the segrule pam_foo.");
		metaStringInfo(BW_SERVER_DIR_PROP,bwRootDir,"Directory that is the root of the BackWeb Server installation");
	}

	public void putProduct(Product prod) {
		if(!initialized)
		{
			buildSegmentationTable();
			initialized = true;
		}
		String bwDir = getStringProperty(BW_SERVER_DIR_PROP,"");
		if(!bwDir.endsWith(File.separator))
			bwDir += File.separator;
		String cdaUploadDir = bwDir + "data" + File.separator + "import" + File.separator;
		String dirName = FileRef.getNewFileId("bw");
		File myDir = null;
		try
		{
			myDir = new File(cdaUploadDir + dirName);
			myDir.mkdirs();
		}
		catch(Exception e)
		{
			reportException(e);
			return;
		}
		/*
		 * In the following line we make instance of AsBWInfopak so that
		 * the life of the instance is the duration of the doSend() call.
		 * This ensures that the finalizer does not garbage collect
		 * references to the temp files created by AsBWInfopak until we
		 * are done.
		 */
		AsBWInfopak pacMaker = new AsBWInfopak(reporter, myDir);
		makeInfopak(prod,pacMaker);
		String topic = null;
		String preTopic = getStringProperty(BW_TOPIC_NAME,"");
		topic = buildTopic(preTopic);
		if(topic == null || topic.length() == 0)
		{
			preTopic = prod.getTopic();
			topic = buildTopic(preTopic);
		}
		if(topic != null && topic.length() > 0)
		{
			log("Using topic: " + preTopic + ", seg rule var = " + topic);
		}
		String publishDate = (new PAMDate()).asMonthDay();
		String title = sanitize(prod.getId()) + " " + publishDate;
		publishToServer(topic,title,myDir,dirName,prod);
		/*
		 * do this here so an optimizing compiler
		 * does not null pacMaker until we are done.  if
		 * we don't have this reference to pacMaker
		 * we might get in trouble.
		 */
		pacMaker = null;
		
		FileRef myDirFileRef = new FileRef(myDir,1,myDir.getName());
		myDirFileRef.setKeepFile(true); // we will clean up directly
		currentInfoPakFileRefs.addElement(myDirFileRef);
		cleanUpInfoPakFiles(currentInfoPakFileRefs);
		currentInfoPakFileRefs = null;
	}

	public String buildTopic(String s)
	{
		if(s == null || s.length() == 0)
			return null;
		if(s.startsWith("pam_"))
		{
			return s;
		}
		StringBuffer id = new StringBuffer(80);
		id.append("pam_");
		id.append(sanitizeTopicName(s));
		return id.toString();
	}

	/*
	 * change text into a backweb acceptable title
	 */
	private String sanitize(String id)
	{
		StringBuffer result = new StringBuffer(80);
		for(int i = 0;i < id.length();i++)
		{
			char c = id.charAt(i);
			if(c >= 'a' && c <= 'z')
			{
			//lowercase first for speed
			}
			else
				if(c >= 'A' && c <= 'Z')
				{
				}
				else
					if(",- :/".indexOf(c) >= 0)
					{
					}
					else
						if(c >= '0' && c <= '9')
						{
						}
						else
						{
							c = '_';
						}
			result.append(c);
		}
		return result.toString();
	}

	/*
	 * sanitizeTopic is used by class BackWebSubscriber
	 */
	public static String sanitizeTopicName(String topic)
	{
		StringBuffer result = new StringBuffer(80);
		for(int i = 0;i < topic.length();i++)
		{
			char c = topic.charAt(i);
			if(c >= 'a' && c <= 'z')
			{
			//lowercase first for speed
			}
			else
				if(c >= 'A' && c <= 'Z')
				{
				}
				else
					if(c >= '0' && c <= '9')
					{
					}
					else
					{
						c = '_';
					}
			result.append(c);
		}
		return result.toString();
	}

	private void makeInfopak(Product p, AsBWInfopak pacMaker)
	{
		pacMaker.visit(p);
		filelist = pacMaker.getInfopakEnumeration();
		while(filelist.hasMoreElements())
		{
			FileRef fr = (FileRef)filelist.nextElement();
			if(fr.getFile().getName().endsWith(".bis"))
			{
				scriptFile = fr.getFile().getName();
			}
		}
		filelist = null;
		currentInfoPakFileRefs = pacMaker.getInfopakFileRefVector();
	}

	private void publishToServer(String topicName, String title, File myDir, String dirName, Product prod)
	{
		String bwDir = getStringProperty(BW_SERVER_DIR_PROP,"");
		if(!bwDir.endsWith(File.separator))
			bwDir += File.separator;

		//    String sDBInsertProgramName = getProperty(DB_INSERT_PROGRAM_PROP);
		String sDBInsertProgramName = DB_INSERT_PROGRAM_VALUE;
		sDBInsertProgramName = bwDir + "bin" + File.separator + sDBInsertProgramName;
		String sChannelName = getStringProperty(BW_CHANNEL_NAME_PROP,"");
		String sExposureGroup = getStringProperty(EXPOSURE_GROUP_PROP,"");
		String sHostName = getStringProperty(BW_HOST_DIM_PROP,"");
		String sBackWebConsolePort = getStringProperty(BW_CONSOLE_PORT_NUMBER_PROP,"");
		String sPasswordFile = bwDir + "conf" + File.separator + "passwd" + File.separator + BW_PASSWORD_VALUE;
		
	/*
	* Add required commands to the command line.
	*/
		NoSpaceVector commandVector = new NoSpaceVector(15);
		commandVector.addElement(sDBInsertProgramName);
		commandVector.addElement("-d");
		commandVector.addElement(dirName);
		commandVector.addElement("-s");
		commandVector.addElement(sChannelName);
		commandVector.addElement("-g");
		commandVector.addElement(sExposureGroup);
		commandVector.addElement("-f");
		commandVector.addElement(sPasswordFile);

	/*
	* Add optional commands to the command line.
	*/
		if(sHostName != null && !sHostName.equals(""))
		{
			commandVector.addElement("-n");
			commandVector.addElement(sHostName);
		}
		if(sBackWebConsolePort != null && !sBackWebConsolePort.equals(""))
		{
			commandVector.addElement("-p");
			commandVector.addElement(sBackWebConsolePort);
		}
		if(scriptFile == null)
		{
			reporter.assert(false,"BackWebLocalDIM: Failed to receive a script file!",null,null,null);
			return;
		}
		else
		{
			/*
			 * Need to build an info.iad file here, based on the seg. rule for
			 * topicName
			 * According to BW convention, segrules=-1 means that the
			 * infopak is to have no audience, but -2 means everyone
			 */
			int ruleNum = -2;
			if(topicName != null && topicName.length() > 0)
			{
				String ruleString = (String)segTable.get(topicName);
				if (ruleString != null)
				try
				{
					ruleNum = Integer.parseInt(ruleString);
				}
				catch(Exception e) {}
			}
			if(ruleNum == -2)
			{
				log("Product sent to all subscribers");
			}
			//System.out.println("Building info file for rule number: "+ruleNum);
			String infopakName = null;
			if(title != null && title.length()!=0)
			{
				infopakName = title;
			}
			else
			{
				if(topicName != null && topicName.startsWith("pam_"))
				{
					infopakName = topicName.substring(4,topicName.length()) + (System.currentTimeMillis() % 1000000);
				}
				else
				{
					infopakName = PAM_DEFAULT + (System.currentTimeMillis() % 1000000);
				}
			}
			try
			{
				File f = new File(myDir,"info.iad");
				FileRef fr = new FileRef(f,System.currentTimeMillis(),f.getName());
				fr.setKeepFile(true); // will clean up ourselves, not gc'd
				currentInfoPakFileRefs.addElement(fr);
				buildInfoFile(f,infopakName,ruleNum,scriptFile,prod);
			}
			catch(IOException e)
			{
				reportException(e);
				return;
			}
		/*
		* exec to the system and pass the command line
		*/
			StringBuffer sCommandLine = new StringBuffer(300);
			for (int i = 0; i< commandVector.size(); i++) {
				if (i > 0)
					sCommandLine.append(" ");
				sCommandLine.append((String)commandVector.elementAt(i));
			}
			reporter.debugMsg("Exec'ing command string: " , sCommandLine, null,null, DEBUG_KEY);
			try
			{
				int len = commandVector.size();
				String args[] = new String[len];
				commandVector.copyInto(args);
				Process theProcess = Runtime.getRuntime().exec(args);
				theProcess.waitFor();
			}
			catch(Exception e)
			{
				warning("Failed to execute command [" + sCommandLine.toString() + "]");
				reportException(e);
				return;
			}
		}
	}

	//end publishToServer()
	/*
	 * lifetime is in seconds
	 */
	private void buildInfoFile(File f, String name, int ruleNum, String sf, Product prod) throws IOException
	{
		//Build a bare bones info.iad file:
		//build expiration date strings for p.getLifetime() milliseconds from now (one tells
		// the client not to auto-display the infopak after this date, and
		// other tells the server not to broadcast the infopak after this date)
		long lifetime = prod.getLifetime();
		PAMDate expDate = new PAMDate(System.currentTimeMillis() + lifetime * 1000);
		String expString = Integer.toString(1900 + expDate.getYear());
		String lastBroadcastString = Integer.toString(1900 + expDate.getYear());
		int month = expDate.getMonth();
		month++;
		String sMonth = Integer.toString(month);
		if(month < 10)
			sMonth = "0" + sMonth;
		int day = expDate.getDate();
		String sDay = Integer.toString(day);
		if(day < 10)
			sDay = "0" + sDay;
		expString += "/" + sMonth;
		lastBroadcastString += sMonth;
		expString += "/" + sDay;
		lastBroadcastString += sDay;
		FileOutputStream fout = new FileOutputStream(f);
		PrintWriter pin = new PrintCRLFWriter(fout,true);
		pin.println("");
		pin.println("[presentation]");
		pin.println("name=\"" + name + "\"");
		pin.println("expiration=\"" + expString + "\"");
		pin.println("max_exposures=0");
		pin.println("max_interactions=1");
		pin.print("display_asap=");
		pin.println(prod.getStringProperty(BW_URGENT,"").equalsIgnoreCase("true") ? '1' : '0');
		pin.print("must_see=");
		pin.println(prod.getStringProperty(BW_MUSTSEE,"").equalsIgnoreCase("true") ? '1' : '0');
		pin.println("tile_wallpaper=\"Default\"");
		pin.println("see_at=\"\"");
		pin.println("method=\"Animation\"");
		pin.println("required_platform=\"Windows 3.1\",\"Windows 95\",\"Macintosh\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\"");
		pin.println("required_resolution=\"640\",\"800\",\"1024\",\"1280\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\"");
		pin.println("required_colors=\"B/W\",\"16 Grayscale\",\"16 Colors\",\"256 Grayscale\",\"256 Colors\",\"16 Bit\",\"24 Bit\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\"");
		pin.println("required_sound=\"None\",\"Exist\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\",\"Undefined\"");
		pin.println("client_level=100");
		//Segmentation rules
		if(ruleNum >= -1)
			pin.println("seg_rules=" + Integer.toString(ruleNum));
		pin.println("first_broadcast_date=0");
		pin.println("last_broadcast_date=" + lastBroadcastString);
		pin.println("min_downloads=0");
		pin.println("max_downloads=0");
		pin.println("id=2");
		//pin.println("installation_time=\"1997/07/23 11:25:53\"");
		pin.println("");
		pin.println("[files]");
		//the scriptFile:
		pin.println("run=\"" + sf + "\"");
		pin.println("");
		pin.println("[publisher]");
		pin.println("name=\"\"");
		pin.println("email=\"\"");
		pin.println("phone=\"\"");
		pin.println("fax=\"\"");
		pin.println("web=\"\"");
		pin.println("comment=\"\"");
		pin.println("");
		pin.println("[plugins]");
		pin.println("; Do not delete or change this line __4");
		pin.close();
	}

	//end buildInfoFile()
	/**
	 * returns null except when we are copying files to our local directory
	 */
	/*
	   public Enumeration getCurrentFileRefs(){
	   return filelist;
	   }
	 */
	private void buildSegmentationTable()
	{
		segTable = new Hashtable();
		String bwDir = getStringProperty(BW_SERVER_DIR_PROP,"");
		if(!bwDir.endsWith(File.separator))
			bwDir += File.separator;
		String BWServerDataDir = bwDir + "data" + File.separator;
		//    String segRulesFile = getProperty(SEG_RULES_FILE_PROP);
		String segRulesFile = SEG_RULES_FILE_VALUE;
		int blockStart = -1;
		int blockEnd = -1;
		blockStart = getIntegerProperty(BLOCK_START_PROP,1);
		blockEnd = getIntegerProperty(BLOCK_END_PROP,1000);
		BufferedReader rdr = null;
		String key;
		String value;
		String temp;
		String nextLine;
		StringTokenizer tok;
		int ctr;
		int n;
		try
		{
			rdr = new BufferedReader(new FileReader(BWServerDataDir + "client_accessible" + File.separator + segRulesFile));
			while((nextLine = rdr.readLine()) != null)
			{
				if(nextLine.indexOf("#") != -1)
					continue;
				tok = new StringTokenizer(nextLine,"=");
				value = tok.nextToken();
				ctr = Integer.parseInt(value);
				if((ctr < blockStart) || (ctr > blockEnd))
					continue;
				// We are assuming that the segmentation rules are of the form:
				//      <num>=[Exists(<varname>) and] <varname>="subscribed"
				//  where <num> is the rule number and <varname> is the topicName
				key = tok.nextToken().trim();
				n = key.indexOf("and");
				if(n != -1)
				{
					key = key.substring(n + 3,key.length()).trim();
				}
				if(segTable.get(key) != null)
				{
					reporter.assert(false,"BackWebLocalDIM: Invalid Rule Error!", ((String)segTable.get(key)) , "\n",null);
					continue;
				}
				segTable.put(key,value);
			}
		}
		catch(IOException e)
		{
			warning("Couldn't read segmentation rules due to io error");
			reportException(e);
		}
		catch(NumberFormatException e)
		{
			warning("Couldn't read segmentation rules due to number format error (probably nonexisting)");
		}
		catch(Exception e)
		{
			reportException(e);
		}
		finally
		{
			if(rdr != null)
				try
				{
					rdr.close();
				}
				catch(IOException ignore)
				{
				}
		}
		reporter.debugMsg("buildSegmentationTable: ", segTable, null,null, DEBUG_KEY);
	}

//end buildSegmentationTable()


	/**
	* delete all infopak files when done, assume
	* that the directory is the last thing in the vector,
	* so it will be empty by the time we get to it
	*/
	private void cleanUpInfoPakFiles(Vector fileRefVector) {
		for (int i = 0; i< fileRefVector.size(); i++) {
			FileRef fr = (FileRef) fileRefVector.elementAt(i);
			fr.getFile().delete();
		}
	}

 final class NoSpaceVector extends Vector {
	public NoSpaceVector(int i) {
		super(i);
	}
	public void addElement(String o) {
		if (o.indexOf(' ') >= 0)
			super.addElement("\""+o+"\"");
		else
			super.addElement(o);
	}
}

}

//end class BackWebLocalDIM
