/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.util.*;
import java.io.*;
import java.net.*;

import org.mitre.pam.aggregator.ChanSchedBase;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductWriter;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.productVisitor.AsPlainText;
import org.mitre.pam.productVisitor.AsPlainTextNoRootId;
import org.mitre.pam.productVisitor.AsTickerText;
import org.mitre.pam.productVisitor.AsTickerTextNoRootId;
import org.mitre.pam.productVisitor.AsFlatFileList;
import org.mitre.pam.putter.DIMShell1;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.PrintCRLFWriter;
import org.mitre.pam.util.smtp.SendMail;
import org.mitre.pam.util.smtp.MIMEc;
import org.mitre.pdif.eventParser.ValueList;

/**
* This class creates a DIM for sending email with attachments.
*
* @author John Fantry, Technical Co-op (D510)
* @author Jason Mathews, added outputStyle option for multiple renderers,
*                        use defaultSmtpServer if not defined for this instance.
*
* 6-2-98
* Rewrote PIM to utilize the MIMEc class, to process the root product of a
* ContainerProduct if it exists, and redefined the configuration variables
* ("targetSpec" is now the name of an SMTP server (mbunix.mitre.org) and
* "mailingList" is now a list of users/addresses).
*
* The PIM now supports MIME types for attachments and the redefined configuration
* variables allow for greater ease of use along with increased flexibility.
* -JF
*/
public class PutMail extends DIMShell1 {

  public final static String MAILING_LIST = "mailingList";
  public final static String USERS = "users";
  public final static String OUTPUT_STYLE = "outputStyle";
  public final static String SMTP_HOST = "smtpHostName";
  
  protected SendMail mailer;

	protected static String[] RENDER_OPTIONS = {"plainText","plainTextNoRootId","tickerText","tickerTextNoRootId"};
  /**
  * Creates a new <tt>PutMail</tt> object.
  */
  public PutMail () {
    setDescription("Send mail via SMTP account");

    setProperty(SMTP_HOST,"");
    setProperty(MAILING_LIST, "user@hostname, user2@hostname, ...");
    setProperty(OUTPUT_STYLE, "plainText");
    setProperty(ChanSchedBase.CREATOR_EMAIL,"");
    
    metaStringInfo(SMTP_HOST,"","SMTP mail server host, such as mail.site.com");
    metaStringInfo(MAILING_LIST, "","list of addresses for recepients, such as naz@site.com, bob@site2.com");
    metaEnumerationInfo(OUTPUT_STYLE, "plainText", "style for output", RENDER_OPTIONS);
    metaStringInfo(ChanSchedBase.CREATOR_EMAIL,"","email address of sender");
  }
	
/**
* Callback to receive a product.
* @param p  Contains the mail that will get sent.
*/
  public void putProduct(Product p) {

    setProgress("starting");
    String smtpHost = getSmtpHostName();
    if (smtpHost.length()==0) {
		reporter.error("no smtp server specified");
		return;
	}
    Vector mailingList = SendMail.buildMailingList(getProperty(MAILING_LIST), smtpHost);
    if (mailingList.size() == 0) {
		reporter.error("no mail recipients specified");
		return;
    }

    if (mailer == null) {
      setProgress("creating instance of SendMail");
      mailer = new SendMail(smtpHost);
    }

    // Attempt to find a "from" address
    String from = getStringProperty(ChanSchedBase.CREATOR_EMAIL,"");
    if(from.length() == 0 || from.equals("<user@hostname>") || from.equals("user@hostname")) {
        from = new String("unknown");
    }

    setProgress("getting product");

    String subject = p.getId();

    setProgress("rendering product");

    ProductWriter pw = getProductVisitor();

    StringWriter sw = new StringWriter();
    PrintWriter ps = new PrintCRLFWriter(sw);

    // If a root product exists then we should render it instead of the container
    if(p instanceof ContainerProduct) {
      ContainerProduct cp = (ContainerProduct)p;
      Product root = cp.getRoot();

      if(root != null) {
        // We have a valid root product
        pw.visit(root);
      }
      else {
        // Root product does not exist, render container instead
        pw.visit(p);
      }
    }
    else {
      // Product is not a container, it does not have a root
      pw.visit(p);
    }

    pw.writeResults(ps);
    String body = sw.toString();

    MIMEc message = new MIMEc();
    message.setFrom(from);
    message.setBody(body);
    message.setSubject(subject);

    AsFlatFileList aflat = new AsFlatFileList();
    aflat.visit(p);
    Enumeration e = aflat.getFileProductList();
    while (e.hasMoreElements()) {
      FileRef fileref = ((FileRef)((Product)e.nextElement()).getContent());
      File attachment = fileref.getFile();

      //message.addAttachment(fileref.getMIMEType(), attachment);
      message.addAttachment(fileref.getMIMEType(), attachment, fileref.getSourceName());
    }

    for (int i=0; i < mailingList.size(); i++) {
      String to = (String)mailingList.elementAt(i);
      message.setTo(to);
      try {
        setProgress("sending to user " + i + " of " + mailingList.size());
        mailer.sendMail(message);
      } catch(IOException ioe) {
        reporter.reportException(ioe);
      }
    }

    setProgress("closing smtp connection");
    mailer.closeSMTP();
    setProgress("done");
  }

  /**
  * This dim does not create any transient FileRefs
  */
    public Enumeration getCurrentFileRefs() {
      return null;
    }

    protected ProductWriter getProductVisitor() {
      ProductWriter pw;
      String outputStyle = getStringProperty(OUTPUT_STYLE,"plainText");
      if (outputStyle.equalsIgnoreCase("plainTextNoRootId"))
	pw = new AsPlainTextNoRootId();
      else if (outputStyle.equalsIgnoreCase("tickerTextNoRootId"))
	pw = new AsTickerTextNoRootId();
      else if (outputStyle.equalsIgnoreCase("tickerText"))
	pw = new AsTickerText();
      else // use default visitor
	pw = new AsPlainText();
      return pw;
  }
  
 /*
 * for backward compatibility
 */
	public void setProperty(String attribute, String value) {
		if (attribute.equalsIgnoreCase(USERS)) {
			attribute = MAILING_LIST;
		}
		else if (attribute.equalsIgnoreCase("targetSpec")) {
			attribute = SMTP_HOST;
		}
		super.setProperty(attribute, value);
	}
}