/* Copyright 2000 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.productVisitor.Serializer;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

/**
* An output agent that sends product to another iPAM's acceptSerializedProduct agent.<p>
*
* <pre>
* Config vars:
*
* destinationHost - the IP Address or hostname of the machine where you want to send
*                   the data. (default localhost or 127.0.0.1)
*
* destinationPort - the port number where this agent should send the data.
*                   (default 19099)
*
* compressionLevel - the level of compression to use (0 - 9)
*                     (default level 0 is uncompressed)
* </pre>
*
* @author Jason Mathews   Based on PutAsyncUDP by Jones Leung.
*
* @since iPAM 2.5.101 3/3/2000
*
* @see org.mitre.pam.util.Serializer
* @see org.mitre.pam.getter.remote.GetSerializedProduct
* @see org.mitre.pam.interfaces.AsynchronousAgent
*/

public class PutSerializedProduct extends DIMShell1
  implements org.mitre.pam.getter.remote.SerializedConstants
{
  /** The host name or IP to which the serialized product will be sent to */
  protected final static String DESTINATION_HOST = "destinationHost";

  protected final static String DEFAULT_HOST = "localhost";

  /** The TCP port to which the serialized product will be sent to */
  protected final static String DESTINATION_PORT = "destinationPort";

  /** The level of compression to use.  Valid values are 0 - 9, with 0 meaning uncompressed */
  protected static final String COMPRESSION_LEVEL = "compressionLevel";
  protected static final int DEFAULT_COMPRESSION_LEVEL = 0;

  private static final String DEBUG_KEY = "PutSerializedProduct";

  public PutSerializedProduct() {
    setDescription ("Send Serialized Products");
    metaStringInfo(DESTINATION_HOST, DEFAULT_HOST, "the host that will receive data");
    metaIntegerInfo(DESTINATION_PORT, DEFAULT_PORT, "the port number that the host will be listening on");
    metaIntegerInfo(COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, "the level of compression (0 - 9) to use, level 0 is uncompressed");
  }

  /**
   * Accept a product for output.  
   */
  public void putProduct(Product prod) {
    int destinationPort = getIntegerProperty(DESTINATION_PORT, 0);
    if (destinationPort <= 0) {
      reporter.error("Invalid " + DESTINATION_PORT + ": " + getProperty(DESTINATION_PORT));
      return;
    }

    String destinationHost = getStringProperty(DESTINATION_HOST, "");
    if (destinationHost.length() == 0) {
      reporter.error("Unspecified " + DESTINATION_HOST);
      return;
    }

    int compressionLevel = getIntegerProperty(COMPRESSION_LEVEL, 0);
    if(compressionLevel < 0 || compressionLevel > 9) {
      warning("Illegal compression level: " + compressionLevel + ", setting to 0");
      compressionLevel = 0;
    }

    Socket s = null;
    ObjectOutputStream out = null;
    try {
      s = new Socket(destinationHost, destinationPort);
      reporter.debugMsg("Send product to listener", s, null, null, DEBUG_KEY);
      OutputStream os = new java.io.BufferedOutputStream(s.getOutputStream());
      DeflaterOutputStream dos = new DeflaterOutputStream(os, new Deflater(compressionLevel));
      out = new ObjectOutputStream(dos);
      Serializer.serialize(prod, out);
    } catch (IOException ioe) {
      reporter.reportException(ioe);
    } finally {
      if (out != null)
	try {
	  out.close();
	} catch (IOException ioe) {}
      if (s != null)
	try {
	  s.close();
	} catch (IOException ioe) {}
    }
  } /* PutProduct */
}
