/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import org.mitre.pam.getter.GetAsyncUDP;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.productVisitor.Serializer;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;

import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

/**
* This class is a DIM that will send UDP datagrams. <p>
*
* <pre>
* Config vars:
* destinationHost - the IP Address or hostname of the machine where you want to send
*                   the data. (default localhost or 127.0.0.1)
*
* destinationPort - the UDP port number where this DIM should send the data.
*                   (default 7502)
*
* compressionLevel - the level of compression to use (0 - 9)
*                     (default 0)
*
* Sample PDIF:
* (destinationHost jleung-pc.mitre.org)
* (destinationPort 7502)
* (compressionLevel 1)
* </pre>
*
* NOTE: The maximum reliable length of a UDP datagram is 64900 bytes.
*       If the byte array size exceeds 64900 an AssertionException will occur. <p>
*
* The largest possible UDP datagram is 64K, but some space is taken up
* by the UDP header and IP header. <p>
*
* @author Jones Leung
* @author John R. Fantry - modified class so that it uses DictionaryBaseDIM
*
* @see org.mitre.pam.util.Serializer
* @see org.mitre.pam.getter.GetAsyncUDP
* @see org.mitre.pam.interfaces.AsynchronousAgent
* @see org.mitre.pam.interfaces.Dictionary
*/
public class PutUDP extends DictionaryBaseDIM {
  /** The IP Address or hostname of the machine to which the datagram will be addressed */
  protected final static String DESTINATION_HOST = "destinationHost";
  protected final static String DEFAULT_HOST = "localhost";

  /** The UDP port to which the datagram will be addressed */
  protected final static String DESTINATION_PORT = "destinationPort";

  /** The level of compression to use.  Valid values are 0 - 9, with 0 meaning uncompressed */
  protected static final String COMPRESSION_LEVEL = "compressionLevel";
  protected static final String DEFAULT_COMPRESSION_LEVEL = "0";

  public PutUDP() {
    setDescription ("Send UDP datagram");
    metaStringInfo(DESTINATION_HOST, DEFAULT_HOST, "the host that will receive data");
    metaStringInfo(DESTINATION_PORT, GetAsyncUDP.DEFAULT_PORT, "the port number that the host will be listenting on");
    metaStringInfo(COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, "the level of compression (0 - 9) to use, level 0 is uncompressed");
  }

  /**
  * Use Java serialization and/or compression to turn the object into a byte array.  Pass the
  * byte array to sendByteArray().
  *
  * @param the object to be sent
  */
  protected void sendObject(Object obj) {
    int compressionLevel = getIntegerProperty(COMPRESSION_LEVEL, 0);
    if(compressionLevel < 0 || compressionLevel > 9) {
      warning("Illegal compression level: " + compressionLevel + ", setting to 0");
      compressionLevel = 0;
    }

    try {
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      DeflaterOutputStream dos = new DeflaterOutputStream(bos, new Deflater(compressionLevel));
      ObjectOutputStream out = new ObjectOutputStream(dos);
      Serializer.serialize(obj, out);
      out.flush();
      out.close();
      sendByteArray(bos.toByteArray());
    } catch(IOException ioe) {
      reportException(ioe);
    }
  }

  /**
  * Send a byte array using UDP.
  *
  * @param the byte array to be sent
  */
  protected void sendByteArray(byte[] b) {
    // the byte array is null, so there is nothing to send
    if (b == null) {
      return;
    }

    // if the byte array is too large, do not attempt to send it.
    assert (b.length <= GetAsyncUDP.MAX_PACKET_SIZE,
                        "PutUDP: Product too large to fit in a single UDP datagram!  Unable to send UDP datagram!",
                        null, null, null);

    if (b.length > GetAsyncUDP.MAX_PACKET_SIZE) {
      return;
    }

    String destinationHost = getProperty (DESTINATION_HOST);
    String destinationPort = getProperty (DESTINATION_PORT);

    try {
      InetAddress ia = InetAddress.getByName (destinationHost);
      DatagramPacket dp = new DatagramPacket (b, b.length, ia, Integer.parseInt (destinationPort));
      DatagramSocket ds = new DatagramSocket ();
      ds.send(dp);
      ds.close();
    } catch (Exception e) {
      reportException(e);
    }
  }
}
