/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.aim;

import org.mitre.pam.interfaces.Reporter;

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.net.Socket;

import java.util.StringTokenizer;
import java.util.Vector;

/**
* This class is the driver for the AIMPIM and AIMDIM.  It handles all the protocol issues
* associated with AOL Instant Messenger.
*
* @author John R. Fantry
*/
final public class AIMDriver implements Runnable {
	static String DEBUG = "aolim";
	/**
	* A list of AIM agents that receive messages
	*/
	protected Vector aimListeners = new Vector();

	// AIM only supports English, so there is no need to make this value configurable at this point
	protected static final String LANGUAGE = "english";

	protected static final String TIC_VERSION = "TIC:$Revision: 1.83 $";

	protected static final String MSG_TRUNCATED = "[message truncated]";

	protected Socket AIMServer = null;
	protected DataInputStream in = null;
	protected DataOutputStream out = null;

	protected boolean connected = false;
	protected boolean loginError = false;

	Reporter reporter = null;

	protected Thread listener = null;

	protected int seqNo = (int)Math.floor(Math.random() * 65535.0);

	String AIMTOCServer = null;
	int AIMTOCServerPort = -1;
	String loginName = null;
	String password = null;
	String authHost = null;
	int authPort = -1;

	public AIMDriver(AIMAgent agent, Reporter reporter) {
		this.reporter = reporter;
		addAIMListener(agent);
	}

	/**
	* Open connection to the AIM server and sign on.  Because the login procedure is somewhat
	* asynchronous (due to the listener thread) there is a while loop in the function that
	* will not return until we know the state of the login attempt.
	*/
	public void startService(String AIMTOCServer, int AIMTOCServerPort, String loginName, String password, String authHost, int authPort) {
		if(connected) {
			return;
		}

		this.AIMTOCServer = AIMTOCServer;
		this.AIMTOCServerPort = AIMTOCServerPort;
		this.loginName = loginName;
		this.password = password;
		this.authHost = authHost;
		this.authPort = authPort;

		loginError = false;

		try {
			AIMServer = new Socket(AIMTOCServer, AIMTOCServerPort);
			in = new DataInputStream(AIMServer.getInputStream());
			out = new DataOutputStream(AIMServer.getOutputStream());
		} catch(IOException ioe) {
			stopService();
			reporter.reportException(ioe);
			return;
		}

		listener = new Thread(this);
		listener.start();

		startLogin();

		/*
		* The idea behind this loop is to ensure that we know the result of the login
		* before returning.  If this loop ends because the listener thread is no
		* longer active then the login attempt may have timed out.
		*/
		while(!connected && listener.isAlive() && !loginError) {
			try {
				Thread.currentThread().sleep(3000);
			} catch(InterruptedException ie) {
				stopService();
				reporter.reportException(ie);
				reporter.warning("login interrupted, aborting process");
				return;
			}
		}

		if(loginError) {
			// this error should have been reported by handleInMsg()
			//System.out.println("AIMDIM: login error detected");
			stopService();
			return;
		}
		else if(!listener.isAlive()) {
			// this indicates that our login attempt timed out because
			// the AIM_TOC_SERVER closed the connection

			stopService();
			reporter.warning("login timed out");
			return;
		}

		//System.out.println("AIMDIM: login successful");
	}

	/**
	* Disconnect from server
	*/
	public void stopService() {
		signOff();
		aimListeners.setSize(0);
	}

	/**
	* Closes the connection to the server.  I couldn't find any message passing involved in the sign off,
	* just close the pipe.
	*/
	private void signOff() {
		try {
			/*
			* in is used as a flag to indicate that a connection is available
			*/
			if(in != null) {
				DataInputStream hold = in;
				in = null;
				hold.close();
			}

			if(out != null) {
				out.close();
				out = null;
			}

			if(AIMServer != null) {
				AIMServer.close();
				AIMServer = null;
			}
		} catch(IOException ioe) {
			reporter.reportException(ioe);
		}

		connected = false;
	}

	/**
	* Starts the login sequence
	*/
	private void startLogin() {
		if(connected) {
			return;
		}

		try {
			//System.out.println("AIMDIM: sending FLAPON");
			out.writeBytes("FLAPON\r\n\r\n");
			out.flush();
		} catch(IOException ioe) {
			reporter.reportException(ioe);
		}
	}

	/**
	* Finish the login sequence
	*/
	private void finishLogin() {
		if(connected) {
			return;
		}

		if(loginName.equals("") || password.equals("")) {
			reporter.warning("Login name and/or password is not defined.");
			return;
		}

		try {
			ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
			DataOutputStream dataOutputStream = new DataOutputStream(byteArrayOutputStream);
			dataOutputStream.writeInt(1);
			dataOutputStream.writeShort(1);
			dataOutputStream.writeShort(loginName.length());
			dataOutputStream.writeBytes(loginName);
			//System.out.println("AIMDIM: sending TOC FLAP SIGNON");
			send(1, byteArrayOutputStream.toByteArray());

			//System.out.println("AIMDIM: sending TOC \"toc_signon\" message");
			send("toc_signon " + authHost + " " + Integer.toString(authPort) + " " + normalize(loginName) + " " + roast(password) + " " + LANGUAGE + " " + encode(TIC_VERSION));
		} catch(IOException ioe) {
			reporter.reportException(ioe);
		}
	}

	public boolean isConnected() {
		return connected;
	}

	private void handleInMsg(int type, byte[] msg) {
		//System.out.println("AIMDIM: type = " + type);

		if(type == -1) {
			// this indicates that something has gone wrong and we need to sign off
			signOff();
			return;
		}

		if(type == 1) {
			// this type is received as a result of sending the FLAPON message
			// it indicates that the FLAPON message was received successfully and that we
			// can proceed with the login
			//log("AIMDIM: received FLAP SIGNON");
			//System.out.println("AIMDIM: received FLAP SIGNON");
			finishLogin();
			return;
		}

		if(msg == null) {
			reporter.warning("received null message, type = " + type);
			return;
		}

		String str = new String(msg);
		String msgs[] = split(str, ':');
		if (msgs.length == 0) {
			return;
		}

		if(msgs[0].equals("SIGN_ON")) {
			// if we supported buddy lists this is where we would upload our list
			// buddy lists are maintained locally and are uploaded during each login

			// this completes the login process
			send("toc_init_done");
			connected = true;
		}
		else if(msgs[0].equals("IM_IN")) {
			// we have received an Instant Message
			// turn it into a product and push it through the system
			sendMsgToListeners(str);
		}
		else if(msgs[0].equals("CONFIG")) {
			// this is essentially a request by the server for the client to build
			// and upload their buddy list
		}
		else if(msgs[0].equals("NICK")) {
			// for whatever reason the server sends us our login name in msgs[1]
			// I suppose we could check it to confirm we logged in correctly, but
			// I don't know if mismatched names actually represents a failed login
		}
		else if(msgs[0].equals("ERROR")) {
			if(msgs.length == 1) {
				reporter.warning("received error with no error code");
				return;
			}

			// msgs[1] holds the error code
			// known error codes:
			// 901: the receiver is not available (not logged on, doesn't exist)
			// 980: we were not able to login properly (bad user name / password)

			switch(Integer.parseInt(msgs[1])) {
				case 901 : {
						// msgs[2] holds the name of the user who is unavailable
						reporter.log("error 901: user \"" + msgs[2] + "\" is unavailable");
						break;
					}
				case 980 : {
						reporter.warning("error 980: login error, check user name and password");
						loginError = true;
						break;
					}
				default : {
						reporter.warning("unknown error: \"" + msgs[1] + "\"");
					}
			}
		}
	}
	/**
	* Send a received message to any listeners that we have
	*/
	private void sendMsgToListeners(String str){
		for (int i = 0; i< aimListeners.size(); i++) {
			AIMAgent agent = (AIMAgent) aimListeners.elementAt(i);
			agent.processIM(str);
		}
	}
	/**
	* Add a listener
	*/
	public void addAIMListener(AIMAgent agent) {
		if (aimListeners.contains(agent))
			return;
		aimListeners.addElement(agent);
	}
	/**
	* Remove a listener
	*/
	public void removeAIMListener(AIMAgent agent) {
		aimListeners.removeElement(agent);
	}
	/**
	* Remove a listener
	*/
	public int getAIMListenerCount() {
		return aimListeners.size();
	}
	/**
	* Send an Instant Message by specifying the reciever and the message.
	*/
	public void sendIM(String userName, String msg) {
		if(!connected) {
			return;
		}

		int count = 0;
		String encdMsg = encode(msg);
		while(encdMsg.length() > 1000) {

			/*
			* The max encoded length of a message is 1024.  The idea here is to preserve as much of
			* the original msg as possible after encoding and appending the MSG_TRUNCATED string.
			* Encoding adds a minimum of 2 chars, but will add an additional 2 for every special
			* char that it finds in the string.  This is why (count * 10) is included in maxTruncLength.
			* After the first iteration an additional 10 chars will be removed from each successive
			* iteration until we get an encoded message that has an acceptable length.
			*/
			int maxTruncLength = 1024 - 2 - MSG_TRUNCATED.length() + 1 - (count * 10);

			StringBuffer truncMsg = new StringBuffer(msg.substring(0, maxTruncLength));
			truncMsg.append("\n" +  MSG_TRUNCATED);

			encdMsg = encode(truncMsg.toString());
			count++;
		}

		//System.out.println("AIMDIM: sending = \"" + msg + "\"");
		//System.out.println("AIMDIM: length = " + msg.length());
		//log("\"" + msg + "\"");
		//System.out.println("AIMDIM: legnth = " + encdMsg.length());

		//send("toc_send_im " +  normalize(userName) + " " + encode(msg) + "");
		send("toc_send_im " +  normalize(userName) + " " + encdMsg + "");
	}

	/**
	* Send a string by converting it into a byte array and calling the corresponding
	* send function.
	*/
	private void send(String msg) {
		try {
			ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
			DataOutputStream dataOutputStream = new DataOutputStream(byteArrayOutputStream);
			dataOutputStream.writeBytes(msg);
			dataOutputStream.write(0);
			send(2, byteArrayOutputStream.toByteArray());
		} catch (IOException ioe) {
			reporter.reportException(ioe);
		}
	}

	/**
	* Send a message.
	*
	* @param type The type or class of this messsge
	* @param msg The message to send
	*/
	private void send(int type, byte[] msg) {
		try {
			int j = 0;
			if(msg != null) {
				j += msg.length;
			}

			out.writeByte(42);                // all msg's must be prefixed
			out.writeByte(type);
			out.writeShort(seqNo);
			seqNo = seqNo + 1 & 65535;
			out.writeShort(j);

			if(msg != null) {
				out.write(msg);
			}

			out.flush();
		} catch (IOException ioe) {
			reporter.reportException(ioe);
		}
	}

	/**
	* This function will roast a string.
	*/
	private String roast(String str) {
		if(str == null) {
			return str;
		}

		String roastStr = "0x";
		int an[] = {84, 105, 99, 47, 84, 111, 99};

		for(int i = 0; i < str.length(); i++) {
			int j = str.charAt(i);
			j ^= an[i % an.length];
			if(j < 16) {
				roastStr = new StringBuffer(String.valueOf(roastStr)).append("0").toString();
			}
			roastStr = new StringBuffer(String.valueOf(roastStr)).append(Integer.toHexString(j)).toString();
		}

		return roastStr;
	}

	/**
	* This function removes white space and ensures that all characters are lower case.
	*/
	private static String normalize(String str) {
		if(str == null) {
			return str;
		}

		int i = str.length();
		StringBuffer stringBuffer = new StringBuffer(i);
		for(int j = 0; j < i; j++) {
			char ch = str.charAt(j);
			if(Character.isLetterOrDigit(ch)) {
				stringBuffer.append(Character.toLowerCase(ch));
			}
		}

		return stringBuffer.toString();
	}

	/**
	* This function inserts '\\' before any special characters.
	*/
	private String encode(String str) {
		int i = str.length();
		StringBuffer stringBuffer = new StringBuffer(i + 10);
		stringBuffer.append('"');

		for(int j = 0; j < i; j++) {
			char ch = str.charAt(j);
			if(ch == 92 || ch == 34 || ch == 123 || ch == 125 || ch == 40 || ch == 41 || ch == 91 || ch == 93 || ch == 36) {
				stringBuffer.append('\\');
			}

			stringBuffer.append(ch);
		}

		stringBuffer.append('"');
		return stringBuffer.toString();
	}

	/**
	* This function breaks a string apart based on a special character and
	* stores each piece in an array.
	*/
	private String[] split(String str, char ch) {
		Vector v = new Vector();
		StringTokenizer st = new StringTokenizer(str, String.valueOf(ch));
		while(st.hasMoreTokens()) {
			String elt = st.nextToken();
			v.addElement(elt);
		}

		String strs[] = new String[v.size()];
		v.copyInto(strs);

		return strs;
	}

	/**
	* Read a message from the input stream.  This method will block.  All messages are
	* passed to the handleInMsg() function.
	*/
	private void read() throws IOException {
		//System.out.println("waiting");
		byte b = in.readByte();
		if(b != 42) {                       // all msgs must begin with 42
			throw new IOException(new StringBuffer("Got wrong start marker: ").append((char)b).toString());
		}

		int i = in.readByte() & 255;        // msg type
		int j = in.readUnsignedShort();     // seqNo
		int k = in.readUnsignedShort();     // msg length

		byte msg[] = null;
		if(k > 0) {
			msg = new byte[k];
			in.readFully(msg);
		}

		handleInMsg(i, msg);
	}

	/**
	* Call read() until there is an error, or signOff() is called.
	*/
	public void run() {
		try {
			while(in != null) {
				read();
			}
		} catch(IOException ioe) {
			// in == null on normal close
			if (in != null) {
				reporter.reportException(ioe);
				handleInMsg(-1, null);
			}
		}
		listener = null;
		connected = false;
	}
}
