/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.aim;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFactory;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.BasicProductAttributes;
import org.mitre.pam.util.PAMDate;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Date;
import java.util.StringTokenizer;
import java.util.Vector;

/**
* This PIM uses the AOL Instant Messenger service to receive messages from AIM users.
*
* @author John R. Fantry
*/
public class AIMPIM extends org.mitre.pam.getter.PIMShell1 implements AIMAgent, AsynchronousAgent {

	static String DEBUG = "aolim";

	protected static final int DEFAULT_AIM_TOC_SERVER_PORT = 21;  // FTP port

	// must be unique, multiple AIMAgents can NOT share loginNames
	protected static final String DEFAULT_LOGIN_NAME = "";

	protected static final String DEFAULT_PASSWORD = "";

	protected static final int DEFAULT_AUTH_HOST_PORT = (int)(Math.random() * 2000.0 + 1025.0);  // any port will work

	// if true, a timestamp will be included in the ID of the product
	protected static final String TIMESTAMP = "timestamp";
	protected static final boolean DEFAULT_TIMESTAMP = false;

	protected static final String DESCRIPTION_LENGTH = "descriptionLength";
	protected static final int DEFAULT_DESCRIPTION_LENGTH = 60;

	// if true, a BasicProduct with URL content will be created for each URL in the IM
	// the BasicProducts and the IM will be packaged in a VectorProduct
	protected static final String EXTRACT_URLS = "extractURLs";
	protected static final boolean DEFAULT_EXTRACT_URLS = true;

	// if true, the HTML tags found in the IM will be removed
	protected static final String REMOVE_HTML_TAGS = "removeHTMLTags";
	protected static final boolean DEFAULT_REMOVE_HTML_TAGS = true;

	protected static final String[] REMOVABLE_WHOLE_TAGS = {"HTML","/HTML","I", "/I", "B", "/B", "U", "/U", "/A", "/FONT","/BODY"};

	protected static final String[] REMOVABLE_PARTIAL_TAGS = {"FONT","BODY"};

	protected static final String URL_TAG = "A HREF=\"http://";

	protected AIMDriverProxy driver = new AIMDriverProxy(this, reporter);

	protected ProductFactory pfab;

	public AIMPIM() {
		metaStringInfo(AIMDIM.AIM_TOC_SERVER, AIMDIM.DEFAULT_AIM_TOC_SERVER, "The AIM TOC server to connect with.");
		metaIntegerInfo(AIMDIM.AIM_TOC_SERVER_PORT, DEFAULT_AIM_TOC_SERVER_PORT, "The port to talk to the AIM TOC server on.");
		metaStringInfo(AIMDIM.LOGIN_NAME, DEFAULT_LOGIN_NAME, "The AIM login name (screen name).");
		metaSetIsPassword(AIMDIM.PASSWORD, "AOL IM login password.");
		metaStringInfo(AIMDIM.AUTH_HOST, AIMDIM.DEFAULT_AUTH_HOST, "The server that will perform the authorization on the loginName and password.");
		metaIntegerInfo(AIMDIM.AUTH_HOST_PORT, DEFAULT_AUTH_HOST_PORT, "The port to talk to the authorization server on.");
		metaBooleanInfo(TIMESTAMP, DEFAULT_TIMESTAMP, "If true, a timestamp will be included in the ID of the product.");
		metaIntegerInfo(DESCRIPTION_LENGTH, DEFAULT_DESCRIPTION_LENGTH, "The number of characters to include in the product description.");
		metaBooleanInfo(EXTRACT_URLS, DEFAULT_EXTRACT_URLS, "If true, a BasicProduct with URL content will be created for each URL in the IM.");
		metaBooleanInfo(REMOVE_HTML_TAGS, DEFAULT_REMOVE_HTML_TAGS, "If true, the HTML tags found in the IM will be removed.");
	}

	/**
	* Ask the driver to login to the AIM service.
	*/
	public void enableService() {
		if (pfab == null)
			pfab = exec.getProductFactory();

		if(driver.isConnected()) {
			return;
		}

		String AIMTOCServer = getStringProperty(AIMDIM.AIM_TOC_SERVER, AIMDIM.DEFAULT_AIM_TOC_SERVER);
		int AIMTOCServerPort = getIntegerProperty(AIMDIM.AIM_TOC_SERVER_PORT, DEFAULT_AIM_TOC_SERVER_PORT);
		String loginName = getStringProperty(AIMDIM.LOGIN_NAME, "");
		String password = getStringProperty(AIMDIM.PASSWORD, "");
		String authHost = getStringProperty(AIMDIM.AUTH_HOST, AIMDIM.DEFAULT_AUTH_HOST);
		int authPort = getIntegerProperty(AIMDIM.AUTH_HOST_PORT, DEFAULT_AUTH_HOST_PORT);

		if(loginName.length()==0 || password.length()==0) {
			warning("loginName/password not defined");
			setEnabled(false);
			return;
		}

		driver.startService(AIMTOCServer, AIMTOCServerPort, loginName, password, authHost, authPort);
	}

	/**
	* Disconnect from server
	*/
	public void disableService() {
		driver.stopService();
	}

	public Product doFetch() {
		return null;
	}

	/**
	* Turn an incoming IM into a product.  The arg contains four ':' delimted fields ("IM_IN", sender, boolean, message).
	*/
	public void processIM(String str) {
		boolean timestamp = getBooleanProperty(TIMESTAMP, DEFAULT_TIMESTAMP);
		int descLength = getIntegerProperty(DESCRIPTION_LENGTH, DEFAULT_DESCRIPTION_LENGTH);
		boolean descTruncated = false;

		StringTokenizer st = new StringTokenizer(str.substring(str.indexOf(":") + 1), ":");

		String sender = st.nextToken();
		boolean bool = st.nextToken().equals("T");

		/*
		* The message may contain the ':' char.  Since that char is what the StringTokenizer is
		* parsing on we may need to loop in order to get the entire message.  This happens whenever
		* a URL is included in the message (i.e. <A HREF>http://www.javasoft.com</A>)
		*/
		int remainingTokens = st.countTokens();
		String rawMsg = null;
		Vector urls = new Vector();
		if(remainingTokens > 1) {
			StringBuffer sb = new StringBuffer();
			sb.append(st.nextToken());
			for(int i=1; i<remainingTokens; i++) {
				sb.append(":" + st.nextToken());
			}
			rawMsg = sb.toString();
		}
		else {
			rawMsg = st.nextToken();
		}

		boolean removeTags = getBooleanProperty(REMOVE_HTML_TAGS, DEFAULT_REMOVE_HTML_TAGS);
		boolean extractURLs = getBooleanProperty(EXTRACT_URLS, DEFAULT_EXTRACT_URLS);

		String parsedMsg = null;

		if(extractURLs || removeTags) {
			// originally this call was only made if the tags were to be removed, however because it is a good
			// spot to detect URL's it performs the side effect of filling in the Vector argument
			parsedMsg = parseIM(rawMsg, urls);
		}

		String id = sender;
		if(timestamp) {
			id = sender + " (" + new PAMDate(System.currentTimeMillis()) + ")";
		}

		String description = null;
		if(removeTags) {                    // use the parsed version of the IM
			if(parsedMsg.length() < descLength) {
				description = parsedMsg;
			}
			else {
				description = parsedMsg.substring(0, descLength);
				descTruncated = true;
			}
		}
		else {                              // use the unparsed version of the IM
			if(rawMsg.length() < descLength) {
				description = rawMsg;
			}
			else {
				description = rawMsg.substring(0, descLength);
				descTruncated = true;
			}
		}

		// if there are no URLs, or we are not supposed to extract the existing ones, we only need
		// a BasicProduct
		if(urls.size() == 0 || !extractURLs) {
			Product bp = null;
			if(descTruncated) {               // fill in the content appropriately
				if(removeTags) {
					bp = pfab.makeProduct(id, description, parsedMsg);
				}
				else {
					bp = pfab.makeProduct(id, description, rawMsg);
				}
			}
			else {
				bp = pfab.makeProduct(id, description, "");
			}

			pushProductToAllClients(bp);
		}
		else {
			ContainerProduct vp = pfab.makeVectorProduct(id, description);
			if(descTruncated) {               // need to include a BasicProduct with the IM as content
				if(removeTags) {
					vp.addElement(pfab.makeProduct(id, "", parsedMsg));
				}
				else {
					vp.addElement(pfab.makeProduct(id, "Instant Message", rawMsg));
				}
			}
			for(int i=0; i<urls.size(); i++) {
				try {
					vp.addElement(pfab.makeProduct(id, "", new URL((String)urls.elementAt(i))));
				} catch(MalformedURLException mue) {
					reportException(mue);
				}
			}

			pushProductToAllClients(vp);
		}
	}

	/**
	* Strip all the html tags out of a raw IM.
	*
	* Incoming IM's seem to have the following format:
	*
	*   <HTML><BODY BGCOLOR="#ffffff">hello world</HTML>
	*
	* The message body may also contain <A HREF></A> tags.  As a side effect the urls Vector
	* will be filled with a string for every URL that is found.
	*/
	protected String parseIM(String rawMsg, Vector urls) {
		reporter.debugMsg("recv str:",rawMsg,null,null,DEBUG);
		int startIndex = rawMsg.indexOf('<');
		if(startIndex == -1) {              // there are no tags in the IM
			return rawMsg;
		}

		StringBuffer sb = new StringBuffer(rawMsg.length());
		int stopIndex = -1;

		while(startIndex != -1) {
			// append up to the start of the tag
			sb.append(rawMsg.substring(stopIndex+1, startIndex));
			// find the next tag
			stopIndex = rawMsg.indexOf('>', startIndex);
			String tag = rawMsg.substring(startIndex+1, stopIndex);
			reporter.debugMsg("tag=",tag,null,null,DEBUG);
			// examine the tag to see it is a url
			if(tag.startsWith(URL_TAG)) {
				String url = tag.substring(tag.indexOf("\"")+1, tag.lastIndexOf("\""));
				urls.addElement(url);
			}
			// if not url then see if we should remove it
			else if(!removeTag(tag)) {
				sb.append(rawMsg.substring(startIndex, stopIndex+1));
			}
			// search for start of next tag
			startIndex = rawMsg.indexOf('<', stopIndex);
		}

		if(stopIndex < rawMsg.length()) {   // append the rest of the IM
			sb.append(rawMsg.substring(stopIndex+1));
		}

		return sb.toString();
	}

	/**
	* Determines whether or not a tag is removable.
	*/
	protected boolean removeTag(String tag) {
		for(int i=0; i<REMOVABLE_WHOLE_TAGS.length; i++) {
			if(tag.equals(REMOVABLE_WHOLE_TAGS[i])) {
				return true;
			}
		}

		for(int i=0; i<REMOVABLE_PARTIAL_TAGS.length; i++) {
			if(tag.startsWith(REMOVABLE_PARTIAL_TAGS[i])) {
				return true;
			}
		}

		return false;
	}

}
