/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.resource;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Reporter;
import java.util.Vector;

public class RateConsumer {
	public static final String BUDGET = "bytesPerMinuteBudget";
	public static final String HISTORICAL_LEVEL = "bytesPerMinuteConsumption";
	public static final String USE_POOL_WEIGHT = "useBytesPerMinutePoolWeight";
	public static final String POOL_WEIGHT = "bytesPerMinutePoolWeight";
	public static final String OVERAGE = "bytesPerMinuteBidOverBudget";

/*
* Time history buffer area
*/
	double budget=1.0;
	double consumption;
	double totalByteCount;
	long timeSpan;
	private int histLen;	
	private int theHead; // newest item
	private static final int THE_TAIL = 0;	//oldest item
	private Slot[] history ;
	
	protected AccessibleAgent agent;
	private String debugKey;
	boolean usePoolWeight; // used by the pool for delegated budets
	double poolWeight = 1.0; // used by the pool for delegated budets
	private static final double MIN_TIME = 60000.0;

/**
* Update statistics but no more
* often than every 500 milliseconds
*/
	private long lastTimeHack = 0;
/**
* Info about files transfered 
*/
	class Slot {
		AccessibleAgent requestor;	// identity of agent making request
		long time;		// time of entry
		double count;	// number of bytes
	}
	RateConsumer(AccessibleAgent agent, String debugKey) {
		this.agent = agent;
		this.debugKey = debugKey;
		setHistorySize(8);
	}
	private double calcConsumption(long timeSpan, double byteCount){
		return (timeSpan > MIN_TIME) ? 
				(60000.0 * byteCount / timeSpan) :
				(byteCount) ;
	}
/**
* Delay time is the number of milliseconds
* to wait to obtain budget within a certain level.
*
* The goal is: <p>
*
*    cumByteCount / (delayTime + timeSpan ) <=  budget * (minutes / milliseconds) <p>
*
* Budget is in bytes per minute.  The answer is in milliseconds.
*/
	private synchronized long calcDelay(double byteCount, long timeSpan) {
		if (budget <= 0.0)
			budget = 1.0;
		long delayTime = (long) (byteCount / (budget /60000.0) - timeSpan);
		if (delayTime < 0) {
			delayTime = 0;
		}
		return delayTime;
	}
/**
* Calculate metrics
*/
	protected synchronized long calculateDelay(long time) {
		return calcDelay(getCumByteCount(), time - history[THE_TAIL].time);
	}
/**
* Calculate metrics
*/
	protected void calculateMetrics(long currentTime) {
	/*
	* update span numbers
	*/
		double bc = getCumByteCount();
		timeSpan = currentTime - history[THE_TAIL].time;
		consumption = calcConsumption(timeSpan, bc);
		
		Reporter reporter = agent.getReporter();
		if (reporter.isDebugKey(debugKey)) {
			reporter.debugMsg("timeSpan (msec) = ",Double.toString(timeSpan),"   bytes = ",Double.toString(bc),debugKey);
			reporter.debugMsg("bpm = ",Double.toString(consumption),null,null,debugKey);
		}
		setConsumption(consumption);
	}
/**
* set the buffer history to fudge the consumption history
*/
	public synchronized void fudgeConsumptionHistory(double value) {
	/*
	* initialize our time history to the current time
	*/
		long t = System.currentTimeMillis();
		for (int i = 0; i< histLen; i++) {
			history[i].time = t-1;
			history[i].count = 0.0;
		}
	/*
	* initialize our consumption for the last "sample"
	*/
		history[theHead].count = value;
		history[theHead].time = t;
		calculateMetrics(t);
	}
/**
* Fudge consumption.
*/
	void fudgeConsumptionHistory(String value){
		try {
			fudgeConsumptionHistory(Double.valueOf(value).doubleValue());
		}
		catch (NumberFormatException e) {
			fudgeConsumptionHistory(0.0 );
		}
	}
/**
* Get current consumption settings
*/
	public double getConsumption() {
		return consumption;
	}
/*
* Recalculate cumByteCount if needed
*/
	private double getCumByteCount() {
		if (totalByteCount > 1.0)
			return totalByteCount;
		totalByteCount = 0.0;
		for (int i = 0; i< histLen; i++) {
			totalByteCount += history[i].count;
		}
		return totalByteCount;
	}
/**
* Allow a change in the length of the history.
*/
	synchronized int getHistorySize() {
		return histLen;
	}
/**
* Flush the history of any agent that is delaying and return
* the list of flushed agents.
*/
	protected synchronized void retractUpdate(AccessibleAgent requestor) {
		agent.getReporter().debugMsg("retract requestor ",requestor, null, null,debugKey);
		int pos = -1;
		for (int i = histLen-1; i>=0 ; i--) {
			if (history[i].requestor == requestor) {
				pos = i;
				break;
			}
		}
		if (pos == -1) {
			agent.getReporter().debugMsg(" * requestor not in history",null, null, null,debugKey);
			return;
		}
		for (int i = pos; i>=1 ; i--) {
			history[i] = history[i-1];
		}
		history[0] = new Slot();
		history[0].time = history[1].time-1;
		history[0].requestor = null;
		history[0].count = 0.0;
	/*
	* we intentionally do not calculate metrics
	* because we expect that the retract will be followed
	* by an update or more retracts - avoid unneeded recalculation
	*/
		totalByteCount = -1.0;
	}	
/**
* Set the allowed pool limit for total number of bytes-per-minute; 
*
* @param resourceDimension for multidimensional pools this is the name of the dimension
* @param newBudget the value of this dimension
*/
	synchronized void setBudget(double budgetAmount){
		if (budgetAmount > 1.0)
			budget = budgetAmount;
		else
			budget = 1.0;
		agent.superSetProperty(BUDGET, Double.toString(budget));
	}
/**
* Set budgeted level.
*/
	void setBudget(String value){
		try {
			setBudget(Double.valueOf(value).doubleValue());
		}
		catch (NumberFormatException e) {
			setBudget(100000.0 );
		}
	}
/**
* Set the current comsumption level
*/
	protected synchronized void setConsumption(double value) {
		consumption = value;
		if (consumption > budget && usePoolWeight) {
			agent.superSetProperty(HISTORICAL_LEVEL,Double.toString(budget));
			agent.superSetProperty(OVERAGE,Double.toString(consumption-budget));
		}
		else {
			agent.superSetProperty(HISTORICAL_LEVEL,Double.toString(consumption));
			agent.superSetProperty(OVERAGE,"0");
		}
	}
/**
* Allow a change in the length of the history.
*/
	synchronized void setHistorySize(int length) {
		histLen = length;
		theHead = histLen-1; 
		if (history == null || length > history.length) {
			history = new Slot[length];
			for (int i = 0; i< length; i++) {
				history[i] = new Slot();
			}
		}
		fudgeConsumptionHistory(budget);
	}
/**
* Set initial byte-per-minute historical value.
*/
	public void setPoolWeight(double value){
		if (value < 0.00001) {
			agent.getReporter().warning("specified pool weight of "+value+" is too small; pool weight set to 1.0");
			value = 1.0;
		}
		agent.superSetProperty(POOL_WEIGHT, value);
		poolWeight = value;
	}
/**
* Set initial byte-per-minute historical value.
*/
	public void setPoolWeight(String value){
		try {
			setPoolWeight(Double.valueOf(value).doubleValue());
		}
		catch (NumberFormatException e) {
			setPoolWeight(1.0 );
		}
	}
/**
* Set whether to use the pool weight to
* divide up the pool's budget or not.
*
* @param value if true means proportional algorithm, false means first come, first served
*/
	void setUsePoolWeight(boolean value){
		usePoolWeight = value;
		agent.superSetProperty(USE_POOL_WEIGHT, value ? "true" : "false");
	}
/**
* Update statistics but no more
* often than every 500 milliseconds
*/
	public void timeHack() {
		long time = System.currentTimeMillis();
		if (time > lastTimeHack + 500) {
			calculateMetrics(System.currentTimeMillis());
			lastTimeHack = time;
		}
	}
/**
* Update history and calculate metrics
*/
	protected synchronized long updateAndCalculateDelay(AccessibleAgent requestor, long time, double byteCount) {
		long delay = calcDelay(getCumByteCount(), time - history[THE_TAIL].time);
		updateHistory(requestor, time, byteCount);
		return delay;
	}
/**
* Update the time/consumption history
*
* @return amount of time to delay to obtain budget
*/
	protected synchronized void updateHistory(AccessibleAgent requestor, long time, double byteCount) {
		Slot head = history[THE_TAIL];
		for (int i = 0; i< histLen-1; i++) {
			history[i] = history[i+1];
		}
		history[theHead] = head;
		head.time = time;
		head.requestor = requestor;
		head.count = byteCount;
		totalByteCount = -1.0; // flag that recalc is needed
		calculateMetrics(time);
	}
}