/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.productVisitor.AsXML;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.smtp.MIMEc;

import org.mitre.pdif.eventParser.ValueList;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Vector;

/**
 * This filter creates an XML representation of a PAM product structure.
 *
 * The filter uses the AsXML visitor to generate the XML.  Below is an example
 * of the XML that this filter generates.  This XML was created by piping two
 * getPIM1 agents through the filter.
 *
 * <?xml version="1.0" ?>
 * <Container>
 *   <Property name="knownContentProperties">
 *    <Value>description</Value>
 *    <Value>id</Value>
 *    <Value>lifeTimeLong</Value>
 *  </Property>
 *  <Property name="description">
 *    <Value />
 *  </Property>
 *  <Property name="id">
 *    <Value>Multiple products</Value>
 *  </Property>
 *  <Property name="lifeTimeLong">
 *    <Value>259200</Value>
 *  </Property>
 *  <Content>
 *    <Property name="description">
 *      <Value>Creates a string product</Value>
 *    </Property>
 *    <Property name="id">
 *      <Value>AdmingetPIM12</Value>
 *    </Property>
 *    <Property name="lifeTimeLong">
 *      <Value>259200</Value>
 *    </Property>
 *    <StringValue><content></StringValue>
 *  </Content>
 *  <Content>
 *    <Property name="description">
 *      <Value>Creates a string product</Value>
 *    </Property>
 *    <Property name="id">
 *      <Value>AdmingetPIM11</Value>
 *    </Property>
 *    <Property name="lifeTimeLong">
 *      <Value>259200</Value>
 *    </Property>
 *    <StringValue><content></StringValue>
 *  </Content>
 * </Container>
 *
 * BasicProducts are represented in the XML as <Content> and ContainerProducts
 * are represented as <Container>.  It is perfectly legal to nest containers
 * inside containers.
 *
 * Each <Content> tag will include entries for all the products attributes
 * (except for those defined in the exclusion list) along with an entry
 * for the products content.
 *
 * Each <Container> tag will include a field called "knownContentProperties".
 * This list defines all the properties that are used by the <Content> tags
 * defined later.
 *
 * The DTD for the XML document is the following:
 *
 * <!ELEMENT Container (Property+,(Content|Container)*)>
 * <!ELEMENT Content (Property+,(StringValue|A|Filename|ByteValue)?)>
 * <!ELEMENT Property (Value*)>
 * <!ATTLIST Property name CDATA #IMPLIED>
 * <!ELEMENT A (#PCDATA)>
 * <!ATTLIST A HREF CDATA #IMPLIED>
 * <!ELEMENT ByteValue (#PCDATA)>
 * <!ELEMENT Filename (#PCDATA)>
 * <!ELEMENT StringValue (#PCDATA)>
 * <!ELEMENT Value (#PCDATA)>
 *
 * @author John R. Fantry
 */
public class ForwardAsXMLFile extends NoSchedBase {
  // a list of product properties that should not appear in the XML
  public static final String EXCLUSION_LIST = "productAttributesToExclude";
  public static final String[] DEFAULT_EXCLUSION_LIST = {
	Product.PROPERTY_LIFETIME, Product.PROPERTY_DESCRIPTION , Memento.MEMENTO_UID
	};

  protected static final String OUTPUT_FILE_NAME = "xmlOutputFileName";
  
  public ForwardAsXMLFile() {
  	setProperty(OUTPUT_FILE_NAME, "");
  	metaStringInfo(OUTPUT_FILE_NAME, "", "optional name for the output XML file");
    ValueList vl = new ValueList(DEFAULT_EXCLUSION_LIST);
    setPropertyValues(EXCLUSION_LIST, vl);
    metaMultiStringInfo(EXCLUSION_LIST, vl, "A list of product properties that should not appear in the XML.");
  }

  protected Product buildProduct(Product jugglersProduct) {
  	String outFileName = getStringProperty(OUTPUT_FILE_NAME,"");
  	if (outFileName.length() == 0)
  		outFileName = FileRef.getNewFileId("fx", "xml");
    File file = new File(getWorkingDir(), outFileName);

    FileWriter writer = null;
    try {
      writer = new FileWriter(file);
    }
    catch(IOException ioe) {
      reportException(ioe);
      return null;
    }

    PrintWriter out = new PrintWriter(writer);

    // read in the exclusion list and turn it into a Vector
    ValueList vl = getPropertyValues(EXCLUSION_LIST);
    Enumeration enum = vl.elements();
    Vector exclusionList = new Vector();

    while(enum.hasMoreElements()) {
      String property = (String)enum.nextElement();
      if(!exclusionList.contains(property)) {
        exclusionList.addElement(property);
      }
    }

    AsXML visitor = new AsXML();
    visitor.visit(jugglersProduct, exclusionList);
    visitor.writeResults(out);

    out.close();

    try {
      writer.close();
    }
    catch(IOException ioe) {
      reportException(ioe);
    }

    BasicProduct bp = new BasicProduct(getId(), getDescription());
    FileRef fr = new FileRef(file, file.lastModified(), file.getName());
    fr.setMIMEType(MIMEc.TEXT_XML);
    bp.setContent(fr);

    return bp;
  }
}
