/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.Product;

import FESI.jslib.JSGlobalObject;
import FESI.jslib.JSException;
import FESI.jslib.JSUtil;

import java.io.*;

/**
 *
 * This filter uses the FESI library to apply ECMAScript (aka JavaScript) code
 * to the incoming product where the script can return the same product or
 * produce its own product. Users of this filter are expected to write their
 * own custom scripts.
 *
 * The ECMA intrepreter is configured to use the JavaAccess, BasicIO, and
 * FileIO extensions by default. Other extensions (e.g. OptionalRegExp, Database)
 * must be explicitly loaded by the script via the loadExtension command.
 *
 * The agent (instance of activated ScriptFilter) and its received product 
 * are available as named variables in the script so the appropriate java
 * methods on those objects are defined (e.g. agent.getId() or
 * product.getDescription()).
 *
 * @author Jason Mathews
 * @since iPAM 2.4.85 (9/27/99)
 */

public class ProcessViaECMA extends NoSchedBase {

  public static final String SCRIPT_FILE = "scriptFile";

  public static final String SUPPORT_SUB_DIR = "ECMAScriptLib";

  private static final String DEFAULT_SCRIPT = "sample.es";

  protected static JSGlobalObject jsGlobal = null;

  private static final String[] jsExtensions =
  {
    "FESI.Extensions.JavaAccess",
    "FESI.Extensions.BasicIO",
    "FESI.Extensions.FileIO"
  };

  /**
   * This agent is not scheduled.
   */
  public ProcessViaECMA () {
    setDescription("Applies a script to the product");
    String defaultScriptFile = ExecGlobal.getSupportDir(SUPPORT_SUB_DIR).getAbsolutePath() +
      File.separator + DEFAULT_SCRIPT;
    setProperty(SCRIPT_FILE, defaultScriptFile);
    metaStringInfo(SCRIPT_FILE, defaultScriptFile, "ECMAScript to execute");
  }

  /**
   * A channel can build and change the product that it's pims provide
   */
  protected Product buildProduct(Product jugglersProduct) {

    String script = getStringProperty(SCRIPT_FILE, "");
    if (script.length() == 0) {
      reporter.error("missing value for " + SCRIPT_FILE);
      return null;
    }
    File file = new File(script);
    if (!file.exists()) {
      reporter.error("script '" + script + "' does not exist");
      return null;
    }

    BufferedReader reader = null;
    try {
      init(); // initialize script interpreter
      reader = new BufferedReader(new FileReader(file));
      /*
       * copy script file into a string buffer
       */
      String line;
      StringBuffer buf = new StringBuffer();
      while (null != (line = reader.readLine())) {
	if (line.length() != 0)
	  buf.append(line);
	buf.append('\n');
      }

      if (buf.length() == 0) {
	reporter.warning("no script to evaluate");
	return null;
      }

      return executeScript(buf, jugglersProduct);

    } catch(Exception e) {
      reportException(e);
    } finally {
      if (null != reader)
	try {
	  reader.close();
	} catch(IOException ioe){}
    }

    return null;
  }


/**
 * Executes script and returns its product
 *
 * @param scriptBuf StringBuffer containing the script code
 * @param jugglersProduct incoming product to examine
 * @return product
 */
  protected Product executeScript(StringBuffer scriptBuf, Product jugglersProduct) throws JSException {

    // Execute script as a function passing it two arguments: a reference
    // to this juggler, and its input product, which are accessed
    // via the script by the 'agent' and 'product' variable names.

    Object o = jsGlobal.evalAsFunction(scriptBuf.toString(),
				       new String[] { "agent", "product" },
				       new Object[] { this, jugglersProduct });
    if (o != null) {
      if (o instanceof Product)
	return (Product) o;
      else reporter.warning("returned object is not a product: " + o);
    }

    return null;
  }


/**
 * Initialize ECMAScript interpreter with defined extensions
 */
  protected void init() throws JSException {
    if (jsGlobal != null) return;
    
    synchronized(ProcessViaECMA.class) {
      // Create interpreter with defined extensions
      jsGlobal = JSUtil.makeEvaluator(jsExtensions);
    }
  }
	
}
