/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import com.kav.xsl.XSLProcessor;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.productVisitor.AsXML;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.smtp.MIMEc;

import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.executive.ExecGlobal;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import java.util.Enumeration;
import java.util.Vector;

/*
 * This filter uses XSLP to apply an XSL stylesheet to an XML file.  The
 * stylesheet basically controls what this filter produces.  Users of this
 * filter are expected to write their own stylesheets for use with this
 * filter.
 *
 * This filter includes an exclusion list attribute, however this list will
 * only be aplied to products that pass through the AsXML visitor.  In other
 * words if the incoming product contains an XML file the exclusion list
 * will not be applied.
 *
 * @author John R. Fantry
 */
public class ProcessViaXSL extends NoSchedBase {
  public static final String SUPPORT_SUB_DIR = "viaXMLorXSL";
  protected static final String DEFAULT_XSL_FILE = "0supportDir"+File.separator+SUPPORT_SUB_DIR+File.separator+"productToTable.xsl";
  protected static final String XSL_FILE = "XSLFile";

  // if true the resultant file will be parsed
  protected static final String POST_PROCESS = "postProcessHTMLFile";
  protected static final boolean DEFAULT_POST_PROCESS = true;

  // these tags will be stripped out of the resultant file created by XSLP and replace
  // them with '<' and '>'
  protected static final String LEFT_BRACKET = "%leftOpenBracket";
  protected static final String RIGHT_BRACKET = "%rightCloseBracket";

  // the name of the resultant file
  protected static final String TARGET_XSL_OUTPUT_NAME = "targetXSLOutputName";
  protected static final String DEFAULT_TARGET_XSL_OUTPUT_NAME = "noname.html";

  public ProcessViaXSL() {
    String defaultXslFile = ExecGlobal.getSupportDir(SUPPORT_SUB_DIR).getAbsolutePath() +
	File.separator + "productToTable.xsl";
    setProperty(XSL_FILE, defaultXslFile);
    ValueList vl = new ValueList(ForwardAsXMLFile.DEFAULT_EXCLUSION_LIST);
    setPropertyValues(ForwardAsXMLFile.EXCLUSION_LIST, vl);
    setProperty(POST_PROCESS, DEFAULT_POST_PROCESS);
    setProperty(TARGET_XSL_OUTPUT_NAME, DEFAULT_TARGET_XSL_OUTPUT_NAME);

    metaStringInfo(XSL_FILE, defaultXslFile, "The XSL stylesheet to use when converting from XML to HTML.");
    metaMultiStringInfo(ForwardAsXMLFile.EXCLUSION_LIST, vl, "A list of product properties to ignore.");
    metaBooleanInfo(POST_PROCESS, DEFAULT_POST_PROCESS, "If true the resultant file will be parsed.");
    metaStringInfo(TARGET_XSL_OUTPUT_NAME, DEFAULT_TARGET_XSL_OUTPUT_NAME, "The output file name.");
  }

  protected Product buildProduct(Product jinputProduct) {
  	try {
  		return myBuildProduct(jinputProduct);
  	}
  	catch (Exception e) {
  		reporter.reportException(e);
  	}
  		return null;
  }

  protected Product myBuildProduct(Product jinputProduct) {
    String XSLFileName = getStringProperty(XSL_FILE, DEFAULT_XSL_FILE);
    File XSLFile = new File(XSLFileName);
    if(!XSLFile.exists()) {
      warning("XSLFile does not exist, returning null");
      return null;
    }

    File XMLFile = null;

    // check to see if the product contains an XML file as content
    // if not, use the AsXML visitor to generate the XML
    Object content = jinputProduct.getContent();
    if(content instanceof FileRef) {
      String MIMEType = ((FileRef)content).getMIMEType();
      if(MIMEType.equals(MIMEc.TEXT_XML)) {
        XMLFile = ((FileRef)content).getFile();
      }
      else {
        // check the first line of the file to see if it contains an XML header
        try {
          BufferedReader in = new BufferedReader(new FileReader(((FileRef)content).getFile()));
          String line = in.readLine();
          if(line.indexOf("<?xml version") != -1) {
            XMLFile = ((FileRef)content).getFile();
          }
          in.close();
        }
        catch(IOException ioe) {
          reportException(ioe);
          return null;
        }
      }
    }

    boolean deleteTempFile = false;

    try {
      if(XMLFile == null) {
        XMLFile = createXMLFile(jinputProduct);
	deleteTempFile = true;
      }

      // check the product to see if it specifies a target file name
      String targetFileName = jinputProduct.getStringProperty(TARGET_XSL_OUTPUT_NAME, null);

      // next check the agents properties for a target file name
      if(targetFileName == null || targetFileName.length() == 0) {
        targetFileName = getStringProperty(TARGET_XSL_OUTPUT_NAME, null);
      }

      // if all else fails use the default
      if(targetFileName == null || targetFileName.length() == 0) {
        //reporter.log("no output file name defined, using default name");
        targetFileName = DEFAULT_TARGET_XSL_OUTPUT_NAME;
      }

      File outputFile = new File(getWorkingDir(), targetFileName);

      applyXSLStylesheet(XMLFile, XSLFile, outputFile);

      boolean replaceTags = getBooleanProperty(POST_PROCESS, DEFAULT_POST_PROCESS);
      if(replaceTags) {
        replaceTags(outputFile);
      }

      BasicProduct bp = new BasicProduct(getId(), getDescription());
      FileRef fileContent = new FileRef(outputFile, outputFile.lastModified(), outputFile.getName());
      bp.setContent(fileContent);

      return bp;
    }
    catch(IOException ioe) {
      reportException(ioe);
      return null;
    }
    finally {
    	if (deleteTempFile && XMLFile != null) {
    		XMLFile.delete();
    	}
    }
  }

  /*
   * Use the AsXML visitor to generate XML
   */
  protected File createXMLFile(Product prod) throws IOException {
    File workingDir = getWorkingDir();
    File XMLFile = new File(workingDir, FileRef.getNewFileId("xml", "tmp"));

    FileWriter writer = new FileWriter(XMLFile);
    PrintWriter out = new PrintWriter(writer);

    ValueList vl = getPropertyValues(ForwardAsXMLFile.EXCLUSION_LIST);
    Enumeration enum = vl.elements();
    Vector exclusionList = new Vector();

    while(enum.hasMoreElements()) {
      String property = (String)enum.nextElement();
      if(!exclusionList.contains(property)) {
        exclusionList.addElement(property);
      }
    }

    AsXML XMLVisitor = new AsXML();
    XMLVisitor.visit(prod, exclusionList);
    XMLVisitor.writeResults(out);

    out.close();
    writer.close();

    return XMLFile;
  }

  /*
   * Use XSLP to apply an XSL stylesheet to the XML file
   */
  protected void applyXSLStylesheet(File XMLFile, File XSLFile, File outputFile) throws IOException {
    FileInputStream XMLIn = new FileInputStream(XMLFile);
    FileInputStream XSLIn = new FileInputStream(XSLFile);
    FileWriter writer = new FileWriter(outputFile);
    PrintWriter out = new PrintWriter(writer);

    XSLProcessor XSLP = new XSLProcessor();
    XSLP.process(XMLIn, XSLIn, out);

    out.close();
    writer.close();
    XSLIn.close();
    XMLIn.close();
  }

  /*
   * Replace LEFT_BRACKET and RIGHT_BRACKET with '<' and '>'
   */
  protected void replaceTags(File inputFile) throws IOException {
    File tempFile = new File(getWorkingDir(), FileRef.getNewFileId("", "tmp"));

    BufferedReader in = new BufferedReader(new FileReader(inputFile));
    BufferedWriter out = new BufferedWriter(new FileWriter(tempFile));

    String line;

    while((line = in.readLine()) != null) {
      int startIndex = 0;
      int index = -1;
      String outLine = new String();
      while((index = line.indexOf(LEFT_BRACKET, startIndex)) != -1) {
        outLine = outLine + line.substring(startIndex, index) + "<";
        startIndex = index + LEFT_BRACKET.length();
      }
      outLine = outLine + line.substring(startIndex);

      line = outLine;
      outLine = new String();

      startIndex = 0;
      index = -1;

      while((index = line.indexOf(RIGHT_BRACKET, startIndex)) != -1) {
        outLine = outLine + line.substring(startIndex, index) + ">";
        startIndex = index + RIGHT_BRACKET.length();
      }
      outLine = outLine + line.substring(startIndex);

      out.write(outLine);
      out.newLine();
    }

    in.close();
    out.close();

    inputFile.delete();

    if(!tempFile.renameTo(inputFile)) {
      reporter.warning("Unable to replace tags.  Original file deleted.");
    }
  }
}
